package imcacat.imcadr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.awt.Dimension;
import java.io.File;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.junit.Test;

/**
 * I test <code>CliArgumentParser</code>.
 */
public class CliArgumentParserTest {
  /* Bad option. */
  @Test(expected = CliArgumentException.class)
  public void testParser01() throws CliArgumentException {
    String[] args = {"--bad-option"};
    CliArgumentParser.parse(args);
  }

  /* No arguments. */
  @Test
  public void testParser02() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.help());
    assertFalse(cargs.version());
    assertFalse(cargs.server());
    assertNull(cargs.configurationFile());
    assertNull(cargs.debugFile());
    assertNull(cargs.exposureTime());
    assertNull(cargs.cellSize());
    assertNull(cargs.zySize());
    assertNull(cargs.zyPreClipSize());
    assertNull(cargs.zyMask());
    assertNull(cargs.zxSize());
    assertNull(cargs.zxPreClipSize());
    assertNull(cargs.zxMask());
    assertFalse(cargs.zxFirst());
    assertTrue(cargs.flyScanEnabled());
    assertTrue(cargs.chasingEnabled());
    assertNull(cargs.bestScoreCutoff());
    assertNull(cargs.analysisPackage());
    assertFalse(cargs.weakDiffracter());
    assertNull(cargs.highResolutionCutoff());
    assertNull(cargs.lowResolutionCutoff());
    assertNull(cargs.minimumSpotArea());
    assertNull(cargs.minimumSignalHeight());
    assertNull(cargs.imagesDirectory());
  }

  /* Missing required options. */
  @Test(expected = CliArgumentException.class)
  public void testParser03() throws CliArgumentException {
    String[] args = {"--exp-time=0.5"};
    CliArgumentParser.parse(args);
  }

  /* Missing required options. */
  @Test(expected = CliArgumentException.class)
  public void testParser04() throws CliArgumentException {
    String[] args = {"--cell-size=70x35"};
    CliArgumentParser.parse(args);
  }

  /* Missing required options. */
  @Test(expected = CliArgumentException.class)
  public void testParser05() throws CliArgumentException {
    String[] args = {"--zy-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* Missing required options. */
  @Test(expected = CliArgumentException.class)
  public void testParser06() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35"};
    CliArgumentParser.parse(args);
  }

  /* Specifying only one raster size is OK. */
  @Test
  public void testParser07() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* Specifying only one raster size is OK. */
  @Test
  public void testParser08() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zx-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* Help. */
  @Test
  public void testParser09() throws CliArgumentException {
    String[] args = {"--help"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.help());

    args = new String[] {"-h"};
    cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.help());
  }

  /* Version. */
  @Test
  public void testParser10() throws CliArgumentException {
    String[] args = {"--version"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.version());
  }

  /* All options. */
  @Test
  public void testParser11() throws CliArgumentException {
    String[] args = {
        "--conf-file=/nonexistent/imcadr.conf",
        "--debug-file=/nonexistent/debug.txt",
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "--zx-size=4x3",
        "--zy-pre-clip-size=30x24",
        "--zx-pre-clip-size=31x25",
        "--zy-mask=1:1,5,6;2:1;5:1,2,6",
        "--zx-mask=1:1,2;2:1;3:1,2,4",
        "--zx-first",
        "--step-scan",
        "--no-chase",
        "--best-score-cutoff=7",
        "--scoring-pkg=dtrek",
        "--weak-diff",
        "--high-res-cutoff=5.1",
        "--low-res-cutoff=50.1",
        "--min-spot-area=5",
        "--min-signal-height=2.5",
        "/tmp"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.help());
    assertFalse(cargs.version());
    assertFalse(cargs.server());
    assertEquals(new File("/nonexistent/imcadr.conf"), cargs.configurationFile());
    assertEquals(new File("/nonexistent/debug.txt"), cargs.debugFile());
    assertEquals(0.25, cargs.exposureTime(), 1e-6);
    assertEquals(new Dimension(65, 25), cargs.cellSize());
    assertEquals(new Dimension(6, 5), cargs.zySize());
    assertEquals(new Dimension(30, 24), cargs.zyPreClipSize());
    Map<Integer, Set<Integer>> mask = new HashMap<>();
    mask.put(0, new HashSet<Integer>(Arrays.asList(0, 4, 5)));
    mask.put(1, new HashSet<Integer>(Arrays.asList(0)));
    mask.put(4, new HashSet<Integer>(Arrays.asList(0, 1, 5)));
    assertEquals(new Raster2dMask(mask), cargs.zyMask());
    assertEquals(new Dimension(4, 3), cargs.zxSize());
    assertEquals(new Dimension(31, 25), cargs.zxPreClipSize());
    mask = new HashMap<>();
    mask.put(0, new HashSet<Integer>(Arrays.asList(0, 1)));
    mask.put(1, new HashSet<Integer>(Arrays.asList(0)));
    mask.put(2, new HashSet<Integer>(Arrays.asList(0, 1, 3)));
    assertEquals(new Raster2dMask(mask), cargs.zxMask());
    assertTrue(cargs.zxFirst());
    assertFalse(cargs.flyScanEnabled());
    assertFalse(cargs.chasingEnabled());
    assertEquals(Integer.valueOf(7), cargs.bestScoreCutoff());
    assertEquals(AnalysisPackage.DTREK, cargs.analysisPackage());
    assertTrue(cargs.weakDiffracter());
    assertEquals(5.1, cargs.highResolutionCutoff(), 1e-6);
    assertEquals(50.1, cargs.lowResolutionCutoff(), 1e-6);
    assertEquals(Integer.valueOf(5), cargs.minimumSpotArea());
    assertEquals(2.5, cargs.minimumSignalHeight(), 1e-6);
    assertEquals(new File("/tmp"), cargs.imagesDirectory());
  }

  /* Bad cell-size dimension width < 0. */
  @Test(expected = CliArgumentException.class)
  public void testParser12() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=-70x35",
        "--zx-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* Bad cell-size dimension width == 0. */
  @Test(expected = CliArgumentException.class)
  public void testParser13() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=0x35",
        "--zx-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* Single argument images directory usage. */
  @Test
  public void testParser14() throws CliArgumentException {
    String[] args = {"/tmp"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(new File("/tmp"), cargs.imagesDirectory());
    assertFalse(cargs.isEmpty());
    assertTrue(cargs.onlyHasImagesDirectory());
  }

  /* Many options but no images directory. */
  @Test
  public void testParser15() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "--zx-size=4x3",
        "--no-chase",
        "--weak-diff"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.help());
    assertFalse(cargs.version());
    assertEquals(0.25, cargs.exposureTime(), 1e-6);
    assertEquals(new Dimension(65, 25), cargs.cellSize());
    assertEquals(new Dimension(6, 5), cargs.zySize());
    assertEquals(new Dimension(4, 3), cargs.zxSize());
    assertFalse(cargs.chasingEnabled());
    assertTrue(cargs.weakDiffracter());
    assertNull(cargs.imagesDirectory());
  }

  /* Server. */
  @Test
  public void testParser16() throws CliArgumentException {
    String[] args = {"--server"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.server());
  }

  /* Weak diffracter flag is off by default. */
  @Test
  public void testParser17() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.weakDiffracter());
  }

  /* Weak diffracter flag. */
  @Test
  public void testParser18() throws CliArgumentException {
    String[] args = {
        "--weak-diff",
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.weakDiffracter());
  }

  /* No-chasing flag is off by default. */
  @Test
  public void testParser19() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.chasingEnabled());
  }

  /* No-chasing flag. */
  @Test
  public void testParser20() throws CliArgumentException {
    String[] args = {
        "--no-chase",
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.chasingEnabled());
  }

  /* High resolution cutoff is unspecified by default. */
  @Test
  public void testParser21() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertNull(cargs.highResolutionCutoff());
  }

  /* High resolution cutoff. */
  @Test
  public void testParser22() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "--high-res-cutoff=3.1"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(3.1, cargs.highResolutionCutoff(), 1e-6);
  }

  /* Low resolution cutoff is unspecified by default. */
  @Test
  public void testParser23() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertNull(cargs.lowResolutionCutoff());
  }

  /* Low resolution cutoff. */
  @Test
  public void testParser24() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "--low-res-cutoff=50.1"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(50.1, cargs.lowResolutionCutoff(), 1e-6);
  }

  /* Minimum spot area is unspecified by default. */
  @Test
  public void testParser25() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertNull(cargs.minimumSpotArea());
  }

  /* Minimum spot area. */
  @Test
  public void testParser26() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "-min-spot-area=5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(Integer.valueOf(5), cargs.minimumSpotArea());
  }

  /* Minimum signal height is unspecified by default. */
  @Test
  public void testParser27() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertNull(cargs.minimumSignalHeight());
  }

  /* Minimum signal height. */
  @Test
  public void testParser28() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5",
        "--min-signal-height=2.5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(2.5, cargs.minimumSignalHeight(), 1e-6);
  }

  /* ZX-first flag is off by default. */
  @Test
  public void testParser29() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.zxFirst());
  }

  /* ZX-first flag. */
  @Test
  public void testParser30() throws CliArgumentException {
    String[] args = {
        "--zx-first",
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.zxFirst());
  }

  /* ZY-pre-clip-size flag requires ZY-size flag. */
  @Test(expected = CliArgumentException.class)
  public void testParser31() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-pre-clip-size=9x8",
        "--zx-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* ZX-pre-clip-size flag requires ZX-size flag. */
  @Test(expected = CliArgumentException.class)
  public void testParser32() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zx-pre-clip-size=9x8",
        "--zy-size=5x5"};
    CliArgumentParser.parse(args);
  }

  /* ZY-mask flag requires ZY-size flag. */
  @Test(expected = CliArgumentException.class)
  public void testParser33() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-mask=1:1,2"};
    CliArgumentParser.parse(args);
  }

  /* ZX-mask flag requires ZX-size flag. */
  @Test(expected = CliArgumentException.class)
  public void testParser34() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zx-mask=1:1,2"};
    CliArgumentParser.parse(args);
  }

  /* Mask with just one entry. */
  @Test
  public void testParser35() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:1"};
    CliArguments a = CliArgumentParser.parse(args);
    Raster2dMask m = a.zyMask();
    assertTrue(m.get(0, 0));
  }

  /* Mask that is empty is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser36() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask="};
    CliArgumentParser.parse(args);
  }

  /* Mask with row with no columns is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser37() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:"};
    CliArgumentParser.parse(args);
  }

  /* Mask with empty row is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser38() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:1;;3:1"};
    CliArgumentParser.parse(args);
  }

  /* Mask indices are 1-based, so row 0 is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser39() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=0:2"};
    CliArgumentParser.parse(args);
  }

  /* Mask indices are 1-based, so column 0 is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser40() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:0"};
    CliArgumentParser.parse(args);
  }

  /* Mask outside raster row boundary is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser41() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=6:1"};
    CliArgumentParser.parse(args);
  }

  /* Mask outside raster column boundary is invalid. */
  @Test(expected = CliArgumentException.class)
  public void testParser42() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:6"};
    CliArgumentParser.parse(args);
  }

  /* Mask with one row. */
  @Test
  public void testParser47() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:1,5"};
    CliArguments a = CliArgumentParser.parse(args);
    Raster2dMask m = a.zyMask();
    /* Raster2dMask indices are 0-based, not 1-based, and the order is
     * (column,row), not (row,column). */
    assertTrue(m.get(0, 0));
    assertFalse(m.get(1, 0));
    assertFalse(m.get(2, 0));
    assertFalse(m.get(3, 0));
    assertTrue(m.get(4, 0));
  }

  /* Mask with one row, but masked internal gap is converted to unmasked. */
  @Test
  public void testParser48() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:1,3,5"};
    CliArguments a = CliArgumentParser.parse(args);
    Raster2dMask m = a.zyMask();
    /* Raster2dMask indices are 0-based, not 1-based, and the order is
     * (column,row), not (row,column). */
    assertTrue(m.get(0, 0));
    assertFalse(m.get(1, 0));
    assertFalse(m.get(2, 0));
    assertFalse(m.get(3, 0));
    assertTrue(m.get(4, 0));
  }

  /* Mask with two rows. */
  @Test
  public void testParser43() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--cell-size=70x35",
        "--zy-size=5x5",
        "--zy-mask=1:1,5;5:1,4,5"};
    CliArguments a = CliArgumentParser.parse(args);
    Raster2dMask m = a.zyMask();
    /* Raster2dMask indices are 0-based, not 1-based, and the order is
     * (column,row), not (row,column). */
    assertTrue(m.get(0, 0));
    assertFalse(m.get(1, 0));
    assertFalse(m.get(2, 0));
    assertFalse(m.get(3, 0));
    assertTrue(m.get(4, 0));
    assertTrue(m.get(0, 4));
    assertFalse(m.get(1, 4));
    assertFalse(m.get(2, 4));
    assertTrue(m.get(3, 4));
    assertTrue(m.get(4, 4));
  }

  /* Step-scan flag is off by default. */
  @Test
  public void testParser44() throws CliArgumentException {
    String[] args = {};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertTrue(cargs.flyScanEnabled());
  }

  /* Step-scan flag. */
  @Test
  public void testParser45() throws CliArgumentException {
    String[] args = {
        "--step-scan",
        "--exp-time=0.25",
        "--cell-size=65x25",
        "--zy-size=6x5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertFalse(cargs.flyScanEnabled());
  }

  /* Debug-file flag indicates not GUI mode and so requires other options. */
  @Test(expected = CliArgumentException.class)
  public void testParser46() throws CliArgumentException {
    String[] args = {
        "--debug-file=/nonexistent/debug.txt"};
    CliArgumentParser.parse(args);
  }

  /* Deprecated "dstartrek" analysis package is still supported. */
  /* Specifying only one raster size is OK. */
  @Test
  public void testParser49() throws CliArgumentException {
    String[] args = {
        "--exp-time=0.5",
        "--scoring-pkg=dstartrek",
        "--cell-size=70x35",
        "--zy-size=5x5"};
    CliArguments cargs = CliArgumentParser.parse(args);
    assertEquals(AnalysisPackage.DTREK, cargs.analysisPackage());
  }
}
