package imcacat.imcadr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * I test <code>ClusterService</code>.
 */
public class ClusterServiceTest {
  private ClusterService cs;

  @Before
  public void setUp() {
    cs = null;
  }

  @After
  public void tearDown() {
    if (cs != null) {
      cs.shutDownNow();
      cs = null;
    }
  }

  @Test(timeout = 2500)
  public void testExecute() throws InterruptedException {
    cs = newService("vader:1, jabba:1, bobafett:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    cs.execute(r3);
    r1.awaitRunning();
    r2.awaitRunning();
    r3.awaitRunning();
    assertEquals("vader", r1.getClusterNode());
    assertEquals("jabba", r2.getClusterNode());
    assertEquals("bobafett", r3.getClusterNode());
  }

  @Test(timeout = 2500)
  public void testExecute_scheduleOnLowestLoad() throws InterruptedException {
    cs = newService("vader:1, jabba:1, bobafett:2");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r4 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    cs.execute(r3);
    cs.execute(r4);
    r1.awaitRunning();
    r2.awaitRunning();
    r3.awaitRunning();
    r4.awaitRunning();
    assertEquals("vader", r1.getClusterNode());
    assertEquals("jabba", r2.getClusterNode());
    assertEquals("bobafett", r3.getClusterNode());
    assertEquals("bobafett", r4.getClusterNode());
  }

  @Test(timeout = 2500)
  public void testExecute_loadGoesDownWhenJobFinishes() throws InterruptedException {
    cs = newService("vader:1, jabba:1, bobafett:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r4 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    cs.execute(r3);
    cs.execute(r4);
    r1.awaitRunning();
    r2.awaitRunning();
    r3.awaitRunning();
    assertFalse(r4.isRunning());
    assertEquals("vader", r1.getClusterNode());
    assertEquals("jabba", r2.getClusterNode());
    assertEquals("bobafett", r3.getClusterNode());
    r2.finish();
    r4.awaitRunning();
    assertEquals("jabba", r4.getClusterNode());
  }

  @Test(timeout = 2500)
  public void testExecute_jobsWaitForClusterCapacity() throws InterruptedException {
    cs = newService("vader:1, jabba:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r4 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    cs.execute(r3);
    cs.execute(r4);
    r1.awaitRunning();
    r2.awaitRunning();
    assertFalse(r3.isRunning());
    assertFalse(r4.isRunning());
    assertEquals("vader", r1.getClusterNode());
    assertEquals("jabba", r2.getClusterNode());
    r2.finish();
    r3.awaitRunning();
    assertEquals("jabba", r3.getClusterNode());
    r1.finish();
    r4.awaitRunning();
    assertEquals("vader", r4.getClusterNode());
  }

  @Test(timeout = 2500)
  public void testExecute_oneNodeCluster() throws InterruptedException {
    cs = newService("vader:2");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    cs.execute(r3);
    r1.awaitRunning();
    r2.awaitRunning();
    assertFalse(r3.isRunning());
    assertEquals("vader", r1.getClusterNode());
    assertEquals("vader", r2.getClusterNode());
    r1.finish();
    r3.awaitRunning();
    assertEquals("vader", r3.getClusterNode());
  }

  @Test(expected = RejectedExecutionException.class)
  public void testExecute_afterShutDown() throws InterruptedException {
    cs = newService("vader:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    cs.shutDown();
    cs.execute(r1);
  }

  @Test(expected = RejectedExecutionException.class)
  public void testExecute_afterShutDownWithRunning() throws InterruptedException {
    cs = newService("vader:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    r1.awaitRunning();
    cs.shutDown();
    cs.execute(r2);
  }

  @Test(expected = RejectedExecutionException.class)
  public void testExecute_afterShutDownWithWaiting() throws InterruptedException {
    cs = newService("vader:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r3 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    r1.awaitRunning();
    assertFalse(r2.isRunning());
    cs.shutDown();
    cs.execute(r3);
  }

  @Test(expected = NullPointerException.class)
  public void testConstructor_nullNodesList() {
    cs = new ClusterService(null);
  }

  @Test(expected = IllegalArgumentException.class)
  public void testConstructor_emptyNodesList() {
    cs = new ClusterService(Collections.<ClusterNode>emptyList());
  }

  @Test(timeout = 5000)
  public void testShutDown() throws InterruptedException {
    cs = newService("vader:1");
    assertFalse(cs.isShutDown());
    assertFalse(cs.isTerminated());
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    r1.awaitRunning();
    cs.shutDown();
    assertTrue(cs.isShutDown());
    assertFalse(cs.isTerminated());
    r1.finish();
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  @Test(timeout = 5000)
  public void testShutDown_neverExecutedAnything() throws InterruptedException {
    cs = newService("vader:1");
    cs.shutDown();
    assertTrue(cs.isShutDown());
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  @Test(timeout = 5000)
  public void testShutDown_withWaiting() throws InterruptedException {
    cs = newService("vader:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    r1.awaitRunning();
    assertFalse(r2.isRunning());
    cs.shutDown();
    assertTrue(cs.isShutDown());
    assertFalse(cs.isTerminated());
    assertTrue(r1.isRunning());
    r1.finish();
    r2.awaitRunning();
    r2.finish();
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  @Test(timeout = 5000)
  public void testShutDownNow() throws InterruptedException {
    cs = newService("vader:1");
    assertFalse(cs.isShutDown());
    assertFalse(cs.isTerminated());
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    r1.awaitRunning();
    List<ClusterRunnable> neverRan = cs.shutDownNow();
    assertNotNull(neverRan);
    assertTrue(neverRan.isEmpty());
    assertTrue(cs.isShutDown());
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  @Test(timeout = 5000)
  public void testShutDownNow_neverExecutedAnything() throws InterruptedException {
    cs = newService("vader:1");
    List<ClusterRunnable> neverRan = cs.shutDownNow();
    assertNotNull(neverRan);
    assertTrue(neverRan.isEmpty());
    assertTrue(cs.isShutDown());
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  @Test(timeout = 5000)
  public void testShutDownNow_withWaiting() throws InterruptedException {
    cs = newService("vader:1");
    ClusterRunnableFake r1 = new ClusterRunnableFake(5000);
    ClusterRunnableFake r2 = new ClusterRunnableFake(5000);
    cs.execute(r1);
    cs.execute(r2);
    r1.awaitRunning();
    assertFalse(r2.isRunning());
    List<ClusterRunnable> neverRan = cs.shutDownNow();
    assertTrue(cs.isShutDown());
    assertNotNull(neverRan);
    assertEquals(1, neverRan.size());
    assertSame(r2, neverRan.get(0));
    assertTrue(cs.awaitTermination(2500, TimeUnit.MILLISECONDS));
    assertTrue(cs.isTerminated());
  }

  private static ClusterService newService(String nodeList) {
    return new ClusterService(newNodes(nodeList));
  }

  private static List<ClusterNode> newNodes(String s) {
    try (Reader r = new StringReader("ClusterNodes = " + s)) {
      return (new ImcadrConfigurationParser(r)).parse().clusterNodes();
    } catch (IOException | ConfigurationException e) {
      if (e instanceof InterruptedIOException) Thread.currentThread().interrupt();
      throw new RuntimeException(e);
    }
  }

  private static class ClusterRunnableFake implements ClusterRunnable {
    private final long runtimeInMs;
    private final CountDownLatch runningLatch;
    private final CountDownLatch finishLatch;
    private volatile String hostname;

    public ClusterRunnableFake(long runtimeInMs) {
      super();
      this.runtimeInMs = runtimeInMs;
      runningLatch = new CountDownLatch(1);
      finishLatch = new CountDownLatch(1);
      hostname = null;
    }

    @Override
    public void run() {
      runningLatch.countDown();
      try {
        finishLatch.await(runtimeInMs, TimeUnit.MILLISECONDS);
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        return;
      } finally {
        finishLatch.countDown();
      }
    }

    public String getClusterNode() {
      return hostname;
    }

    @Override
    public void setClusterNode(String hostname) {
      this.hostname = hostname;
    }

    public boolean isRunning() throws InterruptedException {
      return runningLatch.await(0L, TimeUnit.MILLISECONDS) &&
          !finishLatch.await(0L, TimeUnit.MILLISECONDS);
    }

    public void awaitRunning() throws InterruptedException {
      runningLatch.await();
    }

    public void finish() {
      finishLatch.countDown();
    }

    @Override
    public String toString() {
      StringBuilder result = new StringBuilder();
      result.append("{");
      result.append("runtimeInMs=");
      result.append(runtimeInMs);
      result.append(", ");
      result.append("hostname=");
      result.append(hostname);
      result.append(", ");
      result.append("running=");
      try {
        result.append(isRunning());
      } catch (InterruptedException e) {
        result.append("? (interrupted)");
        Thread.currentThread().interrupt();
      }
      result.append(", ");
      result.append("runningLatchCount=");
      result.append(runningLatch.getCount());
      result.append(", ");
      result.append("finishLatchCount=");
      result.append(finishLatch.getCount());
      result.append("}");
      return result.toString();
    }
  }
}
