package imcacat.imcadr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.ByteArrayInputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.charset.Charset;

import org.junit.Test;

/**
 * I test <code>DTrekDtrankerParser</code>.
 */
public class DTrekDtrankerParserTest {
  private static final Charset UTF8 = Charset.forName("UTF-8");

  /* Typical. */
  @Test
  public void testParse01() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDtrankerOutput01());
      DTrekDtrankerParser parser = new DTrekDtrankerParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(185, result.getScore().intValue());
    assertEquals(1.23, result.getEstimatedResolution1().doubleValue(), 0.001);
    assertEquals(0.3, result.getMosaicity().doubleValue(), 0.001);
    assertEquals(SpaceGroup.P4, result.getSpaceGroup());
    assertEquals(new UnitCell(57.6451, 57.6451, 150.0757, 90.0, 90.0, 90.0),
        result.getUnitCell());
  }

  /*
   * Resolution stats row with I/sig average less than 2, and negative score.
   */
  @Test
  public void testParse02() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDtrankerOutput02());
      DTrekDtrankerParser parser = new DTrekDtrankerParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(-185, result.getScore().intValue());
    assertEquals(1.39, result.getEstimatedResolution1().doubleValue(), 0.001);
    assertEquals(0.3, result.getMosaicity().doubleValue(), 0.001);
    assertEquals(SpaceGroup.P4, result.getSpaceGroup());
    assertEquals(new UnitCell(57.6451, 57.6451, 150.0757, 90.0, 90.0, 90.0),
        result.getUnitCell());
  }

  /*
   * Error from dtindex.
   */
  @Test
  public void testParse03() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDtrankerOutput03());
      DTrekDtrankerParser parser = new DTrekDtrankerParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertNotNull(result.getScore());
    assertEquals(-80, result.getScore().intValue());
  }

  /* Output for a saturated image. */
  @Test
  public void testParse04() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDtrankerOutput04());
      DTrekDtrankerParser parser = new DTrekDtrankerParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(-208, result.getScore().intValue());
    assertNull(result.getEstimatedResolution1());
    assertNull(result.getMosaicity());
    assertNull(result.getSpaceGroup());
    assertNull(result.getUnitCell());
  }

  /* Typical. */
  private static byte[] createDtrankerOutput01() {
    /* @formatter:off */
    String[] output =
      new String[] {
        "",
        "dtranker:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "",
        "Command line:",
        "dtranker image=image_0003.cbf",
        "",
        "",
        "dtextractheader:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "     Header of file image_0003.cbf successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "",
        "dtfind:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dfind dtranker_input.head -reso edge -seq 3 3 -sigma 3 -min 50 -2D \\",
        "    -filter 6 -window 0 0 -rank -ref dtranker_find.ref -out \\",
        "    dtranker_find.head",
        "",
        "     Header of file dtranker_input.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "",
        "     File image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "",
        "Command line string: >>-reso<<",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Command line string: >>-seq<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-min<<",
        "Command line string: >>-2D<<",
        "Command line string: >>-filter<<",
        "Command line string: >>-window<<",
        "Command line string: >>-rank<<",
        "Command line string: >>-ref<<",
        "Command line string: >>-out<<",
        "",
        "Resolution limits of an image are 999.99 to 1.04645",
        "Resolution limits of peak search are 999.99 to 1.23108",
        "",
        "",
        "dtfind: 2D method  used",
        "...reading image /tmp/image_0003.cbf...",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "Processing image '/tmp/image_0003.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Resolution rings found.",
        "--------------------------------------------------",
        " Peak 2Theta 2Theta 2Theta    Reso    Reso    Peak ",
        "   #     Mid  Start    End   Start     End Percent ",
        "--------------------------------------------------",
        "    0  15.75  14.25  17.25  4.0311  3.3341    0.58",
        "--------------------------------------------------",
        "Find object listing:",
        "     Sigma: 3",
        "Resolution: 999.99 to 1.23108",
        "   Minimum: 50",
        "Circle lim: 1231, 1263, 0, 1763",
        "  Rect lim: 24, 25, 2439, 2502",
        "Spot wind.: 0, 0",
        "Peak filt.: 6",
        "Back. tile: 153, 157",
        " Seq. num.: 3, 3",
        "Image dim.: 2463, 2527",
        "   3D dump: 0",
        "251 preliminary spots found in 2D search with rotation angle 1.25 degs.",
        "",
        "INFO: 31 reflns deleted out of 194 that might be in ice rings.",
        "",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 2.65     64    0   37.8   36.51      5.6      0.06",
        "  2.65 - 2.11     78    0   21.4   25.32      5.0      0.07",
        "  2.11 - 1.84     19    0   17.1   18.42      4.2      0.08",
        "  1.84 - 1.67      2    0   19.5   12.61      3.6      0.09",
        "  1.67 - 1.55      0    0    0.0    0.00      0.0      0.00",
        "  1.55 - 1.46      0    0    0.0    0.00      0.0      0.00",
        "  1.46 - 1.39      0    0    0.0    0.00      0.0      0.00",
        "  1.39 - 1.33      0    0    0.0    0.00      0.0      0.00",
        "  1.33 - 1.28      0    0    0.0    0.00      0.0      0.00",
        "  1.28 - 1.23      0    0    0.0    0.00      0.0      0.00",
        "-----------------------------------------------------------",
        " 99.99 - 1.23    163    0   27.3   28.75      5.1      0.07",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    27 -  >  30     47    0  2.12   34.57      5.3      0.06",
        "    24 -     27      9    0  2.16   31.26      5.3      0.06",
        "    21 -     24     11    0  1.80   23.93      4.9      0.08",
        "    18 -     21     18    0  1.89   26.10      5.2      0.06",
        "    15 -     18     22    0  1.88   25.64      5.1      0.08",
        "    12 -     15     30    0  1.95   26.49      5.0      0.07",
        "     9 -     12     17    0  1.83   26.96      5.0      0.07",
        "     6 -      9      7    0  2.03   24.93      5.0      0.06",
        "     3 -      6      2    0  2.39   28.12      5.2      0.05",
        " <   0 -      3      0    0  0.00    0.00      0.0      0.00",
        "------------------------------------------------------------",
        " <   0 -  >  30    163    0  1.80   28.75      5.1      0.07",
        "------------------------------------------------------------",
        "",
        "dtfind: There were 163 spots found.",
        "There were 194 preliminary spots of which 0 were marked as saturated,",
        " or  0.00% of them.",
        "Number of reflections written in 'dtranker_find.ref': 163",
        "dtfind: Spots written to dtranker_find.ref",
        "dtfind - Wrote header file dtranker_find.head",
        "",
        "INFO - The suggested next step is to run dtindex with",
        "       dtranker_find.head as the input .head file and",
        "       dtranker_find.ref as the input .ref file:",
        "",
        "       dtindex dtranker_find.head dtranker_find.ref ...",
        "",
        "",
        "dtindex:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtindex dtranker_find.head dtranker_find.ref -maxresid 3.0 -sigma 5 \\",
        "    -out dtranker_index.head",
        "",
        "     Header of file dtranker_find.head successfully read.",
        "Reflection list: dtranker_find.ref",
        "Creflnlist::nRead with filename: dtranker_find.ref",
        "INFO in Creflnlist::nRead, EOF after 163 reflections read in",
        "                                    (163 total now in list).",
        "Command line string: >>-maxresid<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-out<<",
        "INFO: deleted 0 reflns outside of resolution bounds.",
        "      This leaves 163 reflns for indexing.",
        "",
        "INFO: 51 reflns deleted out of 163 that might be in ice rings.",
        "",
        "Max cell length allowed for reciprocal lattice vectors: 307.677",
        "",
        "",
        "Method:     1D FFT with DPS algorithm",
        "Out header: dtranker_index.head",
        "Max cell:   307.677",
        "Num vecs:   1000",
        "Spacegroup: 0",
        "Verbose:    1",
        "",
        "Performing 1D FFT indexing (not cell reduction) with the DPS algorithm...see",
        "  Steller, Bolotovsky, & Rossmann (1997) J. Appl. Cryst. 30, 1036-1040.",
        "",
        "Max cell is:         307.677",
        "Number of reflections/vectors used: 100",
        ".....................................",
        "...refining best 30 directions and lengths...",
        "..............................  done.",
        "Number of vectors used for integer residual calculation: 100",
        "",
        "     a      b      c  alpha   beta  gamma   Volume  Remarks #Indexed %Residual",
        "==============================================================================",
        "128.67  81.40 281.47  43.37 123.51 161.50   498709     Okay       59     0.064",
        "128.67  81.40 229.21  57.48 109.94 161.50   498692     Okay       59     0.064",
        "281.44  57.71  81.41  45.03  43.37  34.92   498781     Okay       58     0.062",
        " 57.71  81.41 281.44  43.37  34.92  45.03   498781     Okay       58     0.062",
        " 81.41 281.44  57.71  34.92  45.03  43.37   498781     Okay       58     0.062",
        "229.21  57.71  81.40  45.04  57.48  40.89   498692     Okay       59     0.064",
        " 57.71  81.40 229.21  57.48  40.89  45.04   498692     Okay       59     0.064",
        " 81.40 229.21  57.71  40.89  45.04  57.48   498692     Okay       59     0.064",
        "281.45  81.66  57.71  44.84  34.90  64.28   498105     Okay       59     0.063",
        " 81.66  57.71 281.45  34.90  64.28  44.84   498104     Okay       59     0.063",
        " 57.71 281.45  81.66  64.28  44.84  34.90   498104     Okay       59     0.063",
        " 81.67  57.71 229.18  40.83  57.83  44.83   497826     Okay       57     0.062",
        " 57.71 229.18  81.67  57.83  44.83  40.83   497834     Okay       57     0.062",
        "229.18  81.67  57.71  44.83  40.83  57.83   497829     Okay       57     0.062",
        "",
        "",
        "Executing beam refinement with",
        "beam search radius, acceptable shift radius: 10 9 ...",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Original (input header) beam center: [1252.0 1288.0]",
        "",
        "First Pass.  Search dim0 in [1242 1262] dim1 in [1278 1298]",
        ".............................................",
        "Second Pass.  Search dim0 in [1252 1254] dim1 in [1287 1289]",
        "Calculated pre-reduced cell solution is in agreement with detector beam center!",
        "Original (input header) beam center: [1252.0 1288.0]",
        "New      (calculated)   beam center: [1252.9 1288.5]",
        "Header updated to reflect beam center change.",
        "INFO: Restart with adjusted beam center.",
        "Max cell length allowed for reciprocal lattice vectors: 307.677",
        "",
        "",
        "Method:     1D FFT with DPS algorithm",
        "Out header: dtranker_index.head",
        "Max cell:   307.677",
        "Num vecs:   1000",
        "Spacegroup: 0",
        "Verbose:    1",
        "",
        "Performing 1D FFT indexing (not cell reduction) with the DPS algorithm...see",
        "  Steller, Bolotovsky, & Rossmann (1997) J. Appl. Cryst. 30, 1036-1040.",
        "",
        "Max cell is:         307.677",
        "Number of reflections/vectors used: 100",
        ".....................................",
        "...refining best 30 directions and lengths...",
        "..............................  done.",
        "Number of vectors used for integer residual calculation: 100",
        "",
        "     a      b      c  alpha   beta  gamma   Volume  Remarks #Indexed %Residual",
        "==============================================================================",
        " 57.69  81.59 281.59  64.17 101.75  44.92   498636     Okay       47     0.035",
        " 81.59 281.59  57.69 101.75  44.92  64.17   498636     Okay       47     0.035",
        "281.62  57.69  81.59  44.92  64.17 101.75   498727     Okay       50     0.037",
        "129.03  81.59 281.59  64.17  79.37  18.40   498636     Okay       47     0.035",
        "128.87  81.46 281.60  43.55 123.32 161.55   498542     Okay       49     0.036",
        " 81.46 281.60  57.69 101.74 134.99  43.55   498542     Okay       49     0.036",
        "281.60  57.69  81.46 134.99  43.55 101.74   498542     Okay       49     0.036",
        " 57.69  81.46 281.60  43.55 101.74 134.99   498542     Okay       49     0.036",
        "281.60 128.87  57.69  26.55 101.74 123.32   498540     Okay       49     0.036",
        "128.87  57.69 281.60 101.74 123.32  26.55   498542     Okay       49     0.036",
        " 57.68 129.02 281.58  79.37 101.75  26.52   498687     Okay       51     0.038",
        "",
        "",
        "Least square fit to lattice characters...see",
        "  Andrews & Bernstein (1988) Acta Cryst. A44, 1009-1018 and",
        "  Paciorek & Bonin (1992) J. Appl. Cryst. 25, 632-637.",
        "............................................................",
        "............................................................",
        "done.",
        "",
        "Least-squares fit of reduced primitive cell to 44 lattice characters",
        "sorted on decreasing (highest to lowest) symmetry.",
        "Only solutions with residuals <=   3.0 are listed.",
        "=======================================================================",
        " Soln  LeastSq Spgrp Cent    Bravais type         a         b         c",
        "  num residual  num* type     Cell volume     alpha      beta     gamma",
        "=======================================================================",
        "   7     0.415    75    P      tetragonal    57.645    57.645   150.076",
        "                                   498696    90.000    90.000    90.000",
        "",
        "   9     0.413    21    C    orthorhombic    81.465    81.580   150.076",
        "                                   997391    90.000    90.000    90.000",
        "",
        "  11     0.414    16    P    orthorhombic    57.607    57.684   150.076",
        "                                   498695    90.000    90.000    90.000",
        "",
        "  12     0.290     5    C      monoclinic    81.580    81.465   150.076",
        "                                   997383    90.000    90.224    90.000",
        "",
        "  13     0.041     3    P      monoclinic    57.607    57.684   150.076",
        "                                   498688    90.000    90.314    90.000",
        "",
        "  14     0.000     1    P       triclinic    57.607    57.684   150.076",
        "                                   498687    89.997    89.686    89.919",
        "",
        "=======================================================================",
        "*Suggested spacegroup number until systematic absences are examined.",
        "...determining orientation angles...",
        ".",
        "",
        "Unit cell parameters and orientation angles",
        "======================================================================",
        "      Integer         a         b         c",
        "Num  residual     alpha      beta     gamma     Rot1     Rot2     Rot3",
        "======================================================================",
        "  1     0.001    57.645    57.645   150.076  -19.868  -13.111  -60.866",
        "                 90.000    90.000    90.000",
        "",
        "  2     0.001    57.645    57.645   150.076   19.868   13.111  119.134",
        "                 90.000    90.000    90.000",
        "",
        "  3     0.001    57.645    57.645   150.076  160.132  -13.111  -60.866",
        "                 90.000    90.000    90.000",
        "",
        "  4     0.001    57.645    57.645   150.076 -160.132   13.111  119.134",
        "                 90.000    90.000    90.000",
        "",
        "  5     0.001    57.645    57.645   150.076  -13.909   19.329   24.448",
        "                 90.000    90.000    90.000",
        "",
        "  6     0.001    57.645    57.645   150.076   13.909  -19.329 -155.552",
        "                 90.000    90.000    90.000",
        "",
        "  7     0.001    57.645    57.645   150.076  166.091   19.329   24.448",
        "                 90.000    90.000    90.000",
        "",
        "  8     0.001    57.645    57.645   150.076 -166.091  -19.329 -155.552",
        "                 90.000    90.000    90.000",
        "",
        "======================================================================",
        "The above table shows symmetry EQUIVALENT crystal orientation angles",
        "for the indexing orientation.  All the solutions are equivalent for",
        "the selected Bravais lattice.  The default selection usually has the",
        "values closest to crystal orientation found in the input .head file or",
        "the one where (|Rot1| + |Rot2| + |Rot3|) is a minimum.",
        "Orientation angles choice 5 selected.",
        "",
        "Crystal listing:",
        "",
        " Unit cell lengths:   57.6451   57.6451  150.0757",
        " Unit cell  angles:   90.0000   90.0000   90.0000",
        " Unit cell  volume: 498695.877",
        "Orientation angles:  -13.9092   19.3288   24.4484",
        "         Mosaicity:     0.300",
        "       Description: unknown",
        "",
        "Spacegroup number: 75",
        "             name: P4",
        "Num. equiv. posns: 4",
        "",
        "dtindex - Wrote header file dtranker_index.head",
        "",
        "INFO - The above indexing solution is ONLY a hypothesis.",
        "       One must confirm the hypothesis by examining symmetry",
        "       of observed intensities as well as possible systematic",
        "       absences.",
        "       One does this by collecting data, processing the data",
        "       and using dtcell and dtscaleaverage.",
        "",
        "INFO - The suggested next step is to run dtrefine with",
        "       dtranker_index.head as the input .head file: ",
        "",
        "       dtrefine dtranker_index.head dtranker_find.ref ...",
        "",
        "",
        "dtrefine:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrefine dtranker_index.head -seq 3 -sigma 5 -rank +CrysAll +DetAll \\",
        "    -rej 0.5 0.5 0.5 -cycles 500 -go -go -reso edge -out \\",
        "    dtranker_refine.head",
        "",
        "     Header of file dtranker_index.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "Reflection centroids will come from a search of image(s),",
        "and not from a reflnlist file.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "... predicting reflections and getting centroids for image 3 ...",
        "",
        "Total number of reflns in m_poReflnlist: 15855",
        "",
        "",
        "Reflection status after merging",
        "-------------------------------",
        "Raw 2D centroids                      1391",
        "Merged                                 465",
        "",
        "Reflns. on   1 consecutive image(s)    465",
        "",
        "",
        "Total number of predicted merged reflections: 465",
        "Refine listing:",
        "",
        "Crystal listing:",
        "",
        " Unit cell lengths:   57.6451   57.6451  150.0757",
        " Unit cell  angles:   90.0000   90.0000   90.0000",
        " Unit cell  volume: 498695.181",
        "Orientation angles:  -13.9092   19.3288   24.4484",
        "         Mosaicity:     0.300",
        "       Description: unknown",
        "",
        "Spacegroup number: 75",
        "             name: P4",
        "Num. equiv. posns: 4",
        "",
        "",
        "CRYSTAL_ Goniometer listing: ",
        "",
        "        Description: Pilatus 3-circle",
        "   Number of values: 3",
        "",
        "         Name    Datum   Current  Units   Vector",
        "=====================================================================",
        "        Omega    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "          Chi    0.000   Unknown    deg   (  0.000,   0.000,   1.000)",
        "          Phi    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "=====================================================================",
        "",
        "Source listing: ",
        "",
        "  Single wavelength: 1",
        "       Polarization: 0.99, 0, 1, 0",
        "          Intensity: 0",
        "",
        "PILT_Detector listing: ",
        "",
        "   Pixel dimensions:     2463, 2527",
        " Nominal size in mm:     423.636, 434.644",
        "        Description:     PILATUS conversion",
        "    Spatial distortion descriptive text:",
        "    Simple pixel scaling spatial distortion",
        "    Simple spatial distortion:",
        "    Center of primary beam: 1252.9, 1288.5",
        "           Pixel size (mm): 0.172, 0.172",
        "         Direction vectors: 1, 0, 0, -1",
        "",
        "PILT_ Goniometer listing: ",
        "",
        "        Description: UNKNOWN",
        "   Number of values: 6",
        "",
        "         Name    Datum   Current  Units   Vector",
        "=====================================================================",
        "         RotZ    0.000   Unknown    deg   (  0.000,   0.000,   1.000)",
        "   RotX/Swing    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "         RotY    0.000   Unknown    deg   (  0.000,   1.000,   0.000)",
        "       TransX    0.000   Unknown     mm   (  1.000,   0.000,   0.000)",
        "       TransY    0.000   Unknown     mm   (  0.000,   1.000,   0.000)",
        "  TransZ/Dist  199.990   Unknown     mm   (  0.000,   0.000,  -1.000)",
        "=====================================================================",
        "DetResolution min:  999.99",
        "DetResolution max:  1.04621",
        "DetResolution edge: 1.23081",
        "",
        "Refine resol min: 99999",
        "Refine resol max: 1e-05",
        "I/sigma cutoff:   5",
        "Sharpness cutoff: 0.2",
        "Rejection limits: 0.5, 0.5, 0.5",
        "Weighting scheme: Unit weights",
        "",
        "........................................",
        "................INFO: converged after 56 cycles.",
        "",
        "",
        "Number of found reflections and best mosaicity vs Trial rotation offset",
        "------------------------------------------------------------------------------",
        "                             % of Max number of observed reflections*",
        " Offset Mosaicity^ reflns 0...10...20...30...40...50...60...70...80...90...100",
        "------------------------------------------------------------------------------",
        "  -0.40      0.74    139 ***********",
        "  -0.35      0.62    177 *****************",
        "  -0.30      0.53    224 ************************",
        "  -0.25      0.43    258 ******************************",
        "  -0.20      0.34    289 **********************************",
        "  -0.15      0.22    330 *****************************************",
        "  -0.10      0.13    370 ***********************************************",
        "  -0.05      0.06    401 ****************************************************",
        "  -0.00      0.06    405 ****************************************************",
        "   0.05      0.11    391 **************************************************",
        "   0.10      0.20    360 *********************************************",
        "   0.15      0.29    329 *****************************************",
        "   0.20      0.39    295 ***********************************",
        "   0.25      0.48    260 ******************************",
        "   0.30      0.57    219 ***********************",
        "   0.35      0.67    187 ******************",
        "   0.40      0.76    146 ************",
        "------------------------------------------------------------------------------",
        " ^Mosaicity value needed to get maximum number of reflns with this offset.",
        " *When mosaicity is set to 0.06 degrees.",
        "",
        "",
        "INFO: Mosaicity set to 0.06 degrees.",
        "INFO: Rotation Offset of 0.00 degrees applied to obtain lowest mosaicity.",
        "",
        "",
        "Refinement results",
        "==================================================================",
        "Crystal",
        "                         a, b, c:    57.7326    57.7326   149.9026",
        "                          Sigmas:     0.0636     0.0636     0.3682",
        "                          Shifts:     0.0875     0.0875    -0.1730",
        "",
        "              alpha, beta, gamma:    90.0000    90.0000    90.0000",
        "                          Sigmas:     0.0000     0.0000     0.0000",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "           Crys Rot1, Rot2, Rot3:   -13.943     19.296     24.470",
        "                          Sigmas:     0.0198     0.0214     0.0136",
        "                          Shifts:    -0.0336    -0.0326     0.0220",
        "",
        "                       Mosaicity:     0.0616",
        "                           Sigma:     0.1000 ",
        "                           Shift:    -0.2384 ",
        "==================================================================",
        "Detector:  0",
        "DetTrans: TransX, TransY, TransZ:    -0.612     -0.266    200.076",
        "                          Sigmas:     0.0230     0.0287     0.2963",
        "                          Shifts:    -0.6118    -0.2656     0.0858",
        "",
        " DetRots: RotZ, RotX/Swing, RotY:    -0.021      0.074     -0.187",
        "                          Sigmas:     0.0119     0.0750     0.0477",
        "                          Shifts:    -0.0210     0.0741    -0.1873",
        "==================================================================",
        "Source",
        "          Wavelength, Rot1, Rot2:     1.000000   0.000      0.000",
        "                          Sigmas:     fixed      fixed      fixed ",
        "                          Shifts:     fixed      fixed      fixed ",
        "==================================================================",
        "Refinement residuals ",
        "rmsResid (A-1) =    0.00084",
        "rmsResid (mm)  =    0.0699  EXCELLENT, less than or equal to 1 pixel (0.172 mm).",
        "rmsResid (Deg) =    0.1693  GOOD, less than 2/3rds the image rot width.",
        "==================================================================",
        "",
        "Reflections in list:         465",
        "Reflections accepted [blue]: 415 (Resolution limits of these: 17.1479 to 1.31894 A)",
        "Reflections rejected [ red]: 31   (|Xobs-Xcalc| >= 0.5 mm), or",
        "                                 (|Yobs-Ycalc| >= 0.5 mm)",
        "Reflections rejected [ red]: 16   (|Rotobs-Rotcalc| >= 0.5 deg)",
        "Reflections ignored [green]: 3   (I/sigI < 5",
        "                                 or outside resolution of 99999 to 1e-05,",
        "                                 or Lorentz factor too large,",
        "                                 or sharpness > 0.2, etc)",
        "",
        "....INFO: converged after 4 cycles.",
        "",
        "",
        "Number of found reflections and best mosaicity vs Trial rotation offset",
        "------------------------------------------------------------------------------",
        "                             % of Max number of observed reflections*",
        " Offset Mosaicity^ reflns 0...10...20...30...40...50...60...70...80...90...100",
        "------------------------------------------------------------------------------",
        "  -0.40      0.74    139 ***********",
        "  -0.35      0.62    177 *****************",
        "  -0.30      0.53    224 ************************",
        "  -0.25      0.43    258 ******************************",
        "  -0.20      0.34    289 **********************************",
        "  -0.15      0.22    330 *****************************************",
        "  -0.10      0.13    370 ***********************************************",
        "  -0.05      0.06    401 ****************************************************",
        "  -0.00      0.06    405 ****************************************************",
        "   0.05      0.11    391 **************************************************",
        "   0.10      0.20    360 *********************************************",
        "   0.15      0.29    329 *****************************************",
        "   0.20      0.39    295 ***********************************",
        "   0.25      0.48    260 ******************************",
        "   0.30      0.57    219 ***********************",
        "   0.35      0.67    187 ******************",
        "   0.40      0.76    146 ************",
        "------------------------------------------------------------------------------",
        " ^Mosaicity value needed to get maximum number of reflns with this offset.",
        " *When mosaicity is set to 0.06 degrees.",
        "",
        "",
        "INFO: Mosaicity set to 0.06 degrees.",
        "INFO: Rotation Offset of 0.00 degrees applied to obtain lowest mosaicity.",
        "",
        "",
        "Refinement results",
        "==================================================================",
        "Crystal",
        "                         a, b, c:    57.7326    57.7326   149.9026",
        "                          Sigmas:     0.0636     0.0636     0.3682",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "              alpha, beta, gamma:    90.0000    90.0000    90.0000",
        "                          Sigmas:     0.0000     0.0000     0.0000",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "           Crys Rot1, Rot2, Rot3:   -13.943     19.296     24.470",
        "                          Sigmas:     0.0198     0.0214     0.0136",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "                       Mosaicity:     0.0616",
        "                           Sigma:     0.1000 ",
        "                           Shift:     0.0000 ",
        "==================================================================",
        "Detector:  0",
        "DetTrans: TransX, TransY, TransZ:    -0.612     -0.266    200.076",
        "                          Sigmas:     0.0230     0.0287     0.2963",
        "                          Shifts:    -0.0000     0.0000     0.0000",
        "",
        " DetRots: RotZ, RotX/Swing, RotY:    -0.021      0.074     -0.187",
        "                          Sigmas:     0.0119     0.0750     0.0477",
        "                          Shifts:     0.0000    -0.0000    -0.0000",
        "==================================================================",
        "Source",
        "          Wavelength, Rot1, Rot2:     1.000000   0.000      0.000",
        "                          Sigmas:     fixed      fixed      fixed ",
        "                          Shifts:     fixed      fixed      fixed ",
        "==================================================================",
        "Refinement residuals ",
        "rmsResid (A-1) =    0.00084",
        "rmsResid (mm)  =    0.0699  EXCELLENT, less than or equal to 1 pixel (0.172 mm).",
        "rmsResid (Deg) =    0.1693  GOOD, less than 2/3rds the image rot width.",
        "==================================================================",
        "",
        "Reflections in list:         465",
        "Reflections accepted [blue]: 415 (Resolution limits of these: 17.1479 to 1.31894 A)",
        "Reflections rejected [ red]: 31   (|Xobs-Xcalc| >= 0.5 mm), or",
        "                                 (|Yobs-Ycalc| >= 0.5 mm)",
        "Reflections rejected [ red]: 16   (|Rotobs-Rotcalc| >= 0.5 deg)",
        "Reflections ignored [green]: 3   (I/sigI < 5",
        "                                 or outside resolution of 99999 to 1e-05,",
        "                                 or Lorentz factor too large,",
        "                                 or sharpness > 0.2, etc)",
        "",
        "",
        "Listing for all refinement iterations",
        "-----------------------------------------------------------------",
        "Iter  Twin   Volume    Rms     Rms  Total Ignored Rejects Rejects",
        "                      (MM)   (DEG)  refln   refln    (MM)   (DEG)",
        "-----------------------------------------------------------------",
        "  58 01/01   499634 0.0699  0.1693    415       3      31      16",
        "   6 01/01   499634 0.0699  0.1693    415       3      31      16",
        "-----------------------------------------------------------------",
        "",
        "",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "...predicting reflections for image sequence number 3",
        "Number of reflections written in 'dtrefpredict.ref': 3489",
        "",
        "Processing image '/tmp/image_0003.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Resolution rings found.",
        "--------------------------------------------------",
        " Peak 2Theta 2Theta 2Theta    Reso    Reso    Peak ",
        "   #     Mid  Start    End   Start     End Percent ",
        "--------------------------------------------------",
        "    0  15.85  14.35  17.35  4.0032  3.3150    0.82",
        "--------------------------------------------------",
        "",
        "Predicting reflections using refined values and performing quick 2D integration...",
        "",
        "Reflection centroids will come from a search of image(s),",
        "and not from a reflnlist file.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "... predicting reflections and getting centroids for image 3 ...",
        "",
        "Total number of predicted (non-merged) reflections: 3193",
        "",
        "",
        "Deleted 2333 reflections not found on images(s).",
        "",
        "",
        "Found 0 overlapped reflections.",
        "",
        "INFO: 42 reflns deleted out of 860 that might be in ice rings.",
        "",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 2.65    207    0   30.8   34.94      5.0      0.07",
        "  2.65 - 2.10    220    0   15.2   23.99      4.6      0.08",
        "  2.10 - 1.84    153    0   10.5   16.36      3.7      0.09",
        "  1.84 - 1.67    103    0    7.4   10.98      3.2      0.13",
        "  1.67 - 1.55     77    0    7.1    7.85      2.6      0.12",
        "  1.55 - 1.46     32    0    6.3    6.97      2.6      0.10",
        "  1.46 - 1.39     11    0    6.8    5.49      2.3      0.29",
        "  1.39 - 1.33     10    0    4.7    4.38      2.2      0.09",
        "  1.33 - 1.27      4    0    5.1    4.32      1.9      0.20",
        "  1.27 - 1.23      1    0    4.6    2.80      1.6      0.06",
        "-----------------------------------------------------------",
        " 99.99 - 1.23    818    0   15.9   20.90      4.0      0.09",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    18 -  >  20    174    0  1.80   31.55      4.9      0.07",
        "    16 -     18     32    0  1.44   25.15      4.3      0.09",
        "    14 -     16     56    0  1.73   26.35      4.6      0.08",
        "    12 -     14     40    0  1.83   24.94      4.1      0.08",
        "    10 -     12     60    0  1.54   22.59      3.9      0.08",
        "     8 -     10     82    0  1.59   19.88      3.9      0.07",
        "     6 -      8     90    0  1.45   16.63      3.7      0.11",
        "     4 -      6    126    0  1.33   14.57      3.5      0.11",
        "     2 -      4    135    0  1.26   12.49      3.4      0.14",
        " <   0 -      2     23    0  1.33   14.04      3.6      0.10",
        "------------------------------------------------------------",
        " <   0 -  >  20    818    0  1.26   20.90      4.0      0.09",
        "------------------------------------------------------------",
        "dtrefine - INFO wrote header file: dtranker_refine.head",
        "",
        "INFO - The suggested next step is to run dtrefine again:",
        "         dtrefine dtranker_refine.head -seq ...",
        "      -or-",
        "       Run dtpredict to check that the results predict the",
        "       first few images in each scan correctly: ",
        "         dtpredict dtranker_refine.head -seq ...",
        "      -or-",
        "       Run dtmultistrategy or dtintegrate with ",
        "       dtranker_refine.head as the input .head file:",
        "",
        "        dtmultistrategy dtranker_refine.head ...",
        "        dtintegrate dtranker_refine.head ...",
        "",
        "",
        "dtrank:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrank dtranker_refine.head -param 2 6 0 -param 11 6 0 -param 12 1 0",
        "Command line:",
        " dtranker_refine.head -param 2 6 0 -param 11 6 0 -param 12 1 0 ",
        "Processing Option:  dtranker_refine.head ",
        "     Header of file dtranker_refine.head successfully read.",
        "Processing Option: -param 2 6 0 ",
        "Processing Option: -param 11 6 0 ",
        "Processing Option: -param 12 1 0 ",
        "",
        "-----------------------------------------------------------------------",
        "Ranking procedure awards/penalizes points based on a set of rules.",
        "A description of each rule that is used is given below.",
        "",
        "Rule  1: Spot count in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell, a [10] point bonus is awarded.",
        "However, if there are less than 5 reflections found in the highest",
        "resolution shell, [0] points are subtracted from the award.",
        "",
        "Rule  2: I/Sigma in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule  3: Spot sharpness",
        "   Rule  3a: Spot asymmetry",
        "   A penalty of [10] times the average spot profile asymmetry is made.",
        "   Rule  3b: Spot intensity per pixel",
        "   An award of [10] times <Ipix>/[300] is made, where Ipix is an average",
        "   peak intensity per pixel",
        "   Rule  3c: Spot anisotropy",
        "   A penalty of [10] times the average spot anisotropy is made.",
        "   Rule  3d: Spot amorphousness",
        "   A penalty of [100] times the average spot amorphousness is made.",
        "",
        "Rule  4: Strong ice rings",
        "For each strong (peak to background ratio > [4.00] percent) ice ring, [10]",
        "points are deducted.",
        "",
        "Rule  5: Diffuse ice rings",
        "For each diffuse (peak to background ratio > [1.00] percent) ice ring, [5]",
        "points are deducted.",
        "",
        "Rule  6: Percentage of spots indexed",
        "If more than [20] reflections are indexed,",
        "[100] * <Indexed reflections / Used Reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are indexed,",
        "[0] points are subtracted.",
        "",
        "Rule  7: RMS residual after refinement",
        "If a sample fails to refine, or has an RMS residual > [0.30] mm,",
        "9 points are deducted. Otherwise, a penalty of",
        "[30] * <RMS-MM-Residual> is made.",
        "",
        "Rule  8: Mosaicity",
        "If a sample fails to refine, or has a mosaicity > [1.00] deg,",
        "70 points are deducted. Otherwise, a penalty of",
        "[70] * Mosaicity is made.",
        "",
        "Rule  9: Percentage of spots refined",
        "If more than [20] reflections are refined,",
        "[100] * <Refined reflections / Predicted reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are refined,",
        "[0] points are subtracted.",
        "",
        "Rule 10: Spot count in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more predicted reflections found in a shell, a [10] point",
        "bonus is awarded. However, if there are less than 5 predicted reflections",
        "found in the highest resolution shell, [0] points are subtracted from",
        "the award.",
        "",
        "Rule 11: I/Sigma in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule 12: Strategy total rotation width",
        "A penalty of [0] times RotWidth/[360.00] is applied, where RotWidth is the",
        "total rotation width suggested by strategy.",
        "",
        "Each number in braces is a modifiable parameter which can ",
        "be altered by using the -param command.  If for example",
        "you wish to modify the first parameter in Rule (3) to be 100, ",
        "you specify -param 3 1 100.",
        "",
        "-----------------------------------------------------------------------",
        "--------------------------------------------------------------------------------",
        "Rule Category                                                  Pts Cumul    Max",
        "--------------------------------------------------------------------------------",
        "  3  Penalty for peak profile asymmetry of 0.09                 -1    -1    -10",
        "  3  Award for peak intensity per pixel of 43.45                 1     0     10",
        "  3  Penalty for spot shape anisotropy of 0.09                   0     0    -10",
        "  3  Penalty for spot shape amorphousness of 0.000               0     0   -100",
        "  6  Indexed 44% of all reflections used in indexing            43    43    100",
        "  7  Penalty for RMS residual value of 0.070                    -2    41     -8",
        "  8  Penalty for Mosaicity value of 0.06                        -4    37    -70",
        "  9  Refined 89% of all reflections used in refinement          89   126    100",
        " 10  Predicted&Found: >=5 reflns in HRS #8 (1.39-1.33)\u00C5         10   136     10",
        " 10  Predicted&Found: >=5 reflns in HRS #7 (1.46-1.39)\u00C5         10   146     10",
        " 10  Predicted&Found: >=5 reflns in HRS #6 (1.55-1.46)\u00C5         10   156     10",
        " 10  Predicted&Found: >=5 reflns in HRS #5 (1.67-1.55)\u00C5         10   166     10",
        " 10  Predicted&Found: >=5 reflns in HRS #4 (1.84-1.67)\u00C5         10   176     10",
        " 11  Predicted&Found: I/S=4.7 in HRS #8 (1.39-1.33)\u00C5             1   177     20",
        " 11  Predicted&Found: I/S=6.8 in HRS #7 (1.46-1.39)\u00C5             2   179     20",
        " 11  Predicted&Found: I/S=6.3 in HRS #6 (1.55-1.46)\u00C5             2   181     20",
        " 11  Predicted&Found: I/S=7.1 in HRS #5 (1.67-1.55)\u00C5             2   183     20",
        " 11  Predicted&Found: I/S=7.4 in HRS #4 (1.84-1.67)\u00C5             2   185     20",
        "--------------------------------------------------------------------------------",
        "     Cumulative for /tmp/image_0003.cbf                              185",
        "",
        "",
        "",
        "     Header of file dtranker.head successfully read."
      };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\r\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /*
   * Resolution stats row with I/sig average less than 2, and negative score.
   */
  private static byte[] createDtrankerOutput02() {
    /* @formatter:off */
    String[] output =
      new String[] {
        "",
        "dtranker:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "",
        "Command line:",
        "dtranker image=image_0003.cbf",
        "",
        "",
        "dtextractheader:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "     Header of file image_0003.cbf successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "",
        "dtfind:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dfind dtranker_input.head -reso edge -seq 3 3 -sigma 3 -min 50 -2D \\",
        "    -filter 6 -window 0 0 -rank -ref dtranker_find.ref -out \\",
        "    dtranker_find.head",
        "",
        "     Header of file dtranker_input.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "",
        "     File image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "",
        "Command line string: >>-reso<<",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Command line string: >>-seq<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-min<<",
        "Command line string: >>-2D<<",
        "Command line string: >>-filter<<",
        "Command line string: >>-window<<",
        "Command line string: >>-rank<<",
        "Command line string: >>-ref<<",
        "Command line string: >>-out<<",
        "",
        "Resolution limits of an image are 999.99 to 1.04645",
        "Resolution limits of peak search are 999.99 to 1.23108",
        "",
        "",
        "dtfind: 2D method  used",
        "...reading image /tmp/image_0003.cbf...",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "Processing image '/tmp/image_0003.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Resolution rings found.",
        "--------------------------------------------------",
        " Peak 2Theta 2Theta 2Theta    Reso    Reso    Peak ",
        "   #     Mid  Start    End   Start     End Percent ",
        "--------------------------------------------------",
        "    0  15.75  14.25  17.25  4.0311  3.3341    0.58",
        "--------------------------------------------------",
        "Find object listing:",
        "     Sigma: 3",
        "Resolution: 999.99 to 1.23108",
        "   Minimum: 50",
        "Circle lim: 1231, 1263, 0, 1763",
        "  Rect lim: 24, 25, 2439, 2502",
        "Spot wind.: 0, 0",
        "Peak filt.: 6",
        "Back. tile: 153, 157",
        " Seq. num.: 3, 3",
        "Image dim.: 2463, 2527",
        "   3D dump: 0",
        "251 preliminary spots found in 2D search with rotation angle 1.25 degs.",
        "",
        "INFO: 31 reflns deleted out of 194 that might be in ice rings.",
        "",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 2.65     64    0   37.8   36.51      5.6      0.06",
        "  2.65 - 2.11     78    0   21.4   25.32      5.0      0.07",
        "  2.11 - 1.84     19    0   17.1   18.42      4.2      0.08",
        "  1.84 - 1.67      2    0   19.5   12.61      3.6      0.09",
        "  1.67 - 1.55      0    0    0.0    0.00      0.0      0.00",
        "  1.55 - 1.46      0    0    0.0    0.00      0.0      0.00",
        "  1.46 - 1.39      0    0    0.0    0.00      0.0      0.00",
        "  1.39 - 1.33      0    0    0.0    0.00      0.0      0.00",
        "  1.33 - 1.28      0    0    0.0    0.00      0.0      0.00",
        "  1.28 - 1.23      0    0    0.0    0.00      0.0      0.00",
        "-----------------------------------------------------------",
        " 99.99 - 1.23    163    0   27.3   28.75      5.1      0.07",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    27 -  >  30     47    0  2.12   34.57      5.3      0.06",
        "    24 -     27      9    0  2.16   31.26      5.3      0.06",
        "    21 -     24     11    0  1.80   23.93      4.9      0.08",
        "    18 -     21     18    0  1.89   26.10      5.2      0.06",
        "    15 -     18     22    0  1.88   25.64      5.1      0.08",
        "    12 -     15     30    0  1.95   26.49      5.0      0.07",
        "     9 -     12     17    0  1.83   26.96      5.0      0.07",
        "     6 -      9      7    0  2.03   24.93      5.0      0.06",
        "     3 -      6      2    0  2.39   28.12      5.2      0.05",
        " <   0 -      3      0    0  0.00    0.00      0.0      0.00",
        "------------------------------------------------------------",
        " <   0 -  >  30    163    0  1.80   28.75      5.1      0.07",
        "------------------------------------------------------------",
        "",
        "dtfind: There were 163 spots found.",
        "There were 194 preliminary spots of which 0 were marked as saturated,",
        " or  0.00% of them.",
        "Number of reflections written in 'dtranker_find.ref': 163",
        "dtfind: Spots written to dtranker_find.ref",
        "dtfind - Wrote header file dtranker_find.head",
        "",
        "INFO - The suggested next step is to run dtindex with",
        "       dtranker_find.head as the input .head file and",
        "       dtranker_find.ref as the input .ref file:",
        "",
        "       dtindex dtranker_find.head dtranker_find.ref ...",
        "",
        "",
        "dtindex:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtindex dtranker_find.head dtranker_find.ref -maxresid 3.0 -sigma 5 \\",
        "    -out dtranker_index.head",
        "",
        "     Header of file dtranker_find.head successfully read.",
        "Reflection list: dtranker_find.ref",
        "Creflnlist::nRead with filename: dtranker_find.ref",
        "INFO in Creflnlist::nRead, EOF after 163 reflections read in",
        "                                    (163 total now in list).",
        "Command line string: >>-maxresid<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-out<<",
        "INFO: deleted 0 reflns outside of resolution bounds.",
        "      This leaves 163 reflns for indexing.",
        "",
        "INFO: 51 reflns deleted out of 163 that might be in ice rings.",
        "",
        "Max cell length allowed for reciprocal lattice vectors: 307.677",
        "",
        "",
        "Method:     1D FFT with DPS algorithm",
        "Out header: dtranker_index.head",
        "Max cell:   307.677",
        "Num vecs:   1000",
        "Spacegroup: 0",
        "Verbose:    1",
        "",
        "Performing 1D FFT indexing (not cell reduction) with the DPS algorithm...see",
        "  Steller, Bolotovsky, & Rossmann (1997) J. Appl. Cryst. 30, 1036-1040.",
        "",
        "Max cell is:         307.677",
        "Number of reflections/vectors used: 100",
        ".....................................",
        "...refining best 30 directions and lengths...",
        "..............................  done.",
        "Number of vectors used for integer residual calculation: 100",
        "",
        "     a      b      c  alpha   beta  gamma   Volume  Remarks #Indexed %Residual",
        "==============================================================================",
        "128.67  81.40 281.47  43.37 123.51 161.50   498709     Okay       59     0.064",
        "128.67  81.40 229.21  57.48 109.94 161.50   498692     Okay       59     0.064",
        "281.44  57.71  81.41  45.03  43.37  34.92   498781     Okay       58     0.062",
        " 57.71  81.41 281.44  43.37  34.92  45.03   498781     Okay       58     0.062",
        " 81.41 281.44  57.71  34.92  45.03  43.37   498781     Okay       58     0.062",
        "229.21  57.71  81.40  45.04  57.48  40.89   498692     Okay       59     0.064",
        " 57.71  81.40 229.21  57.48  40.89  45.04   498692     Okay       59     0.064",
        " 81.40 229.21  57.71  40.89  45.04  57.48   498692     Okay       59     0.064",
        "281.45  81.66  57.71  44.84  34.90  64.28   498105     Okay       59     0.063",
        " 81.66  57.71 281.45  34.90  64.28  44.84   498104     Okay       59     0.063",
        " 57.71 281.45  81.66  64.28  44.84  34.90   498104     Okay       59     0.063",
        " 81.67  57.71 229.18  40.83  57.83  44.83   497826     Okay       57     0.062",
        " 57.71 229.18  81.67  57.83  44.83  40.83   497834     Okay       57     0.062",
        "229.18  81.67  57.71  44.83  40.83  57.83   497829     Okay       57     0.062",
        "",
        "",
        "Executing beam refinement with",
        "beam search radius, acceptable shift radius: 10 9 ...",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Original (input header) beam center: [1252.0 1288.0]",
        "",
        "First Pass.  Search dim0 in [1242 1262] dim1 in [1278 1298]",
        ".............................................",
        "Second Pass.  Search dim0 in [1252 1254] dim1 in [1287 1289]",
        "Calculated pre-reduced cell solution is in agreement with detector beam center!",
        "Original (input header) beam center: [1252.0 1288.0]",
        "New      (calculated)   beam center: [1252.9 1288.5]",
        "Header updated to reflect beam center change.",
        "INFO: Restart with adjusted beam center.",
        "Max cell length allowed for reciprocal lattice vectors: 307.677",
        "",
        "",
        "Method:     1D FFT with DPS algorithm",
        "Out header: dtranker_index.head",
        "Max cell:   307.677",
        "Num vecs:   1000",
        "Spacegroup: 0",
        "Verbose:    1",
        "",
        "Performing 1D FFT indexing (not cell reduction) with the DPS algorithm...see",
        "  Steller, Bolotovsky, & Rossmann (1997) J. Appl. Cryst. 30, 1036-1040.",
        "",
        "Max cell is:         307.677",
        "Number of reflections/vectors used: 100",
        ".....................................",
        "...refining best 30 directions and lengths...",
        "..............................  done.",
        "Number of vectors used for integer residual calculation: 100",
        "",
        "     a      b      c  alpha   beta  gamma   Volume  Remarks #Indexed %Residual",
        "==============================================================================",
        " 57.69  81.59 281.59  64.17 101.75  44.92   498636     Okay       47     0.035",
        " 81.59 281.59  57.69 101.75  44.92  64.17   498636     Okay       47     0.035",
        "281.62  57.69  81.59  44.92  64.17 101.75   498727     Okay       50     0.037",
        "129.03  81.59 281.59  64.17  79.37  18.40   498636     Okay       47     0.035",
        "128.87  81.46 281.60  43.55 123.32 161.55   498542     Okay       49     0.036",
        " 81.46 281.60  57.69 101.74 134.99  43.55   498542     Okay       49     0.036",
        "281.60  57.69  81.46 134.99  43.55 101.74   498542     Okay       49     0.036",
        " 57.69  81.46 281.60  43.55 101.74 134.99   498542     Okay       49     0.036",
        "281.60 128.87  57.69  26.55 101.74 123.32   498540     Okay       49     0.036",
        "128.87  57.69 281.60 101.74 123.32  26.55   498542     Okay       49     0.036",
        " 57.68 129.02 281.58  79.37 101.75  26.52   498687     Okay       51     0.038",
        "",
        "",
        "Least square fit to lattice characters...see",
        "  Andrews & Bernstein (1988) Acta Cryst. A44, 1009-1018 and",
        "  Paciorek & Bonin (1992) J. Appl. Cryst. 25, 632-637.",
        "............................................................",
        "............................................................",
        "done.",
        "",
        "Least-squares fit of reduced primitive cell to 44 lattice characters",
        "sorted on decreasing (highest to lowest) symmetry.",
        "Only solutions with residuals <=   3.0 are listed.",
        "=======================================================================",
        " Soln  LeastSq Spgrp Cent    Bravais type         a         b         c",
        "  num residual  num* type     Cell volume     alpha      beta     gamma",
        "=======================================================================",
        "   7     0.415    75    P      tetragonal    57.645    57.645   150.076",
        "                                   498696    90.000    90.000    90.000",
        "",
        "   9     0.413    21    C    orthorhombic    81.465    81.580   150.076",
        "                                   997391    90.000    90.000    90.000",
        "",
        "  11     0.414    16    P    orthorhombic    57.607    57.684   150.076",
        "                                   498695    90.000    90.000    90.000",
        "",
        "  12     0.290     5    C      monoclinic    81.580    81.465   150.076",
        "                                   997383    90.000    90.224    90.000",
        "",
        "  13     0.041     3    P      monoclinic    57.607    57.684   150.076",
        "                                   498688    90.000    90.314    90.000",
        "",
        "  14     0.000     1    P       triclinic    57.607    57.684   150.076",
        "                                   498687    89.997    89.686    89.919",
        "",
        "=======================================================================",
        "*Suggested spacegroup number until systematic absences are examined.",
        "...determining orientation angles...",
        ".",
        "",
        "Unit cell parameters and orientation angles",
        "======================================================================",
        "      Integer         a         b         c",
        "Num  residual     alpha      beta     gamma     Rot1     Rot2     Rot3",
        "======================================================================",
        "  1     0.001    57.645    57.645   150.076  -19.868  -13.111  -60.866",
        "                 90.000    90.000    90.000",
        "",
        "  2     0.001    57.645    57.645   150.076   19.868   13.111  119.134",
        "                 90.000    90.000    90.000",
        "",
        "  3     0.001    57.645    57.645   150.076  160.132  -13.111  -60.866",
        "                 90.000    90.000    90.000",
        "",
        "  4     0.001    57.645    57.645   150.076 -160.132   13.111  119.134",
        "                 90.000    90.000    90.000",
        "",
        "  5     0.001    57.645    57.645   150.076  -13.909   19.329   24.448",
        "                 90.000    90.000    90.000",
        "",
        "  6     0.001    57.645    57.645   150.076   13.909  -19.329 -155.552",
        "                 90.000    90.000    90.000",
        "",
        "  7     0.001    57.645    57.645   150.076  166.091   19.329   24.448",
        "                 90.000    90.000    90.000",
        "",
        "  8     0.001    57.645    57.645   150.076 -166.091  -19.329 -155.552",
        "                 90.000    90.000    90.000",
        "",
        "======================================================================",
        "The above table shows symmetry EQUIVALENT crystal orientation angles",
        "for the indexing orientation.  All the solutions are equivalent for",
        "the selected Bravais lattice.  The default selection usually has the",
        "values closest to crystal orientation found in the input .head file or",
        "the one where (|Rot1| + |Rot2| + |Rot3|) is a minimum.",
        "Orientation angles choice 5 selected.",
        "",
        "Crystal listing:",
        "",
        " Unit cell lengths:   57.6451   57.6451  150.0757",
        " Unit cell  angles:   90.0000   90.0000   90.0000",
        " Unit cell  volume: 498695.877",
        "Orientation angles:  -13.9092   19.3288   24.4484",
        "         Mosaicity:     0.300",
        "       Description: unknown",
        "",
        "Spacegroup number: 75",
        "             name: P4",
        "Num. equiv. posns: 4",
        "",
        "dtindex - Wrote header file dtranker_index.head",
        "",
        "INFO - The above indexing solution is ONLY a hypothesis.",
        "       One must confirm the hypothesis by examining symmetry",
        "       of observed intensities as well as possible systematic",
        "       absences.",
        "       One does this by collecting data, processing the data",
        "       and using dtcell and dtscaleaverage.",
        "",
        "INFO - The suggested next step is to run dtrefine with",
        "       dtranker_index.head as the input .head file: ",
        "",
        "       dtrefine dtranker_index.head dtranker_find.ref ...",
        "",
        "",
        "dtrefine:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrefine dtranker_index.head -seq 3 -sigma 5 -rank +CrysAll +DetAll \\",
        "    -rej 0.5 0.5 0.5 -cycles 500 -go -go -reso edge -out \\",
        "    dtranker_refine.head",
        "",
        "     Header of file dtranker_index.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "Reflection centroids will come from a search of image(s),",
        "and not from a reflnlist file.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "... predicting reflections and getting centroids for image 3 ...",
        "",
        "Total number of reflns in m_poReflnlist: 15855",
        "",
        "",
        "Reflection status after merging",
        "-------------------------------",
        "Raw 2D centroids                      1391",
        "Merged                                 465",
        "",
        "Reflns. on   1 consecutive image(s)    465",
        "",
        "",
        "Total number of predicted merged reflections: 465",
        "Refine listing:",
        "",
        "Crystal listing:",
        "",
        " Unit cell lengths:   57.6451   57.6451  150.0757",
        " Unit cell  angles:   90.0000   90.0000   90.0000",
        " Unit cell  volume: 498695.181",
        "Orientation angles:  -13.9092   19.3288   24.4484",
        "         Mosaicity:     0.300",
        "       Description: unknown",
        "",
        "Spacegroup number: 75",
        "             name: P4",
        "Num. equiv. posns: 4",
        "",
        "",
        "CRYSTAL_ Goniometer listing: ",
        "",
        "        Description: Pilatus 3-circle",
        "   Number of values: 3",
        "",
        "         Name    Datum   Current  Units   Vector",
        "=====================================================================",
        "        Omega    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "          Chi    0.000   Unknown    deg   (  0.000,   0.000,   1.000)",
        "          Phi    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "=====================================================================",
        "",
        "Source listing: ",
        "",
        "  Single wavelength: 1",
        "       Polarization: 0.99, 0, 1, 0",
        "          Intensity: 0",
        "",
        "PILT_Detector listing: ",
        "",
        "   Pixel dimensions:     2463, 2527",
        " Nominal size in mm:     423.636, 434.644",
        "        Description:     PILATUS conversion",
        "    Spatial distortion descriptive text:",
        "    Simple pixel scaling spatial distortion",
        "    Simple spatial distortion:",
        "    Center of primary beam: 1252.9, 1288.5",
        "           Pixel size (mm): 0.172, 0.172",
        "         Direction vectors: 1, 0, 0, -1",
        "",
        "PILT_ Goniometer listing: ",
        "",
        "        Description: UNKNOWN",
        "   Number of values: 6",
        "",
        "         Name    Datum   Current  Units   Vector",
        "=====================================================================",
        "         RotZ    0.000   Unknown    deg   (  0.000,   0.000,   1.000)",
        "   RotX/Swing    0.000   Unknown    deg   (  1.000,   0.000,   0.000)",
        "         RotY    0.000   Unknown    deg   (  0.000,   1.000,   0.000)",
        "       TransX    0.000   Unknown     mm   (  1.000,   0.000,   0.000)",
        "       TransY    0.000   Unknown     mm   (  0.000,   1.000,   0.000)",
        "  TransZ/Dist  199.990   Unknown     mm   (  0.000,   0.000,  -1.000)",
        "=====================================================================",
        "DetResolution min:  999.99",
        "DetResolution max:  1.04621",
        "DetResolution edge: 1.23081",
        "",
        "Refine resol min: 99999",
        "Refine resol max: 1e-05",
        "I/sigma cutoff:   5",
        "Sharpness cutoff: 0.2",
        "Rejection limits: 0.5, 0.5, 0.5",
        "Weighting scheme: Unit weights",
        "",
        "........................................",
        "................INFO: converged after 56 cycles.",
        "",
        "",
        "Number of found reflections and best mosaicity vs Trial rotation offset",
        "------------------------------------------------------------------------------",
        "                             % of Max number of observed reflections*",
        " Offset Mosaicity^ reflns 0...10...20...30...40...50...60...70...80...90...100",
        "------------------------------------------------------------------------------",
        "  -0.40      0.74    139 ***********",
        "  -0.35      0.62    177 *****************",
        "  -0.30      0.53    224 ************************",
        "  -0.25      0.43    258 ******************************",
        "  -0.20      0.34    289 **********************************",
        "  -0.15      0.22    330 *****************************************",
        "  -0.10      0.13    370 ***********************************************",
        "  -0.05      0.06    401 ****************************************************",
        "  -0.00      0.06    405 ****************************************************",
        "   0.05      0.11    391 **************************************************",
        "   0.10      0.20    360 *********************************************",
        "   0.15      0.29    329 *****************************************",
        "   0.20      0.39    295 ***********************************",
        "   0.25      0.48    260 ******************************",
        "   0.30      0.57    219 ***********************",
        "   0.35      0.67    187 ******************",
        "   0.40      0.76    146 ************",
        "------------------------------------------------------------------------------",
        " ^Mosaicity value needed to get maximum number of reflns with this offset.",
        " *When mosaicity is set to 0.06 degrees.",
        "",
        "",
        "INFO: Mosaicity set to 0.06 degrees.",
        "INFO: Rotation Offset of 0.00 degrees applied to obtain lowest mosaicity.",
        "",
        "",
        "Refinement results",
        "==================================================================",
        "Crystal",
        "                         a, b, c:    57.7326    57.7326   149.9026",
        "                          Sigmas:     0.0636     0.0636     0.3682",
        "                          Shifts:     0.0875     0.0875    -0.1730",
        "",
        "              alpha, beta, gamma:    90.0000    90.0000    90.0000",
        "                          Sigmas:     0.0000     0.0000     0.0000",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "           Crys Rot1, Rot2, Rot3:   -13.943     19.296     24.470",
        "                          Sigmas:     0.0198     0.0214     0.0136",
        "                          Shifts:    -0.0336    -0.0326     0.0220",
        "",
        "                       Mosaicity:     0.0616",
        "                           Sigma:     0.1000 ",
        "                           Shift:    -0.2384 ",
        "==================================================================",
        "Detector:  0",
        "DetTrans: TransX, TransY, TransZ:    -0.612     -0.266    200.076",
        "                          Sigmas:     0.0230     0.0287     0.2963",
        "                          Shifts:    -0.6118    -0.2656     0.0858",
        "",
        " DetRots: RotZ, RotX/Swing, RotY:    -0.021      0.074     -0.187",
        "                          Sigmas:     0.0119     0.0750     0.0477",
        "                          Shifts:    -0.0210     0.0741    -0.1873",
        "==================================================================",
        "Source",
        "          Wavelength, Rot1, Rot2:     1.000000   0.000      0.000",
        "                          Sigmas:     fixed      fixed      fixed ",
        "                          Shifts:     fixed      fixed      fixed ",
        "==================================================================",
        "Refinement residuals ",
        "rmsResid (A-1) =    0.00084",
        "rmsResid (mm)  =    0.0699  EXCELLENT, less than or equal to 1 pixel (0.172 mm).",
        "rmsResid (Deg) =    0.1693  GOOD, less than 2/3rds the image rot width.",
        "==================================================================",
        "",
        "Reflections in list:         465",
        "Reflections accepted [blue]: 415 (Resolution limits of these: 17.1479 to 1.31894 A)",
        "Reflections rejected [ red]: 31   (|Xobs-Xcalc| >= 0.5 mm), or",
        "                                 (|Yobs-Ycalc| >= 0.5 mm)",
        "Reflections rejected [ red]: 16   (|Rotobs-Rotcalc| >= 0.5 deg)",
        "Reflections ignored [green]: 3   (I/sigI < 5",
        "                                 or outside resolution of 99999 to 1e-05,",
        "                                 or Lorentz factor too large,",
        "                                 or sharpness > 0.2, etc)",
        "",
        "....INFO: converged after 4 cycles.",
        "",
        "",
        "Number of found reflections and best mosaicity vs Trial rotation offset",
        "------------------------------------------------------------------------------",
        "                             % of Max number of observed reflections*",
        " Offset Mosaicity^ reflns 0...10...20...30...40...50...60...70...80...90...100",
        "------------------------------------------------------------------------------",
        "  -0.40      0.74    139 ***********",
        "  -0.35      0.62    177 *****************",
        "  -0.30      0.53    224 ************************",
        "  -0.25      0.43    258 ******************************",
        "  -0.20      0.34    289 **********************************",
        "  -0.15      0.22    330 *****************************************",
        "  -0.10      0.13    370 ***********************************************",
        "  -0.05      0.06    401 ****************************************************",
        "  -0.00      0.06    405 ****************************************************",
        "   0.05      0.11    391 **************************************************",
        "   0.10      0.20    360 *********************************************",
        "   0.15      0.29    329 *****************************************",
        "   0.20      0.39    295 ***********************************",
        "   0.25      0.48    260 ******************************",
        "   0.30      0.57    219 ***********************",
        "   0.35      0.67    187 ******************",
        "   0.40      0.76    146 ************",
        "------------------------------------------------------------------------------",
        " ^Mosaicity value needed to get maximum number of reflns with this offset.",
        " *When mosaicity is set to 0.06 degrees.",
        "",
        "",
        "INFO: Mosaicity set to 0.06 degrees.",
        "INFO: Rotation Offset of 0.00 degrees applied to obtain lowest mosaicity.",
        "",
        "",
        "Refinement results",
        "==================================================================",
        "Crystal",
        "                         a, b, c:    57.7326    57.7326   149.9026",
        "                          Sigmas:     0.0636     0.0636     0.3682",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "              alpha, beta, gamma:    90.0000    90.0000    90.0000",
        "                          Sigmas:     0.0000     0.0000     0.0000",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "           Crys Rot1, Rot2, Rot3:   -13.943     19.296     24.470",
        "                          Sigmas:     0.0198     0.0214     0.0136",
        "                          Shifts:     0.0000     0.0000     0.0000",
        "",
        "                       Mosaicity:     0.0616",
        "                           Sigma:     0.1000 ",
        "                           Shift:     0.0000 ",
        "==================================================================",
        "Detector:  0",
        "DetTrans: TransX, TransY, TransZ:    -0.612     -0.266    200.076",
        "                          Sigmas:     0.0230     0.0287     0.2963",
        "                          Shifts:    -0.0000     0.0000     0.0000",
        "",
        " DetRots: RotZ, RotX/Swing, RotY:    -0.021      0.074     -0.187",
        "                          Sigmas:     0.0119     0.0750     0.0477",
        "                          Shifts:     0.0000    -0.0000    -0.0000",
        "==================================================================",
        "Source",
        "          Wavelength, Rot1, Rot2:     1.000000   0.000      0.000",
        "                          Sigmas:     fixed      fixed      fixed ",
        "                          Shifts:     fixed      fixed      fixed ",
        "==================================================================",
        "Refinement residuals ",
        "rmsResid (A-1) =    0.00084",
        "rmsResid (mm)  =    0.0699  EXCELLENT, less than or equal to 1 pixel (0.172 mm).",
        "rmsResid (Deg) =    0.1693  GOOD, less than 2/3rds the image rot width.",
        "==================================================================",
        "",
        "Reflections in list:         465",
        "Reflections accepted [blue]: 415 (Resolution limits of these: 17.1479 to 1.31894 A)",
        "Reflections rejected [ red]: 31   (|Xobs-Xcalc| >= 0.5 mm), or",
        "                                 (|Yobs-Ycalc| >= 0.5 mm)",
        "Reflections rejected [ red]: 16   (|Rotobs-Rotcalc| >= 0.5 deg)",
        "Reflections ignored [green]: 3   (I/sigI < 5",
        "                                 or outside resolution of 99999 to 1e-05,",
        "                                 or Lorentz factor too large,",
        "                                 or sharpness > 0.2, etc)",
        "",
        "",
        "Listing for all refinement iterations",
        "-----------------------------------------------------------------",
        "Iter  Twin   Volume    Rms     Rms  Total Ignored Rejects Rejects",
        "                      (MM)   (DEG)  refln   refln    (MM)   (DEG)",
        "-----------------------------------------------------------------",
        "  58 01/01   499634 0.0699  0.1693    415       3      31      16",
        "   6 01/01   499634 0.0699  0.1693    415       3      31      16",
        "-----------------------------------------------------------------",
        "",
        "",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "...predicting reflections for image sequence number 3",
        "Number of reflections written in 'dtrefpredict.ref': 3489",
        "",
        "Processing image '/tmp/image_0003.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/image_0003.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Resolution rings found.",
        "--------------------------------------------------",
        " Peak 2Theta 2Theta 2Theta    Reso    Reso    Peak ",
        "   #     Mid  Start    End   Start     End Percent ",
        "--------------------------------------------------",
        "    0  15.85  14.35  17.35  4.0032  3.3150    0.82",
        "--------------------------------------------------",
        "",
        "Predicting reflections using refined values and performing quick 2D integration...",
        "",
        "Reflection centroids will come from a search of image(s),",
        "and not from a reflnlist file.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: image_0003.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "     File /tmp/image_0003.cbf successfully opened.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 33275",
        "",
        "... predicting reflections and getting centroids for image 3 ...",
        "",
        "Total number of predicted (non-merged) reflections: 3193",
        "",
        "",
        "Deleted 2333 reflections not found on images(s).",
        "",
        "",
        "Found 0 overlapped reflections.",
        "",
        "INFO: 42 reflns deleted out of 860 that might be in ice rings.",
        "",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 2.65    207    0   30.8   34.94      5.0      0.07",
        "  2.65 - 2.10    220    0   15.2   23.99      4.6      0.08",
        "  2.10 - 1.84    153    0   10.5   16.36      3.7      0.09",
        "  1.84 - 1.67    103    0    7.4   10.98      3.2      0.13",
        "  1.67 - 1.55     77    0    7.1    7.85      2.6      0.12",
        "  1.55 - 1.46     32    0    6.3    6.97      2.6      0.10",
        "  1.46 - 1.39     11    0    6.8    5.49      2.3      0.29",
        "  1.39 - 1.33     10    0    0.0    4.38      2.2      0.09",
        "  1.33 - 1.27      4    0    0.0    4.32      1.9      0.20",
        "  1.27 - 1.23      1    0    0.0    2.80      1.6      0.06",
        "-----------------------------------------------------------",
        " 99.99 - 1.23    818    0   15.9   20.90      4.0      0.09",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    18 -  >  20    174    0  1.80   31.55      4.9      0.07",
        "    16 -     18     32    0  1.44   25.15      4.3      0.09",
        "    14 -     16     56    0  1.73   26.35      4.6      0.08",
        "    12 -     14     40    0  1.83   24.94      4.1      0.08",
        "    10 -     12     60    0  1.54   22.59      3.9      0.08",
        "     8 -     10     82    0  1.59   19.88      3.9      0.07",
        "     6 -      8     90    0  1.45   16.63      3.7      0.11",
        "     4 -      6    126    0  1.33   14.57      3.5      0.11",
        "     2 -      4    135    0  1.26   12.49      3.4      0.14",
        " <   0 -      2     23    0  1.33   14.04      3.6      0.10",
        "------------------------------------------------------------",
        " <   0 -  >  20    818    0  1.26   20.90      4.0      0.09",
        "------------------------------------------------------------",
        "dtrefine - INFO wrote header file: dtranker_refine.head",
        "",
        "INFO - The suggested next step is to run dtrefine again:",
        "         dtrefine dtranker_refine.head -seq ...",
        "      -or-",
        "       Run dtpredict to check that the results predict the",
        "       first few images in each scan correctly: ",
        "         dtpredict dtranker_refine.head -seq ...",
        "      -or-",
        "       Run dtmultistrategy or dtintegrate with ",
        "       dtranker_refine.head as the input .head file:",
        "",
        "        dtmultistrategy dtranker_refine.head ...",
        "        dtintegrate dtranker_refine.head ...",
        "",
        "",
        "dtrank:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrank dtranker_refine.head -param 2 6 0 -param 11 6 0 -param 12 1 0",
        "Command line:",
        " dtranker_refine.head -param 2 6 0 -param 11 6 0 -param 12 1 0 ",
        "Processing Option:  dtranker_refine.head ",
        "     Header of file dtranker_refine.head successfully read.",
        "Processing Option: -param 2 6 0 ",
        "Processing Option: -param 11 6 0 ",
        "Processing Option: -param 12 1 0 ",
        "",
        "-----------------------------------------------------------------------",
        "Ranking procedure awards/penalizes points based on a set of rules.",
        "A description of each rule that is used is given below.",
        "",
        "Rule  1: Spot count in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell, a [10] point bonus is awarded.",
        "However, if there are less than 5 reflections found in the highest",
        "resolution shell, [0] points are subtracted from the award.",
        "",
        "Rule  2: I/Sigma in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule  3: Spot sharpness",
        "   Rule  3a: Spot asymmetry",
        "   A penalty of [10] times the average spot profile asymmetry is made.",
        "   Rule  3b: Spot intensity per pixel",
        "   An award of [10] times <Ipix>/[300] is made, where Ipix is an average",
        "   peak intensity per pixel",
        "   Rule  3c: Spot anisotropy",
        "   A penalty of [10] times the average spot anisotropy is made.",
        "   Rule  3d: Spot amorphousness",
        "   A penalty of [100] times the average spot amorphousness is made.",
        "",
        "Rule  4: Strong ice rings",
        "For each strong (peak to background ratio > [4.00] percent) ice ring, [10]",
        "points are deducted.",
        "",
        "Rule  5: Diffuse ice rings",
        "For each diffuse (peak to background ratio > [1.00] percent) ice ring, [5]",
        "points are deducted.",
        "",
        "Rule  6: Percentage of spots indexed",
        "If more than [20] reflections are indexed,",
        "[100] * <Indexed reflections / Used Reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are indexed,",
        "[0] points are subtracted.",
        "",
        "Rule  7: RMS residual after refinement",
        "If a sample fails to refine, or has an RMS residual > [0.30] mm,",
        "9 points are deducted. Otherwise, a penalty of",
        "[30] * <RMS-MM-Residual> is made.",
        "",
        "Rule  8: Mosaicity",
        "If a sample fails to refine, or has a mosaicity > [1.00] deg,",
        "70 points are deducted. Otherwise, a penalty of",
        "[70] * Mosaicity is made.",
        "",
        "Rule  9: Percentage of spots refined",
        "If more than [20] reflections are refined,",
        "[100] * <Refined reflections / Predicted reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are refined,",
        "[0] points are subtracted.",
        "",
        "Rule 10: Spot count in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more predicted reflections found in a shell, a [10] point",
        "bonus is awarded. However, if there are less than 5 predicted reflections",
        "found in the highest resolution shell, [0] points are subtracted from",
        "the award.",
        "",
        "Rule 11: I/Sigma in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule 12: Strategy total rotation width",
        "A penalty of [0] times RotWidth/[360.00] is applied, where RotWidth is the",
        "total rotation width suggested by strategy.",
        "",
        "Each number in braces is a modifiable parameter which can ",
        "be altered by using the -param command.  If for example",
        "you wish to modify the first parameter in Rule (3) to be 100, ",
        "you specify -param 3 1 100.",
        "",
        "-----------------------------------------------------------------------",
        "--------------------------------------------------------------------------------",
        "Rule Category                                                  Pts Cumul    Max",
        "--------------------------------------------------------------------------------",
        "  3  Penalty for peak profile asymmetry of 0.09                 -1    -1    -10",
        "  3  Award for peak intensity per pixel of 43.45                 1     0     10",
        "  3  Penalty for spot shape anisotropy of 0.09                   0     0    -10",
        "  3  Penalty for spot shape amorphousness of 0.000               0     0   -100",
        "  6  Indexed 44% of all reflections used in indexing            43    43    100",
        "  7  Penalty for RMS residual value of 0.070                    -2    41     -8",
        "  8  Penalty for Mosaicity value of 0.06                        -4    37    -70",
        "  9  Refined 89% of all reflections used in refinement          89   126    100",
        " 10  Predicted&Found: >=5 reflns in HRS #8 (1.39-1.33)\u00C5         10   136     10",
        " 10  Predicted&Found: >=5 reflns in HRS #7 (1.46-1.39)\u00C5         10   146     10",
        " 10  Predicted&Found: >=5 reflns in HRS #6 (1.55-1.46)\u00C5         10   156     10",
        " 10  Predicted&Found: >=5 reflns in HRS #5 (1.67-1.55)\u00C5         10   166     10",
        " 10  Predicted&Found: >=5 reflns in HRS #4 (1.84-1.67)\u00C5         10   176     10",
        " 11  Predicted&Found: I/S=4.7 in HRS #8 (1.39-1.33)\u00C5             1   177     20",
        " 11  Predicted&Found: I/S=6.8 in HRS #7 (1.46-1.39)\u00C5             2   179     20",
        " 11  Predicted&Found: I/S=6.3 in HRS #6 (1.55-1.46)\u00C5             2   181     20",
        " 11  Predicted&Found: I/S=7.1 in HRS #5 (1.67-1.55)\u00C5             2   183     20",
        " 11  Predicted&Found: I/S=7.4 in HRS #4 (1.84-1.67)\u00C5             2   185     20",
        "--------------------------------------------------------------------------------",
        "     Cumulative for /tmp/image_0003.cbf                             -185",
        "",
        "",
        "",
        "     Header of file dtranker.head successfully read."
      };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\r\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /*
   * Error from dtindex.
   */
  private static byte[] createDtrankerOutput03() {
    /* @formatter:off */
    String[] output =
      new String[] {
        "",
        "dtranker:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "",
        "Command line:",
        "dtranker image=/tmp/cell_zy_01_01.cbf",
        "",
        "",
        "dtextractheader:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "WARNING: rotation start == rotation end.",
        "     Header of file /tmp/cell_zy_01_01.cbf successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "",
        "dtfind:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dfind dtranker_input.head -reso edge -seq 1 1 -sigma 3 -min 50 -2D \\",
        "    -filter 6 -window 0 0 -rank -ref dtranker_find.ref -out \\",
        "    dtranker_find.head",
        "",
        "     Header of file dtranker_input.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: /tmp/cell_zy_01_01.cbf ...",
        "",
        "     File /tmp/cell_zy_01_01.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "",
        "Command line string: >>-reso<<",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: /tmp/cell_zy_01_01.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Command line string: >>-seq<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-min<<",
        "Command line string: >>-2D<<",
        "Command line string: >>-filter<<",
        "Command line string: >>-window<<",
        "Command line string: >>-rank<<",
        "Command line string: >>-ref<<",
        "Command line string: >>-out<<",
        "",
        "Resolution limits of an image are 999.99 to 1.41262",
        "Resolution limits of peak search are 999.99 to 1.79514",
        "",
        "",
        "dtfind: 2D method  used",
        "...reading image /tmp/cell_zy_01_01.cbf...",
        "",
        "     File /tmp/cell_zy_01_01.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "",
        "Processing image '/tmp/cell_zy_01_01.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/cell_zy_01_01.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/cell_zy_01_01.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/cell_zy_01_01.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "INFO:  No powder rings detected.",
        "--------------------------------------------------",
        "Find object listing:",
        "     Sigma: 3",
        "Resolution: 999.99 to 1.79514",
        "   Minimum: 50",
        "Circle lim: 1231, 1263, 0, 1763",
        "  Rect lim: 24, 25, 2439, 2502",
        "Spot wind.: 0, 0",
        "Peak filt.: 6",
        "Back. tile: 153, 157",
        " Seq. num.: 1, 1",
        "Image dim.: 2463, 2527",
        "   3D dump: 0",
        "2 preliminary spots found in 2D search with rotation angle 0 degs.",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 3.87      0    0    0.0    0.00      0.0      0.00",
        "  3.87 - 3.07      1    0   21.6   10.56      3.3      0.11",
        "  3.07 - 2.68      0    0    0.0    0.00      0.0      0.00",
        "  2.68 - 2.44      0    0    0.0    0.00      0.0      0.00",
        "  2.44 - 2.26      0    0    0.0    0.00      0.0      0.00",
        "  2.26 - 2.13      0    0    0.0    0.00      0.0      0.00",
        "  2.13 - 2.02      1    0   24.9    3.56      1.9      0.01",
        "  2.02 - 1.93      0    0    0.0    0.00      0.0      0.00",
        "  1.93 - 1.86      0    0    0.0    0.00      0.0      0.00",
        "  1.86 - 1.80      0    0    0.0    0.00      0.0      0.00",
        "-----------------------------------------------------------",
        " 99.99 - 1.80      2    0   23.2    7.06      2.6      0.06",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    27 -  >  30      0    0  0.00    0.00      0.0      0.00",
        "    24 -     27      0    0  0.00    0.00      0.0      0.00",
        "    21 -     24      1    0  2.08    3.56      1.9      0.01",
        "    18 -     21      1    0  3.33   10.56      3.3      0.11",
        "    15 -     18      0    0  0.00    0.00      0.0      0.00",
        "    12 -     15      0    0  0.00    0.00      0.0      0.00",
        "     9 -     12      0    0  0.00    0.00      0.0      0.00",
        "     6 -      9      0    0  0.00    0.00      0.0      0.00",
        "     3 -      6      0    0  0.00    0.00      0.0      0.00",
        " <   0 -      3      0    0  0.00    0.00      0.0      0.00",
        "------------------------------------------------------------",
        " <   0 -  >  30      2    0  2.08    7.06      2.6      0.06",
        "------------------------------------------------------------",
        "WARNING:  No strong spots found on first pass of integration:  Using estimated defaults!",
        "",
        "dtfind: There were 2 spots found.",
        "There were 2 preliminary spots of which 0 were marked as saturated,",
        " or  0.00% of them.",
        "Number of reflections written in 'dtranker_find.ref': 2",
        "dtfind: Spots written to dtranker_find.ref",
        "dtfind - Wrote header file dtranker_find.head",
        "",
        "INFO - Not very many spots were found. The suggested",
        "       next step is to re-run dtfind with different",
        "       arguments to help find more spots.",
        "",
        "",
        "dtindex:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtindex dtranker_find.head dtranker_find.ref -maxresid 3.0 -sigma 5 \\",
        "    -out dtranker_index.head",
        "",
        "     Header of file dtranker_find.head successfully read.",
        "Reflection list: dtranker_find.ref",
        "Creflnlist::nRead with filename: dtranker_find.ref",
        "INFO in Creflnlist::nRead, EOF after 2 reflections read in",
        "                                    (2 total now in list).",
        "Command line string: >>-maxresid<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-out<<",
        "INFO: deleted 0 reflns outside of resolution bounds.",
        "      This leaves 2 reflns for indexing.",
        "",
        "INFO: 1 reflns deleted out of 2 that might be in ice rings.",
        "",
        "ERROR:  dtindex cannot index with 1 reflections.",
        "      Terminating.",
        "Max cell length allowed for reciprocal lattice vectors: 538.446",
        "ERROR:  Not enough reflections!",
        "",
        "dtrank:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrank dtranker_find.head -param 2 6 0 -param 11 6 0 -param 12 1 0",
        "Command line:",
        " dtranker_find.head -param 2 6 0 -param 11 6 0 -param 12 1 0 ",
        "Processing Option:  dtranker_find.head ",
        "     Header of file dtranker_find.head successfully read.",
        "Processing Option: -param 2 6 0 ",
        "Processing Option: -param 11 6 0 ",
        "Processing Option: -param 12 1 0 ",
        "",
        "-----------------------------------------------------------------------",
        "Ranking procedure awards/penalizes points based on a set of rules.",
        "A description of each rule that is used is given below.",
        "",
        "Rule  1: Spot count in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell, a [10] point bonus is awarded.",
        "However, if there are less than 5 reflections found in the highest",
        "resolution shell, [0] points are subtracted from the award.",
        "",
        "Rule  2: I/Sigma in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule  3: Spot sharpness",
        "   Rule  3a: Spot asymmetry",
        "   A penalty of [10] times the average spot profile asymmetry is made.",
        "   Rule  3b: Spot intensity per pixel",
        "   An award of [10] times <Ipix>/[300] is made, where Ipix is an average",
        "   peak intensity per pixel",
        "   Rule  3c: Spot anisotropy",
        "   A penalty of [10] times the average spot anisotropy is made.",
        "   Rule  3d: Spot amorphousness",
        "   A penalty of [100] times the average spot amorphousness is made.",
        "",
        "Rule  4: Strong ice rings",
        "For each strong (peak to background ratio > [4.00] percent) ice ring, [10]",
        "points are deducted.",
        "",
        "Rule  5: Diffuse ice rings",
        "For each diffuse (peak to background ratio > [1.00] percent) ice ring, [5]",
        "points are deducted.",
        "",
        "Rule  6: Percentage of spots indexed",
        "If more than [20] reflections are indexed,",
        "[100] * <Indexed reflections / Used Reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are indexed,",
        "[0] points are subtracted.",
        "",
        "Rule  7: RMS residual after refinement",
        "If a sample fails to refine, or has an RMS residual > [0.30] mm,",
        "9 points are deducted. Otherwise, a penalty of",
        "[30] * <RMS-MM-Residual> is made.",
        "",
        "Rule  8: Mosaicity",
        "If a sample fails to refine, or has a mosaicity > [1.00] deg,",
        "70 points are deducted. Otherwise, a penalty of",
        "[70] * Mosaicity is made.",
        "",
        "Rule  9: Percentage of spots refined",
        "If more than [20] reflections are refined,",
        "[100] * <Refined reflections / Predicted reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are refined,",
        "[0] points are subtracted.",
        "",
        "Rule 10: Spot count in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more predicted reflections found in a shell, a [10] point",
        "bonus is awarded. However, if there are less than 5 predicted reflections",
        "found in the highest resolution shell, [0] points are subtracted from",
        "the award.",
        "",
        "Rule 11: I/Sigma in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule 12: Strategy total rotation width",
        "A penalty of [0] times RotWidth/[360.00] is applied, where RotWidth is the",
        "total rotation width suggested by strategy.",
        "",
        "Each number in braces is a modifiable parameter which can ",
        "be altered by using the -param command.  If for example",
        "you wish to modify the first parameter in Rule (3) to be 100, ",
        "you specify -param 3 1 100.",
        "",
        "-----------------------------------------------------------------------",
        "--------------------------------------------------------------------------------",
        "Rule Category                                                  Pts Cumul    Max",
        "--------------------------------------------------------------------------------",
        "  3  Penalty for peak profile asymmetry of 0.06                 -1    -1    -10",
        "  3  Award for peak intensity per pixel of 33.55                 1     0     10",
        "  3  Penalty for spot shape anisotropy of 0.27                  -2    -2    -10",
        "  3  Penalty for spot shape amorphousness of 0.000               0    -2   -100",
        "  7  Penalty for RMS residual > 0.30 (or no refinement)         -8   -10     -8",
        "  8  Penalty for Mosaicity > 1.00 (or no refinement)           -70   -80    -70",
        "--------------------------------------------------------------------------------",
        "     Cumulative for /tmp/cell_zy_01_01.cbf                           -80",
        "",
        "",
        "",
        "     Header of file dtranker.head successfully read.",
        ""
      };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\r\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /* Saturated image. */
  private static byte[] createDtrankerOutput04() {
    /* @formatter:off */
    String[] output =
      new String[] {
        "",
        "dtranker:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "",
        "Command line:",
        "dtranker image=cell_zy_04_05.cbf",
        "",
        "",
        "dtextractheader:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "WARNING: rotation start == rotation end.",
        "     Header of file cell_zy_04_05.cbf successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "     Header of file dtranker_input.head successfully read.",
        "",
        "dtfind:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dfind dtranker_input.head -reso edge -seq 5 5 -sigma 3 -min 50 -2D \\",
        "    -filter 6 -window 0 0 -rank -ref dtranker_find.ref -out \\",
        "    dtranker_find.head",
        "",
        "     Header of file dtranker_input.head successfully read.",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: cell_zy_04_05.cbf ...",
        "",
        "     File cell_zy_04_05.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "",
        "",
        "Command line string: >>-reso<<",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "...reading mask file: cell_zy_04_05.cbf ...",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Command line string: >>-seq<<",
        "Command line string: >>-sigma<<",
        "Command line string: >>-min<<",
        "Command line string: >>-2D<<",
        "Command line string: >>-filter<<",
        "Command line string: >>-window<<",
        "Command line string: >>-rank<<",
        "Command line string: >>-ref<<",
        "Command line string: >>-out<<",
        "",
        "Resolution limits of an image are 999.99 to 1.54738",
        "Resolution limits of peak search are 999.99 to 1.99755",
        "",
        "",
        "dtfind: 2D method  used",
        "...reading image /tmp/t1/cell_zy_04_05.cbf...",
        "",
        "     File /tmp/t1/cell_zy_04_05.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "",
        "Processing image '/tmp/t1/cell_zy_04_05.cbf' for powder ring detection.",
        "Search starts at 2Theta = 10.0 deg. Expected ring width = 3.0 deg",
        "",
        "",
        "     File /tmp/t1/cell_zy_04_05.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/t1/cell_zy_04_05.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "",
        "     File /tmp/t1/cell_zy_04_05.cbf successfully opened.",
        "WARNING: rotation start == rotation end.",
        "INFO: Min, Max pixel values in pseudo-CBF image: -2, 1048575",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "PILT_NONUNF_TYPE: >>Simple_mask<<",
        "INFO in Cnonunf: using",
        "     /tmp/t1/cell_zy_04_05.cbf",
        "     as the simple mask/nonunf file, was FirstScanImage.",
        "Min raw image pixel OK value in mask/nonunf/image file: 1",
        "Resolution rings found.",
        "--------------------------------------------------",
        " Peak 2Theta 2Theta 2Theta    Reso    Reso    Peak ",
        "   #     Mid  Start    End   Start     End Percent ",
        "--------------------------------------------------",
        "    0  14.15  12.65  15.65  4.5385  3.6725    3.99",
        "    1  16.35  14.85  17.85  3.8691  3.2229    2.65",
        "    2  20.75  19.25  22.25  2.9904  2.5913    1.11",
        "    3  24.25  22.75  25.75  2.5351  2.2439    2.07",
        "    4  25.95  24.45  27.45  2.3613  2.1074    3.84",
        "    5  27.95  26.45  29.45  2.1855  1.9671    0.62",
        "    6  30.05  28.55  31.55  2.0278  1.8392    0.55",
        "    7  32.45  30.95  33.95  1.8739  1.7126    0.92",
        "    8  34.25  32.75  35.75  1.7735  1.6290    1.90",
        "--------------------------------------------------",
        "Find object listing:",
        "     Sigma: 3",
        "Resolution: 999.99 to 1.99755",
        "   Minimum: 50",
        "Circle lim: 1231, 1263, 0, 1763",
        "  Rect lim: 24, 25, 2439, 2502",
        "Spot wind.: 0, 0",
        "Peak filt.: 6",
        "Back. tile: 153, 157",
        " Seq. num.: 5, 5",
        "Image dim.: 2463, 2527",
        "   3D dump: 0",
        "0 preliminary spots found in 2D search with rotation angle 0 degs.",
        "",
        "INFO: No reflns deleted out of 0 that might be in ice rings.",
        "",
        "Resolution statistics",
        "-----------------------------------------------------------",
        " Resolution      Num  Num  I/sig Backgnd B/sig(B) sharpness",
        "    range     reflns rejs    avg     avg      avg       avg",
        "-----------------------------------------------------------",
        " 99.99 - 4.30      0    0    0.0    0.00      0.0      0.00",
        "  4.30 - 3.42      0    0    0.0    0.00      0.0      0.00",
        "  3.42 - 2.98      0    0    0.0    0.00      0.0      0.00",
        "  2.98 - 2.71      0    0    0.0    0.00      0.0      0.00",
        "  2.71 - 2.52      0    0    0.0    0.00      0.0      0.00",
        "  2.52 - 2.37      0    0    0.0    0.00      0.0      0.00",
        "  2.37 - 2.25      0    0    0.0    0.00      0.0      0.00",
        "  2.25 - 2.15      0    0    0.0    0.00      0.0      0.00",
        "  2.15 - 2.07      0    0    0.0    0.00      0.0      0.00",
        "  2.07 - 2.00      0    0    0.0    0.00      0.0      0.00",
        "-----------------------------------------------------------",
        " 99.99 - 2.00      0    0    0.0    0.00      0.0      0.00",
        "-----------------------------------------------------------",
        "Intensity statistics",
        "------------------------------------------------------------",
        "     Int/sigmaI    Num  Num  Reso Backgnd B/sig(B) sharpness",
        "        range   reflns rejs   max     avg      avg       avg",
        "------------------------------------------------------------",
        "    18 -  >  20      0    0  0.00    0.00      0.0      0.00",
        "    16 -     18      0    0  0.00    0.00      0.0      0.00",
        "    14 -     16      0    0  0.00    0.00      0.0      0.00",
        "    12 -     14      0    0  0.00    0.00      0.0      0.00",
        "    10 -     12      0    0  0.00    0.00      0.0      0.00",
        "     8 -     10      0    0  0.00    0.00      0.0      0.00",
        "     6 -      8      0    0  0.00    0.00      0.0      0.00",
        "     4 -      6      0    0  0.00    0.00      0.0      0.00",
        "     2 -      4      0    0  0.00    0.00      0.0      0.00",
        " <   0 -      2      0    0  0.00    0.00      0.0      0.00",
        "------------------------------------------------------------",
        " <   0 -  >  20      0    0  0.00    0.00      0.0      0.00",
        "------------------------------------------------------------",
        "WARNING:  No strong spots found on first pass of integration:  Using estimated defaults!",
        "There were no spots found.",
        "dtfind - Wrote header file dtranker_find.head",
        "",
        "INFO - Not very many spots were found. The suggested",
        "       next step is to re-run dtfind with different",
        "       arguments to help find more spots.",
        "",
        "",
        "dtrank:  Copyright (c) 2007 Rigaku",
        "d*TREK version 9.9.3L -- Jun 25 2008",
        "Please see the file ${DTREK_ROOT}/ACKNOWLEDGEMENTS ",
        "for further acknowledgements, copyrights and license information.",
        "",
        "Command line:",
        " dtrank dtranker_input.head -param 2 6 0 -param 11 6 0 -param 12 1 0",
        "Command line:",
        " dtranker_input.head -param 2 6 0 -param 11 6 0 -param 12 1 0 ",
        "Processing Option:  dtranker_input.head ",
        "     Header of file dtranker_input.head successfully read.",
        "Processing Option: -param 2 6 0 ",
        "Processing Option: -param 11 6 0 ",
        "Processing Option: -param 12 1 0 ",
        "",
        "-----------------------------------------------------------------------",
        "Ranking procedure awards/penalizes points based on a set of rules.",
        "A description of each rule that is used is given below.",
        "",
        "Rule  1: Spot count in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell, a [10] point bonus is awarded.",
        "However, if there are less than 5 reflections found in the highest",
        "resolution shell, [0] points are subtracted from the award.",
        "",
        "Rule  2: I/Sigma in resolution shells (found spots)",
        "Divide the total resolution range into [10] shells of equal volume.",
        "Consider [7] highest and [0] lowest shells. If there are [5]",
        "or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule  3: Spot sharpness",
        "   Rule  3a: Spot asymmetry",
        "   A penalty of [10] times the average spot profile asymmetry is made.",
        "   Rule  3b: Spot intensity per pixel",
        "   An award of [10] times <Ipix>/[300] is made, where Ipix is an average",
        "   peak intensity per pixel",
        "   Rule  3c: Spot anisotropy",
        "   A penalty of [10] times the average spot anisotropy is made.",
        "   Rule  3d: Spot amorphousness",
        "   A penalty of [100] times the average spot amorphousness is made.",
        "",
        "Rule  4: Strong ice rings",
        "For each strong (peak to background ratio > [4.00] percent) ice ring, [10]",
        "points are deducted.",
        "",
        "Rule  5: Diffuse ice rings",
        "For each diffuse (peak to background ratio > [1.00] percent) ice ring, [5]",
        "points are deducted.",
        "",
        "Rule  6: Percentage of spots indexed",
        "If more than [20] reflections are indexed,",
        "[100] * <Indexed reflections / Used Reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are indexed,",
        "[0] points are subtracted.",
        "",
        "Rule  7: RMS residual after refinement",
        "If a sample fails to refine, or has an RMS residual > [0.30] mm,",
        "9 points are deducted. Otherwise, a penalty of",
        "[30] * <RMS-MM-Residual> is made.",
        "",
        "Rule  8: Mosaicity",
        "If a sample fails to refine, or has a mosaicity > [1.00] deg,",
        "70 points are deducted. Otherwise, a penalty of",
        "[70] * Mosaicity is made.",
        "",
        "Rule  9: Percentage of spots refined",
        "If more than [20] reflections are refined,",
        "[100] * <Refined reflections / Predicted reflections> points are awarded.",
        "If less than 20 or less than [0] percent reflections are refined,",
        "[0] points are subtracted.",
        "",
        "Rule 10: Spot count in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more predicted reflections found in a shell, a [10] point",
        "bonus is awarded. However, if there are less than 5 predicted reflections",
        "found in the highest resolution shell, [0] points are subtracted from",
        "the award.",
        "",
        "Rule 11: I/Sigma in resolution shells (predicted and found spots)",
        "After refinement, divide the total resolution range into [10] shells of",
        "equal volume. Consider [7] highest and [0] lowest shells. If there are",
        "[5] or more reflections found in a shell,",
        "([20] * <I/sig>  - [0] * dev) / [60.00]",
        "points are awarded, where <I/sig> is the average I/sig value in that shell,",
        "and dev is a standard deviation of I/sig with respect to the rotation",
        "angle.",
        "",
        "Rule 12: Strategy total rotation width",
        "A penalty of [0] times RotWidth/[360.00] is applied, where RotWidth is the",
        "total rotation width suggested by strategy.",
        "",
        "Each number in braces is a modifiable parameter which can ",
        "be altered by using the -param command.  If for example",
        "you wish to modify the first parameter in Rule (3) to be 100, ",
        "you specify -param 3 1 100.",
        "",
        "-----------------------------------------------------------------------",
        "--------------------------------------------------------------------------------",
        "Rule Category                                                  Pts Cumul    Max",
        "--------------------------------------------------------------------------------",
        "  3  Penalty for absent profile asymmetry info                 -10   -10    -10",
        "  3  Penalty for absent peak intensity density info            -10   -20     10",
        "  3  Penalty for absent spot shape anisotropy info             -10   -30    -10",
        "  3  Penalty for absent spot shape amorphousness info         -100  -130   -100",
        "  7  Penalty for RMS residual > 0.30 (or no refinement)         -8  -138     -8",
        "  8  Penalty for Mosaicity > 1.00 (or no refinement)           -70  -208    -70",
        "--------------------------------------------------------------------------------",
        "     Cumulative for /tmp/t1/cell_zy_04_05.cbf                       -208",
        "",
        "",
        "",
        "     Header of file dtranker.head successfully read.",
      };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\r\n");
    }
    return result.toString().getBytes(UTF8);
  }

  private static void closeSilently(Closeable closeable) {
    if (closeable == null) return;
    try {
      closeable.close();
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (IOException | RuntimeException e) {
      /* IGNORED */
    }
  }
}
