package imcacat.imcadr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.Arrays;

import org.junit.Test;

import imcacat.imcadr.util.NetworkUtils;
import imcacat.imcadr.util.StringUtils;

/**
 * I test <code>ImcadrConfigurationParser</code>.
 */
public class ImcadrConfigurationParserTest {
  @Test
  public void testClusterKeyFile() throws IOException, ConfigurationException {
    String conf = "ClusterKeyFile = /nonexistent/id_imcadr";
    ImcadrConfiguration c = parse(conf);
    assertEquals(new File("/nonexistent/id_imcadr"), c.clusterKeyFile());
  }

  @Test(expected = ConfigurationException.class)
  public void testClusterKeyFile_notAbsolutePath() throws IOException, ConfigurationException {
    String conf = "ClusterKeyFile = id_imcadr";
    parse(conf);
  }

  @Test
  public void testClusterKeyFile_unset() throws IOException, ConfigurationException {
    String conf = "";
    ImcadrConfiguration c = parse(conf);
    assertNull(c.clusterKeyFile());
  }

  @Test
  public void testClusterNodes() throws IOException, ConfigurationException {
    String conf = "ClusterNodes = vader:4, jabba:8, bobafett:16";
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(3, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals("vader", node.hostname());
    assertEquals(4, node.jobCapacity());
    node = c.clusterNodes().get(1);
    assertEquals("jabba", node.hostname());
    assertEquals(8, node.jobCapacity());
    node = c.clusterNodes().get(2);
    assertEquals("bobafett", node.hostname());
    assertEquals(16, node.jobCapacity());
  }

  @Test
  public void testClusterNodes_noWhitespace() throws IOException, ConfigurationException {
    String conf = "ClusterNodes=vader:4,jabba:8";
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(2, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals("vader", node.hostname());
    assertEquals(4, node.jobCapacity());
    node = c.clusterNodes().get(1);
    assertEquals("jabba", node.hostname());
    assertEquals(8, node.jobCapacity());
  }

  @Test
  public void testClusterNodes_lineContinuation() throws IOException, ConfigurationException {
    String conf = StringUtils.join("\n", new String[] {
        "ClusterNodes = vader, jabba, bobafett, \\",
        "    tarkin"});
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(4, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(3);
    assertEquals("tarkin", node.hostname());
  }

  @Test
  public void testClusterNodes_jobCapacityDefaultsToOne() throws IOException, ConfigurationException {
    String conf = "ClusterNodes = vader, jabba:2, bobafett";
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(3, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals("vader", node.hostname());
    assertEquals(1, node.jobCapacity());
    node = c.clusterNodes().get(1);
    assertEquals("jabba", node.hostname());
    assertEquals(2, node.jobCapacity());
    node = c.clusterNodes().get(2);
    assertEquals("bobafett", node.hostname());
    assertEquals(1, node.jobCapacity());
  }

  @Test
  public void testClusterNodes_empty() throws IOException, ConfigurationException {
    String conf = "ClusterNodes =";
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(1, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals(NetworkUtils.LOCALHOST_NAME, node.hostname());
    assertEquals(Runtime.getRuntime().availableProcessors(), node.jobCapacity());
  }

  @Test
  public void testClusterNodes_unset() throws IOException, ConfigurationException {
    String conf = "";
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(1, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals(NetworkUtils.LOCALHOST_NAME, node.hostname());
    assertEquals(Runtime.getRuntime().availableProcessors(), node.jobCapacity());
  }

  @Test
  public void testClusterNodes_maxNumScoresJobsAsDefaultJobCapacity()
      throws IOException, ConfigurationException {
    String conf = StringUtils.join("\n", new String[] {
        "MaxNumScoreJobs = 3",
        "ClusterNodes = vader"});
    ImcadrConfiguration c = parse(conf);
    assertNotNull(c.clusterNodes());
    assertEquals(1, c.clusterNodes().size());
    ClusterNode node = c.clusterNodes().get(0);
    assertEquals("vader", node.hostname());
    assertEquals(3, node.jobCapacity());
  }

  @Test
  public void testLabelitDistSignalStrengthArgs() throws IOException, ConfigurationException {
    String conf = "Labelit.DistlSignalStrengthArgs = $image distl.verbose=True $more";
    ImcadrConfiguration c = parse(conf);
    assertEquals(Arrays.asList("$image", "distl.verbose=True", "$more"),
        c.analysisProgramArguments(AnalysisProgram.LABELIT_DISTL_SIGNAL_STRENGTH));
  }

  @Test(expected = ConfigurationException.class)
  public void testLabelitDistSignalStrengthArgs_noImageVar()
      throws IOException, ConfigurationException {
    parse("Labelit.DistlSignalStrengthArgs = distl.verbose=True $more");
  }

  @Test(expected = ConfigurationException.class)
  public void testLabelitDistSignalStrengthArgs_noMoreVar()
      throws IOException, ConfigurationException {
    parse("Labelit.DistlSignalStrengthArgs = $image distl.verbose=True");
  }

  @Test(expected = ConfigurationException.class)
  public void testLabelitDistSignalStrengthArgs_noMoreVarWord()
      throws IOException, ConfigurationException {
    parse("Labelit.DistlSignalStrengthArgs = $image distl.verbose=True x=$more");
  }

  @Test
  public void testDTrekDtrankerArgs() throws IOException, ConfigurationException {
    String conf = "DTrek.DtrankerArgs = image=$image foo=bar $more";
    ImcadrConfiguration c = parse(conf);
    assertEquals(Arrays.asList("image=$image", "foo=bar", "$more"),
        c.analysisProgramArguments(AnalysisProgram.DTREK_DTRANKER));
  }

  @Test(expected = ConfigurationException.class)
  public void testDTrekDtrankerArgs_noImageVar() throws IOException, ConfigurationException {
    parse("DTrek.DtrankerArgs = image=whatever foo=bar $more");
  }

  @Test(expected = ConfigurationException.class)
  public void testDTrekDtrankerArgs_noMoreVar() throws IOException, ConfigurationException {
    parse("DTrek.DtrankerArgs = image=$image foo=bar");
  }

  @Test(expected = ConfigurationException.class)
  public void testDTrekDtrankerArgs_noMoreVarWord() throws IOException, ConfigurationException {
    parse("DTrek.DtrankerArgs = image=$image foo=bar x=$more");
  }

  @Test
  public void testMaxNumScoreJobs_deprecatedButAllowed() throws IOException, ConfigurationException {
    String conf = "MaxNumScoreJobs = 2";
    parse(conf);
  }

  @Test(expected = ConfigurationException.class)
  public void testMaxNumScoreJobs_lessThanOne() throws IOException, ConfigurationException {
    String conf = "MaxNumScoreJobs = 0";
    parse(conf);
  }

  private static ImcadrConfiguration parse(String conf) throws IOException, ConfigurationException {
    try (Reader r = new StringReader(conf)) {
      return (new ImcadrConfigurationParser(r)).parse();
    }
  }
}
