package imcacat.imcadr;

import java.io.BufferedWriter;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeoutException;
import java.util.regex.Pattern;

/**
 * I am a client for the diffraction rastering standard input and standard
 * output server API.
 * <p>
 * Units: exposure time [s], cell size [micrometers], raster size [cells].
 */
public class ImcadrServerClient {
  private static final Charset UTF_8 = Charset.forName("UTF-8");
  private static final String NIL = "nil";
  private static final String TRUE = "true";
  private static final String FALSE = "false";
  private static final String OK = "+OK";
  private static final String OK_WITH = OK + " ";
  private static final String NEWLINE = "\n";
  private static final Pattern SPACE_RE = Pattern.compile(" ");
  private static final Pattern SEMICOLON_RE = Pattern.compile(";");
  private static final Pattern COMMA_RE = Pattern.compile(",");
  private static final Pattern SLASH_RE = Pattern.compile("/");

  private static final HashMap<String, SpaceGroup> DATA_TO_SPACE_GROUP_MAP;

  static {
    DATA_TO_SPACE_GROUP_MAP = new HashMap<String, SpaceGroup>();
    DATA_TO_SPACE_GROUP_MAP.put("P1", SpaceGroup.P1);
    DATA_TO_SPACE_GROUP_MAP.put("P2", SpaceGroup.P2);
    DATA_TO_SPACE_GROUP_MAP.put("P222", SpaceGroup.P222);
    DATA_TO_SPACE_GROUP_MAP.put("P4", SpaceGroup.P4);
    DATA_TO_SPACE_GROUP_MAP.put("P6", SpaceGroup.P6);
    DATA_TO_SPACE_GROUP_MAP.put("C222", SpaceGroup.C222);
    DATA_TO_SPACE_GROUP_MAP.put("C2", SpaceGroup.C2);
    DATA_TO_SPACE_GROUP_MAP.put("I4", SpaceGroup.I4);
    DATA_TO_SPACE_GROUP_MAP.put("I222", SpaceGroup.I222);
    DATA_TO_SPACE_GROUP_MAP.put("F222", SpaceGroup.F222);
    if (DATA_TO_SPACE_GROUP_MAP.size() != SpaceGroup.values().length) {
      throw new BugException();
    }
  }

  private LineReader in = null;
  private BufferedWriter out = null;

  /**
   * Connects this client to a server via the specified input and output
   * streams.
   *
   * @param in stream for reading server responses
   * @param out stream for writing server commands
   */
  public void connect(final InputStream in, final OutputStream out) {
    this.in = new LineReader(new InputStreamReader(in, UTF_8));
    this.in.closePrerequisite(new Callable<Void>() {
      @Override
      public Void call() throws Exception {
        in.close();
        return null;
      }
    });
    this.out = new BufferedWriter(new OutputStreamWriter(out, UTF_8));
  }

  /**
   * Disconnects this client from the server and closes the input and output
   * streams.
   */
  public void disconnect() {
    closeSilently(in);
    in = null;
    closeSilently(out);
    out = null;
  }

  public boolean isConnected() {
    return in != null && out != null;
  }

  public File imagesDirectory() throws ImcadrServerClientException, IOException {
    writeLine("IMAGES_DIR");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("images directory", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    return new File(valueString);
  }

  public void imagesDirectory(File dir) throws ImcadrServerClientException, IOException {
    String valueString = (dir == null) ? NIL : dir.getAbsolutePath();
    writeLine("IMAGES_DIR " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("images directory", line);
  }

  public Double exposureTime() throws ImcadrServerClientException, IOException {
    writeLine("EXP_TIME");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("exposure time", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    try {
      return Double.valueOf(valueString);
    } catch (NumberFormatException e) {
      throw createParseException("exposure time", valueString, e);
    }
  }

  public void exposureTime(Double seconds) throws ImcadrServerClientException, IOException {
    String valueString = (seconds == null) ? NIL : format("%.3f", seconds);
    writeLine("EXP_TIME " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("exposure time", line);
  }

  public Rectangle cellSize() throws ImcadrServerClientException, IOException {
    writeLine("CELL_SIZE");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("cell size", line);
    String valueString = line.substring(OK_WITH.length());
    return parseRectangle(valueString, LengthUnit.MICROMETER, "cell size");
  }

  public void cellSize(Rectangle size) throws ImcadrServerClientException, IOException {
    Rectangle sizeInUm = (size == null) ? null : size.convertTo(LengthUnit.MICROMETER);
    String valueString =
        (sizeInUm == null) ? NIL
            : format("%.0fx%.0f", sizeInUm.getWidth(), sizeInUm.getHeight());
    writeLine("CELL_SIZE " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("cell size", line);
  }

  public Rectangle zySize() throws ImcadrServerClientException, IOException {
    writeLine("ZY_SIZE");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("ZY size", line);
    String valueString = line.substring(OK_WITH.length());
    return parseRectangle(valueString, LengthUnit.CELL, "ZY size");
  }

  public void zySize(Rectangle size) throws ImcadrServerClientException, IOException {
    Rectangle sizeInCell = (size == null) ? null : size.convertTo(LengthUnit.CELL);
    String valueString =
        (sizeInCell == null) ? NIL : format("%.0fx%.0f", sizeInCell.getWidth(),
            sizeInCell.getHeight());
    writeLine("ZY_SIZE " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("ZY size", line);
  }

  public Rectangle zxSize() throws ImcadrServerClientException, IOException {
    writeLine("ZX_SIZE");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("ZX size", line);
    String valueString = line.substring(OK_WITH.length());
    return parseRectangle(valueString, LengthUnit.CELL, "ZX size");
  }

  public void zxSize(Rectangle size) throws ImcadrServerClientException, IOException {
    Rectangle sizeInCell = (size == null) ? null : size.convertTo(LengthUnit.CELL);
    String valueString =
        (sizeInCell == null) ? NIL : format("%.0fx%.0f", sizeInCell.getWidth(),
            sizeInCell.getHeight());
    writeLine("ZX_SIZE " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("ZX size", line);
  }

  public Boolean flyScanEnabled() throws ImcadrServerClientException, IOException {
    writeLine("FLY_SCAN");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("fly scan enabled", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    return TRUE.equalsIgnoreCase(valueString) ? Boolean.TRUE : Boolean.FALSE;
  }

  public void flyScanEnabled(Boolean enabled) throws ImcadrServerClientException, IOException {
    String valueString = (enabled == null) ? NIL : (enabled.booleanValue() ? TRUE : FALSE);
    writeLine("FLY_SCAN " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("fly scan enabled", line);
  }

  public Boolean chasingEnabled() throws ImcadrServerClientException, IOException {
    writeLine("CHASING");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("chasing enabled", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    return TRUE.equalsIgnoreCase(valueString) ? Boolean.TRUE : Boolean.FALSE;
  }

  public void chasingEnabled(Boolean enabled) throws ImcadrServerClientException, IOException {
    String valueString = (enabled == null) ? NIL : (enabled.booleanValue() ? TRUE : FALSE);
    writeLine("CHASING " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("chasing enabled", line);
  }

  public Boolean weakDiffracter() throws ImcadrServerClientException, IOException {
    writeLine("WEAK_DIFF");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("weak diffracter", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    return TRUE.equalsIgnoreCase(valueString) ? Boolean.TRUE : Boolean.FALSE;
  }

  public void weakDiffracter(Boolean enabled) throws ImcadrServerClientException, IOException {
    String valueString = (enabled == null) ? NIL : (enabled.booleanValue() ? TRUE : FALSE);
    writeLine("WEAK_DIFF " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("weak diffracter", line);
  }

  public AnalysisPackage analysisPackage() throws ImcadrServerClientException, IOException {
    writeLine("ANALYSIS_PACKAGE");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("analysis package", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    if (valueString.equalsIgnoreCase("labelit")) {
      return AnalysisPackage.LABELIT;
    } else if (valueString.equalsIgnoreCase("dtrek")) {
      return AnalysisPackage.DTREK;
    } else if (valueString.equalsIgnoreCase("dstartrek")) {
      /*
       * Backward compatibility: accept deprecated name.
       */
      return AnalysisPackage.DTREK;
    } else {
      throw createParseException("analysis package", valueString);
    }
  }

  public void analysisPackage(AnalysisPackage analysisPackage)
      throws ImcadrServerClientException, IOException {
    String valueString;
    if (analysisPackage == null) {
      valueString = NIL;
    } else {
      switch (analysisPackage) {
        case LABELIT:
          valueString = "labelit";
          break;
        case DTREK:
          valueString = "dtrek";
          break;
        default:
          throw new BugException();
      }
    }
    writeLine("ANALYSIS_PACKAGE " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("analysis package", line);
  }

  public Double highResolutionCutoff() throws ImcadrServerClientException, IOException {
    writeLine("HIGH_RES_CUTOFF");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("high resolution cutoff", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    try {
      return Double.valueOf(valueString);
    } catch (NumberFormatException e) {
      throw createParseException("high resolution cutoff", valueString, e);
    }
  }

  public void highResolutionCutoff(Double angstroms) throws ImcadrServerClientException,
      IOException {
    String valueString = (angstroms == null) ? NIL : format("%.2f", angstroms);
    writeLine("HIGH_RES_CUTOFF " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("high resolution cutoff", line);
  }

  public Double lowResolutionCutoff() throws ImcadrServerClientException, IOException {
    writeLine("LOW_RES_CUTOFF");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("low resolution cutoff", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    try {
      return Double.valueOf(valueString);
    } catch (NumberFormatException e) {
      throw createParseException("low resolution cutoff", valueString, e);
    }
  }

  public void lowResolutionCutoff(Double angstroms) throws ImcadrServerClientException,
      IOException {
    String valueString = (angstroms == null) ? NIL : format("%.2f", angstroms);
    writeLine("LOW_RES_CUTOFF " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("low resolution cutoff", line);
  }

  public Integer minimumSpotArea() throws ImcadrServerClientException, IOException {
    writeLine("MIN_SPOT_AREA");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("minimum spot area", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    try {
      return Integer.valueOf(valueString);
    } catch (NumberFormatException e) {
      throw createParseException("minimum spot area", valueString, e);
    }
  }

  public void minimumSpotArea(Integer pixels) throws ImcadrServerClientException, IOException {
    String valueString = (pixels == null) ? NIL : format("%d", pixels);
    writeLine("MIN_SPOT_AREA " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("minimum spot area", line);
  }

  public Double minimumSignalHeight() throws ImcadrServerClientException, IOException {
    writeLine("MIN_SIGNAL_HEIGHT");
    String line = readLine();
    if (!line.startsWith(OK_WITH)) throw createGetException("minimum signal height", line);
    String valueString = line.substring(OK_WITH.length());
    if (NIL.equals(valueString)) return null;
    try {
      return Double.valueOf(valueString);
    } catch (NumberFormatException e) {
      throw createParseException("minimum signal height", valueString, e);
    }
  }

  public void minimumSignalHeight(Double sigma) throws ImcadrServerClientException, IOException {
    String valueString = (sigma == null) ? NIL : format("%.2f", sigma);
    writeLine("MIN_SIGNAL_HEIGHT " + valueString);
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createSetException("minimum signal height", line);
  }

  public ImcadrServerStatus status() throws ImcadrServerClientException, IOException {
    writeLine("STAT");
    String line = readLine();
    if (!line.startsWith(OK)) throw createGetException("status", line);
    ImcadrServerStatus result = new ImcadrServerStatus();
    for (String each = readLine(); !each.equals("."); each = readLine()) {
      StatusField field = parseStatusField(each);
      if (field.type() == null || field.value() == null) continue;
      String value = field.value();
      String description = field.type().description();
      switch (field.type()) {
        case PROGRESS:
          result.setProgress(parseStatusProgress(value));
          break;
        case STATE:
          Object[] stateAndMessage = parseStatusState(value);
          result.setState((ImcadrServerState)stateAndMessage[0]);
          result.setMessage((String)stateAndMessage[1]);
          break;
        case ORIGIN_ZY:
          result.setOriginZy(parseStatusOrigin(value, description));
          break;
        case ORIGIN_ZX:
          result.setOriginZx(parseStatusOrigin(value, description));
          break;
        case SCORES_ZY:
          result.setScoresZy(parseStatusScores(value, description));
          break;
        case SCORES_ZX:
          result.setScoresZx(parseStatusScores(value, description));
          break;
        case SPOT_TOT_ZY:
          result.setSpotTotalsZy(parseStatusSpotTotals(value, description));
          break;
        case SPOT_TOT_ZX:
          result.setSpotTotalsZx(parseStatusSpotTotals(value, description));
          break;
        case IN_RES_TOT_ZY:
          result.setInResolutionTotalsZy(parseStatusInResolutionTotals(value, description));
          break;
        case IN_RES_TOT_ZX:
          result.setInResolutionTotalsZx(parseStatusInResolutionTotals(value, description));
          break;
        case BRAGG_TOT_ZY:
          result.setGoodBraggTotalsZy(parseStatusGoodBraggTotals(value, description));
          break;
        case BRAGG_TOT_ZX:
          result.setGoodBraggTotalsZx(parseStatusGoodBraggTotals(value, description));
          break;
        case ICE_RING_TOT_ZY:
          result.setIceRingTotalsZy(parseStatusIceRingTotals(value, description));
          break;
        case ICE_RING_TOT_ZX:
          result.setIceRingTotalsZx(parseStatusIceRingTotals(value, description));
          break;
        case RES1_ZY:
          result.setEstimatedResolutions1Zy(parseStatusEstimatedResolutions(value, description));
          break;
        case RES1_ZX:
          result.setEstimatedResolutions1Zx(parseStatusEstimatedResolutions(value, description));
          break;
        case RES2_ZY:
          result.setEstimatedResolutions2Zy(parseStatusEstimatedResolutions(value, description));
          break;
        case RES2_ZX:
          result.setEstimatedResolutions2Zx(parseStatusEstimatedResolutions(value, description));
          break;
        case INT_SIG_TOT_ZY:
          result.setTotalIntegratedSignalsZy(parseStatusTotalIntegratedSignals(value,
              description));
          break;
        case INT_SIG_TOT_ZX:
          result.setTotalIntegratedSignalsZx(parseStatusTotalIntegratedSignals(value,
              description));
          break;
        case MOSAICITY_ZY:
          result.setMosaicitiesZy(parseStatusMosaicities(value, description));
          break;
        case MOSAICITY_ZX:
          result.setMosaicitiesZx(parseStatusMosaicities(value, description));
          break;
        case SPACE_GRP_ZY:
          result.setSpaceGroupsZy(parseStatusSpaceGroups(value, description));
          break;
        case SPACE_GRP_ZX:
          result.setSpaceGroupsZx(parseStatusSpaceGroups(value, description));
          break;
        case UNIT_CELL_ZY:
          result.setUnitCellsZy(parseStatusUnitCells(value, description));
          break;
        case UNIT_CELL_ZX:
          result.setUnitCellsZx(parseStatusUnitCells(value, description));
          break;
        case CENTER_ZY:
          result.setCenterZy(parseStatusCenter2d(value, description));
          break;
        case CENTER_ZX:
          result.setCenterZx(parseStatusCenter2d(value, description));
          break;
        case CENTER:
          result.setCenter(parseStatusCenter3d(value, description));
          break;
        case TIME:
          result.setTime(parseStatusTime(value));
          break;
      }
    }
    return result;
  }

  public String errorDetail() throws ImcadrServerClientException, IOException {
    writeLine("ERROR_DETAIL");
    String line = readLine();
    if (!line.startsWith(OK)) throw createGetException("error detail", line);
    StringBuilder result = new StringBuilder(1024);
    boolean first = true;
    for (;;) {
      String each = readLine();
      if (each == null) {
        throw createParseException("error detail", "", "end of stream before dot line");
      }
      if (each.equals(".")) break;
      if (first) {
        first = false;
      } else {
        result.append(NEWLINE);
      }
      result.append(each.startsWith(".") ? each.substring(1) : each);
    }
    String valueString = result.toString();
    if (NIL.equals(valueString)) return null;
    return valueString;
  }

  public void start() throws ImcadrServerClientException, IOException {
    writeLine("START");
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createCommandException("start", line);
  }

  public void stop() throws ImcadrServerClientException, IOException {
    writeLine("STOP");
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createCommandException("stop", line);
  }

  public boolean stop(long msecTimeout) throws ImcadrServerClientException, IOException {
    writeLine("STOP");
    String line = null;
    try {
      line = readLine(msecTimeout);
    } catch (TimeoutException e) {
      return false;
    }
    if (line.startsWith(OK)) return true;
    throw createCommandException("stop", line);
  }

  public void reset() throws ImcadrServerClientException, IOException {
    writeLine("RESET");
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createCommandException("reset", line);
  }

  public void quit() throws ImcadrServerClientException, IOException {
    writeLine("QUIT");
    String line = readLine();
    if (line.startsWith(OK)) return;
    throw createCommandException("quit", line);
  }

  public boolean quit(long msecTimeout) throws ImcadrServerClientException, IOException {
    writeLine("QUIT");
    String line = null;
    try {
      line = readLine(msecTimeout);
    } catch (TimeoutException e) {
      return false;
    }
    if (line.startsWith(OK)) return true;
    throw createCommandException("quit", line);
  }

  private int parseStatusProgress(String text) throws ImcadrServerClientException {
    int result = 0;
    try {
      result = Integer.parseInt(text);
    } catch (NumberFormatException e) {
      throw createParseException("status progress", text, e);
    }
    if (result >= 0 && result <= 100) return result;
    throw createParseException("status progress", text, "not in range [0,100]");
  }

  private Object[] parseStatusState(String text) throws ImcadrServerClientException {
    Object[] result = new Object[2];
    int spaceIndex = text.indexOf(' ');
    String stateString = text.substring(0, (spaceIndex >= 0) ? spaceIndex : text.length());
    if ("Idle".equals(stateString)) {
      result[0] = ImcadrServerState.IDLE;
    } else if ("Running".equals(stateString)) {
      result[0] = ImcadrServerState.RUNNING;
    } else if ("Paused".equals(stateString)) {
      result[0] = ImcadrServerState.PAUSED;
    } else if ("Error".equals(stateString)) {
      result[0] = ImcadrServerState.ERROR;
    } else {
      throw createParseException("status state", text);
    }
    if (spaceIndex < 0) return result;
    int openParenIndex = text.indexOf('(', spaceIndex);
    if (openParenIndex < 0) return result;
    int closeParenIndex = text.lastIndexOf(')');
    if (closeParenIndex < 0) return result;
    if (closeParenIndex < openParenIndex) return result;
    String message = text.substring(openParenIndex + 1, closeParenIndex);
    if (message.isEmpty()) return result;
    result[1] = message;
    return result;
  }

  private Position2d parseStatusOrigin(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] origin = COMMA_RE.split(text, -1);
    if (origin.length != 2) throw createParseException(attribute, text, "not two values");
    try {
      return Position2d.mm(Double.parseDouble(origin[0]), Double.parseDouble(origin[1]));
    } catch (NumberFormatException e) {
      throw createParseException(attribute, text, e);
    }
  }

  private Integer[][] parseStatusScores(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] rows = SEMICOLON_RE.split(text, -1);
    String[] firstRow = COMMA_RE.split(rows[0], -1);
    Integer[][] result = new Integer[rows.length][firstRow.length];
    for (int r = 0; r < rows.length; r++) {
      String[] row = COMMA_RE.split(rows[r], -1);
      if (row.length != firstRow.length) {
        throw createParseException(attribute, text, "rows of different length");
      }
      for (int c = 0; c < row.length; c++) {
        try {
          result[r][c] = row[c].isEmpty() ? null : Integer.valueOf(row[c]);
        } catch (NumberFormatException e) {
          throw createParseException(attribute, text, e);
        }
      }
    }
    return result;
  }

  private Integer[][] parseStatusSpotTotals(String text, String attribute)
      throws ImcadrServerClientException {
    Integer[][] result = parse2dIntegerArray(text, attribute);
    ensureWithinRange(result, 0, null, text, attribute);
    return result;
  }

  private Integer[][] parseStatusInResolutionTotals(String text, String attribute)
      throws ImcadrServerClientException {
    Integer[][] result = parse2dIntegerArray(text, attribute);
    ensureWithinRange(result, 0, null, text, attribute);
    return result;
  }

  private Integer[][] parseStatusGoodBraggTotals(String text, String attribute)
      throws ImcadrServerClientException {
    Integer[][] result = parse2dIntegerArray(text, attribute);
    ensureWithinRange(result, 0, null, text, attribute);
    return result;
  }

  private Integer[][] parseStatusIceRingTotals(String text, String attribute)
      throws ImcadrServerClientException {
    Integer[][] result = parse2dIntegerArray(text, attribute);
    ensureWithinRange(result, 0, null, text, attribute);
    return result;
  }

  private Double[][] parseStatusEstimatedResolutions(String text, String attribute)
      throws ImcadrServerClientException {
    Double[][] result = parse2dDoubleArray(text, attribute);
    ensureWithinRange(result, 0.0, null, text, attribute);
    return result;
  }

  private Integer[][] parseStatusTotalIntegratedSignals(String text, String attribute)
      throws ImcadrServerClientException {
    Integer[][] result = parse2dIntegerArray(text, attribute);
    ensureWithinRange(result, 0, null, text, attribute);
    return result;
  }

  private Double[][] parseStatusMosaicities(String text, String attribute)
      throws ImcadrServerClientException {
    Double[][] result = parse2dDoubleArray(text, attribute);
    ensureWithinRange(result, 0.0, null, text, attribute);
    return result;
  }

  private SpaceGroup[][] parseStatusSpaceGroups(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] rows = SEMICOLON_RE.split(text, -1);
    String[] firstRow = COMMA_RE.split(rows[0], -1);
    SpaceGroup[][] result = new SpaceGroup[rows.length][firstRow.length];
    for (int r = 0; r < rows.length; r++) {
      String[] row = COMMA_RE.split(rows[r], -1);
      if (row.length != firstRow.length) {
        throw createParseException(attribute, text, "rows of different length");
      }
      for (int c = 0; c < row.length; c++) {
        result[r][c] = row[c].isEmpty() ? null : parseSpaceGroup(row[c], attribute);
      }
    }
    return result;
  }

  private UnitCell[][] parseStatusUnitCells(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] rows = SEMICOLON_RE.split(text, -1);
    String[] firstRow = COMMA_RE.split(rows[0], -1);
    UnitCell[][] result = new UnitCell[rows.length][firstRow.length];
    for (int r = 0; r < rows.length; r++) {
      String[] row = COMMA_RE.split(rows[r], -1);
      if (row.length != firstRow.length) {
        throw createParseException(attribute, text, "rows of different length");
      }
      for (int c = 0; c < row.length; c++) {
        result[r][c] = row[c].isEmpty() ? null : parseUnitCell(row[c], attribute);
      }
    }
    return result;
  }

  private Position2d parseStatusCenter2d(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] center = COMMA_RE.split(text, -1);
    if (center.length != 2) throw createParseException(attribute, text, "not two values");
    try {
      return Position2d.mm(Double.parseDouble(center[0]), Double.parseDouble(center[1]));
    } catch (NumberFormatException e) {
      throw createParseException(attribute, text, e);
    }
  }

  private Position3d parseStatusCenter3d(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] center = COMMA_RE.split(text, -1);
    if (center.length != 3) throw createParseException(attribute, text, "not three values");
    try {
      return Position3d.mm(Double.parseDouble(center[0]), Double.parseDouble(center[1]),
          Double.parseDouble(center[2]));
    } catch (NumberFormatException e) {
      throw createParseException(attribute, text, e);
    }
  }

  private Integer parseStatusTime(String text) throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    int result;
    try {
      result = Integer.parseInt(text);
    } catch (NumberFormatException e) {
      throw createParseException("status time", text, e);
    }
    if (result < 0) throw createParseException("status time", text, "negative time");
    return new Integer(result);
  }

  private void writeLine(String data) throws ImcadrServerClientException, IOException {
    if (!isConnected()) throw createNotConnectedException();
    out.write(data);
    out.write('\n');
    out.flush();
  }

  private String readLine() throws ImcadrServerClientException, IOException {
    if (!isConnected()) throw createNotConnectedException();
    return in.readLine();
  }

  private String readLine(long msecTimeout) throws ImcadrServerClientException, IOException,
      TimeoutException {
    if (!isConnected()) throw createNotConnectedException();
    return in.readLine(msecTimeout);
  }

  private static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }

  private static Rectangle parseRectangle(String dim, LengthUnit unit, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(dim)) return null;
    String[] parts = dim.split("[xX]");
    if (parts.length != 2) throw createParseException(attribute, dim);
    try {
      double width = Double.parseDouble(parts[0].trim());
      if (width <= 0.0) {
        throw createParseException(attribute, dim, "width <= 0");
      }
      double height = Double.parseDouble(parts[1].trim());
      if (height <= 0.0) {
        throw createParseException(attribute, dim, "height <= 0");
      }
      return new Rectangle(width, height, unit);
    } catch (NumberFormatException e) {
      throw createParseException(attribute, dim);
    }
  }

  private static StatusField parseStatusField(String statusLine) {
    String[] parts = SPACE_RE.split(statusLine, 2);
    if (parts.length == 2) {
      return new StatusField(StatusFieldType.valueFor(parts[0]), parts[1]);
    } else {
      return new StatusField(StatusFieldType.valueFor(parts[0]), null);
    }
  }

  private static Integer[][] parse2dIntegerArray(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] rows = SEMICOLON_RE.split(text, -1);
    String[] firstRow = COMMA_RE.split(rows[0], -1);
    Integer[][] result = new Integer[rows.length][firstRow.length];
    for (int r = 0; r < rows.length; r++) {
      String[] row = COMMA_RE.split(rows[r], -1);
      if (row.length != firstRow.length) {
        throw createParseException(attribute, text, "rows of different length");
      }
      for (int c = 0; c < row.length; c++) {
        try {
          result[r][c] = row[c].isEmpty() ? null : Integer.valueOf(row[c]);
        } catch (NumberFormatException e) {
          throw createParseException(attribute, text, e);
        }
      }
    }
    return result;
  }

  private static Double[][] parse2dDoubleArray(String text, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(text)) return null;
    String[] rows = SEMICOLON_RE.split(text, -1);
    String[] firstRow = COMMA_RE.split(rows[0], -1);
    Double[][] result = new Double[rows.length][firstRow.length];
    for (int r = 0; r < rows.length; r++) {
      String[] row = COMMA_RE.split(rows[r], -1);
      if (row.length != firstRow.length) {
        throw createParseException(attribute, text, "rows of different length");
      }
      for (int c = 0; c < row.length; c++) {
        try {
          result[r][c] = row[c].isEmpty() ? null : Double.valueOf(row[c]);
        } catch (NumberFormatException e) {
          throw createParseException(attribute, text, e);
        }
      }
    }
    return result;
  }

  private static SpaceGroup parseSpaceGroup(String spaceGroup, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(spaceGroup)) return null;
    String capitalizedSpaceGroup = spaceGroup.toUpperCase(Locale.US);
    SpaceGroup result = DATA_TO_SPACE_GROUP_MAP.get(capitalizedSpaceGroup);
    if (result != null) return result;
    throw createParseException(attribute, spaceGroup, "bad value");
  }

  private static UnitCell parseUnitCell(String unitCell, String attribute)
      throws ImcadrServerClientException {
    if (NIL.equals(unitCell)) return null;
    String[] parts = SLASH_RE.split(unitCell, -1);
    if (parts.length != 6) throw createParseException(attribute, unitCell);
    try {
      double[] lengths = new double[3];
      double[] angles = new double[3];
      for (int i = 0; i < lengths.length; i++) {
        lengths[i] = Double.parseDouble(parts[i].trim());
        if (lengths[i] <= 0.0) throw createParseException(attribute, unitCell, "length <= 0");
      }
      for (int i = 0; i < angles.length; i++) {
        angles[i] = Double.parseDouble(parts[i + lengths.length].trim());
        if (angles[i] <= 0.0) throw createParseException(attribute, unitCell, "angle <= 0");
        if (angles[i] >= 180.0) throw createParseException(attribute, unitCell, "angle >= 180");
      }
      return new UnitCell(lengths, angles);
    } catch (NumberFormatException e) {
      throw createParseException(attribute, unitCell);
    }
  }

  private static void ensureWithinRange(Integer[][] values, Integer min, Integer max,
      String text, String attribute) throws ImcadrServerClientException {
    if (values == null) return;
    if (min == null && max == null) return;
    for (int r = 0; r < values.length; r++) {
      for (int c = 0; c < values[0].length; c++) {
        if (values[r][c] == null) continue;
        int each = values[r][c].intValue();
        if ((min != null && each < min.intValue()) || (max != null && each > max.intValue())) {
          throw createParseException(attribute, text, createNotInRangeMessage(min, max) +
              " at (" + c + "," + r + ")");
        }
      }
    }
  }

  private static void ensureWithinRange(Double[][] values, Double min, Double max, String text,
      String attribute) throws ImcadrServerClientException {
    if (values == null) return;
    if (min == null && max == null) return;
    for (int r = 0; r < values.length; r++) {
      for (int c = 0; c < values[0].length; c++) {
        if (values[r][c] == null) continue;
        double each = values[r][c].doubleValue();
        if ((min != null && each < min.doubleValue()) ||
            (max != null && each > max.doubleValue())) {
          throw createParseException(attribute, text, createNotInRangeMessage(min, max) +
              " at (" + c + "," + r + ")");
        }
      }
    }
  }

  private static String createNotInRangeMessage(Number min, Number max) {
    if (min == null && max == null) {
      throw new IllegalArgumentException("min and max are both null");
    }
    if (min == null) {
      return "> " + max;
    } else if (max == null) {
      return "< " + max;
    } else {
      return "not in range [" + min + "," + max + "]";
    }
  }

  private static ImcadrServerClientException createGetException(String attribute, String response) {
    return new ImcadrServerClientException("Failed to query '" + attribute +
        "'; response was '" + response + "'");
  }

  private static ImcadrServerClientException createSetException(String attribute, String response) {
    return new ImcadrServerClientException("Failed to set '" + attribute + "'; response was '" +
        response + "'");
  }

  private static ImcadrServerClientException createCommandException(String command,
      String response) {
    return new ImcadrServerClientException("Failed to command '" + command +
        "'; response was '" + response + "'");
  }

  private static ImcadrServerClientException createParseException(String attribute, String text,
      Throwable cause) {
    return new ImcadrServerClientException("Failed to parse '" + attribute + "' from '" + text +
        "'", cause);
  }

  private static ImcadrServerClientException createParseException(String attribute, String text) {
    return createParseException(attribute, text, null, null);
  }

  private static ImcadrServerClientException createParseException(String attribute, String text,
      String reason) {
    return createParseException(attribute, text, reason, null);
  }

  private static ImcadrServerClientException createParseException(String attribute, String text,
      String reason, Throwable cause) {
    String message = "Failed to parse '" + attribute + "' from '" + text + "'";
    if (reason != null) message += "; " + reason;
    return (cause == null) ? new ImcadrServerClientException(message)
        : new ImcadrServerClientException(message, cause);
  }

  private static ImcadrServerClientException createNotConnectedException() {
    return new ImcadrServerClientException("Not connected");
  }

  private static void closeSilently(Closeable c) {
    if (c == null) return;
    try {
      c.close();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
    } catch (IOException | RuntimeException e) {
      /* IGNORED */
    }
  }

  private static class StatusField {
    private final StatusFieldType type;
    private final String value;

    public StatusField(StatusFieldType type, String value) {
      super();
      this.type = type;
      this.value = value;
    }

    public StatusFieldType type() {
      return type;
    }

    public String value() {
      return value;
    }
  }

  private static enum StatusFieldType {
    PROGRESS("PROGRESS", "status progress"),
    STATE("STATE", "status state"),
    ORIGIN_ZY("ORIGIN_ZY", "status ZY origin"),
    ORIGIN_ZX("ORIGIN_ZX", "status ZX origin"),
    SCORES_ZY("SCORES_ZY", "status ZY center"),
    SCORES_ZX("SCORES_ZX", "status ZX center"),
    SPOT_TOT_ZY("SPOT_TOT_ZY", "status ZY spot totals"),
    SPOT_TOT_ZX("SPOT_TOT_ZX", "status ZX spot totals"),
    IN_RES_TOT_ZY("IN_RES_TOT_ZY", "status ZY in-resolution totals"),
    IN_RES_TOT_ZX("IN_RES_TOT_ZX", "status ZX in-resolution totals"),
    BRAGG_TOT_ZY("BRAGG_TOT_ZY", "status ZY good Bragg candidate totals"),
    BRAGG_TOT_ZX("BRAGG_TOT_ZX", "status ZX good Bragg candidate totals"),
    ICE_RING_TOT_ZY("ICE_RING_TOT_ZY", "status ZY ice ring totals"),
    ICE_RING_TOT_ZX("ICE_RING_TOT_ZX", "status ZX ice ring totals"),
    RES1_ZY("RES1_ZY", "status ZY estimated resolutions 1"),
    RES1_ZX("RES1_ZX", "status ZX estimated resolutions 1"),
    RES2_ZY("RES2_ZY", "status ZY estimated resolutions 2"),
    RES2_ZX("RES2_ZX", "status ZX estimated resolutions 2"),
    INT_SIG_TOT_ZY("INT_SIG_TOT_ZY", "status ZY total integrated signals"),
    INT_SIG_TOT_ZX("INT_SIG_TOT_ZX", "status ZX total integrated signals"),
    MOSAICITY_ZY("MOSAICITY_ZY", "status ZY mosaicities"),
    MOSAICITY_ZX("MOSAICITY_ZX", "status ZX mosaicities"),
    SPACE_GRP_ZY("SPACE_GRP_ZY", "status ZY space groups"),
    SPACE_GRP_ZX("SPACE_GRP_ZX", "status ZX space groups"),
    UNIT_CELL_ZY("UNIT_CELL_ZY", "status ZY unit cells"),
    UNIT_CELL_ZX("UNIT_CELL_ZX", "status ZX unit cells"),
    CENTER_ZY("CENTER_ZY", "status ZY center"),
    CENTER_ZX("CENTER_ZX", "status ZX center"),
    CENTER("CENTER", "status center"),
    TIME("TIME", "status time");

    private static final HashMap<String, StatusFieldType> FIELD_NAME_TO_PROPERTY_MAP;

    static {
      FIELD_NAME_TO_PROPERTY_MAP = new HashMap<String, StatusFieldType>();
      for (StatusFieldType each : StatusFieldType.values()) {
        FIELD_NAME_TO_PROPERTY_MAP.put(each.fieldName(), each);
      }
    }

    private final String fieldName;
    private final String description;

    StatusFieldType(String fieldName, String description) {
      this.fieldName = fieldName;
      this.description = description;
    }

    public String fieldName() {
      return fieldName;
    }

    public String description() {
      return description;
    }

    public static StatusFieldType valueFor(String fieldName) {
      return FIELD_NAME_TO_PROPERTY_MAP.get(fieldName);
    }
  }
}
