package imcacat.imcadr;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import imcacat.imcadr.util.Pair;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.channels.Channels;
import java.nio.channels.Pipe;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * I test <code>ImcadrServer</code>.
 */
public class ImcadrServerTest {
  protected ImcadrServerClient client;
  protected MockImcadrServer server;
  protected Thread serverThread;

  @Before
  public void setUp() throws IOException {
    Pair<ImcadrServerClient, MockImcadrServer> clientAndServer = createClientAndServer();
    client = clientAndServer.getFirst();
    server = clientAndServer.getSecond();
    serverThread = createServerThread(server);
    serverThread.start();
  }

  @After
  public void tearDown() throws InterruptedException {
    if (client != null) {
      client.disconnect();
    }
    if (serverThread != null) {
      serverThread.interrupt();
      serverThread.join();
    }
    MockImcadrServerSession.setScoreCreationRange(null);
  }

  @Test
  public void testImagesDirectory01() throws Exception {
    assertNull(client.imagesDirectory());
    File tmpDir = getTmpDir();
    client.imagesDirectory(tmpDir);
    assertEquals(tmpDir, client.imagesDirectory());
    client.imagesDirectory(null);
    assertNull(client.imagesDirectory());
  }

  @Test
  public void testExposureTime01() throws Exception {
    assertNull(client.exposureTime());
    Double expTime = new Double(0.251);
    client.exposureTime(expTime);
    assertEquals(expTime, client.exposureTime());
    client.exposureTime(null);
    assertNull(client.exposureTime());
  }

  @Test
  public void testCellSize01() throws Exception {
    assertNull(client.cellSize());
    Rectangle cellSize = Rectangle.um(57.0, 23.0);
    client.cellSize(cellSize);
    assertEquals(cellSize, client.cellSize());
    client.cellSize(null);
    assertNull(client.cellSize());
  }

  @Test
  public void testZySize01() throws Exception {
    assertNull(client.zySize());
    Rectangle zySize = new Rectangle(7.0, 5.0, LengthUnit.CELL);
    client.zySize(zySize);
    assertEquals(zySize, client.zySize());
    client.zySize(null);
    assertNull(client.zySize());
  }

  @Test
  public void testZxSize01() throws Exception {
    assertNull(client.zxSize());
    Rectangle zxSize = new Rectangle(7.0, 5.0, LengthUnit.CELL);
    client.zxSize(zxSize);
    assertEquals(zxSize, client.zxSize());
    client.zxSize(null);
    assertNull(client.zxSize());
  }

  @Test
  public void testChasingEnabled01() throws Exception {
    assertNull(client.chasingEnabled());
    client.chasingEnabled(Boolean.FALSE);
    assertEquals(Boolean.FALSE, client.chasingEnabled());
    client.chasingEnabled(Boolean.TRUE);
    assertEquals(Boolean.TRUE, client.chasingEnabled());
    client.chasingEnabled(null);
    assertNull(client.chasingEnabled());
  }

  @Test
  public void testAnalysisPackage01() throws Exception {
    assertNull(client.analysisPackage());
    AnalysisPackage analysisPackage = AnalysisPackage.DTREK;
    client.analysisPackage(analysisPackage);
    assertEquals(analysisPackage, client.analysisPackage());
    client.analysisPackage(null);
    assertNull(client.analysisPackage());
  }

  @Test
  public void testWeakDiffracter01() throws Exception {
    assertNull(client.weakDiffracter());
    client.weakDiffracter(Boolean.FALSE);
    assertEquals(Boolean.FALSE, client.weakDiffracter());
    client.weakDiffracter(Boolean.TRUE);
    assertEquals(Boolean.TRUE, client.weakDiffracter());
    client.weakDiffracter(null);
    assertNull(client.weakDiffracter());
  }

  @Test
  public void testHighResolutionCutoff01() throws Exception {
    assertNull(client.highResolutionCutoff());
    Double highResolutionCutoff = new Double(2.1);
    client.highResolutionCutoff(highResolutionCutoff);
    assertEquals(highResolutionCutoff, client.highResolutionCutoff());
    client.highResolutionCutoff(null);
    assertNull(client.highResolutionCutoff());
  }

  @Test
  public void testLowResolutionCutoff01() throws Exception {
    assertNull(client.lowResolutionCutoff());
    Double lowResolutionCutoff = Double.valueOf(2.1);
    client.lowResolutionCutoff(lowResolutionCutoff);
    assertEquals(lowResolutionCutoff, client.lowResolutionCutoff());
    client.lowResolutionCutoff(null);
    assertNull(client.lowResolutionCutoff());
  }

  @Test
  public void testMinimumSpotArea01() throws Exception {
    assertNull(client.minimumSpotArea());
    Integer minimumSpotArea = Integer.valueOf(2);
    client.minimumSpotArea(minimumSpotArea);
    assertEquals(minimumSpotArea, client.minimumSpotArea());
    client.minimumSpotArea(null);
    assertNull(client.minimumSpotArea());
  }

  @Test
  public void testMinimumSignalHeight01() throws Exception {
    assertNull(client.minimumSignalHeight());
    Double minimumSignalHeight = Double.valueOf(2.1);
    client.minimumSignalHeight(minimumSignalHeight);
    assertEquals(minimumSignalHeight, client.minimumSignalHeight());
    client.minimumSignalHeight(null);
    assertNull(client.minimumSignalHeight());
  }

  /* Before running. */
  @Test
  public void testStat01() throws Exception {
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertNull(status.getMessage());
    assertEquals(0, status.getProgress());
    assertNull(status.getOriginZy());
    assertNull(status.getOriginZx());
    assertNull(status.getScoresZy());
    assertNull(status.getScoresZx());
    assertNull(status.getSpotTotalsZx());
    assertNull(status.getSpotTotalsZy());
    assertNull(status.getInResolutionTotalsZx());
    assertNull(status.getInResolutionTotalsZy());
    assertNull(status.getGoodBraggTotalsZx());
    assertNull(status.getGoodBraggTotalsZy());
    assertNull(status.getIceRingTotalsZx());
    assertNull(status.getIceRingTotalsZy());
    assertNull(status.getEstimatedResolutions1Zx());
    assertNull(status.getEstimatedResolutions1Zy());
    assertNull(status.getEstimatedResolutions2Zx());
    assertNull(status.getEstimatedResolutions2Zy());
    assertNull(status.getTotalIntegratedSignalsZy());
    assertNull(status.getTotalIntegratedSignalsZx());
    assertNull(status.getMosaicitiesZy());
    assertNull(status.getMosaicitiesZx());
    assertNull(status.getSpaceGroupsZy());
    assertNull(status.getSpaceGroupsZx());
    assertNull(status.getUnitCellsZy());
    assertNull(status.getUnitCellsZx());
    assertNull(status.getCenterZy());
    assertNull(status.getCenterZx());
    assertNull(status.getCenter());
    assertNull(status.getTime());
  }

  /* ZY raster. */
  @Test
  public void testStat02() throws Exception {
    MockImcadrServerSession.setScoreCreationRange(new Range<Integer>(0, 100));
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    assertEquals(Position2d.mm(-0.14, -0.045), status.getOriginZy());
    assertNull(status.getOriginZx());
    /* @formatter:off */
    Integer[][] zyScores = new Integer[][] {
      { 0,  0,   0, 0 },
      { 0, 66, 100, 0 },
      { 0,  0,   0, 0 }
    };
    /* @formatter:on */
    assertArrayEquals(zyScores, status.getScoresZy());
    assertNull(status.getScoresZx());
    assertEquals(Position2d.mm(0.007, 0.0), status.getCenterZy());
    assertNull(status.getCenterZx());
    assertNull(status.getCenter());
    assertNotNull(status.getTime());
  }

  /* Bigger ZY raster. */
  @Test
  public void testStat03() throws Exception {
    MockImcadrServerSession.setScoreCreationRange(new Range<Integer>(0, 100));
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(5.0, 5.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    assertEquals(Position2d.mm(-0.175, -0.075), status.getOriginZy());
    assertNull(status.getOriginZx());
    /* @formatter:off */
    Integer[][] zyScores = new Integer[][] {
      { 0,  0,  0,   0, 0 },
      { 0, 33, 50,  66, 0 },
      { 0, 50, 66,  83, 0 },
      { 0, 66, 83, 100, 0 },
      { 0,  0,  0,   0, 0 }
    };
    /* @formatter:on */
    assertArrayEquals(zyScores, status.getScoresZy());
    assertNull(status.getScoresZx());
    assertEquals(Position2d.mm(0.039, 0.017), status.getCenterZy());
    assertNull(status.getCenterZx());
    assertNull(status.getCenter());
    assertNotNull(status.getTime());
  }

  /* ZX raster. */
  @Test
  public void testStat04() throws Exception {
    MockImcadrServerSession.setScoreCreationRange(new Range<Integer>(0, 100));
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    assertNull(status.getOriginZy());
    assertEquals(Position2d.mm(-0.105, -0.075), status.getOriginZx());
    /* @formatter:off */
    Integer[][] zxScores = new Integer[][] {
      { 0,   0, 0 },
      { 0,  50, 0 },
      { 0,  75, 0 },
      { 0, 100, 0 },
      { 0,   0, 0 }
    };
    /* @formatter:on */
    assertNull(status.getScoresZy());
    assertArrayEquals(zxScores, status.getScoresZx());
    assertNull(status.getCenterZy());
    assertEquals(Position2d.mm(0.0, 0.017), status.getCenterZx());
    assertNull(status.getCenter());
    assertNotNull(status.getTime());
  }

  /* ZY and ZX raster. */
  @Test
  public void testStat05() throws Exception {
    MockImcadrServerSession.setScoreCreationRange(new Range<Integer>(0, 100));
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    assertEquals(Position2d.mm(-0.14, -0.045), status.getOriginZy());
    assertEquals(Position2d.mm(-0.098, -0.075), status.getOriginZx());
    /* @formatter:off */
    Integer[][] zyScores = new Integer[][] {
      { 0,  0,   0, 0 },
      { 0, 66, 100, 0 },
      { 0,  0,   0, 0 }
    };
    /* @formatter:on */
    assertArrayEquals(zyScores, status.getScoresZy());
    /* @formatter:off */
    Integer[][] zxScores = new Integer[][] {
      { 0,   0, 0 },
      { 0,  50, 0 },
      { 0,  75, 0 },
      { 0, 100, 0 },
      { 0,   0, 0 }
    };
    /* @formatter:on */
    assertArrayEquals(zxScores, status.getScoresZx());
    assertEquals(Position2d.mm(0.007, 0.0), status.getCenterZy());
    assertEquals(Position2d.mm(0.007, 0.017), status.getCenterZx());
    assertEquals(Position3d.mm(0.017, 0.0, 0.007), status.getCenter());
    assertNotNull(status.getTime());
  }

  /* ZY raster using LABELIT. */
  @Test
  public void testStat06() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.analysisPackage(AnalysisPackage.LABELIT);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    /* @formatter:off */
    Integer[][] zySpotTotals = new Integer[][] {
      { 1, 2, 3, 4 },
      { 2, 3, 4, 5 },
      { 3, 4, 5, 6 }
    };
    /* @formatter:on */
    assertArrayEquals(zySpotTotals, status.getSpotTotalsZy());
    assertNull(status.getSpotTotalsZx());
    /* @formatter:off */
    Integer[][] zyInResolutionTotals = new Integer[][] {
      { 2, 3, 4, 5 },
      { 3, 4, 5, 6 },
      { 4, 5, 6, 7 }
    };
    /* @formatter:on */
    assertArrayEquals(zyInResolutionTotals, status.getInResolutionTotalsZy());
    assertNull(status.getInResolutionTotalsZx());
    /* @formatter:off */
    Integer[][] zyGoodBraggTotals = new Integer[][] {
      { 3, 4, 5, 6 },
      { 4, 5, 6, 7 },
      { 5, 6, 7, 8 }
    };
    /* @formatter:on */
    assertArrayEquals(zyGoodBraggTotals, status.getGoodBraggTotalsZy());
    assertNull(status.getGoodBraggTotalsZx());
    /* @formatter:off */
    Integer[][] zyIceRingTotals = new Integer[][] {
      { 4, 5, 6, 7 },
      { 5, 6, 7, 8 },
      { 6, 7, 8, 9 }
    };
    /* @formatter:on */
    assertArrayEquals(zyIceRingTotals, status.getIceRingTotalsZy());
    assertNull(status.getIceRingTotalsZx());
    /* @formatter:off */
    Double[][] zyEstimatedResolutions1 = new Double[][] {
      {  0.50, 20.20, 39.90, 59.60},
      { 20.20, 39.90, 59.60, 79.30},
      { 39.90, 59.60, 79.30, 99.00}
    };
    /* @formatter:on */
    assertArrayEquals(zyEstimatedResolutions1, status.getEstimatedResolutions1Zy());
    assertNull(status.getEstimatedResolutions1Zx());
    assertArrayEquals(new Double[3][4], status.getEstimatedResolutions2Zy());
    assertNull(status.getEstimatedResolutions2Zx());
    /* @formatter:off */
    Integer[][] zyTotalIntegratedSignal = new Integer[][] {
      { 5, 6, 7,  8 },
      { 6, 7, 8,  9 },
      { 7, 8, 9, 10 }
    };
    /* @formatter:on */
    assertArrayEquals(zyTotalIntegratedSignal, status.getTotalIntegratedSignalsZy());
    assertNull(status.getTotalIntegratedSignalsZx());
    assertArrayEquals(new Double[3][4], status.getMosaicitiesZy());
    assertNull(status.getMosaicitiesZx());
    assertArrayEquals(new Double[3][4], status.getSpaceGroupsZy());
    assertNull(status.getSpaceGroupsZx());
    assertArrayEquals(new Double[3][4], status.getUnitCellsZy());
    assertNull(status.getUnitCellsZx());
  }

  /* ZY raster using d*TREK. */
  @Test
  public void testStat07() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.analysisPackage(AnalysisPackage.DTREK);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.start();
    while (client.status().isRunningOrPaused()) {
      Thread.sleep(10);
    }
    ImcadrServerStatus status = client.status();
    assertEquals(ImcadrServerState.IDLE, status.getState());
    assertEquals(100, status.getProgress());
    assertArrayEquals(new Integer[3][4], status.getSpotTotalsZy());
    assertNull(status.getSpotTotalsZx());
    assertArrayEquals(new Integer[3][4], status.getInResolutionTotalsZy());
    assertNull(status.getInResolutionTotalsZx());
    assertArrayEquals(new Integer[3][4], status.getGoodBraggTotalsZy());
    assertNull(status.getGoodBraggTotalsZx());
    assertArrayEquals(new Integer[3][4], status.getIceRingTotalsZy());
    assertNull(status.getIceRingTotalsZx());
    /* @formatter:off */
    Double[][] zyEstimatedResolutions1 = new Double[][] {
      {  0.50, 20.20, 39.90, 59.60},
      { 20.20, 39.90, 59.60, 79.30},
      { 39.90, 59.60, 79.30, 99.00}
    };
    /* @formatter:on */
    assertArrayEquals(zyEstimatedResolutions1, status.getEstimatedResolutions1Zy());
    assertNull(status.getEstimatedResolutions1Zx());
    assertArrayEquals(new Double[3][4], status.getEstimatedResolutions2Zy());
    assertNull(status.getEstimatedResolutions2Zx());
    assertArrayEquals(new Integer[3][4], status.getTotalIntegratedSignalsZy());
    assertNull(status.getTotalIntegratedSignalsZx());
    /* @formatter:off */
    Double[][] mosaicities = new Double[][] {
      { 0.00, 2.00, 4.00, 6.00},
      { 2.00, 4.00, 6.00, 8.00},
      { 4.00, 6.00, 8.00, 10.00}
    };
    /* @formatter:on */
    assertArrayEquals(mosaicities, status.getMosaicitiesZy());
    assertNull(status.getMosaicitiesZx());
    /* @formatter:off */
    SpaceGroup[][] groups = new SpaceGroup[][] {
      { SpaceGroup.P1, SpaceGroup.P2,   SpaceGroup.P4,   SpaceGroup.C222},
      { SpaceGroup.P2, SpaceGroup.P4,   SpaceGroup.C222, SpaceGroup.I4},
      { SpaceGroup.P4, SpaceGroup.C222, SpaceGroup.I4,   SpaceGroup.F222}
    };
    /* @formatter:on */
    assertArrayEquals(groups, status.getSpaceGroupsZy());
    assertNull(status.getSpaceGroupsZx());
    UnitCell[][] actual = status.getUnitCellsZy();
    for (int r = 0; r < actual.length; r++) {
      for (int c = 0; c < actual[0].length; c++) {
        assertNotNull("col=" + c + ",r=" + r, actual[r][c]);
      }
    }
    assertNull(status.getUnitCellsZx());
  }

  @Test
  public void testErrorDetail01() throws Exception {
    Logger logger = Logger.getLogger(BaseRasterizer.class.getName());
    Level originalLoggerLevel = logger.getLevel();
    try {
      logger.setLevel(Level.OFF);
      final String exceptionMessage = "Bad motivator";
      MockBeamlineDriver driver = (MockBeamlineDriver)server.getModel().getBeamlineDriver();
      driver.setRealisticExposureTimeEnabled(true);
      client.flyScanEnabled(Boolean.FALSE);
      client.imagesDirectory(getTmpDir());
      client.exposureTime(new Double(0.1));
      client.cellSize(Rectangle.um(70, 30));
      client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
      client.start();
      for (;;) {
        List<MockRasterizer2d> rasterizers = MockRasterizer2d.getInstances();
        if (rasterizers.isEmpty()) {
          Thread.sleep(10);
        } else {
          for (MockRasterizer2d each : rasterizers) {
            each.setAnalyzerCreationException(new RuntimeException(exceptionMessage));
          }
          break;
        }
      }
      while (client.status().isRunningOrPaused()) {
        Thread.sleep(10);
      }
      ImcadrServerStatus status = client.status();
      assertEquals(ImcadrServerState.ERROR, status.getState());
      String errorDetail = client.errorDetail();
      assertNotNull(errorDetail);
      assertTrue(errorDetail.contains(exceptionMessage));
    } finally {
      logger.setLevel(originalLoggerLevel);
    }
  }

  /* Error detail with line that starts with a dot. */
  @Test
  public void testErrorDetail02() throws Exception {
    Logger logger = Logger.getLogger(BaseRasterizer.class.getName());
    Level originalLoggerLevel = logger.getLevel();
    try {
      logger.setLevel(Level.OFF);
      final String exceptionMessage = "Bad\n.motivator";
      MockBeamlineDriver driver = (MockBeamlineDriver)server.getModel().getBeamlineDriver();
      driver.setRealisticExposureTimeEnabled(true);
      client.flyScanEnabled(Boolean.FALSE);
      client.imagesDirectory(getTmpDir());
      client.exposureTime(new Double(0.1));
      client.cellSize(Rectangle.um(70, 30));
      client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
      client.start();
      for (;;) {
        List<MockRasterizer2d> rasterizers = MockRasterizer2d.getInstances();
        if (rasterizers.isEmpty()) {
          Thread.sleep(10);
        } else {
          for (MockRasterizer2d each : rasterizers) {
            each.setAnalyzerCreationException(new RuntimeException(exceptionMessage));
          }
          break;
        }
      }
      while (client.status().isRunningOrPaused()) {
        Thread.sleep(10);
      }
      ImcadrServerStatus status = client.status();
      assertEquals(ImcadrServerState.ERROR, status.getState());
      String errorDetail = client.errorDetail();
      assertNotNull(errorDetail);
      assertTrue(errorDetail.contains(exceptionMessage));
    } finally {
      logger.setLevel(originalLoggerLevel);
    }
  }

  /* OK to start. */
  @Test
  public void testStart01() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    assertEquals(ImcadrServerState.RUNNING, client.status().getState());
  }

  /* Not OK to start due to missing exposure time. */
  @Test(expected = ImcadrServerClientException.class)
  public void testStart02() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
  }

  /* Not OK to start when already running. */
  @Test(expected = ImcadrServerClientException.class)
  public void testStart03() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    client.start();
  }

  /* OK to stop. */
  @Test
  public void testStop01() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    client.stop();
    assertEquals(ImcadrServerState.IDLE, client.status().getState());
  }

  /* OK to stop even if not running. */
  @Test
  public void testStop02() throws Exception {
    client.stop();
    assertEquals(ImcadrServerState.IDLE, client.status().getState());
  }

  /* OK to reset. */
  @Test
  public void testReset01() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.reset();
    assertNull(client.imagesDirectory());
    assertNull(client.exposureTime());
    assertNull(client.cellSize());
    assertNull(client.zySize());
    assertNull(client.zxSize());
  }

  /* Not OK to reset when running. */
  @Test(expected = ImcadrServerClientException.class)
  public void testReset02() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    client.reset();
  }

  /* OK to quit. */
  @Test
  public void testQuit01() throws Exception {
    client.quit();
  }

  /* Not OK to quit if running. */
  @Test(expected = ImcadrServerClientException.class)
  public void testQuit02() throws Exception {
    client.flyScanEnabled(Boolean.FALSE);
    client.imagesDirectory(getTmpDir());
    client.exposureTime(new Double(0.25));
    client.cellSize(Rectangle.um(70, 30));
    client.zySize(new Rectangle(4.0, 3.0, LengthUnit.CELL));
    client.zxSize(new Rectangle(3.0, 5.0, LengthUnit.CELL));
    client.start();
    client.quit();
  }

  protected static Pair<ImcadrServerClient, MockImcadrServer> createClientAndServer()
      throws IOException {
    Pipe clientToServerPipe = Pipe.open();
    Pipe serverToClientPipe = Pipe.open();
    InputStream clientIn = Channels.newInputStream(serverToClientPipe.source());
    OutputStream clientOut = Channels.newOutputStream(clientToServerPipe.sink());
    InputStream serverIn = Channels.newInputStream(clientToServerPipe.source());
    OutputStream serverOut = Channels.newOutputStream(serverToClientPipe.sink());
    ImcadrServerClient client = new ImcadrServerClient();
    client.connect(clientIn, clientOut);
    MockImcadrServer server = new MockImcadrServer(serverIn, serverOut);
    return new Pair<ImcadrServerClient, MockImcadrServer>(client, server);
  }

  protected static Thread createServerThread(final ImcadrServer server) {
    return new Thread() {
      @Override
      public void run() {
        server.run();
      }
    };
  }

  protected static File getTmpDir() {
    return new File(System.getProperty("java.io.tmpdir"));
  }
}
