package imcacat.imcadr;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

import java.io.ByteArrayInputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.Charset;

import org.junit.Test;

/**
 * I test <code>LabelitDistSignalStrengthParser</code>.
 */
public class LabelitDistSignalStrengthParserTest {
  private static final Charset UTF8 = Charset.forName("UTF-8");

  /* Typical. */
  @Test
  public void testParse01() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDistSignalStrengthOutput01());
      LabelitDistSignalStrengthParser parser = new LabelitDistSignalStrengthParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(521, result.getScore().intValue());
    assertEquals(561, result.getSpotTotal().intValue());
    assertEquals(529, result.getInResolutionTotal().intValue());
    assertEquals(521, result.getGoodBraggTotal().intValue());
    assertEquals(0, result.getIceRingTotal().intValue());
    assertEquals(1.58, result.getEstimatedResolution1().doubleValue(), 0.001);
    assertEquals(1.57, result.getEstimatedResolution2().doubleValue(), 0.001);
    assertEquals(1088998, result.getTotalIntegratedSignal().intValue());
  }

  /* "NONE" for estimated resolution 1. */
  @Test
  public void testParse02() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDistSignalStrengthOutput02());
      LabelitDistSignalStrengthParser parser = new LabelitDistSignalStrengthParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(521, result.getScore().intValue());
    assertEquals(561, result.getSpotTotal().intValue());
    assertEquals(529, result.getInResolutionTotal().intValue());
    assertEquals(521, result.getGoodBraggTotal().intValue());
    assertEquals(0, result.getIceRingTotal().intValue());
    assertNull(result.getEstimatedResolution1());
    assertEquals(1.57, result.getEstimatedResolution2().doubleValue(), 0.001);
    assertEquals(1088998, result.getTotalIntegratedSignal().intValue());
  }

  /* Output for a saturated image. */
  @Test
  public void testParse03() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDistSignalStrengthOutput03());
      LabelitDistSignalStrengthParser parser = new LabelitDistSignalStrengthParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(0, result.getScore().intValue());
    assertEquals(0, result.getSpotTotal().intValue());
    assertEquals(0, result.getInResolutionTotal().intValue());
    assertEquals(0, result.getGoodBraggTotal().intValue());
    assertEquals(1, result.getIceRingTotal().intValue());
    assertEquals(10.0, result.getEstimatedResolution1().doubleValue(), 0.001);
    assertNull(result.getEstimatedResolution2());
    assertNull(result.getTotalIntegratedSignal());
  }

  /* Relatively weak diffraction. */
  @Test
  public void testParse04() {
    ImageAnalysisResult result = null;
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(createDistSignalStrengthOutput04());
      LabelitDistSignalStrengthParser parser = new LabelitDistSignalStrengthParser(stream);
      result = parser.run();
    } finally {
      closeSilently(stream);
    }
    assertEquals(6, result.getScore().intValue());
    assertEquals(12, result.getSpotTotal().intValue());
    assertEquals(12, result.getInResolutionTotal().intValue());
    assertEquals(6, result.getGoodBraggTotal().intValue());
    assertEquals(0, result.getIceRingTotal().intValue());
    assertEquals(2.48, result.getEstimatedResolution1().doubleValue(), 0.001);
    assertNull(result.getEstimatedResolution2());
    assertEquals(104, result.getTotalIntegratedSignal().intValue());
  }

  /* Typical. */
  private static byte[] createDistSignalStrengthOutput01() {
    /* @formatter:off */
    String[] output = new String[] {
      "distl.signal_strength: characterization of candidate Bragg spots",
      "#Parameters used:",
      "#phil __ON__",
      "",
      "distl {",
      "  image = \"image_0003.cbf\"",
      "  res {",
      "    outer = None",
      "    inner = None",
      "  }",
      "  verbose = False",
      "}",
      "distl {",
      "  minimum_spot_area = None",
      "  minimum_signal_height = None",
      "  pdf_output = None",
      "  port = 8125",
      "  processors = 1",
      "}",
      "",
      "#phil __OFF__",
      "",
      "",
      "                     File : image_0003.cbf",
      "               Spot Total :    561",
      "      In-Resolution Total :    529",
      "    Good Bragg Candidates :    521",
      "                Ice Rings :      0",
      "      Method 1 Resolution :   1.58",
      "      Method 2 Resolution :   1.57",
      "        Maximum unit cell :   79.5",
      "%Saturation, Top 50 Peaks :   0.50",
      "In-Resolution Ovrld Spots :      0",
      "",
      "Bin population cutoff for method 2 resolution: 20%",
      "",
      "Number of focus spots on image #3 within the input resolution range: 529",
      "Total integrated signal, pixel-ADC units above local background (just the good Bragg " +
          "candidates) 1088998",
      "Signals range from 94.1 to 40042.6 with mean integrated signal 2285.2",
      "Saturations range from 0.0% to 2.2% with mean saturation 0.1%"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /* "NONE" for estimated resolution 1. */
  private static byte[] createDistSignalStrengthOutput02() {
    /* @formatter:off */
    String[] output = new String[] {
      "distl.signal_strength: characterization of candidate Bragg spots",
      "#Parameters used:",
      "#phil __ON__",
      "",
      "distl {",
      "  image = \"image_0003.cbf\"",
      "  res {",
      "    outer = None",
      "    inner = None",
      "  }",
      "  verbose = False",
      "}",
      "distl {",
      "  minimum_spot_area = None",
      "  minimum_signal_height = None",
      "  pdf_output = None",
      "  port = 8125",
      "  processors = 1",
      "}",
      "",
      "#phil __OFF__",
      "",
      "",
      "                     File : image_0003.cbf",
      "               Spot Total :    561",
      "      In-Resolution Total :    529",
      "    Good Bragg Candidates :    521",
      "                Ice Rings :      0",
      "      Method 1 Resolution :   NONE",
      "      Method 2 Resolution :   1.57",
      "        Maximum unit cell :   79.5",
      "%Saturation, Top 50 Peaks :   0.50",
      "In-Resolution Ovrld Spots :      0",
      "",
      "Bin population cutoff for method 2 resolution: 20%",
      "",
      "Number of focus spots on image #3 within the input resolution range: 529",
      "Total integrated signal, pixel-ADC units above local background (just the good Bragg " +
          "candidates) 1088998",
      "Signals range from 94.1 to 40042.6 with mean integrated signal 2285.2",
      "Saturations range from 0.0% to 2.2% with mean saturation 0.1%"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /* Saturated image. */
  private static byte[] createDistSignalStrengthOutput03() {
    /* @formatter:off */
    String[] output = new String[] {
      "distl.signal_strength: characterization of candidate Bragg spots",
      "#Parameters used:",
      "#phil __ON__",
      "",
      "distl {",
      "  image = \"cell_zy_04_05.cbf\"",
      "  res {",
      "    outer = None",
      "    inner = None",
      "  }",
      "  verbose = False",
      "  bins {",
      "    verbose = False",
      "    N = 20",
      "    corner = True",
      "  }",
      "}",
      "autoindex_override_beam = None",
      "autoindex_override_distance = None",
      "autoindex_override_wavelength = None",
      "autoindex_override_twotheta = None",
      "autoindex_override_deltaphi = None",
      "image_specific_osc_start = None",
      "codecamp {",
      "}",
      "pdf_output {",
      "}",
      "distl {",
      "  minimum_spot_area = 10",
      "  minimum_signal_height = None",
      "  minimum_spot_height = None",
      "  spot_area_maximum_factor = None",
      "  peak_intensity_maximum_factor = None",
      "  compactness_filter = False",
      "  scanbox_windows = 101 51 51",
      "  peripheral_margin = 20",
      "  pdf_output = None",
      "  port = 8125",
      "  processors = 1",
      "}",
      "",
      "#phil __OFF__",
      "",
      "",
      "                     File : cell_zy_##_##.cbf",
      "               Spot Total :      0",
      "               Remove Ice : None",
      "      In-Resolution Total :      0",
      "    Good Bragg Candidates :      0",
      "                Ice Rings :      1",
      "      Method 1 Resolution :  10.00",
      "      Method 2 Resolution : None",
      "        Maximum unit cell : None",
      "<Spot model eccentricity> : None",
      " %Saturation, Top 0 Peaks :   0.00",
      "In-Resolution Ovrld Spots :      0",
      "",
      "Bin population cutoff for method 2 resolution: 20%",
      "",
      "Number of focus spots on image #4005 within the input resolution range: 0"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return result.toString().getBytes(UTF8);
  }

  /* Relatively weak diffraction. */
  private static byte[] createDistSignalStrengthOutput04() {
    /* @formatter:off */
    String[] output = new String[] {
      "distl.signal_strength: characterization of candidate Bragg spots",
      "#Parameters used:",
      "#phil __ON__",
      "",
      "distl {",
      "  image = \"cell_zy_06_09.cbf\"",
      "  res {",
      "    outer = None",
      "    inner = None",
      "  }",
      "  verbose = False",
      "  bins {",
      "    verbose = False",
      "    N = 20",
      "    corner = True",
      "  }",
      "}",
      "autoindex_override_beam = None",
      "autoindex_override_distance = None",
      "autoindex_override_wavelength = None",
      "autoindex_override_twotheta = None",
      "autoindex_override_deltaphi = None",
      "image_specific_osc_start = None",
      "codecamp {",
      "}",
      "pdf_output {",
      "}",
      "distl {",
      "  minimum_spot_area = 10",
      "  minimum_signal_height = None",
      "  minimum_spot_height = None",
      "  spot_area_maximum_factor = None",
      "  peak_intensity_maximum_factor = None",
      "  compactness_filter = False",
      "  scanbox_windows = 101 51 51",
      "  peripheral_margin = 20",
      "  pdf_output = None",
      "  port = 8125",
      "  processors = 1",
      "}",
      "",
      "#phil __OFF__",
      "",
      "",
      "                     File : cell_zy_##_##.cbf",
      "               Spot Total :     12",
      "               Remove Ice : None",
      "      In-Resolution Total :     12",
      "    Good Bragg Candidates :      6",
      "                Ice Rings :      0",
      "      Method 1 Resolution :   2.48",
      "      Method 2 Resolution : None",
      "        Maximum unit cell :  523.2",
      "<Spot model eccentricity> :   0.853 ",
      "%Saturation, Top 12 Peaks :   0.00",
      "In-Resolution Ovrld Spots :      0",
      "",
      "Bin population cutoff for method 2 resolution: 20%",
      "",
      "Number of focus spots on image #6009 within the input resolution range: 6",
      "Total integrated signal, pixel-ADC units above local background (just the good Bragg " +
          "candidates) 104",
      "Signals range from 12.5 to 26.9 with mean integrated signal 17.5",
      "Saturations range from 0.0% to 0.0% with mean saturation 0.0%"
    };
    /* @formatter:on */
    StringBuilder result = new StringBuilder();
    for (String each : output) {
      result.append(each);
      result.append("\n");
    }
    return result.toString().getBytes(UTF8);
  }

  private static void closeSilently(Closeable closeable) {
    if (closeable == null) return;
    try {
      closeable.close();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
    } catch (IOException | RuntimeException e) {
      /* IGNORED */
    }
  }
}
