package imcacat.imcadr;

import gov.aps.jca.TimeoutException;
import imcacat.imcadr.util.DurationUtils;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * I am a Mock object of <tt>BeamlineDriver</tt>.
 */
public class MockBeamlineDriver implements BeamlineDriver {
  protected long exposureTime = 0;
  protected boolean realisticExposureTimeEnabled = false;
  protected ExperimentState experimentState = ExperimentState.IDLE;
  protected Map<BeamlineAxis, Double> minVelo = new EnumMap<>(BeamlineAxis.class);
  protected Map<BeamlineAxis, Double> maxVelo = new EnumMap<>(BeamlineAxis.class);
  protected Map<BeamlineAxis, Double> velo = new EnumMap<>(BeamlineAxis.class);
  protected Map<BeamlineAxis, Double> accelTime = new EnumMap<>(BeamlineAxis.class);
  protected LinkedList<Double> phiRbvs = new LinkedList<Double>();
  protected LinkedList<Double> xRbvs = new LinkedList<Double>();
  protected LinkedList<Double> yRbvs = new LinkedList<Double>();
  protected LinkedList<Double> zRbvs = new LinkedList<Double>();
  protected List<File> imagesTaken = new ArrayList<>();

  public MockBeamlineDriver() {
    for (BeamlineAxis each : BeamlineAxis.values()) {
      minVelo.put(each, 0.0);
      maxVelo.put(each, 10.0);
      velo.put(each,  5.0);
      accelTime.put(each, 0.25);
    }
    phiRbvs.add(0.0);
    xRbvs.add(0.0);
    yRbvs.add(0.0);
    zRbvs.add(0.0);
  }

  public void release() {
  }

  public boolean isRealisticExposureTimeEnabled() {
    return realisticExposureTimeEnabled;
  }

  public void setRealisticExposureTimeEnabled(boolean enabled) {
    realisticExposureTimeEnabled = enabled;
  }

  public String getImageFileNameExtension() {
    return ".cbf";
  }

  public void disableAutobp() throws InterruptedException, DriverException {
  }

  public void enableAutobp() throws InterruptedException, DriverException {
  }

  public double getDetectorReadoutTime() {
    return 0.0023;
  }

  public double getMinVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    return minVelo.get(axis);
  }

  public double getMaxVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    return maxVelo.get(axis);
  }

  public double getVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    return velo.get(axis);
  }

  public void setVelo(BeamlineAxis axis, double desiredVelo)
      throws InterruptedException, DriverException {
    velo.put(axis, desiredVelo);
  }

  public double getAccelTime(BeamlineAxis axis) throws InterruptedException, DriverException {
    return accelTime.get(axis);
  }

  public double phiRbv() throws InterruptedException, DriverException {
    return phiRbvs.getLast();
  }

  public double[] phiRbvHistory() {
    return doubleArray(phiRbvs);
  }

  public void movePhi(double desiredPosition) throws InterruptedException, DriverException {
    phiRbvs.addLast(desiredPosition);
  }

  public double xRbv() throws InterruptedException, DriverException {
    return xRbvs.getLast();
  }

  public double[] xRbvHistory() {
    return doubleArray(xRbvs);
  }

  public void moveX(double desiredPosition) throws InterruptedException, DriverException {
    xRbvs.addLast(desiredPosition);
  }

  public double yRbv() throws InterruptedException, DriverException {
    return yRbvs.getLast();
  }

  public double[] yRbvHistory() {
    return doubleArray(yRbvs);
  }

  public void moveY(double desiredPosition) throws InterruptedException, DriverException {
    yRbvs.addLast(desiredPosition);
  }

  public double zRbv() throws InterruptedException, DriverException {
    return zRbvs.getLast();
  }

  public double[] zRbvHistory() {
    return doubleArray(zRbvs);
  }

  public void moveZ(double desiredPosition) throws InterruptedException, DriverException {
    zRbvs.addLast(desiredPosition);
  }

  public double rbv(BeamlineAxis axis) throws InterruptedException, DriverException {
    if (BeamlineAxis.PHI.equals(axis)) return phiRbv();
    if (BeamlineAxis.X.equals(axis)) return xRbv();
    if (BeamlineAxis.Y.equals(axis)) return yRbv();
    if (BeamlineAxis.Z.equals(axis)) return zRbv();
    throw new BugException();
  }

  public void move(BeamlineAxis axis, double desiredPosition) throws InterruptedException,
      DriverException {
    if (BeamlineAxis.PHI.equals(axis)) movePhi(desiredPosition);
    else if (BeamlineAxis.X.equals(axis)) moveX(desiredPosition);
    else if (BeamlineAxis.Y.equals(axis)) moveY(desiredPosition);
    else if (BeamlineAxis.Z.equals(axis)) moveZ(desiredPosition);
    else throw new BugException();
  }

  public void enterCollectDataExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException {
    /*
     * Enter something that could be interrupted right away if the current
     * thread has already been interrupted to partially simulate what would
     * really happen for a CA put.
     */
    Thread.sleep(1L);
    experimentState = ExperimentState.COLLECT_DATA;
  }

  public void enterIdleExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException {
    /*
     * Enter something that could be interrupted right away if the current
     * thread has already been interrupted to partially simulate what would
     * really happen for a CA put.
     */
    Thread.sleep(1L);
    experimentState = ExperimentState.IDLE;
  }

  public ExperimentState experimentState() {
    return experimentState;
  }

  public void beginDetectorDataset(double exposureTime)
      throws InterruptedException, DriverException {
    beginDetectorDataset(exposureTime, exposureTime + getDetectorReadoutTime(), BeamlineAxis.PHI, 1,
        null);
  }

  public void beginDetectorDataset(double exposureTime, double exposurePeriod,
      BeamlineAxis exposureSyncAxis, int numImages, File imageSeqFilePath)
          throws InterruptedException, DriverException {
    this.exposureTime = DurationUtils.secondsToMillis(exposureTime);
  }

  public void endDetectorDataset() throws InterruptedException, DriverException {
  }

  public void takeDetectorImage(File imageFilePath, boolean isLastImage)
      throws InterruptedException, DriverException, TimeoutException,
      TooManyDetectorAcquisitionRetriesException {
    if (!realisticExposureTimeEnabled) return;
    Thread.sleep(exposureTime);
  }

  public void publishDetectorImageTaken(File imageFilePath)
      throws InterruptedException, DriverException {
    imagesTaken.add(imageFilePath);
  }

  public List<File> publishedDetectorImagesTaken() {
    return Collections.unmodifiableList(imagesTaken);
  }

  public void beginFlyScan(double shutterOpenPos, double shutterClosePos, double exposureTriggerPos)
      throws InterruptedException, DriverException {
  }

  public void endFlyScan() throws InterruptedException, DriverException {
  }

  public void endFlyScanNow() throws InterruptedException, DriverException {
  }

  public void clearHistories() {
    double last;
    last = phiRbvs.getLast();
    phiRbvs.clear();
    phiRbvs.addLast(last);
    last = xRbvs.getLast();
    xRbvs.clear();
    xRbvs.addLast(last);
    last = yRbvs.getLast();
    yRbvs.clear();
    yRbvs.addLast(last);
    last = zRbvs.getLast();
    zRbvs.clear();
    zRbvs.addLast(last);
    imagesTaken.clear();
  }

  private static double[] doubleArray(List<Double> doubles) {
    double[] result = new double[doubles.size()];
    int i = 0;
    for (Double each : doubles) {
      result[i] = each;
      i++;
    }
    return result;
  }
}
