package imcacat.imcadr;

import imcacat.imcadr.util.MathUtils;

import java.util.HashMap;
import java.util.Map;

/**
 * I am a Mock object of <tt>ImcadrServerSession</tt>.
 */
public class MockImcadrServerSession extends ImcadrServerSession {
  private static Range<Integer> ScoreCreationRange = null;

  public MockImcadrServerSession(BeamlineDriver driver) {
    super(driver);
  }

  @Override
  protected Rasterizer2d createRasterizer2d(Raster2dPlane plane) {
    Raster2dSpec spec = createRaster2dSpec(plane);
    return new MockRasterizer2d(spec, driver, false, createPositionScoresMap(spec),
        Position2d.mm(0.001, 0.001));
  }

  @Override
  protected Rasterizer3d createRasterizer3d() {
    Raster3dSpec spec = createRaster3dSpec();
    return new MockRasterizer3d(spec, driver, false, createPositionScoresMap(spec.getZySpec()),
        createPositionScoresMap(spec.getZxSpec(), computeZxStartingCenter()), Position2d.mm(
            0.001, 0.001));
  }

  protected Position2d computeZxStartingCenter() {
    return Position2d.mm(computeBestZyPosition().getX(), 0.0);
  }

  protected Position2d computeBestZyPosition() {
    boolean realisticEnabled = false;
    MockBeamlineDriver mockDriver = null;
    if (driver instanceof MockBeamlineDriver) {
      mockDriver = (MockBeamlineDriver)driver;
      realisticEnabled = mockDriver.isRealisticExposureTimeEnabled();
    }
    try {
      if (mockDriver != null) {
        mockDriver.setRealisticExposureTimeEnabled(false);
      }
      Rasterizer2d r2d = createRasterizer2d(Raster2dPlane.ZY);
      r2d.run();
      driver.moveY(0.0);
      driver.moveZ(0.0);
      return r2d.getBestPosition();
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw new RuntimeException(e);
    } catch (DriverException e) {
      throw new RuntimeException(e);
    } finally {
      if (mockDriver != null) {
        mockDriver.setRealisticExposureTimeEnabled(realisticEnabled);
      }
    }
  }

  protected static Map<Position2d, Integer> createPositionScoresMap(Raster2dSpec spec) {
    return createPositionScoresMap(spec, Position2d.mm(0.0, 0.0));
  }

  protected static Map<Position2d, Integer> createPositionScoresMap(Raster2dSpec spec,
      Position2d center) {
    int numCols = spec.getWidth();
    int numRows = spec.getHeight();
    int maxCol = numCols - 1;
    int maxRow = numRows - 1;
    Rectangle cellSize = spec.getCellSize().convertTo(LengthUnit.MILLIMETER);
    Position2d centerMm = center.convertTo(LengthUnit.MILLIMETER);
    double scale = 1e12;
    long scaledCellWidth = (long)(cellSize.getWidth() * scale);
    long scaledHalfCellWidth = scaledCellWidth / 2L;
    long scaledCellHeight = (long)(cellSize.getHeight() * scale);
    long scaledHalfCellHeight = scaledCellHeight / 2L;
    long scaledCenterX = (long)(centerMm.getX() * scale);
    long scaledCenterY = (long)(centerMm.getY() * scale);
    long scaledOriginX = scaledCenterX - ((scaledCellWidth * numCols) / 2L);
    long scaledOriginY = scaledCenterY - ((scaledCellHeight * numRows) / 2L);
    long scaledTopLeftX = scaledOriginX + scaledHalfCellWidth;
    long scaledTopLeftY = scaledOriginY + scaledHalfCellHeight;
    Range<Integer> scoreRange = getScoreCreationRange();
    if (scoreRange == null) {
      scoreRange = spec.getAnalysisSpec().getAnalysisPackage().getScoreMinimumRange();
    }
    Map<Position2d, Integer> result = new HashMap<Position2d, Integer>();
    for (int r = 0; r < numRows; r++) {
      double y = (double)(scaledTopLeftY + (r * scaledCellHeight)) / scale;
      for (int c = 0; c < numCols; c++) {
        double x = (double)(scaledTopLeftX + (c * scaledCellWidth)) / scale;
        result.put(Position2d.mm(x, y), createScore(c, r, maxCol, maxRow, scoreRange));
      }
    }
    return result;
  }

  protected static int createScore(int c, int r, int maxCol, int maxRow,
      Range<Integer> scoreRange) {
    if (maxCol < 2 && maxRow < 2) {
      return scoreRange.high();
    } else if (maxCol < 2) {
      if (r == 0 || r == maxRow) {
        return scoreRange.low();
      } else {
        return MathUtils.scale(r, 0, maxRow - 1, scoreRange.low(), scoreRange.high());
      }
    } else if (maxRow < 2) {
      if (c == 0 || c == maxCol) {
        return scoreRange.low();
      } else {
        return MathUtils.scale(c, 0, maxCol - 1, scoreRange.low(), scoreRange.high());
      }
    }
    if (c == 0 || c == maxCol || r == 0 || r == maxRow) {
      return scoreRange.low();
    } else {
      return MathUtils.scale(c + r, 0, maxCol + maxRow - 2, scoreRange.low(), scoreRange.high());
    }
  }

  public static synchronized Range<Integer> getScoreCreationRange() {
    return ScoreCreationRange;
  }

  public static synchronized void setScoreCreationRange(Range<Integer> newRange) {
    ScoreCreationRange = newRange;
  }
}
