package imcacat.imcadr;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * I am a Mock object of <tt>Rasterizer2d</tt>.
 */
public class MockRasterizer2d extends Rasterizer2d {
  private static final long serialVersionUID = 1;

  private static final ArrayList<WeakReference<MockRasterizer2d>> Instances =
      new ArrayList<WeakReference<MockRasterizer2d>>();

  private Map<Position2d, Integer> scores;
  private Position2d scoreLookUpDelta;
  private RuntimeException analyzerCreationException;

  public MockRasterizer2d(Raster2dSpec spec, boolean firePropertyChangesOnEdtEnabled,
      Map<Position2d, Integer> scores) {
    this(spec, new MockBeamlineDriver(), firePropertyChangesOnEdtEnabled, scores);
  }

  public MockRasterizer2d(Raster2dSpec spec, BeamlineDriver driver,
      boolean firePropertyChangesOnEdtEnabled, Map<Position2d, Integer> scores) {
    this(spec, driver, firePropertyChangesOnEdtEnabled, scores, null);
  }

  public MockRasterizer2d(Raster2dSpec spec, BeamlineDriver driver,
      boolean firePropertyChangesOnEdtEnabled, Map<Position2d, Integer> scores,
      Position2d scoreLookUpDelta) {
    super(spec, driver, firePropertyChangesOnEdtEnabled);
    this.scores = scores;
    this.scoreLookUpDelta = scoreLookUpDelta;
    analyzerCreationException = null;
    doNotWriteResultsFile();
    doNotLogException();
    addInstance(this);
  }

  public synchronized RuntimeException getAnalyzerCreationException() {
    return analyzerCreationException;
  }

  public synchronized void setAnalyzerCreationException(
      RuntimeException newAnalyzerCreationException) {
    analyzerCreationException = newAnalyzerCreationException;
  }

  @Override
  protected ScheduledExecutorService createImageTakenPublisherExecutor() {
    return new ZeroDelayScheduledThreadPoolExecutor(1);
  }

  @Override
  public RasterCellAnalyzer createAnalyzer(Raster2dIndex index, long earliestImageAcquiredTime,
      long earliestImageAcquiredMonotonicTime, String flyScanImageName) {
    RuntimeException e = getAnalyzerCreationException();
    if (e != null) throw e;
    return new MockRasterCellAnalyzer(index, raster, Integer.valueOf(lookUpScore(index)));
  }

  private int lookUpScore(Raster2dIndex index) {
    return lookUpScore(raster.getCell(index).getPosition());
  }

  private int lookUpScore(Position2d position) {
    if (scores.containsKey(position)) return scores.get(position);
    if (scoreLookUpDelta != null) {
      for (Position2d each : scores.keySet()) {
        if (each.equals(position, scoreLookUpDelta)) return scores.get(each);
      }
    }
    throw new BugException("No entry for position: " + position);
  }

  private static synchronized void addInstance(MockRasterizer2d rasterizer) {
    for (Iterator<WeakReference<MockRasterizer2d>> i = Instances.iterator(); i.hasNext();) {
      WeakReference<MockRasterizer2d> each = i.next();
      if (each.get() == null) i.remove();
    }
    Instances.add(new WeakReference<MockRasterizer2d>(rasterizer));
  }

  public static synchronized List<MockRasterizer2d> getInstances() {
    ArrayList<MockRasterizer2d> result = new ArrayList<MockRasterizer2d>();
    for (Iterator<WeakReference<MockRasterizer2d>> i = Instances.iterator(); i.hasNext();) {
      WeakReference<MockRasterizer2d> eachRef = i.next();
      MockRasterizer2d each = eachRef.get();
      if (each == null) {
        i.remove();
      } else {
        result.add(each);
      }
    }
    return result;
  }

  private static class ZeroDelayScheduledThreadPoolExecutor extends ScheduledThreadPoolExecutor {
    public ZeroDelayScheduledThreadPoolExecutor(int corePoolSize) {
      super(corePoolSize);
    }

    @Override
    public ScheduledFuture<?> schedule(Runnable command, long delay, TimeUnit unit) {
      return super.schedule(command, 0, unit);
    }

    @Override
    public <V> ScheduledFuture<V> schedule(Callable<V> callable, long delay, TimeUnit unit) {
      return super.schedule(callable, 0, unit);
    }

    @Override
    public ScheduledFuture<?> scheduleAtFixedRate(Runnable command, long initialDelay, long period,
        TimeUnit unit) {
      return super.scheduleAtFixedRate(command, 0, 1, TimeUnit.NANOSECONDS);
    }

    @Override
    public ScheduledFuture<?> scheduleWithFixedDelay(Runnable command, long initialDelay,
        long delay, TimeUnit unit) {
      return super.scheduleWithFixedDelay(command, 0, 1, TimeUnit.NANOSECONDS);
    }
  }
}
