package imcacat.imcadr;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.junit.Test;

/**
 * I test <code>Rasterizer3d</code>.
 */
public class Rasterizer3dTest {
  private static final int DEFAULT_RASTER_WIDTH = 3;
  private static final int DEFAULT_RASTER_HEIGHT = 4;
  private static final double DEFAULT_CELL_WIDTH = 0.3;
  private static final double DEFAULT_CELL_HEIGHT = 0.2;

  /* Ensure Phi gets driven to 0 degrees before ZY raster. */
  @Test
  public void testRun01() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}}
    };
    int[][] scores = new int[][] {
      {40, 40, 40},
      {60, 60, 60},
      {60, 80, 60},
      {60, 60, 60}
    };
    /* @formatter:on */

    Map<Position2d, Integer> positionScores = createPositionScoresMap(positions, scores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    driver.movePhi(45.0);
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent"))), driver, false, positionScores, positionScores);

    rasterizer.run();

    assertArrayEquals(new double[] {
        0.0,
        45.0,
        0.0,
        90.0}, driver.phiRbvHistory(), 1e-6);
  }

  /* Ensure progress is updated through ZY and ZX rasters. */
  @Test
  public void testRun02() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}}
    };
    int[][] scores = new int[][] {
      {40, 40, 40},
      {60, 60, 60},
      {60, 80, 60},
      {60, 60, 60}
    };
    /* @formatter:on */

    Map<Position2d, Integer> positionScores = createPositionScoresMap(positions, scores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent"))), driver, false, positionScores, positionScores);

    final ArrayList<Integer> progresses = new ArrayList<Integer>();
    rasterizer.addPropertyChangeListener(BaseRasterizer.PROPNAME_PROGRESS,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            progresses.add((Integer)event.getNewValue());
          }
        });

    rasterizer.run();

    assertTrue(progresses.size() > 4);
    boolean foundExpected = false;
    for (Integer each : progresses) {
      if (each.intValue() == 0) continue;
      if (each.intValue() == 1) continue;
      if (each.intValue() == 50) continue;
      if (each.intValue() == 99) continue;
      if (each.intValue() == 100) continue;
      foundExpected = true;
    }
    assertTrue(foundExpected);
  }

  /* Ensure ZX raster is rasterized first if spec indicates this. */
  @Test
  public void testRun03() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}}
    };
    int[][] scores = new int[][] {
      {40, 40, 40},
      {60, 60, 60},
      {60, 80, 60},
      {60, 60, 60}
    };
    /* @formatter:on */

    Map<Position2d, Integer> positionScores = createPositionScoresMap(positions, scores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    driver.movePhi(45.0);
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent")), true), driver, false, positionScores, positionScores);

    rasterizer.run();

    assertArrayEquals(new double[] {
        0.0,
        45.0,
        90.0,
        0.0}, driver.phiRbvHistory(), 1e-6);
  }

  /* Ensure best ZX horizontal position is used if score better than ZY. */
  @Test
  public void testRun04() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}, {0.6, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}, {0.6, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}, {0.6,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}, {0.6,  0.3}}
    };
    int[][] zyScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 40, 40},
      {40, 80, 40, 40},
      {40, 40, 40, 40}
    };
    int[][] zxScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 81, 40},
      {40, 40, 40, 40},
      {40, 40, 40, 40}
    };
    /* @formatter:on */
    int bestColumn = 2;
    int bestRow = 1;

    Map<Position2d, Integer> zyPositionScores = createPositionScoresMap(positions, zyScores);
    Map<Position2d, Integer> zxPositionScores = createPositionScoresMap(positions, zxScores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent"))), driver, false, zyPositionScores, zxPositionScores);

    rasterizer.run();

    assertEquals(positions[bestRow][bestColumn][0], rasterizer.getBestPosition().getZ(), 1e-6);
  }

  /* Ensure best ZY horizontal position is used if score better than ZX. */
  @Test
  public void testRun05() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}, {0.6, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}, {0.6, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}, {0.6,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}, {0.6,  0.3}}
    };
    int[][] zyScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 40, 40},
      {40, 81, 40, 40},
      {40, 40, 40, 40}
    };
    int[][] zxScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 80, 40},
      {40, 40, 40, 40},
      {40, 40, 40, 40}
    };
    /* @formatter:on */
    int bestColumn = 1;
    int bestRow = 2;

    Map<Position2d, Integer> zyPositionScores = createPositionScoresMap(positions, zyScores);
    Map<Position2d, Integer> zxPositionScores = createPositionScoresMap(positions, zxScores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent"))), driver, false, zyPositionScores, zxPositionScores);

    rasterizer.run();

    assertEquals(positions[bestRow][bestColumn][0], rasterizer.getBestPosition().getZ(), 1e-6);
  }

  /*
   * Ensure best ZX horizontal position is used if score better than ZY when ZX
   * raster rasterized first.
   */
  @Test
  public void testRun06() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}, {0.6, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}, {0.6, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}, {0.6,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}, {0.6,  0.3}}
    };
    int[][] zyScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 40, 40},
      {40, 80, 40, 40},
      {40, 40, 40, 40}
    };
    int[][] zxScores = new int[][] {
      {40, 40, 40, 40},
      {40, 40, 81, 40},
      {40, 40, 40, 40},
      {40, 40, 40, 40}
    };
    /* @formatter:on */
    int bestColumn = 2;
    int bestRow = 1;

    Map<Position2d, Integer> zyPositionScores = createPositionScoresMap(positions, zyScores);
    Map<Position2d, Integer> zxPositionScores = createPositionScoresMap(positions, zxScores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")), new Raster2dSpec(
            Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2, new File(
                "/nonexistent")), true), driver, false, zyPositionScores, zxPositionScores);

    rasterizer.run();

    assertEquals(positions[bestRow][bestColumn][0], rasterizer.getBestPosition().getZ(), 1e-6);
  }

  /* ZY high score <= cutoff. */
  @Test
  public void testRun07() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}}
    };
    int[][] zyScores = new int[][] {
      {10, 11, 12},
      {13, 13, 13},
      {13, 25, 13},
      {13, 13, 13}
    };
    int[][] zxScores = new int[][] {
      {10, 10, 10},
      {10, 80, 10},
      {10, 10, 10},
      {11, 12, 13}
      };
    /* @formatter:on */

    Map<Position2d, Integer> zyPositionScores = createPositionScoresMap(positions, zyScores);
    Map<Position2d, Integer> zxPositionScores = createPositionScoresMap(positions, zxScores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, true, 25, new File("/nonexistent"), null),
            new Raster2dSpec(Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2,
                true, 25, new File("/nonexistent"), null)), driver, false, zyPositionScores,
            zxPositionScores);

    rasterizer.run();

    assertTrue(rasterizer.hasException());
    assertEquals(BestScoreCutoffException.class, rasterizer.getException().getClass());
    assertTrue(rasterizer.getZyRasterizer().hasException());
    assertEquals(BestScoreCutoffException.class, rasterizer.getZyRasterizer().getException()
        .getClass());
    assertFalse(rasterizer.getZxRasterizer().hasException());
  }

  /* ZX high score <= cutoff. */
  @Test
  public void testRun08() throws Exception {
    /* @formatter:off */
    double[][][] positions = new double[][][] {
      {{-0.3, -0.3}, {0.0, -0.3}, {0.3, -0.3}},
      {{-0.3, -0.1}, {0.0, -0.1}, {0.3, -0.1}},
      {{-0.3,  0.1}, {0.0,  0.1}, {0.3,  0.1}},
      {{-0.3,  0.3}, {0.0,  0.3}, {0.3,  0.3}}
    };
    int[][] zyScores = new int[][] {
      {10, 11, 12},
      {13, 13, 13},
      {13, 26, 13},
      {13, 13, 13}
    };
    int[][] zxScores = new int[][] {
      {10, 10, 10},
      {10, 25, 10},
      {10, 10, 10},
      {11, 12, 13}
    };
    /* @formatter:on */

    Map<Position2d, Integer> zyPositionScores = createPositionScoresMap(positions, zyScores);
    Map<Position2d, Integer> zxPositionScores = createPositionScoresMap(positions, zxScores);

    MockBeamlineDriver driver = new MockBeamlineDriver();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY, 3, 4, 0.3, 0.2,
            LengthUnit.MILLIMETER, 500, 2, true, 25, new File("/nonexistent"), null),
            new Raster2dSpec(Raster2dPlane.ZX, 3, 4, 0.3, 0.2, LengthUnit.MILLIMETER, 500, 2,
                true, 25, new File("/nonexistent"), null)), driver, false, zyPositionScores,
            zxPositionScores);

    rasterizer.run();

    assertTrue(rasterizer.hasException());
    assertEquals(BestScoreCutoffException.class, rasterizer.getException().getClass());
    assertFalse(rasterizer.getZyRasterizer().hasException());
    assertTrue(rasterizer.getZxRasterizer().hasException());
    assertEquals(BestScoreCutoffException.class, rasterizer.getZxRasterizer().getException()
        .getClass());
  }

  /* Experiment state should be restored if interrupted. */
  @Test
  public void testInterrupt01() throws Exception {
    MockBeamlineDriver driver = new MockBeamlineDriver();
    driver.setRealisticExposureTimeEnabled(true);
    Map<Position2d, Integer> scores = createDefaultPositionScoresMap();
    Rasterizer3d rasterizer =
        new MockRasterizer3d(new Raster3dSpec(new Raster2dSpec(Raster2dPlane.ZY,
            DEFAULT_RASTER_WIDTH, DEFAULT_RASTER_HEIGHT, DEFAULT_CELL_WIDTH,
            DEFAULT_CELL_HEIGHT, LengthUnit.MILLIMETER, 500, 2, new File("/nonexistent")),
            new Raster2dSpec(Raster2dPlane.ZX, DEFAULT_RASTER_WIDTH, DEFAULT_RASTER_HEIGHT,
                DEFAULT_CELL_WIDTH, DEFAULT_CELL_HEIGHT, LengthUnit.MILLIMETER, 500, 2,
                new File("/nonexistent"))), driver, false, scores, scores);
    rasterizer.doNotLogException();

    assertEquals(ExperimentState.IDLE, driver.experimentState());

    Thread t = new Thread(rasterizer);
    try {
      t.start();
      Thread.sleep(1000L);
    } finally {
      t.interrupt();
      t.join();
    }

    assertEquals(ExperimentState.IDLE, driver.experimentState());
  }

  private static Map<Position2d, Integer> createDefaultPositionScoresMap() {
    return createPositionScoresMap(
/* @formatter:off */
      new double[][][] {
        {{-0.3, -0.3}, {0, -0.3}, {0.3, -0.3}},
        {{-0.3, -0.1}, {0, -0.1}, {0.3, -0.1}},
        {{-0.3,  0.1}, {0,  0.1}, {0.3,  0.1}},
        {{-0.3,  0.3}, {0,  0.3}, {0.3,  0.3}}
      },
      new int[][] {
        {40, 40, 40},
        {60, 60, 60},
        {60, 80, 60},
        {60, 60, 60}
      });
      /* @formatter:on */
  }

  private static Map<Position2d, Integer> createPositionScoresMap(double[][][] positions,
      int[][] scores) {
    Map<Position2d, Integer> result = new HashMap<Position2d, Integer>();
    for (int r = 0; r < positions.length; r++) {
      for (int c = 0; c < positions[0].length; c++) {
        result.put(Position2d.mm(positions[r][c][0], positions[r][c][1]), scores[r][c]);
      }
    }
    return result;
  }
}
