package imcacat.imcadr.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.Map;

import org.junit.Test;

/**
 * I test <code>ShellUtils</code>.
 */
public class ShellUtilsTest {
  @Test
  public void testExpand_basic() {
    assertEquals("Color: red.", ShellUtils.expand("Color: $color.", newMap("color", "red")));
    assertEquals("Color: red.", ShellUtils.expand("Color: ${color}.", newMap("color", "red")));
    assertEquals("red", ShellUtils.expand("$color", newMap("color", "red")));
    assertEquals("red", ShellUtils.expand("${color}", newMap("color", "red")));
    assertEquals("abc red", ShellUtils.expand("abc $color", newMap("color", "red")));
    assertEquals("abc red", ShellUtils.expand("abc ${color}", newMap("color", "red")));
    assertEquals("red abc", ShellUtils.expand("$color abc", newMap("color", "red")));
    assertEquals("red abc", ShellUtils.expand("${color} abc", newMap("color", "red")));
    assertEquals("red abc", ShellUtils.expand("$c abc", newMap("c", "red")));
    assertEquals("red abc", ShellUtils.expand("${c} abc", newMap("c", "red")));
    assertEquals("abc red", ShellUtils.expand("abc $c", newMap("c", "red")));
    assertEquals("abc red", ShellUtils.expand("abc ${c}", newMap("c", "red")));
    assertEquals("abcred", ShellUtils.expand("abc$color", newMap("color", "red")));
    assertEquals("abcred", ShellUtils.expand("abc${color}", newMap("color", "red")));
    assertEquals("redabc", ShellUtils.expand("${color}abc", newMap("color", "red")));
    assertEquals("1 2 3", ShellUtils.expand("$a $b $c", newMap("a", "1", "b", "2", "c", "3")));
    assertEquals("x 1 2 y", ShellUtils.expand("x $a $b y", newMap("a", "1", "b", "2")));
    assertEquals("1 1 1", ShellUtils.expand("$a $a $a", newMap("a", "1")));
    assertEquals("/home/luke", ShellUtils.expand("/home/$user", newMap("user", "luke")));
  }

  @Test
  public void testExpand_escape() {
    assertEquals("\\", ShellUtils.expand("\\\\", newMap()));
    assertEquals("\\$", ShellUtils.expand("\\\\\\$", newMap()));
    assertEquals("$", ShellUtils.expand("\\$", newMap()));
    assertEquals("x", ShellUtils.expand("\\x", newMap()));
    assertEquals("$color", ShellUtils.expand("\\$color", newMap("color", "red")));
    assertEquals("${color}", ShellUtils.expand("\\${color}", newMap("color", "red")));
    assertEquals("$red", ShellUtils.expand("\\$$color", newMap("color", "red")));
    assertEquals("$red", ShellUtils.expand("\\$${color}", newMap("color", "red")));
    assertEquals("$red$", ShellUtils.expand("\\$$color\\$", newMap("color", "red")));
    assertEquals("$red$", ShellUtils.expand("\\$${color}\\$", newMap("color", "red")));
  }

  @Test
  public void testExpand_varValueRegexEscape() {
    assertEquals("x \\1", ShellUtils.expand("$a $b", newMap("a", "x", "b", "\\1")));
    assertEquals("$x", ShellUtils.expand("$a", newMap("a", "$x")));
  }

  @Test
  public void testExpand_noMapping() {
    assertEquals("", ShellUtils.expand("$color", newMap()));
  }

  @Test(expected = NullPointerException.class)
  public void testExpand_nullInput() {
    ShellUtils.expand(null, newMap());
  }

  @Test(expected = NullPointerException.class)
  public void testExpand_nullVarNamesToValues() {
    ShellUtils.expand("$color", null);
  }

  @Test
  public void testExpand_emptyInput() {
    assertEquals("", ShellUtils.expand("", newMap()));
  }

  @Test
  public void testExpand_noVars() {
    assertEquals("abc", ShellUtils.expand("abc", newMap()));
  }

  @Test
  public void testExpand_invalid() {
    assertEquals("\\", ShellUtils.expand("\\", newMap()));
    assertEquals("$", ShellUtils.expand("$", newMap()));
    assertEquals("${", ShellUtils.expand("${", newMap()));
    assertEquals("${a", ShellUtils.expand("${a", newMap()));
    assertEquals("$$", ShellUtils.expand("$$", newMap()));
    assertEquals("$1", ShellUtils.expand("$1", newMap()));
  }

  @Test
  public void testContainsVariable_basic() {
    assertTrue(ShellUtils.containsVariable("Color: $color.", "color"));
    assertTrue(ShellUtils.containsVariable("Color: ${color}.", "color"));
    assertTrue(ShellUtils.containsVariable("$color", "color"));
    assertTrue(ShellUtils.containsVariable("${color}", "color"));
    assertTrue(ShellUtils.containsVariable("abc $color", "color"));
    assertTrue(ShellUtils.containsVariable("abc ${color}", "color"));
    assertTrue(ShellUtils.containsVariable("$color abc", "color"));
    assertTrue(ShellUtils.containsVariable("${color} abc", "color"));
    assertTrue(ShellUtils.containsVariable("$c abc", "c"));
    assertTrue(ShellUtils.containsVariable("${c} abc", "c"));
    assertTrue(ShellUtils.containsVariable("abc $c", "c"));
    assertTrue(ShellUtils.containsVariable("abc ${c}", "c"));
    assertTrue(ShellUtils.containsVariable("abc$color", "color"));
    assertTrue(ShellUtils.containsVariable("abc${color}", "color"));
    assertTrue(ShellUtils.containsVariable("${color}abc", "color"));
    assertTrue(ShellUtils.containsVariable("$a $b $c", "b"));
    assertTrue(ShellUtils.containsVariable("x $a $b y", "b"));
    assertTrue(ShellUtils.containsVariable("$a $a $a", "a"));
    assertTrue(ShellUtils.containsVariable("/home/$user", "user"));
  }

  @Test
  public void testContainsVariable_escape() {
    assertFalse(ShellUtils.containsVariable("\\$color", "color"));
    assertFalse(ShellUtils.containsVariable("\\${color}", "color"));
    assertTrue(ShellUtils.containsVariable("\\$$color", "color"));
    assertTrue(ShellUtils.containsVariable("\\$${color}", "color"));
    assertTrue(ShellUtils.containsVariable("\\$$color\\$", "color"));
    assertTrue(ShellUtils.containsVariable("\\$${color}\\$", "color"));
  }

  @Test(expected = NullPointerException.class)
  public void testContainsVariable_nullInput() {
    ShellUtils.containsVariable(null, "color");
  }

  @Test(expected = NullPointerException.class)
  public void testContainsVariable_nullVarName() {
    ShellUtils.containsVariable("$color", null);
  }

  @Test
  public void testContainsVariable_emptyInput() {
    assertFalse(ShellUtils.containsVariable("", "color"));
  }

  private static Map<String, String> newMap(String... mapping) {
    if (mapping.length % 2 != 0) {
      throw new IllegalArgumentException("Odd number of arguments");
    }
    Map<String, String> result = new HashMap<>(mapping.length / 2);
    for (int i = 0; i < mapping.length; i += 2) {
      result.put(mapping[i], mapping[i + 1]);
    }
    return result;
  }
}
