package imcacat.imcadr;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;

/**
 * I am a driver for an ADSC detector.
 */
public class AdscDetectorDriver extends ExternalSoftwareTriggerDetectorDriver {
  private static final double[] Q210R_READOUT_TIMES = new double[] {0.9, 0.25};
  private static final double EXPOSURE_TIMEOUT = 10.0;

  protected static final String PHI_PV_NAME = "ADSCPhi";
  protected static final String ROTATION_AXIS_PV_NAME = "ADSCAxis";
  protected static final String IMAGE_WIDTH_PV_NAME = "ADSCImWidth";
  protected static final String DETECTOR_DISTANCE_PV_NAME = "ADSCDistnce";
  protected static final String DETECTOR_2THETA_PV_NAME = "ADSC2Theta";
  protected static final String WAVELENGTH_PV_NAME = "ADSCWavelen";
  protected static final String LAST_IMAGE_PV_NAME = "ADSCLastImage";

  protected static final int LAST_IMAGE_VALUE = 1;

  private volatile boolean released = false;
  protected EpicsPv phiPv;
  protected EpicsPv rotationAxisPv;
  protected EpicsPv imageWidthPv;
  protected EpicsPv detectorDistancePv;
  protected EpicsPv detector2thetaPv;
  protected EpicsPv wavelengthPv;
  protected EpicsPv lastImagePv;

  public AdscDetectorDriver(String pvName) {
    super(pvName);

    phiPv = new EpicsPv(pvName + PHI_PV_NAME, false, false);
    rotationAxisPv =
        new EpicsPv(pvName + ROTATION_AXIS_PV_NAME, false, false,
            new ShortToEnumEpicsCaValueConverter(CrystalRotationAxis.class));
    imageWidthPv = new EpicsPv(pvName + IMAGE_WIDTH_PV_NAME, false, false);
    detectorDistancePv = new EpicsPv(pvName + DETECTOR_DISTANCE_PV_NAME, false, false);
    detector2thetaPv = new EpicsPv(pvName + DETECTOR_2THETA_PV_NAME, false, false);
    wavelengthPv = new EpicsPv(pvName + WAVELENGTH_PV_NAME, false, false);
    lastImagePv = new EpicsPv(pvName + LAST_IMAGE_PV_NAME, false, false);
  }

  @Override
  protected final void finalize() throws Throwable {
    try {
      release();
    } finally {
      super.finalize();
    }
  }

  public void release() {
    if (released) return;
    released = true;

    try {
      if (phiPv != null) phiPv.release();
      if (rotationAxisPv != null) rotationAxisPv.release();
      if (imageWidthPv != null) imageWidthPv.release();
      if (detectorDistancePv != null) detectorDistancePv.release();
      if (detector2thetaPv != null) detector2thetaPv.release();
      if (wavelengthPv != null) wavelengthPv.release();
      if (lastImagePv != null) lastImagePv.release();
    } finally {
      super.release();
    }
  }

  @Override
  public double getReadoutTime(boolean binningEnabled) {
    return Q210R_READOUT_TIMES[binningEnabled ? 1 : 0];
  }

  @Override
  public double getExposureTimeout() {
    return EXPOSURE_TIMEOUT;
  }

  public void lastImage() throws InterruptedException, DriverException {
    try {
      lastImagePv.setValueAndWait(LAST_IMAGE_VALUE);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setDetector2theta(double position) throws InterruptedException, DriverException {
    try {
      detector2thetaPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setDetectorDistance(double position) throws InterruptedException, DriverException {
    try {
      detectorDistancePv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setImageWidth(double width) throws InterruptedException, DriverException {
    try {
      imageWidthPv.setValueAndWait(width);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setPhi(double position) throws InterruptedException, DriverException {
    try {
      phiPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setRotationAxis(CrystalRotationAxis axis) throws InterruptedException,
      DriverException {
    try {
      rotationAxisPv.setValueAndWait(axis);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setWavelength(double position) throws InterruptedException, DriverException {
    try {
      wavelengthPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setAttenuation(double attenuation) throws InterruptedException, DriverException {
  }
}
