package imcacat.imcadr;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.imcadr.util.EpicsCaUtils;

import java.io.File;
import java.util.logging.Logger;

/**
 * I provide some of the basic functionality common to all <a
 * href="http://cars9.uchicago.edu/software/epics/areaDetector.html"
 * >areaDetector</a> driver instances.
 */
public abstract class AreaDetectorDriver implements DetectorDriver {
  private static final Logger LOG = Logger.getLogger(AreaDetectorDriver.class.getName());

  protected static final String MAKE_PV_NAME = "Manufacturer_RBV";
  protected static final String MODEL_PV_NAME = "Model_RBV";
  protected static final String BINNING_X_PV_NAME = "BinX";
  protected static final String BINNING_Y_PV_NAME = "BinY";
  protected static final String EXPOSURE_TIME_PV_NAME = "AcquireTime";
  protected static final String EXPOSURE_PERIOD_PV_NAME = "AcquirePeriod";
  protected static final String NUMBER_OF_IMAGES_PV_NAME = "NumImages";
  protected static final String FILE_PATH_PV_NAME = "FilePath";
  protected static final String FILE_NAME_PV_NAME = "FileName";
  protected static final String FILE_NAME_TEMPLATE_PV_NAME = "FileTemplate";
  protected static final int QUERY_MAKE_MODEL_MAX_RETRIES = 2;

  private volatile boolean released = false;
  protected EpicsPv binningXPv;
  protected EpicsPv binningYPv;
  protected EpicsPv exposureTimePv;
  protected EpicsPv exposurePeriodPv;
  protected EpicsPv numberOfImagesPv;
  protected EpicsPv filePathPv;
  protected EpicsPv fileNamePv;
  protected EpicsPv fileNameTemplatePv;

  public AreaDetectorDriver(String pvName) {
    binningXPv = new EpicsPv(pvName + BINNING_X_PV_NAME, false, false);
    binningYPv = new EpicsPv(pvName + BINNING_Y_PV_NAME, false, false);
    exposureTimePv = new EpicsPv(pvName + EXPOSURE_TIME_PV_NAME, false, false);
    exposurePeriodPv = new EpicsPv(pvName + EXPOSURE_PERIOD_PV_NAME, false, false);
    numberOfImagesPv = new EpicsPv(pvName + NUMBER_OF_IMAGES_PV_NAME, false, false);
    filePathPv = new EpicsPv(pvName + FILE_PATH_PV_NAME, true, false);
    fileNamePv = new EpicsPv(pvName + FILE_NAME_PV_NAME, true, false);
    fileNameTemplatePv = new EpicsPv(pvName + FILE_NAME_TEMPLATE_PV_NAME, true, false);
  }

  public void release() {
    if (released) return;
    released = true;

    if (binningXPv != null) binningXPv.release();
    if (binningYPv != null) binningYPv.release();
    if (exposureTimePv != null) exposureTimePv.release();
    if (exposurePeriodPv != null) exposurePeriodPv.release();
    if (numberOfImagesPv != null) numberOfImagesPv.release();
    if (filePathPv != null) filePathPv.release();
    if (fileNamePv != null) fileNamePv.release();
    if (fileNameTemplatePv != null) fileNameTemplatePv.release();
  }

  public void setBinningEnabled(boolean enabled) throws InterruptedException, DriverException {
    int binning = enabled ? 2 : 1;
    try {
      binningXPv.setValueAndWait(binning);
      binningYPv.setValueAndWait(binning);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getExposureTime() throws InterruptedException, DriverException {
    try {
      return (Double)exposureTimePv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setExposureTime(double exposureTime) throws InterruptedException, DriverException {
    try {
      exposureTimePv.setValueAndWait(exposureTime);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getExposurePeriod() throws InterruptedException, DriverException {
    try {
      return (Double)exposurePeriodPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setExposurePeriod(double exposurePeriod) throws InterruptedException,
      DriverException {
    try {
      exposurePeriodPv.setValueAndWait(exposurePeriod);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public int getNumberOfImages() throws InterruptedException, DriverException {
    try {
      return (Integer)numberOfImagesPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setNumberOfImages(int numberOfImages) throws InterruptedException, DriverException {
    try {
      numberOfImagesPv.setValueAndWait(numberOfImages);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setImageFilePath(File imageFilePath) throws InterruptedException, DriverException {
    try {
      File absImageFilePath = imageFilePath.getAbsoluteFile();
      File absImageDirPath = absImageFilePath.getParentFile();
      /*
       * Some data acquisition systems (e.g. PILATUS acquiring to a RAM disk)
       * need to be told the image directory path because they will create the
       * directory if it does not exist (e.g. on the RAM disk). Set it to
       * satisfy this requirement, and then immediately set it back to the
       * empty string since our file name template requires it to be empty
       * because we construct the full image file path ourselves and set it as
       * the image file name.
       */
      if (absImageDirPath != null) {
        filePathPv.setValueAndWait(caLongString(absImageDirPath.getPath()));
      }
      filePathPv.setValueAndWait(caLongString(""));
      fileNamePv.setValueAndWait(caLongString(absImageFilePath.getPath()));
      fileNameTemplatePv.setValueAndWait(caLongString("%s%s"));
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public static AreaDetectorDriver newInstance(String pvName) {
    String[] makeAndModel = queryMakeAndModel(pvName);
    String make = makeAndModel[0];
    String model = makeAndModel[1];
    if ("ADSC".equalsIgnoreCase(make)) return new AdscDetectorDriver(pvName);
    else if ("Dectris".equalsIgnoreCase(make)) return new PilatusDetectorDriver(pvName);
    else throw new RuntimeException("Unsupported areaDetector driver (" + "BasePv=\"" + pvName +
        "\", Make=\"" + make + "\", Model=\"" + model + "\")");
  }

  protected static String[] queryMakeAndModel(String pvName) {
    Throwable exception = null;
    for (int retryCount = 0; retryCount <= QUERY_MAKE_MODEL_MAX_RETRIES; retryCount++) {
      if (retryCount > 0) {
        LOG.warning("Retrying make-and-model query (retry count: " + retryCount + ")");
      }
      EpicsPv makePv = null;
      EpicsPv modelPv = null;
      try {
        makePv = new EpicsPv(pvName + MAKE_PV_NAME, false, false);
        String makeString = (String)makePv.getValue();
        modelPv = new EpicsPv(pvName + MODEL_PV_NAME, false, false);
        String modelString = (String)modelPv.getValue();
        return new String[] {
            makeString,
            modelString};
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw new RuntimeException(e);
      } catch (EpicsCaException | RuntimeException e) {
        if (exception == null) {
          exception = e;
        } else {
          exception.addSuppressed(e);
        }
      } finally {
        if (makePv != null) makePv.release();
        if (modelPv != null) modelPv.release();
      }
    }
    throw new RuntimeException("Failed to query areaDetector make and model " +
        "with base PV \"" + pvName + "\"", exception);
  }

  protected static byte[] caLongString(String s) {
    return EpicsCaUtils.longStringByteArray(s);
  }
}
