package imcacat.imcadr;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsMotor;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.epicsca4j.util.EpicsUtilities;
import imcacat.imcadr.util.DurationUtils;
import imcacat.imcadr.util.EpicsCaUtils;

/**
 * I am a control interface for the beamline. I am an instance of the Facade
 * pattern.
 */
public class BaseBeamlineDriver implements BeamlineDriver {
  private static final Logger LOG = Logger.getLogger(BaseBeamlineDriver.class.getName());

  protected static final long AUTOBP_REQUEST_PAUSE_TIMEOUT = 185000;
  protected static final long EXPERIMENT_STATE_CHANGE_TIMEOUT = 60000;
  protected static final long EXPERIMENT_STATE_ABORT_TIMEOUT = 10000;
  protected static final int DETECTOR_ACQUISITION_MAX_RETRIES = 2;
  protected static final long SHUTTER_OPEN_CLOSE_TIMEOUT = 5000;
  protected static final int MOVE_PHI_MAX_RETRIES = 2;

  private volatile boolean released = false;
  protected EpicsPv autobpRequestExternalPausePv;
  protected EpicsPv autobpReleaseExternalPausePv;
  protected EpicsPv experimentStatePv;
  protected EpicsPv experimentStateRbvPv;
  protected EpicsPv experimentStateSetPv;
  protected EpicsPv experimentStateStatusPv;
  protected EpicsPv experimentStateAbortPv;
  protected EpicsPv auxiliaryImageFilePathPv;
  protected EpicsMotor phiMotor;
  protected EpicsMotor xMotor;
  protected EpicsMotor yMotor;
  protected EpicsMotor zMotor;
  protected EpicsPv detectorDistanceRbvPv;
  protected EpicsPv detector2thetaRbvPv;
  protected EpicsPv wavelengthRbvPv;
  protected EpicsPv attenuationRbvPv;
  protected DetectorDriver detectorDriver;
  protected ExposyncDriver expoPhiDriver;
  protected ExposyncDriver expoXDriver;
  protected ExposyncDriver expoYDriver;
  protected ExposyncDriver expoZDriver;
  protected Double phiMotorPositionErrorMax;
  protected Double xMotorPositionErrorMax;
  protected Double yMotorPositionErrorMax;
  protected Double zMotorPositionErrorMax;
  protected double datasetExposureTime;
  protected BeamlineAxis datasetExposureSyncAxis;

  public BaseBeamlineDriver() {
    autobpRequestExternalPausePv =
        new EpicsPv(Resources.getInstance().getAutobpRequestExternalPausePvName(), false, false);
    autobpReleaseExternalPausePv =
        new EpicsPv(Resources.getInstance().getAutobpReleaseExternalPausePvName(), false, false);
    experimentStatePv =
        new EpicsPv(Resources.getInstance().getExperimentStatePvName(), false, false,
            new ShortToEnumEpicsCaValueConverter(ExperimentState.class));
    experimentStateRbvPv =
        new EpicsPv(Resources.getInstance().getExperimentStateRbvPvName(), false, false,
            new ShortToEnumEpicsCaValueConverter(ExperimentState.class));
    experimentStateSetPv =
        new EpicsPv(Resources.getInstance().getExperimentStateSetPvName(), false, false);
    experimentStateStatusPv =
        new EpicsPv(Resources.getInstance().getExperimentStateStatusPvName(), false, true,
            new ShortToEnumEpicsCaValueConverter(ExperimentStateStatus.class));
    experimentStateAbortPv =
        new EpicsPv(Resources.getInstance().getExperimentStateAbortPvName(), false, false);
    auxiliaryImageFilePathPv =
        new EpicsPv(Resources.getInstance().getAuxiliaryImageFilePathPvName(), false, false);
    detectorDistanceRbvPv =
        new EpicsPv(Resources.getInstance().getDetectorDistanceRbvPvName(), false, false);
    detector2thetaRbvPv =
        new EpicsPv(Resources.getInstance().getDetector2thetaRbvPvName(), false, false);
    wavelengthRbvPv =
        new EpicsPv(Resources.getInstance().getWavelengthRbvPvName(), false, false);
    attenuationRbvPv =
        new EpicsPv(Resources.getInstance().getAttenuationRbvPvName(), false, false);
    detectorDriver =
        AreaDetectorDriver.newInstance(Resources.getInstance().getAreaDetectorPvName());
    expoPhiDriver =
        new ExposyncDriver(Resources.getInstance().getExposyncPhiPvName(), Resources.getInstance()
            .getExposureShutterRbvPvName());
    expoXDriver =
        new ExposyncDriver(Resources.getInstance().getExposyncXPvName(), Resources.getInstance()
            .getExposureShutterRbvPvName());
    expoYDriver =
        new ExposyncDriver(Resources.getInstance().getExposyncYPvName(), Resources.getInstance()
            .getExposureShutterRbvPvName());
    expoZDriver =
        new ExposyncDriver(Resources.getInstance().getExposyncZPvName(), Resources.getInstance()
            .getExposureShutterRbvPvName());
    phiMotorPositionErrorMax = Resources.getInstance().getPhiMotorPositionErrorMax();
    xMotorPositionErrorMax = Resources.getInstance().getXMotorPositionErrorMax();
    yMotorPositionErrorMax = Resources.getInstance().getYMotorPositionErrorMax();
    zMotorPositionErrorMax = Resources.getInstance().getZMotorPositionErrorMax();
  }

  @Override
  protected final void finalize() throws Throwable {
    try {
      release();
    } finally {
      super.finalize();
    }
  }

  public void release() {
    if (released) return;
    released = true;

    if (autobpRequestExternalPausePv != null) autobpRequestExternalPausePv.release();
    if (autobpReleaseExternalPausePv != null) autobpReleaseExternalPausePv.release();
    if (experimentStatePv != null) experimentStatePv.release();
    if (experimentStateRbvPv != null) experimentStateRbvPv.release();
    if (experimentStateSetPv != null) experimentStateSetPv.release();
    if (experimentStateStatusPv != null) experimentStateStatusPv.release();
    if (experimentStateAbortPv != null) experimentStateAbortPv.release();
    if (auxiliaryImageFilePathPv != null) auxiliaryImageFilePathPv.release();
    if (phiMotor != null) phiMotor.release();
    if (xMotor != null) xMotor.release();
    if (yMotor != null) yMotor.release();
    if (zMotor != null) zMotor.release();
    if (detectorDistanceRbvPv != null) detectorDistanceRbvPv.release();
    if (detector2thetaRbvPv != null) detector2thetaRbvPv.release();
    if (wavelengthRbvPv != null) wavelengthRbvPv.release();
    if (attenuationRbvPv != null) attenuationRbvPv.release();
    if (detectorDriver instanceof AreaDetectorDriver) {
      ((AreaDetectorDriver)detectorDriver).release();
    }
    if (expoPhiDriver != null) expoPhiDriver.release();
    if (expoXDriver != null) expoXDriver.release();
    if (expoYDriver != null) expoYDriver.release();
    if (expoZDriver != null) expoZDriver.release();
  }

  public String getImageFileNameExtension() {
    if (detectorDriver instanceof AdscDetectorDriver) return ".img";
    else if (detectorDriver instanceof PilatusDetectorDriver) return ".cbf";
    else throw new BugException();
  }

  public void disableAutobp() throws InterruptedException, DriverException {
    try {
      autobpRequestExternalPausePv.setValueAndWait(Resources.getInstance()
          .getAutobpRequestExternalPauseValue(), AUTOBP_REQUEST_PAUSE_TIMEOUT);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public void enableAutobp() throws InterruptedException, DriverException {
    try {
      autobpReleaseExternalPausePv.setValue(Resources.getInstance()
          .getAutobpReleaseExternalPauseValue());
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getDetectorReadoutTime() {
    return detectorDriver.getReadoutTime(shouldEnableBinning(detectorDriver));
  }

  public double getMinVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    try {
      return getMotor(axis).getMinVelocity();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getMaxVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    try {
      return getMotor(axis).getMaxVelocity();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getVelo(BeamlineAxis axis) throws InterruptedException, DriverException {
    try {
      return getMotor(axis).getVelocity();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setVelo(BeamlineAxis axis, double desiredVelo)
      throws InterruptedException, DriverException {
    try {
      getMotor(axis).setVelocityAndWait(desiredVelo);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public double getAccelTime(BeamlineAxis axis) throws InterruptedException, DriverException {
    try {
      return getMotor(axis).getAccelerationTime();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double phiRbv() throws InterruptedException, DriverException {
    try {
      return getPhiMotor().getPosition();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void movePhi(double desiredPosition) throws InterruptedException, DriverException {
    EpicsMotor m = getPhiMotor();
    double actualPosition = Double.NaN;
    DriverException exception = null;
    for (int retryCount = 0; retryCount <= MOVE_PHI_MAX_RETRIES; retryCount++) {
      if (retryCount > 0) {
        LOG.warning("Retrying Phi move (retry count: " + retryCount + ")");
      }
      m.setDesiredPosition(desiredPosition);
      try {
        m.moveAndWait();
        if (phiMotorPositionErrorMax == null) return;
        actualPosition = m.getPositionNow();
        if (Math.abs(actualPosition - desiredPosition) <= phiMotorPositionErrorMax) return;
      } catch (EpicsCaException | TimeoutException e) {
        if (exception == null) {
          exception = new DriverException(e);
        } else {
          exception.addSuppressed(e);
        }
      }
    }
    if (exception != null) throw exception;
    throw new PositionErrorException("Phi", desiredPosition, actualPosition,
        phiMotorPositionErrorMax);
  }

  public double xRbv() throws InterruptedException, DriverException {
    try {
      return getXMotor().getPosition();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void moveX(double desiredPosition) throws InterruptedException, DriverException {
    EpicsMotor m = getXMotor();
    m.setDesiredPosition(desiredPosition);
    try {
      m.moveAndWait();
      if (xMotorPositionErrorMax == null) return;
      double actualPosition = m.getPositionNow();
      if (Math.abs(actualPosition - desiredPosition) <= xMotorPositionErrorMax) return;
      throw new PositionErrorException("X", desiredPosition, actualPosition,
          xMotorPositionErrorMax);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public double yRbv() throws InterruptedException, DriverException {
    try {
      return getYMotor().getPosition();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void moveY(double desiredPosition) throws InterruptedException, DriverException {
    EpicsMotor m = getYMotor();
    m.setDesiredPosition(desiredPosition);
    try {
      m.moveAndWait();
      if (yMotorPositionErrorMax == null) return;
      double actualPosition = m.getPositionNow();
      if (Math.abs(actualPosition - desiredPosition) <= yMotorPositionErrorMax) return;
      throw new PositionErrorException("Y", desiredPosition, actualPosition,
          yMotorPositionErrorMax);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public double zRbv() throws InterruptedException, DriverException {
    try {
      return getZMotor().getPosition();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void moveZ(double desiredPosition) throws InterruptedException, DriverException {
    EpicsMotor m = getZMotor();
    m.setDesiredPosition(desiredPosition);
    try {
      m.moveAndWait();
      if (zMotorPositionErrorMax == null) return;
      double actualPosition = m.getPositionNow();
      if (Math.abs(actualPosition - desiredPosition) <= zMotorPositionErrorMax) return;
      throw new PositionErrorException("Z", desiredPosition, actualPosition,
          zMotorPositionErrorMax);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public double rbv(BeamlineAxis axis) throws InterruptedException, DriverException {
    if (BeamlineAxis.PHI.equals(axis)) return phiRbv();
    if (BeamlineAxis.X.equals(axis)) return xRbv();
    if (BeamlineAxis.Y.equals(axis)) return yRbv();
    if (BeamlineAxis.Z.equals(axis)) return zRbv();
    throw new BugException();
  }

  public void move(BeamlineAxis axis, double desiredPosition) throws InterruptedException,
      DriverException {
    if (BeamlineAxis.PHI.equals(axis)) movePhi(desiredPosition);
    else if (BeamlineAxis.X.equals(axis)) moveX(desiredPosition);
    else if (BeamlineAxis.Y.equals(axis)) moveY(desiredPosition);
    else if (BeamlineAxis.Z.equals(axis)) moveZ(desiredPosition);
    else throw new BugException();
  }

  protected double detectorDistanceRbv() throws InterruptedException, DriverException {
    try {
      return (Double)detectorDistanceRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected double detector2thetaRbv() throws InterruptedException, DriverException {
    try {
      return (Double)detector2thetaRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected double wavelengthRbv() throws InterruptedException, DriverException {
    try {
      return (Double)wavelengthRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected double attenuationRbv() throws InterruptedException, DriverException {
    try {
      return (Double)attenuationRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void enterCollectDataExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException {
    enterExperimentState(ExperimentState.COLLECT_DATA);
  }

  public void enterIdleExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException {
    enterExperimentState(ExperimentState.IDLE);
  }

  protected void enterExperimentState(ExperimentState desiredState) throws InterruptedException,
      DriverException, ExperimentStateNotAchievedException {
    try {
      ExperimentStateStatus status = (ExperimentStateStatus)experimentStateStatusPv.getValue();
      if (ExperimentStateStatus.EXECUTING.equals(status)) {
        experimentStateAbortPv.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
        experimentStateStatusPv.waitForAnyOneValue(
            Arrays.asList(ExperimentStateStatus.DONE_OK, ExperimentStateStatus.DONE_ERROR),
            EXPERIMENT_STATE_ABORT_TIMEOUT);
      }
    } catch (TimeoutException e) {
      throw new DriverException(e);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }

    try {
      experimentStatePv.setValueAndWait(desiredState);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }

    final List<ExperimentStateStatus> statuses =
        Collections.synchronizedList(new ArrayList<ExperimentStateStatus>());
    final CountDownLatch statusesLatch = new CountDownLatch(2);
    PropertyChangeListener statusListener = new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        statuses.add((ExperimentStateStatus)event.getNewValue());
        statusesLatch.countDown();
      }
    };

    try {
      experimentStateStatusPv.addPropertyChangeListener(EpicsPv.PROPNAME_VALUE, statusListener);
      experimentStateSetPv.setValue(EpicsUtilities.PROC_FIELD_PUT_VALUE);
      boolean ok = statusesLatch.await(EXPERIMENT_STATE_CHANGE_TIMEOUT, TimeUnit.MILLISECONDS);
      if (!ok) throw new ExperimentStateNotAchievedException();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } finally {
      experimentStateStatusPv.removePropertyChangeListener(EpicsPv.PROPNAME_VALUE,
          statusListener);
    }

    List<ExperimentStateStatus> expectedStatuses =
        Arrays.asList(ExperimentStateStatus.EXECUTING, ExperimentStateStatus.DONE_OK);
    try {
      synchronized (statuses) {
        if (!expectedStatuses.equals(statuses) ||
            !desiredState.equals(experimentStateRbvPv.getValue())) {
          throw new ExperimentStateNotAchievedException();
        }
      }
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void setAuxiliaryImageFilePath(File imageFilePath) throws InterruptedException,
      DriverException {
    try {
      auxiliaryImageFilePathPv.setValueAndWait(caLongString(imageFilePath.getAbsolutePath()));
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void beginDetectorDataset(double exposureTime)
      throws InterruptedException, DriverException {
    beginDetectorDataset(exposureTime, exposureTime + getDetectorReadoutTime(), BeamlineAxis.PHI, 1,
        null);
  }

  public void beginDetectorDataset(double exposureTime, double exposurePeriod,
      BeamlineAxis exposureSyncAxis, int numImages, File imageSeqFilePath)
          throws InterruptedException, DriverException {
    datasetExposureTime = exposureTime * (double)numImages;
    datasetExposureSyncAxis = exposureSyncAxis;
    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);
    /*
     * WORKAROUND: The PILATUS driver requires that the wavelength be set
     * *before* other parameters since it may cause the threshold gain and
     * energy to be changed which will reset certain parameters such as number
     * of images and possibly MX settings.
     */
    detectorDriver.setWavelength(wavelengthRbv());
    detectorDriver.setAttenuation(attenuationRbv());
    detectorDriver.setBinningEnabled(shouldEnableBinning(detectorDriver));
    detectorDriver.setExposureTime(exposureTime);
    if (detectorDriver instanceof AdscDetectorDriver) {
      exposyncDriver.setMaxExposureTime(toMs(exposureTime));
      detectorDriver.setExposurePeriod(0);
    } else if (detectorDriver instanceof PilatusDetectorDriver) {
      exposyncDriver.setMaxExposureTime(
          toMs(datasetExposureTime) + toMs(detectorDriver.getExposureTimeout()));
      detectorDriver.setExposurePeriod(exposurePeriod);
      detectorDriver.setNumberOfImages(numImages);
      if (imageSeqFilePath != null) {
        detectorDriver.setImageFilePath(imageSeqFilePath);
      }
    } else {
      throw new BugException();
    }
    detectorDriver.setDetectorDistance(detectorDistanceRbv());
    detectorDriver.setDetector2theta(detector2thetaRbv());
    detectorDriver.setImageWidth(0.0);
    detectorDriver.setRotationAxis(CrystalRotationAxis.PHI);
    detectorDriver.startDataset();
  }

  public void endDetectorDataset() throws InterruptedException, DriverException {
    detectorDriver.stopDataset();
  }

  public void takeDetectorImage(File imageFilePath, boolean isLastImage)
      throws InterruptedException, DriverException, TimeoutException,
      TooManyDetectorAcquisitionRetriesException {
    setAuxiliaryImageFilePath(imageFilePath);
    detectorDriver.setImageFilePath(imageFilePath);
    detectorDriver.setPhi(phiRbv());
    if (isLastImage) detectorDriver.lastImage();
    if (detectorDriver instanceof AdscDetectorDriver) executeAdscExposure();
    else if (detectorDriver instanceof PilatusDetectorDriver) executePilatusExposure();
    else throw new BugException();
  }

  public void publishDetectorImageTaken(File imageFilePath)
      throws InterruptedException, DriverException {
    setAuxiliaryImageFilePath(imageFilePath);
  }

  public void beginFlyScan(double shutterOpenPos, double shutterClosePos, double exposureTriggerPos)
      throws InterruptedException, DriverException {
    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);
    exposyncDriver.setShutterOpenPosition(shutterOpenPos);
    exposyncDriver.setShutterClosePosition(shutterClosePos);
    exposyncDriver.setExposureTriggerEnabled(true);
    exposyncDriver.setExposureTriggerPosition(exposureTriggerPos);
    /*
     * It would be best to wait for the Exposync driver to become armed
     * after trying to arm it to ensure motors are positioned correctly and
     * the fly-scan will work, but because the armed readback PV might not
     * update right away (because the Exposync EPICS driver does not execute
     * a blocking call into the PLC program, so there's a chance that the PLC
     * program has not updated the armed status value yet when it gets read by
     * the Exposync EPICS driver), we do not wait for it because that would
     * mean waiting until the next Exposync EPICS driver poll event.
     */
    exposyncDriver.setArmed(true);
    detectorDriver.setPhi(phiRbv());
    detectorDriver.startExposing();
  }

  public void endFlyScan() throws InterruptedException, DriverException {
    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);
    boolean ok = exposyncDriver.waitForShutterToClose(SHUTTER_OPEN_CLOSE_TIMEOUT);
    if (!ok) throw new DriverException("Exposure shutter did not close");
    ok = detectorDriver.waitForExposureToEnd(toMs(detectorDriver.getExposureTimeout()));
    if (!ok) throw new DriverException("Exposure never ended");
  }

  public void endFlyScanNow() throws InterruptedException, DriverException {
    boolean interrupted = false;
    Exception caughtException = null;

    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);

    try {
      exposyncDriver.setArmed(false);
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      exposyncDriver.closeShutter();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }

    try {
      detectorDriver.stopExposing();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }

    if (caughtException instanceof InterruptedException) {
      throw (InterruptedException)caughtException;
    }
    if (interrupted) {
      Thread.currentThread().interrupt();
    }
    if (caughtException != null) {
      if (caughtException instanceof DriverException) {
        throw (DriverException)caughtException;
      } else {
        throw new RuntimeException(caughtException);
      }
    }
  }

  protected void executeAdscExposure() throws InterruptedException, DriverException,
      TooManyDetectorAcquisitionRetriesException {
    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);
    boolean shouldRetry = false;
    for (int retryCount = 0; retryCount <= DETECTOR_ACQUISITION_MAX_RETRIES; retryCount++) {
      boolean completed = false;
      try {
        detectorDriver.startExposing();
        exposyncDriver.openShutter();
        boolean ok = exposyncDriver.waitForShutterToClose(
            toMs(datasetExposureTime) + SHUTTER_OPEN_CLOSE_TIMEOUT);
        if (!ok) throw new DriverException("Exposure shutter did not close");
        shouldRetry = detectorDriver.stopExposing();
        completed = true;
      } finally {
        if (!completed) {
          try {
            exposyncDriver.closeShutter();
            detectorDriver.stopExposing();
          } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
          } catch (DriverException e) {
            /* IGNORED */
          }
        }
      }
      if (!shouldRetry) break;
    }
    if (shouldRetry) throw new TooManyDetectorAcquisitionRetriesException();
  }

  protected void executePilatusExposure() throws InterruptedException, DriverException {
    ExposyncDriver exposyncDriver = getExposyncDriver(datasetExposureSyncAxis);
    boolean completed = false;
    try {
      detectorDriver.startExposing();
      exposyncDriver.openShutter();
      boolean ok;
      ok = exposyncDriver.waitForShutterToOpen(SHUTTER_OPEN_CLOSE_TIMEOUT);
      if (!ok) throw new DriverException("Exposure shutter did not open");
      exposyncDriver.triggerExposure();
      ok = detectorDriver.waitForExposureToEnd(
          toMs(datasetExposureTime) + toMs(detectorDriver.getExposureTimeout()));
      if (!ok) throw new DriverException("Exposure never ended");
      exposyncDriver.closeShutter();
      ok = exposyncDriver.waitForShutterToClose(SHUTTER_OPEN_CLOSE_TIMEOUT);
      if (!ok) throw new DriverException("Exposure shutter did not close");
      completed = true;
    } finally {
      if (!completed) {
        exposyncDriver.closeShutter();
        detectorDriver.stopExposing();
      }
    }
  }

  protected EpicsMotor getPhiMotor() throws InterruptedException, DriverException {
    if (phiMotor == null) phiMotor = createMotor(Resources.getInstance().getPhiMotorPvName());
    return phiMotor;
  }

  protected EpicsMotor getXMotor() throws InterruptedException, DriverException {
    if (xMotor == null) xMotor = createMotor(Resources.getInstance().getXMotorPvName());
    return xMotor;
  }

  protected EpicsMotor getYMotor() throws InterruptedException, DriverException {
    if (yMotor == null) yMotor = createMotor(Resources.getInstance().getYMotorPvName());
    return yMotor;
  }

  protected EpicsMotor getZMotor() throws InterruptedException, DriverException {
    if (zMotor == null) zMotor = createMotor(Resources.getInstance().getZMotorPvName());
    return zMotor;
  }

  protected EpicsMotor getMotor(BeamlineAxis axis) throws InterruptedException, DriverException {
    switch (axis) {
      case PHI: return getPhiMotor();
      case X: return getXMotor();
      case Y: return getYMotor();
      case Z: return getZMotor();
      default: throw new BugException();
    }
  }

  protected EpicsMotor createMotor(String recordName) throws InterruptedException,
      DriverException {
    try {
      return new EpicsMotor(recordName);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected static byte[] caLongString(String s) {
    return EpicsCaUtils.longStringByteArray(s);
  }

  protected ExposyncDriver getExposyncDriver(BeamlineAxis axis) {
    switch (axis) {
      case PHI: return expoPhiDriver;
      case X: return expoXDriver;
      case Y: return expoYDriver;
      case Z: return expoZDriver;
      default: throw new BugException();
    }
  }

  protected static long toMs(double s) {
    return DurationUtils.secondsToMillis(s);
  }

  protected static double toS(long ms) {
    return DurationUtils.millisToSeconds(ms);
  }

  protected static boolean shouldEnableBinning(DetectorDriver d) {
    return d instanceof AdscDetectorDriver;
  }
}
