package imcacat.imcadr;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InterruptedIOException;
import java.io.PrintWriter;
import java.nio.channels.ClosedByInterruptException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * I provide behavior supporting performing a rasterization.
 * <p>
 * Subclasses should override my <code>initialize</code>, <code>go</code>,
 * <code>cleanUp</code>, and <code>release</code> methods.
 * <p>
 * When my <code>run</code> method is called, I do the following:
 * <ol>
 * <li>call <code>go</code>
 * <li>call <code>cleanUp</code> if <code>go</code> threw an exception
 * </ol>
 */
public abstract class BaseRasterizer extends EdtAwareModel implements Rasterizer, Runnable {
  private static final long serialVersionUID = 1;
  private static final Logger LOG = Logger.getLogger(BaseRasterizer.class.getName());

  protected static final String SUCCESSFUL_COMPLETION_STATUS = "Completed successfully";

  public static final String PROPNAME_RUNNING = "running";
  public static final String PROPNAME_STATUS = "status";
  public static final String PROPNAME_PROGRESS = "progress";
  public static final String PROPNAME_ACQUIRE_TIME = "acquireTime";
  public static final String PROPNAME_ANALYZE_TIME = "analyzeTime";
  public static final String PROPNAME_WALL_TIME = "wallTime";

  protected Throwable exception = null;
  protected boolean okToWriteResultsFile = true;
  protected boolean okToLogException = true;
  protected boolean running = false;
  protected boolean successful = true;
  protected String status = "";
  protected int progress = 0;
  protected long acquireTime = 0;
  protected long analyzeTime = 0;
  protected long wallTime = 0;
  protected List<Long> cellAnalyzeImageWaitTimes = new ArrayList<>();
  protected List<Long> cellAnalyzeScoreTimes = new ArrayList<>();

  public BaseRasterizer(boolean firePropertyChangesOnEdtEnabled) {
    super(firePropertyChangesOnEdtEnabled);
  }

  public void run() {
    setRunning(true);
    boolean interrupted = false;
    try {
      go();
    } catch (InterruptedException | InterruptedIOException | ClosedByInterruptException e) {
      interrupted = true;
      Thread.interrupted();
      exception = e;
      setSuccessful(false);
      performCleanUp("Stopping", "Stopped", "Stop failed");
    } catch (Exception e) {
      interrupted = Thread.interrupted();
      exception = e;
      setSuccessful(false);
      if (okToLogException) LOG.log(Level.SEVERE, "Rasterization failed", e);
      performCleanUp("Failed", "Failed", "Failed");
    } finally {
      if (Thread.interrupted()) interrupted = true;
      if (okToWriteResultsFile) {
        try {
          writeResultsFile();
        } catch (FileNotFoundException | PrintWriterErrorException e) {
          if (exception == null) {
            exception = e;
            if (okToLogException) LOG.log(Level.SEVERE, "Failed to write results file", e);
          } else {
            LOG.log(Level.SEVERE, "Failed to write results file", e);
          }
          setSuccessful(false);
        }
      }
      setRunning(false);
      if (interrupted) Thread.currentThread().interrupt();
    }
  }

  /**
   * Subclasses should override this method to perform a rasterization.
   */
  protected abstract void go() throws Exception;

  protected void performCleanUp(String beforeCleanUp, String cleanUpSuccess,
      String cleanUpFailure) {
    setStatus(beforeCleanUp + "; restoring good beamline state");
    try {
      cleanUp();
      setStatus(cleanUpSuccess + "; good beamline state restored");
    } catch (RuntimeException e) {
      setStatus(cleanUpFailure + "; beamline in an inconsistent state");
      LOG.log(Level.SEVERE,
          "Failed to perform Rasterizer clean up; beamline in an inconsistent state", e);
    }
  }

  /**
   * This method is invoked if an exception is thrown while I am running
   * (including an <code>InterruptedException</code> if I am interrupted).
   * Subclasses should override this method to perform any work that should be
   * done to restore the beamline to a good state.
   */
  protected abstract void cleanUp();

  public boolean hasException() {
    return exception != null;
  }

  public Throwable getException() {
    return exception;
  }

  public void doNotWriteResultsFile() {
    okToWriteResultsFile = false;
  }

  public void doNotLogException() {
    okToLogException = false;
  }

  public boolean isRunning() {
    return running;
  }

  protected void setRunning(boolean newRunning) {
    if (running == newRunning) return;
    boolean oldRunning = running;
    running = newRunning;
    notifyChanged(PROPNAME_RUNNING, oldRunning, newRunning);
  }

  public String getStatus() {
    return status;
  }

  protected void setStatus(String newStatus) {
    if (status.equals(newStatus)) return;
    String oldStatus = status;
    status = newStatus;
    notifyChanged(PROPNAME_STATUS, oldStatus, newStatus);
  }

  public int getProgress() {
    return progress;
  }

  protected void setProgress(int newProgress) {
    if (progress == newProgress) return;
    int oldProgress = progress;
    progress = newProgress;
    notifyChanged(PROPNAME_PROGRESS, oldProgress, newProgress);
  }

  public long getAcquireTime() {
    return acquireTime;
  }

  protected void setAcquireTime(long newAcquireTime) {
    if (acquireTime == newAcquireTime) return;
    long oldAcquireTime = acquireTime;
    acquireTime = newAcquireTime;
    notifyChanged(PROPNAME_ACQUIRE_TIME, oldAcquireTime, newAcquireTime);
  }

  public long getAnalyzeTime() {
    return analyzeTime;
  }

  protected void setAnalyzeTime(long newAnalyzeTime) {
    if (analyzeTime == newAnalyzeTime) return;
    long oldAnalyzeTime = analyzeTime;
    analyzeTime = newAnalyzeTime;
    notifyChanged(PROPNAME_ANALYZE_TIME, oldAnalyzeTime, newAnalyzeTime);
  }

  public long getWallTime() {
    return wallTime;
  }

  protected void setWallTime(long newWallTime) {
    if (wallTime == newWallTime) return;
    long oldWallTime = wallTime;
    wallTime = newWallTime;
    notifyChanged(PROPNAME_WALL_TIME, oldWallTime, newWallTime);
  }

  public List<Long> getCellAnalyzeImageWaitTimes() {
    return Collections.unmodifiableList(cellAnalyzeImageWaitTimes);
  }

  public List<Long> getCellAnalyzeScoreTimes() {
    return Collections.unmodifiableList(cellAnalyzeScoreTimes);
  }

  public boolean wasSuccessful() {
    return successful;
  }

  protected void setSuccessful(boolean newSuccessful) {
    successful = newSuccessful;
  }

  protected void writeResultsFile() throws FileNotFoundException, PrintWriterErrorException {
    File resultsFile = createResultsFile();
    File resultsFileDir = resultsFile.getParentFile();
    if (resultsFileDir != null && !resultsFileDir.exists()) {
      throw new FileNotFoundException("Results file directory does not exist");
    }
    PrintWriter writer = null;
    try {
      writer = new PrintWriter(resultsFile);
      writeResults(writer);
      if (writer.checkError()) throw new PrintWriterErrorException();
    } finally {
      if (writer != null) writer.close();
    }
  }

  /**
   * Creates the file to which the result of running this rasterizer should be
   * written as HTML.
   *
   * @return file to which result should be written
   */
  protected abstract File createResultsFile();

  /**
   * Writes the result of running this rasterizer as HTML using the specified
   * writer.
   *
   * @param writer writer for writing the result
   */
  protected abstract void writeResults(PrintWriter writer);

  protected static long nanosToMillis(long nanoseconds) {
    return nanoseconds / 1000000L;
  }

  protected static long millisToNanos(long milliseconds) {
    return TimeUnit.MILLISECONDS.toNanos(milliseconds);
  }
}
