package imcacat.imcadr;

import gov.aps.jca.TimeoutException;

import java.io.File;

public interface BeamlineDriver {
  void release();

  String getImageFileNameExtension();

  void disableAutobp() throws InterruptedException, DriverException;

  void enableAutobp() throws InterruptedException, DriverException;

  /**
   * Gets the detector readout time in seconds.
   *
   * @return the readout time of the detector in seconds
   */
  double getDetectorReadoutTime();

  double getMinVelo(BeamlineAxis axis) throws InterruptedException, DriverException;

  double getMaxVelo(BeamlineAxis axis) throws InterruptedException, DriverException;

  double getVelo(BeamlineAxis axis) throws InterruptedException, DriverException;

  void setVelo(BeamlineAxis axis, double desiredVelo) throws InterruptedException, DriverException;

  double getAccelTime(BeamlineAxis axis) throws InterruptedException, DriverException;

  double phiRbv() throws InterruptedException, DriverException;

  void movePhi(double desiredPosition) throws InterruptedException, DriverException;

  double xRbv() throws InterruptedException, DriverException;

  void moveX(double desiredPosition) throws InterruptedException, DriverException;

  double yRbv() throws InterruptedException, DriverException;

  void moveY(double desiredPosition) throws InterruptedException, DriverException;

  double zRbv() throws InterruptedException, DriverException;

  void moveZ(double desiredPosition) throws InterruptedException, DriverException;

  double rbv(BeamlineAxis axis) throws InterruptedException, DriverException;

  void move(BeamlineAxis axis, double desiredPosition) throws InterruptedException,
      DriverException;

  void enterCollectDataExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException;

  void enterIdleExperimentState() throws InterruptedException, DriverException,
      ExperimentStateNotAchievedException;

  /**
   * Starts a detector dataset.
   * <p>
   * Initializes per-dataset parameters, automatically takes dark images if
   * needed, and makes the driver ready for image acquisitions with the
   * {@link #takeDetectorImage(File, boolean)} method.
   * <p>
   * When finished collecting a dataset, must call the
   * {@link #endDetectorDataset()} method.
   *
   * @param exposureTime exposure time in seconds
   */
  void beginDetectorDataset(double exposureTime) throws InterruptedException, DriverException;

  /**
   * Starts a detector dataset.
   * <p>
   * Initializes per-dataset parameters, automatically takes dark images if
   * needed, and makes the driver ready for image acquisitions with the
   * {@link #takeDetectorImage(File, boolean)} method or a move that triggers
   * an acquisition.
   * <p>
   * When finished collecting a dataset, must call the
   * {@link #endDetectorDataset()} method.
   *
   * @param exposureTime exposure time in seconds
   * @param exposurePeriod exposure period in seconds
   * @param exposureSyncAxis axis to use for exposure and exposure-shutter
   *          synchronization
   * @param numImages number of images in dataset
   * @param imageSeqFilePath path indicating where image sequence should be
   *          written; if {@code null}, the path must be specified in the
   *          {@code takeDetectorImage} method when an image is acquired
   */
  void beginDetectorDataset(double exposureTime, double exposurePeriod,
      BeamlineAxis exposureSyncAxis, int numImages, File imageSeqFilePath)
          throws InterruptedException, DriverException;

  /**
   * Ends a detector dataset.
   *
   * @see #beginDetectorDataset(double)
   */
  void endDetectorDataset() throws InterruptedException, DriverException;

  /**
   * Acquires an image from the detector within the dataset started via a
   * previous call to the {@link #beginDetectorDataset(double)} method.
   *
   * @param imageFilePath path of file to write image to
   * @param isLastImage whether is the last image in the dataset
   */
  void takeDetectorImage(File imageFilePath, boolean isLastImage) throws InterruptedException,
      DriverException, TimeoutException, TooManyDetectorAcquisitionRetriesException;

  /**
   * Publishes that the specified detector image has been taken. It is not
   * necessary to call this method for an image acquired via
   * {@link #takeDetectorImage} since it already arranges for publishing that
   * the image has been taken. However, for the case of an image acquisition
   * sequence triggered by a move, no automatic publishing takes place; this
   * method provides a way to publish that such images have been taken.
   *
   * @param imageFilePath path of file that has been taken
   */
  void publishDetectorImageTaken(File imageFilePath) throws InterruptedException, DriverException;

  /**
   * Sets the specified fly-scan parameters and arms the system for a future
   * move that will trigger the fly-scan acquisition sequence.
   * <p>
   * When finished, the caller must call {@link #endFlyScan()} to disarm the
   * system.
   *
   * @param shutterOpenPos the position at which the exposure shutter should be
   *          commanded to open
   * @param shutterClosePos the position at which the exposure shutter should
   *          be commanded to close
   * @param exposureTriggerPos the position at which the detector should be
   *          commanded to start the exposure sequence
   */
  void beginFlyScan(double shutterOpenPos, double shutterClosePos, double exposureTriggerPos)
      throws InterruptedException, DriverException;

  /**
   * Ends a fly-scan started with the
   * {@link #beginFlyScan(double, double, double)} call.
   */
  void endFlyScan() throws InterruptedException, DriverException;

  /**
   * Forcefully ends immediately a fly-scan started with the
   * {@link #beginFlyScan(double, double, double)} call.
   */
  void endFlyScanNow() throws InterruptedException, DriverException;
}
