package imcacat.imcadr;

import java.awt.Dimension;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import joptsimple.OptionException;
import joptsimple.OptionParser;
import joptsimple.OptionSet;

/**
 * I parse this application's command line arguments.
 */
public class CliArgumentParser {
  private static final String CONF_FILE = "conf-file";
  private static final String DEBUG_FILE = "debug-file";
  private static final String EXP_TIME = "exp-time";
  private static final String CELL_SIZE = "cell-size";
  private static final String ZY_SIZE = "zy-size";
  private static final String ZY_PRE_CLIP_SIZE = "zy-pre-clip-size";
  private static final String ZY_MASK = "zy-mask";
  private static final String ZX_SIZE = "zx-size";
  private static final String ZX_PRE_CLIP_SIZE = "zx-pre-clip-size";
  private static final String ZX_MASK = "zx-mask";
  private static final String ZX_FIRST = "zx-first";
  private static final String STEP_SCAN = "step-scan";
  private static final String NO_CHASE = "no-chase";
  private static final String BEST_SCORE_CUTOFF = "best-score-cutoff";
  private static final String SCORING_PKG = "scoring-pkg";
  private static final String WEAK_DIFF = "weak-diff";
  private static final String HIGH_RES_CUTOFF = "high-res-cutoff";
  private static final String LOW_RES_CUTOFF = "low-res-cutoff";
  private static final String MIN_SPOT_AREA = "min-spot-area";
  private static final String MIN_SIGNAL_HEIGHT = "min-signal-height";
  private static final String SERVER = "server";
  private static final String HELP = "help";
  private static final String VERSION = "version";
  private static final HashSet<String> ALL_OPTION_NAMES = new HashSet<String>(
      Arrays.asList(new String[] {
          CONF_FILE,
          DEBUG_FILE,
          EXP_TIME,
          CELL_SIZE,
          ZY_SIZE,
          ZY_PRE_CLIP_SIZE,
          ZY_MASK,
          ZX_SIZE,
          ZX_PRE_CLIP_SIZE,
          ZX_MASK,
          ZX_FIRST,
          STEP_SCAN,
          NO_CHASE,
          BEST_SCORE_CUTOFF,
          SCORING_PKG,
          WEAK_DIFF,
          HIGH_RES_CUTOFF,
          LOW_RES_CUTOFF,
          MIN_SPOT_AREA,
          MIN_SIGNAL_HEIGHT,
          SERVER,
          HELP,
          VERSION}));

  private static final Map<String, AnalysisPackage> STRING_TO_ANALYSIS_PACKAGE_MAP =
      createStringToAnalysisPackageMap();
  private static final OptionParser PARSER = createOptionParser();

  private CliArgumentParser() {
  }

  public static CliArguments parse(String[] args) throws CliArgumentException {
    try {
      OptionSet options = PARSER.parse(args);
      CliArguments result = new CliArguments();
      result.help(parseHelp(options));
      if (result.help()) return result;
      result.version(parseVersion(options));
      if (result.version()) return result;
      result.configurationFile(parseConfigurationFile(options));
      result.debugFile(parseDebugFile(options));
      result.server(parseServer(options));
      if (result.server()) return result;
      result.exposureTime(parseExposureTime(options));
      result.cellSize(parseCellSize(options));
      result.zySize(parseZySize(options));
      result.zyPreClipSize(parseZyPreClipSize(options));
      result.zyMask(parseZyMask(options));
      result.zxSize(parseZxSize(options));
      result.zxPreClipSize(parseZxPreClipSize(options));
      result.zxMask(parseZxMask(options));
      result.zxFirst(parseZxFirst(options));
      result.flyScanEnabled(!parseStepScan(options));
      result.chasingEnabled(!parseNoChase(options));
      result.bestScoreCutoff(parseBestScoreCutoff(options));
      result.analysisPackage(parseScoringPkg(options));
      result.weakDiffracter(parseWeakDiff(options));
      result.highResolutionCutoff(parseHighResCutoff(options));
      result.lowResolutionCutoff(parseLowResCutoff(options));
      result.minimumSpotArea(parseMinSpotArea(options));
      result.minimumSignalHeight(parseMinSignalHeight(options));
      result.imagesDirectory(parseImagesDirectory(options));
      if (hasNone(options)) return result;
      ensureHas(options, EXP_TIME);
      ensureHas(options, CELL_SIZE);
      ensureHasAtLeastOne(options, ZY_SIZE, ZX_SIZE);
      ensureHasIfHas(options, ZY_SIZE, ZY_PRE_CLIP_SIZE);
      ensureBoundedBy(options, ZY_SIZE, ZY_PRE_CLIP_SIZE, result.zySize(),
          result.zyPreClipSize());
      ensureHasIfHas(options, ZY_SIZE, ZY_MASK);
      ensureBoundedBy(options, ZY_MASK, ZY_SIZE, result.zyMask(), result.zySize());
      ensureHasIfHas(options, ZX_SIZE, ZX_PRE_CLIP_SIZE);
      ensureBoundedBy(options, ZX_SIZE, ZX_PRE_CLIP_SIZE, result.zxSize(),
          result.zxPreClipSize());
      ensureHasIfHas(options, ZX_SIZE, ZX_MASK);
      ensureBoundedBy(options, ZX_MASK, ZX_SIZE, result.zxMask(), result.zxSize());
      return result;
    } catch (OptionException e) {
      if (e.getMessage() != null) throw new CliArgumentException(e.getMessage(), e);
      else throw new CliArgumentException("Failed to parse CLI arguments", e);
    }
  }

  private static boolean parseHelp(OptionSet options) {
    return options.has(HELP);
  }

  private static boolean parseVersion(OptionSet options) {
    return options.has(VERSION);
  }

  private static boolean parseServer(OptionSet options) {
    return options.has(SERVER);
  }

  private static File parseConfigurationFile(OptionSet options) throws CliArgumentException {
    String result = (String)options.valueOf(CONF_FILE);
    if (result == null) return null;
    if (result.isEmpty()) {
      throw new CliArgumentException("Option '" + CONF_FILE + "' value must not be empty");
    }
    return new File(result);
  }

  private static File parseDebugFile(OptionSet options) throws CliArgumentException {
    String result = (String)options.valueOf(DEBUG_FILE);
    if (result == null) return null;
    if (result.isEmpty()) {
      throw new CliArgumentException("Option '" + DEBUG_FILE + "' value must not be empty");
    }
    return new File(result);
  }

  private static Double parseExposureTime(OptionSet options) throws CliArgumentException {
    Double result = (Double)options.valueOf(EXP_TIME);
    if (result == null) return null;
    if (result > 0.0) return result;
    throw new CliArgumentException("Option '" + EXP_TIME + "' value must be > 0");
  }

  private static Dimension parseCellSize(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(CELL_SIZE);
    if (value == null) return null;
    return parseDimension(value, CELL_SIZE);
  }

  private static Dimension parseZySize(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZY_SIZE);
    if (value == null) return null;
    return parseDimension(value, ZY_SIZE);
  }

  private static Dimension parseZyPreClipSize(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZY_PRE_CLIP_SIZE);
    if (value == null) return null;
    return parseDimension(value, ZY_PRE_CLIP_SIZE);
  }

  private static Raster2dMask parseZyMask(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZY_MASK);
    if (value == null) return null;
    return parseMask(value, ZY_MASK);
  }

  private static Dimension parseZxSize(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZX_SIZE);
    if (value == null) return null;
    return parseDimension(value, ZX_SIZE);
  }

  private static Dimension parseZxPreClipSize(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZX_PRE_CLIP_SIZE);
    if (value == null) return null;
    return parseDimension(value, ZX_PRE_CLIP_SIZE);
  }

  private static Raster2dMask parseZxMask(OptionSet options) throws CliArgumentException {
    String value = (String)options.valueOf(ZX_MASK);
    if (value == null) return null;
    return parseMask(value, ZX_MASK);
  }

  private static Dimension parseDimension(String dim, String optName)
      throws CliArgumentException {
    String[] parts = dim.split("[xX]");
    if (parts.length != 2) {
      throw new CliArgumentException("Invalid '" + optName + "' option value: " + dim);
    }
    try {
      int width = Integer.parseInt(parts[0].trim());
      if (width <= 0) {
        throw new CliArgumentException("Option '" + optName + "' value width must be > 0");
      }
      int height = Integer.parseInt(parts[1].trim());
      if (height <= 0) {
        throw new CliArgumentException("Option '" + optName + "' value height must be > 0");
      }
      return new Dimension(width, height);
    } catch (NumberFormatException e) {
      throw new CliArgumentException("Invalid '" + optName + "' option value: " + dim, e);
    }
  }

  /*
   * Parses a raster mask string like:
   *
   * 1:1,2,5,6;2:1;5:1,2,6,7
   *
   * Rows are separated by a semicolon. The row number is 1-based (i.e.
   * starting at 1, not 0). The row number is separated from the column numbers
   * by a colon. The column numbers are 1-based (i.e. starting at 1, not 0).
   * The column numbers are separated by a comma. No whitespace is allowed.
   * Empty rows are not allowed. Gaps in unmasked columns and rows are removed.
   * See Raster2dMask.createsGap() for an explanation of what a gap is.
   */
  private static Raster2dMask parseMask(String mask, String optName) throws CliArgumentException {
    Raster2dMask result = new Raster2dMask();
    String[] rows = mask.split(";");
    for (String row : rows) {
      String[] rowParts = row.split(":");
      if (rowParts.length != 2) {
        throw new CliArgumentException("Invalid '" + optName + "' option value row: " + row);
      }
      int rowNum;
      try {
        rowNum = Integer.parseInt(rowParts[0]);
      } catch (NumberFormatException e) {
        throw new CliArgumentException("Invalid '" + optName + "' option value row number: " +
            rowParts[0]);
      }
      if (rowNum <= 0) {
        throw new CliArgumentException("Option '" + optName + "' value row number must be > 0");
      }
      String[] cols = rowParts[1].split(",");
      for (String col : cols) {
        int colNum;
        try {
          colNum = Integer.parseInt(col);
        } catch (NumberFormatException e) {
          throw new CliArgumentException("Invalid '" + optName +
              "' option value column number: " + col);
        }
        if (colNum <= 0) {
          throw new CliArgumentException("Option '" + optName +
              "' value column number must be > 0");
        }
        result.set(colNum - 1, rowNum - 1, true);
      }
    }
    result.removeGaps();
    return result;
  }

  private static boolean parseZxFirst(OptionSet options) {
    return options.has(ZX_FIRST);
  }

  private static boolean parseStepScan(OptionSet options) {
    return options.has(STEP_SCAN);
  }

  private static boolean parseNoChase(OptionSet options) {
    return options.has(NO_CHASE);
  }

  private static Integer parseBestScoreCutoff(OptionSet options) {
    return (Integer)options.valueOf(BEST_SCORE_CUTOFF);
  }

  private static AnalysisPackage parseScoringPkg(OptionSet options) throws CliArgumentException {
    String resultString = (String)options.valueOf(SCORING_PKG);
    if (resultString == null) return null;
    AnalysisPackage result = STRING_TO_ANALYSIS_PACKAGE_MAP.get(resultString);
    if (result != null) return result;
    throw new CliArgumentException(String.format(Locale.US, "Option '%s' value is invalid",
        SCORING_PKG));
  }

  private static boolean parseWeakDiff(OptionSet options) {
    return options.has(WEAK_DIFF);
  }

  private static Double parseHighResCutoff(OptionSet options) throws CliArgumentException {
    Double result = (Double)options.valueOf(HIGH_RES_CUTOFF);
    if (result == null) return null;
    double min = Resources.getInstance().getHighResolutionCutoffMin();
    double max = Resources.getInstance().getHighResolutionCutoffMax();
    if (result >= min && result <= max) return result;
    throw new CliArgumentException(String.format(Locale.US,
        "Option '%s' value must be between %.2f and %.2f", HIGH_RES_CUTOFF, min, max));
  }

  private static Double parseLowResCutoff(OptionSet options) throws CliArgumentException {
    Double result = (Double)options.valueOf(LOW_RES_CUTOFF);
    if (result == null) return null;
    double min = Resources.getInstance().getLowResolutionCutoffMin();
    double max = Resources.getInstance().getLowResolutionCutoffMax();
    if (result >= min && result <= max) return result;
    throw new CliArgumentException(String.format(Locale.US,
        "Option '%s' value must be between %.2f and %.2f", LOW_RES_CUTOFF, min, max));
  }

  private static Integer parseMinSpotArea(OptionSet options) throws CliArgumentException {
    Integer result = (Integer)options.valueOf(MIN_SPOT_AREA);
    if (result == null) return null;
    int min = Resources.getInstance().getMinimumSpotAreaMin();
    int max = Resources.getInstance().getMinimumSpotAreaMax();
    if (result >= min && result <= max) return result;
    throw new CliArgumentException(String.format(Locale.US,
        "Option '%s' value must be between %d and %d", MIN_SPOT_AREA, min, max));
  }

  private static Double parseMinSignalHeight(OptionSet options) throws CliArgumentException {
    Double result = (Double)options.valueOf(MIN_SIGNAL_HEIGHT);
    if (result == null) return null;
    double min = Resources.getInstance().getMinimumSignalHeightMin();
    double max = Resources.getInstance().getMinimumSignalHeightMax();
    if (result >= min && result <= max) return result;
    throw new CliArgumentException(String.format(Locale.US,
        "Option '%s' value must be between %.2f and %.2f", MIN_SIGNAL_HEIGHT, min, max));
  }

  private static File parseImagesDirectory(OptionSet options) throws CliArgumentException {
    List<String> nonOptionArgs = options.nonOptionArguments();
    if (nonOptionArgs.isEmpty()) return null;
    if (nonOptionArgs.size() == 1) {
      String imagesDir = nonOptionArgs.get(0);
      if (!imagesDir.isEmpty()) return new File(imagesDir);
      throw new CliArgumentException("Images directory must not be the empty string");
    }
    throw new CliArgumentException("Unexpected command line arguments");
  }

  private static void ensureHas(OptionSet options, String optName) throws CliArgumentException {
    if (options.has(optName)) return;
    throw new CliArgumentException("Option '" + optName + "' is required");
  }

  private static void ensureHasAtLeastOne(OptionSet options, String... optNames)
      throws CliArgumentException {
    if (hasAny(options, optNames)) return;
    StringBuilder buffer = new StringBuilder();
    int lastIndex = optNames.length - 1;
    for (int i = 0; i < optNames.length; i++) {
      if (i != 0) {
        buffer.append((optNames.length == 2) ? " " : ", ");
        if (i == lastIndex) buffer.append("or ");
      }
      buffer.append('\'');
      buffer.append(optNames[i]);
      buffer.append('\'');
    }
    throw new CliArgumentException("At least one of the following options is required: " +
        buffer.toString());
  }

  private static void ensureHasIfHas(OptionSet options, String optName, String ifOptName)
      throws CliArgumentException {
    if (!options.has(ifOptName)) return;
    if (options.has(optName)) return;
    throw new CliArgumentException("Option '" + optName + "' is required by option '" +
        ifOptName + "'");
  }

  private static void ensureBoundedBy(OptionSet options, String boundeeOptName,
      String bounderOptName, Dimension boundee, Dimension bounder) throws CliArgumentException {
    if (boundee == null || bounder == null) return;
    if (boundee.getWidth() > bounder.getWidth()) {
      throw new CliArgumentException("Option '" + boundeeOptName +
          "' width must be <= option '" + bounderOptName + "' width");
    }
    if (boundee.getHeight() > bounder.getHeight()) {
      throw new CliArgumentException("Option '" + boundeeOptName +
          "' height must be <= option '" + bounderOptName + "' height");
    }
  }

  private static void ensureBoundedBy(OptionSet options, String boundeeOptName,
      String bounderOptName, Raster2dMask boundee, Dimension bounder)
      throws CliArgumentException {
    if (boundee == null || bounder == null) return;
    if ((boundee.maxColumn() + 1) > bounder.getWidth()) {
      throw new CliArgumentException("Option '" + boundeeOptName +
          "' width must be <= option '" + bounderOptName + "' width");
    }
    if ((boundee.maxRow() + 1) > bounder.getHeight()) {
      throw new CliArgumentException("Option '" + boundeeOptName +
          "' height must be <= option '" + bounderOptName + "' height");
    }
  }

  private static boolean hasAny(OptionSet options, String... optNames) {
    for (String each : optNames)
      if (options.has(each)) return true;
    return false;
  }

  private static boolean hasNone(OptionSet options) {
    return hasNoneOtherThan(options);
  }

  private static boolean hasNoneOtherThan(OptionSet options, String... optNames) {
    HashSet<String> optNamesSet = new HashSet<String>(Arrays.asList(optNames));
    for (String each : ALL_OPTION_NAMES) {
      if (optNamesSet.contains(each)) continue;
      if (options.has(each)) return false;
    }
    return true;
  }

  public static void printHelp(OutputStream sink) throws IOException {
    Writer sinkWriter = new OutputStreamWriter(sink);
    try {
      printHelp(sinkWriter);
    } finally {
      sinkWriter.close();
    }
  }

  public static void printHelp(Writer sink) throws IOException {
    String newline = Resources.getLineSeparator();
    String[] lines = {
        "usage: imcadr [--best-score-cutoff=number] [--cell-size=size]",
        "         [--conf-file=file] [--debug-file=file] [--exp-time=number]",
        "         [--help] [--high-res-cutoff=number] [--low-res-cutoff=number]",
        "         [--min-signal-height=number] [--min-spot-area=number]",
        "         [--no-chase] [--scoring-pkg=package] [--server] [--step-scan]",
        "         [--version] [--weak-diff] [--zx-first] [--zx-mask=mask]",
        "         [--zx-pre-clip-size=size] [--zx-size=size] [--zy-mask=mask]",
        "         [--zy-pre-clip-size=size] [--zy-size=size] [directory]"};
    for (String each : lines) {
      sink.write(each);
      sink.write(newline);
    }
  }

  private static Map<String, AnalysisPackage> createStringToAnalysisPackageMap() {
    HashMap<String, AnalysisPackage> result = new HashMap<String, AnalysisPackage>();
    for (AnalysisPackage each : AnalysisPackage.values()) {
      switch (each) {
        case LABELIT:
          result.put("labelit", each);
          break;
        case DTREK:
          result.put("dtrek", each);
          /*
           * Backward compatibility: accept deprecated name.
           */
          result.put("dstartrek", each);
          break;
      }
    }
    return result;
  }

  private static OptionParser createOptionParser() {
    OptionParser result = new OptionParser();
    result.accepts(CONF_FILE).withRequiredArg();
    result.accepts(DEBUG_FILE).withRequiredArg();
    result.accepts(EXP_TIME).withRequiredArg().ofType(Double.class);
    result.accepts(CELL_SIZE).withRequiredArg();
    result.accepts(ZY_SIZE).withRequiredArg();
    result.accepts(ZY_PRE_CLIP_SIZE).withRequiredArg();
    result.accepts(ZY_MASK).withRequiredArg();
    result.accepts(ZX_SIZE).withRequiredArg();
    result.accepts(ZX_PRE_CLIP_SIZE).withRequiredArg();
    result.accepts(ZX_MASK).withRequiredArg();
    result.accepts(ZX_FIRST);
    result.accepts(STEP_SCAN);
    result.accepts(NO_CHASE);
    result.accepts(BEST_SCORE_CUTOFF).withRequiredArg().ofType(Integer.class);
    result.accepts(SCORING_PKG).withRequiredArg();
    result.accepts(WEAK_DIFF);
    result.accepts(HIGH_RES_CUTOFF).withRequiredArg().ofType(Double.class);
    result.accepts(LOW_RES_CUTOFF).withRequiredArg().ofType(Double.class);
    result.accepts(MIN_SPOT_AREA).withRequiredArg().ofType(Integer.class);
    result.accepts(MIN_SIGNAL_HEIGHT).withRequiredArg().ofType(Double.class);
    result.accepts(SERVER);
    result.acceptsAll(Arrays.asList("h", HELP));
    result.accepts(VERSION);
    return result;
  }
}
