package imcacat.imcadr;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * I am a service providing job execution on a computer cluster.
 * <p>
 * It is the responsibility of each {@link ClusterRunnable} submitted for
 * execution to catch and handle (e.g., save, log, etc.) any exceptions it
 * wants to in its {@code run()} method when it runs. If its {@code run()}
 * method throws an exception, this service will catch it and silently drop it.
 */
public class ClusterService {
  private static final Logger LOG = Logger.getLogger(ClusterService.class.getName());
  private static final ClusterRunnable SHUT_DOWN_SENTINEL = new NullClusterRunnable();

  private final List<Node> nodes;
  private final Semaphore clusterJobPermit;
  private final BlockingQueue<ClusterRunnable> waiting;
  private final ExecutorService running;
  private final ExecutorService helpers;
  private final Object lock;
  /* Guarded by lock. */
  private boolean initialized;
  /* Guarded by lock. */
  private boolean shutDown;

  /**
   * @throws IllegalArgumentException if {@code nodes} is empty
   * @throws NullPointerException if {@code nodes} is {@code null}
   */
  public ClusterService(List<ClusterNode> nodes) {
    super();
    if (Objects.requireNonNull(nodes).isEmpty()) {
      throw new IllegalArgumentException("nodes is empty");
    }
    this.nodes = new ArrayList<>(nodes.size());
    for (ClusterNode each : nodes) {
      this.nodes.add(new Node(each));
    }
    clusterJobPermit = new Semaphore(computeClusterJobCapacity(nodes));
    waiting = new LinkedBlockingQueue<>();
    running = Executors.newCachedThreadPool();
    helpers = Executors.newCachedThreadPool();
    lock = new Object();
    initialized = false;
    shutDown = false;
  }

  private void initialize() {
    synchronized (lock) {
      if (initialized || shutDown) return;
      helpers.execute(newScheduleHelper(this));
      initialized = true;
    }
  }

  /**
   * @throws NullPointerException if {@code r} is {@code null}
   * @throws RejectedExecutionException if this service is shut down
   */
  public void execute(ClusterRunnable r) {
    initialize();
    synchronized (lock) {
      if (shutDown) {
        throw new RejectedExecutionException("Service is shut down");
      }
      waiting.add(r);
    }
  }

  public boolean awaitTermination(long timeout, TimeUnit unit) throws InterruptedException {
    return running.awaitTermination(timeout, unit);
  }

  public boolean isShutDown() {
    synchronized (lock) {
      return shutDown;
    }
  }

  public boolean isTerminated() {
    return running.isTerminated();
  }

  public void shutDown() {
    synchronized (lock) {
      if (shutDown) return;
      if (initialized) {
        waiting.add(SHUT_DOWN_SENTINEL);
      } else {
        running.shutdown();
        helpers.shutdownNow();
      }
      shutDown = true;
    }
  }

  public List<ClusterRunnable> shutDownNow() {
    synchronized (lock) {
      boolean interrupted = false;
      helpers.shutdownNow();
      try {
        /*
         * Wait for the helpers to stop before draining the queues to ensure
         * no more jobs get scheduled.  All helpers terminate immediately,
         * so the wait should be short.
         */
        helpers.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
      } catch (InterruptedException e) {
        interrupted = true;
      }
      List<ClusterRunnable> result = new ArrayList<>(waiting.size());
      waiting.drainTo(result);
      /* Remove the shut-down sentinel if shutDown() was called. */
      result.remove(SHUT_DOWN_SENTINEL);
      running.shutdownNow();
      shutDown = true;
      if (interrupted) Thread.currentThread().interrupt();
      return result;
    }
  }

  private Node findNodeWithLowestLoad() {
    Node result = nodes.get(0);
    double resultLoad = result.load();
    if (resultLoad == 0.0) return result;
    for (Node each : nodes.subList(1, nodes.size())) {
      double eachLoad = each.load();
      if (eachLoad == 0.0) return each;
      if (eachLoad < resultLoad) {
        result = each;
        resultLoad = eachLoad;
      }
    }
    return result;
  }

  /*
   * Helper: runs in a separate thread.
   */
  private void schedule() throws InterruptedException {
    try {
      for (;;) {
        clusterJobPermit.acquire();
        ClusterRunnable r = waiting.take();
        if (r == SHUT_DOWN_SENTINEL) {
          running.shutdown();
          /* Interrupts this thread, but that's OK because it's done. */
          helpers.shutdownNow();
          return;
        }
        Node node = findNodeWithLowestLoad();
        r.setClusterNode(node.hostname());
        node.incrementJobCount();
        running.execute(new Job(r, node, clusterJobPermit));
      }
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw e;
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      LOG.log(Level.SEVERE, "schedule thread failed", t);
    }
  }

  private static int computeClusterJobCapacity(List<ClusterNode> nodes) {
    int result = 0;
    for (ClusterNode each : nodes) {
      result += each.jobCapacity();
    }
    return result;
  }

  private static Runnable newScheduleHelper(final ClusterService s) {
    return new Runnable() {
      @Override
      public void run() {
        try {
          s.schedule();
        } catch (InterruptedException e) {
          Thread.currentThread().interrupt();
        }
      }
    };
  }

  /**
   * I am a Null Object {@code ClusterRunnable} that does nothing.
   */
  private static class NullClusterRunnable implements ClusterRunnable {
    @Override
    public void run() {
    }

    @Override
    public void setClusterNode(String hostname) {
    }
  }

  /**
   * I am a computer cluster node that knows my current job count and load.
   */
  private static class Node {
    private final String hostname;
    private final int jobCapacity;
    private final Object lock;
    /* Guarded by lock. */
    private int jobCount;
    /* Guarded by lock. */
    private double load;

    public Node(ClusterNode n) {
      this(n.hostname(), n.jobCapacity());
    }

    public Node(String hostname, int jobCapacity) {
      this.hostname = hostname;
      this.jobCapacity = jobCapacity;
      lock = new Object();
      jobCount = 0;
      load = 0.0;
    }

    public String hostname() {
      return hostname;
    }

    @SuppressWarnings("unused")
    public int jobCapacity() {
      return jobCapacity;
    }

    public int jobCount() {
      synchronized (lock) {
        return jobCount;
      }
    }

    public double load() {
      synchronized (lock) {
        return load;
      }
    }

    public void incrementJobCount() {
      synchronized (lock) {
        jobCount++;
        updateLoad();
      }
    }

    public void decrementJobCount() {
      synchronized (lock) {
        jobCount--;
        updateLoad();
      }
    }

    private void updateLoad() {
      synchronized (lock) {
        load = (double)jobCount / (double)jobCapacity;
      }
    }

    @Override
    public boolean equals(Object o) {
      if (o == this) return true;
      if (!(o instanceof Node)) return false;
      Node n = (Node)o;
      if (!hostname.equals(n.hostname)) return false;
      if (jobCapacity != n.jobCapacity) return false;
      return jobCount() == n.jobCount();
    }

    @Override
    public int hashCode() {
      int result = 17;
      result = 31 * result + hostname.hashCode();
      result = 31 * result + jobCapacity;
      result = 31 * result + jobCount();
      return result;
    }

    @Override
    public String toString() {
      return (new StringBuilder())
          .append("hostname=")
          .append(hostname)
          .append(", ")
          .append("jobCapacity=")
          .append(jobCapacity)
          .append(", ")
          .append("jobCount=")
          .append(jobCount())
          .append(", ")
          .append("load=")
          .append(load())
          .toString();
    }
  }

  /**
   * I am a {@code Runnable} Decorator that knows which {@code Node} my
   * component {@code Runnable} runs on and which {@code Semaphore} is managing
   * the computer cluster job permits so that I can decrement the job count of
   * the {@code Node} and release a {@code Semaphore} job permit when my
   * component finishes.
   */
  private static class Job implements Runnable {
    private final Runnable component;
    private final Node node;
    private final Semaphore clusterJobPermit;

    public Job(Runnable component, Node node, Semaphore clusterJobPermit) {
      super();
      this.component = Objects.requireNonNull(component);
      this.node = Objects.requireNonNull(node);
      this.clusterJobPermit = Objects.requireNonNull(clusterJobPermit);
    }

    @Override
    public void run() {
      try {
        component.run();
      } catch (ThreadDeath td) {
        throw td;
      } catch (Throwable t) {
        /* IGNORED */
      } finally {
        node.decrementJobCount();
        clusterJobPermit.release();
      }
    }

    @Override
    public String toString() {
      return (new StringBuilder()).append("component=").append(component).toString();
    }
  }
}
