package imcacat.imcadr;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * I parse the output of the <a
 * href="http://www.rigaku.com/products/protein/dtrek/">d*TREK</a>
 * <tt>dtranker</tt> program.
 */
public class DTrekDtrankerParser extends ImageAnalysisParser {
  private static final Charset UTF8 = Charset.forName("UTF-8");

  private static final Pattern SECTION_START_RE =
      newPattern("^\\s*([a-z]+)\\s*:\\s*Copyright.*");
  private static final Pattern UNIT_CELL_LENGTHS_RE =
      newPattern("^\\s*Unit cell\\s+lengths:\\s*([0-9.]+)\\s+([0-9.]+)\\s+([0-9.]+).*");
  private static final Pattern UNIT_CELL_ANGLES_RE =
      newPattern("^\\s*Unit cell\\s+angles:\\s*([0-9.]+)\\s+([0-9.]+)\\s+([0-9.]+).*");
  private static final Pattern MOSAICITY_RE = newPattern("^\\s*Mosaicity:\\s*([0-9.]+).*");
  private static final Pattern SPACE_GROUP_NUMBER_RE =
      newPattern("^\\s*Space\\s*group\\s+number:.*");
  private static final Pattern SPACE_GROUP_NAME_RE =
      newPattern("^\\s*name:\\s*([a-zA-Z0-9]+).*");
  private static final Pattern RESOLUTION_STATS_HEADING_RE =
      newPattern("^\\s*Resolution statistics(?:(?:\\s.*)|$)");
  private static final Pattern DASH_LINE_RE = newPattern("^\\s*-+\\s*$");
  private static final Pattern RESOLUTION_STATS_ROW_RE =
      newPattern("^\\s*[0-9.]+\\s*-\\s*([0-9.]+)\\s+\\d+\\s+\\d+\\s+([0-9.]+).*");
  private static final Pattern POINT_TOTAL_RE =
      newPattern("^\\s*Cumulative for\\s.*\\s+([0-9-]+)\\s*$");

  private static final HashMap<String, SpaceGroup> DATA_TO_SPACE_GROUP_MAP;

  static {
    DATA_TO_SPACE_GROUP_MAP = new HashMap<String, SpaceGroup>();
    DATA_TO_SPACE_GROUP_MAP.put("P1", SpaceGroup.P1);
    DATA_TO_SPACE_GROUP_MAP.put("P2", SpaceGroup.P2);
    DATA_TO_SPACE_GROUP_MAP.put("P222", SpaceGroup.P222);
    DATA_TO_SPACE_GROUP_MAP.put("P4", SpaceGroup.P4);
    DATA_TO_SPACE_GROUP_MAP.put("P6", SpaceGroup.P6);
    DATA_TO_SPACE_GROUP_MAP.put("C222", SpaceGroup.C222);
    DATA_TO_SPACE_GROUP_MAP.put("C2", SpaceGroup.C2);
    DATA_TO_SPACE_GROUP_MAP.put("I4", SpaceGroup.I4);
    DATA_TO_SPACE_GROUP_MAP.put("I222", SpaceGroup.I222);
    DATA_TO_SPACE_GROUP_MAP.put("F222", SpaceGroup.F222);
    if (DATA_TO_SPACE_GROUP_MAP.size() != SpaceGroup.values().length) {
      throw new BugException();
    }
  }

  private BufferedReader reader = null;
  private InputStreamReader inputStreamReader = null;
  private Section section = null;

  public DTrekDtrankerParser(InputStream in) {
    super(in);
  }

  @Override
  public ImageAnalysisResult run() {
    try {
      initializeReader();
      return parse();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
      return newAnalysisResult("Interrupted", e);
    } catch (IOException | RuntimeException e) {
      return newAnalysisResult("Failed to parse analysis program output", e);
    } finally {
      closeReaderSilently();
    }
  }

  private void initializeReader() {
    inputStreamReader = new InputStreamReader(in, UTF8);
    reader = new BufferedReader(inputStreamReader);
  }

  private void closeReaderSilently() {
    closeSilently(reader);
    closeSilently(inputStreamReader);
  }

  private ImageAnalysisResult parse() throws IOException {
    ImageAnalysisResult result = new ImageAnalysisResult();
    result.setUnitCell(parseUnitCell());
    result.setMosaicity(parseMosaicity());
    result.setSpaceGroup(parseSpaceGroup());
    result.setEstimatedResolution1(parseEstimatedResolution1());
    Integer pointTotal = parsePointTotal();
    result.setScore(computeScore(pointTotal));
    return result;
  }

  private UnitCell parseUnitCell() throws IOException {
    double[] lengths = null;
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = UNIT_CELL_LENGTHS_RE.matcher(line);
      if (!m.matches()) continue;
      lengths = new double[3];
      int groupOffset = 1;
      for (int i = 0; i < lengths.length; i++) {
        Double value = getDouble(m.group(i + groupOffset));
        if (value == null) return null;
        lengths[i] = value.doubleValue();
      }
      break;
    }
    double[] angles = null;
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = UNIT_CELL_ANGLES_RE.matcher(line);
      if (!m.matches()) continue;
      angles = new double[3];
      int groupOffset = 1;
      for (int i = 0; i < angles.length; i++) {
        Double value = getDouble(m.group(i + groupOffset));
        if (value == null) return null;
        angles[i] = value.doubleValue();
      }
      break;
    }
    return (lengths == null || angles == null) ? null : new UnitCell(lengths, angles);
  }

  private Double parseMosaicity() throws IOException {
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = MOSAICITY_RE.matcher(line);
      if (m.matches()) return getDouble(m.group(1));
    }
    return null;
  }

  private SpaceGroup parseSpaceGroup() throws IOException {
    boolean foundSpaceGroupNumber = false;
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = SPACE_GROUP_NUMBER_RE.matcher(line);
      if (!m.matches()) continue;
      foundSpaceGroupNumber = true;
      break;
    }
    if (!foundSpaceGroupNumber) return null;
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = SPACE_GROUP_NAME_RE.matcher(line);
      if (m.matches()) return DATA_TO_SPACE_GROUP_MAP.get(m.group(1));
    }
    return null;
  }

  private Double parseEstimatedResolution1() throws IOException {
    readResolutionStatisticsHeading(Section.DTREFINE);
    readDashLine(Section.DTREFINE);
    readDashLine(Section.DTREFINE);
    ArrayList<Double> resRangeMin = new ArrayList<Double>();
    ArrayList<Double> iOverSigmaAvg = new ArrayList<Double>();
    for (String line = readLineInDtrefine(); line != null; line = readLineInDtrefine()) {
      Matcher m = RESOLUTION_STATS_ROW_RE.matcher(line);
      if (!m.matches()) break;
      resRangeMin.add(getDouble(m.group(1)));
      iOverSigmaAvg.add(getDouble(m.group(2)));
    }
    if (resRangeMin.isEmpty()) return null;
    final double iOverSigmaAvgCutoff = 2.0;
    final int resRangeMinSize = resRangeMin.size();
    for (int i = 1; i < resRangeMinSize; i++) {
      Double eachIOverSigmaAvg = iOverSigmaAvg.get(i);
      if (eachIOverSigmaAvg == null) continue;
      if (eachIOverSigmaAvg.doubleValue() >= iOverSigmaAvgCutoff) continue;
      return resRangeMin.get(i - 1);
    }
    return resRangeMin.get(resRangeMinSize - 1);
  }

  private Integer parsePointTotal() throws IOException {
    for (String line = readLineInDtrank(); line != null; line = readLineInDtrank()) {
      Matcher m = POINT_TOTAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private String readLineInDtrefine() throws IOException {
    return readLine(Section.DTREFINE);
  }

  private String readLineInDtrank() throws IOException {
    return readLine(Section.DTRANK);
  }

  private void readResolutionStatisticsHeading(Section target) throws IOException {
    for (String line = readLine(target); line != null; line = readLine(target)) {
      Matcher m = RESOLUTION_STATS_HEADING_RE.matcher(line);
      if (m.matches()) return;
    }
  }

  private void readDashLine(Section target) throws IOException {
    for (String line = readLine(target); line != null; line = readLine(target)) {
      Matcher m = DASH_LINE_RE.matcher(line);
      if (m.matches()) return;
    }
  }

  private String readLine(Section target) throws IOException {
    for (;;) {
      if (beyond(target)) return null;
      String line = readLine();
      if (line == null) return null;
      if (in(target)) return line;
    }
  }

  private String readLine() throws IOException {
    String line = reader.readLine();
    if (line == null) return null;
    Matcher m = SECTION_START_RE.matcher(line);
    if (m.matches()) entered(Section.getInstance(m.group(1)));
    return line;
  }

  private boolean beyond(Section aSection) {
    if (section == null) return false;
    if (aSection == null) return false;
    return section.ordinal() > aSection.ordinal();
  }

  private boolean in(Section aSection) {
    if (section == null) return false;
    if (aSection == null) return false;
    return section.equals(aSection);
  }

  private void entered(Section newSection) {
    if (newSection == null) return;
    section = newSection;
  }

  private static Integer getInteger(String value) {
    try {
      return Integer.valueOf(value);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  private static Double getDouble(String value) {
    try {
      return Double.valueOf(value);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  private static Integer computeScore(Integer pointTotal) {
    return pointTotal;
  }

  private static enum Section {
    DTRANKER("dtranker"),
    DTEXTRACTHEADER("dtextractheader"),
    DTFIND("dtfind"),
    DTINDEX("dtindex"),
    DTREFINE("dtrefine"),
    DTRANK("dtrank");

    private static final Map<String, Section> ID_TO_SECTION;

    static {
      ID_TO_SECTION = new HashMap<String, Section>();
      for (Section each : values()) {
        ID_TO_SECTION.put(each.getId(), each);
      }
    }

    private final String id;

    Section(String id) {
      this.id = id;
    }

    public String getId() {
      return id;
    }

    public static Section getInstance(String id) {
      return ID_TO_SECTION.get(id);
    }
  }
}
