package imcacat.imcadr;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToBooleanEpicsCaValueConverter;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;

/**
 * I am a control interface for J. Lewis Muir's Exposync software.
 */
public class ExposyncDriver {
  protected static final String SHUTTER_OPEN_POS_PV_NAME = "OpenPos";
  protected static final String SHUTTER_OPEN_POS_RBV_PV_NAME = "OpenPosRbv";
  protected static final String SHUTTER_CLOSE_POS_PV_NAME = "ClosePos";
  protected static final String SHUTTER_CLOSE_POS_RBV_PV_NAME = "ClosePosRbv";
  protected static final String EXPOSURE_TRIGGER_ENABLED_PV_NAME = "ExpTrEna";
  protected static final String EXPOSURE_TRIGGER_ENABLED_RBV_PV_NAME = "ExpTrEnaRbv";
  protected static final String EXPOSURE_TRIGGER_POS_PV_NAME = "ExpTrPos";
  protected static final String EXPOSURE_TRIGGER_POS_RBV_PV_NAME = "ExpTrPosRbv";
  protected static final String ARMED_PV_NAME = "Armed";
  protected static final String ARMED_RBV_PV_NAME = "ArmedRbv";
  protected static final String MAX_EXPOSURE_TIME_PV_NAME = "MaxExpTm";
  protected static final String MAX_EXPOSURE_TIME_RBV_PV_NAME = "MaxExpTmRbv";
  protected static final String SHUTTER_PV_NAME = "Shutter";
  protected static final String EXPOSURE_TRIGGER_PV_NAME = "ExpTrig";

  protected static final short EXPOSURE_TRIGGER_ACTIVATE = 1;

  private volatile boolean released = false;
  protected EpicsPv shutterOpenPosPv;
  protected EpicsPv shutterOpenPosRbvPv;
  protected EpicsPv shutterClosePosPv;
  protected EpicsPv shutterClosePosRbvPv;
  protected EpicsPv exposureTriggerEnabledPv;
  protected EpicsPv exposureTriggerEnabledRbvPv;
  protected EpicsPv exposureTriggerPosPv;
  protected EpicsPv exposureTriggerPosRbvPv;
  protected EpicsPv armedPv;
  protected EpicsPv armedRbvPv;
  protected EpicsPv maxExposureTimePv;
  protected EpicsPv maxExposureTimeRbvPv;
  protected EpicsPv shutterPv;
  protected EpicsPv shutterRbvPv;
  protected EpicsPv exposureTriggerPv;

  public ExposyncDriver(String pvName, String shutterRbvPvName) {
    shutterOpenPosPv = new EpicsPv(pvName + SHUTTER_OPEN_POS_PV_NAME, false, false);
    shutterOpenPosRbvPv = new EpicsPv(pvName + SHUTTER_OPEN_POS_RBV_PV_NAME, false, true);
    shutterClosePosPv = new EpicsPv(pvName + SHUTTER_CLOSE_POS_PV_NAME, false, false);
    shutterClosePosRbvPv = new EpicsPv(pvName + SHUTTER_CLOSE_POS_RBV_PV_NAME, false, true);
    exposureTriggerEnabledPv = new EpicsPv(pvName + EXPOSURE_TRIGGER_ENABLED_PV_NAME, false, false,
        new ShortToBooleanEpicsCaValueConverter());
    exposureTriggerEnabledRbvPv = new EpicsPv(pvName + EXPOSURE_TRIGGER_ENABLED_RBV_PV_NAME, false,
        true, new ShortToBooleanEpicsCaValueConverter());
    exposureTriggerPosPv = new EpicsPv(pvName + EXPOSURE_TRIGGER_POS_PV_NAME, false, false);
    exposureTriggerPosRbvPv = new EpicsPv(pvName + EXPOSURE_TRIGGER_POS_RBV_PV_NAME, false, true);
    armedPv = new EpicsPv(pvName + ARMED_PV_NAME, false, false,
        new ShortToBooleanEpicsCaValueConverter());
    armedRbvPv = new EpicsPv(pvName + ARMED_RBV_PV_NAME, false, true,
        new ShortToBooleanEpicsCaValueConverter());
    maxExposureTimePv = new EpicsPv(pvName + MAX_EXPOSURE_TIME_PV_NAME, false, false);
    maxExposureTimeRbvPv = new EpicsPv(pvName + MAX_EXPOSURE_TIME_RBV_PV_NAME, false, true);
    shutterPv =
        new EpicsPv(pvName + SHUTTER_PV_NAME, false, false,
            new ShortToEnumEpicsCaValueConverter(ShutterState.class));
    shutterRbvPv =
        new EpicsPv(shutterRbvPvName, false, true, new ShortToEnumEpicsCaValueConverter(
            ShutterState.class));
    exposureTriggerPv = new EpicsPv(pvName + EXPOSURE_TRIGGER_PV_NAME, false, false);
  }

  @Override
  protected final void finalize() throws Throwable {
    try {
      release();
    } finally {
      super.finalize();
    }
  }

  public void release() {
    if (released) return;
    released = true;

    if (shutterOpenPosPv != null) shutterOpenPosPv.release();
    if (shutterOpenPosRbvPv != null) shutterOpenPosRbvPv.release();
    if (shutterClosePosPv != null) shutterClosePosPv.release();
    if (shutterClosePosRbvPv != null) shutterClosePosRbvPv.release();
    if (exposureTriggerEnabledPv != null) exposureTriggerEnabledPv.release();
    if (exposureTriggerEnabledRbvPv != null) exposureTriggerEnabledRbvPv.release();
    if (exposureTriggerPosPv != null) exposureTriggerPosPv.release();
    if (exposureTriggerPosRbvPv != null) exposureTriggerPosRbvPv.release();
    if (armedPv != null) armedPv.release();
    if (armedRbvPv != null) armedRbvPv.release();
    if (maxExposureTimePv != null) maxExposureTimePv.release();
    if (maxExposureTimeRbvPv != null) maxExposureTimeRbvPv.release();
    if (shutterPv != null) shutterPv.release();
    if (shutterRbvPv != null) shutterRbvPv.release();
    if (exposureTriggerPv != null) exposureTriggerPv.release();
  }

  public double getShutterOpenPosition() throws InterruptedException, DriverException {
    try {
      return (Double)shutterOpenPosRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setShutterOpenPosition(double pos) throws InterruptedException, DriverException {
    try {
      shutterOpenPosPv.setValueAndWait(pos);
      shutterOpenPosRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getShutterClosePosition() throws InterruptedException, DriverException {
    try {
      return (Double)shutterClosePosRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setShutterClosePosition(double pos) throws InterruptedException, DriverException {
    try {
      shutterClosePosPv.setValueAndWait(pos);
      shutterClosePosRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public boolean isExposureTriggerEnabled() throws InterruptedException, DriverException {
    try {
      return (Boolean)exposureTriggerEnabledRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setExposureTriggerEnabled(boolean ena) throws InterruptedException, DriverException {
    try {
      exposureTriggerEnabledPv.setValueAndWait(ena);
      exposureTriggerEnabledRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public double getExposureTriggerPosition() throws InterruptedException, DriverException {
    try {
      return (Double)exposureTriggerPosRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setExposureTriggerPosition(double pos) throws InterruptedException, DriverException {
    try {
      exposureTriggerPosPv.setValueAndWait(pos);
      exposureTriggerPosRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public boolean isArmed() throws InterruptedException, DriverException {
    try {
      return (Boolean)armedRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setArmed(boolean armed) throws InterruptedException, DriverException {
    try {
      armedPv.setValueAndWait(armed);
      armedRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public long getMaxExposureTime() throws InterruptedException, DriverException {
    try {
      return ((Integer)maxExposureTimeRbvPv.getValue()).longValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setMaxExposureTime(long maxExposureTime) throws InterruptedException,
      DriverException {
    if (maxExposureTime > Integer.MAX_VALUE) throw new IllegalArgumentException(
        "maxExposureTime > Integer.MAX_VALUE");
    try {
      maxExposureTimePv.setValueAndWait((int)maxExposureTime);
      maxExposureTimeRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void openShutter() throws InterruptedException, DriverException {
    setShutter(ShutterState.OPEN);
  }

  public void closeShutter() throws InterruptedException, DriverException {
    setShutter(ShutterState.CLOSED);
  }

  protected boolean isShutterOpen() throws InterruptedException, DriverException {
    try {
      return ShutterState.OPEN.equals(shutterRbvPv.getValue());
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void setShutter(ShutterState desiredState) throws InterruptedException,
      DriverException {
    try {
      shutterPv.setValueAndWait(desiredState);
      shutterRbvPv.getValueNowAndWait();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public boolean waitForShutterToOpen(long timeout) throws InterruptedException, DriverException {
    return waitForShutterState(ShutterState.OPEN, timeout);
  }

  public boolean waitForShutterToClose(long timeout) throws InterruptedException,
      DriverException {
    return waitForShutterState(ShutterState.CLOSED, timeout);
  }

  protected boolean waitForShutterState(ShutterState state, long timeout)
      throws InterruptedException, DriverException {
    try {
      shutterRbvPv.waitForValue(state, timeout);
      return true;
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      return false;
    }
  }

  public void triggerExposure() throws InterruptedException, DriverException {
    try {
      exposureTriggerPv.setValueAndWait(EXPOSURE_TRIGGER_ACTIVATE);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }
}
