package imcacat.imcadr;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.model.EpicsPv;

/**
 * I provide an interface to the IMCA-CAT External Software Trigger Control
 * API.
 */
public abstract class ExternalSoftwareTriggerDetectorDriver extends AreaDetectorDriver {
  protected static final String DATASET_CONTROL_PV_NAME = "ExSwTrAcq";
  protected static final String OK_TO_EXPOSE_PV_NAME = "ExSwTrOkToExp";
  protected static final String EXPOSURE_CONTROL_PV_NAME = "ExSwTrCtl";
  protected static final String EXPOSURE_CONTROL_RESPONSE_PV_NAME = "ExSwTrCtlRsp";

  protected static final long WAIT_OK_TO_EXPOSE_TIMEOUT = 15000;
  protected static final long WAIT_EXPOSURE_CONTROL_RESPONSE_TIMEOUT = 15000;

  protected static final short OK_TO_EXPOSE_NO = 0;
  protected static final short OK_TO_EXPOSE_YES = 1;

  protected static final short DATASET_CONTROL_STOP = 0;
  protected static final short DATASET_CONTROL_START = 1;

  protected static final short EXPOSURE_CONTROL_STOP = 0;
  protected static final short EXPOSURE_CONTROL_START = 1;

  protected static final short EXPOSURE_CONTROL_RESPONSE_STOP = 0;
  protected static final short EXPOSURE_CONTROL_RESPONSE_START = 1;
  protected static final short EXPOSURE_CONTROL_RESPONSE_OK = 2;
  protected static final short EXPOSURE_CONTROL_RESPONSE_AGAIN = 3;

  private volatile boolean released = false;
  protected EpicsPv datasetControlPv;
  protected EpicsPv okToExposePv;
  protected EpicsPv exposureControlPv;
  protected EpicsPv exposureControlResponsePv;

  public ExternalSoftwareTriggerDetectorDriver(String pvName) {
    super(pvName);

    datasetControlPv = new EpicsPv(pvName + DATASET_CONTROL_PV_NAME, false, false);
    okToExposePv = new EpicsPv(pvName + OK_TO_EXPOSE_PV_NAME, false, true);
    exposureControlPv = new EpicsPv(pvName + EXPOSURE_CONTROL_PV_NAME, false, false);
    exposureControlResponsePv =
        new EpicsPv(pvName + EXPOSURE_CONTROL_RESPONSE_PV_NAME, false, true);
  }

  public void release() {
    if (released) return;
    released = true;

    try {
      if (datasetControlPv != null) datasetControlPv.release();
      if (okToExposePv != null) okToExposePv.release();
      if (exposureControlPv != null) exposureControlPv.release();
      if (exposureControlResponsePv != null) exposureControlResponsePv.release();
    } finally {
      super.release();
    }
  }

  public void startDataset() throws InterruptedException, DriverException {
    try {
      datasetControlPv.setValue(DATASET_CONTROL_START);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void stopDataset() throws InterruptedException, DriverException {
    try {
      datasetControlPv.setValue(DATASET_CONTROL_STOP);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void startExposing() throws InterruptedException, DriverException {
    try {
      okToExposePv.waitForValue(OK_TO_EXPOSE_YES, WAIT_OK_TO_EXPOSE_TIMEOUT);
      exposureControlPv.setValue(EXPOSURE_CONTROL_START);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }

    try {
      exposureControlResponsePv.waitForValue(EXPOSURE_CONTROL_RESPONSE_START,
          WAIT_EXPOSURE_CONTROL_RESPONSE_TIMEOUT);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException("Response never became Start");
    }
  }

  public boolean stopExposing() throws InterruptedException, DriverException {
    final List<Short> responses = Collections.synchronizedList(new ArrayList<Short>());
    final CountDownLatch okOrAgainLatch = new CountDownLatch(1);
    PropertyChangeListener responseListener = new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Short response = (Short)event.getNewValue();
        responses.add(response);
        if (response == EXPOSURE_CONTROL_RESPONSE_OK ||
            response == EXPOSURE_CONTROL_RESPONSE_AGAIN) okOrAgainLatch.countDown();
      }
    };

    try {
      exposureControlResponsePv.addPropertyChangeListener(EpicsPv.PROPNAME_VALUE,
          responseListener);
      exposureControlPv.setValue(EXPOSURE_CONTROL_STOP);
      boolean succeeded =
          okOrAgainLatch.await(WAIT_EXPOSURE_CONTROL_RESPONSE_TIMEOUT, TimeUnit.MILLISECONDS);
      if (!succeeded) throw new DriverException("Response never became OK nor Again");
      short response = (Short)exposureControlResponsePv.getValue();
      if (response == EXPOSURE_CONTROL_RESPONSE_OK) return false;
      if (response == EXPOSURE_CONTROL_RESPONSE_AGAIN) return true;
      String responsesString = "";
      boolean isFirst = true;
      for (Short each : responses) {
        responsesString += (isFirst ? "" : ", ") + each;
        if (isFirst) isFirst = false;
      }
      throw new DriverException("Response not OK nor Again; was " + response +
          "; responses were " + responsesString);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } finally {
      exposureControlResponsePv.removePropertyChangeListener(EpicsPv.PROPNAME_VALUE,
          responseListener);
    }
  }

  public boolean waitForExposureToEnd(long timeout) throws InterruptedException, DriverException {
    try {
      exposureControlResponsePv.waitForAnyOneValue(
          Arrays.asList(EXPOSURE_CONTROL_RESPONSE_OK, EXPOSURE_CONTROL_RESPONSE_AGAIN), timeout);
      return true;
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      return false;
    }
  }
}
