package imcacat.imcadr;

import imcacat.imcadr.util.FileUtils;
import imcacat.imcadr.util.NetworkUtils;
import imcacat.imcadr.util.ShellUtils;
import imcacat.procrun.ProcessRunner;
import imcacat.procrun.ProcessRunnerResult;

import java.io.ByteArrayInputStream;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

/**
 * I analyze a macromolecular crystallography X-ray diffraction image.
 */
public abstract class ImageAnalysis {
  protected final ImageAnalysisSpec spec;
  protected final File image;
  protected final File clusterKeyFile;
  protected final String clusterNode;

  protected ImageAnalysis(ImageAnalysisSpec spec, File image) {
    this(spec, image, null, null);
  }

  protected ImageAnalysis(ImageAnalysisSpec spec, File image, File clusterKeyFile,
      String clusterNode) {
    super();
    this.spec = spec;
    this.image = image;
    this.clusterKeyFile = clusterKeyFile;
    this.clusterNode = (clusterNode == null) ? NetworkUtils.LOCALHOST_NAME : clusterNode;
  }

  public ImageAnalysisSpec getSpec() {
    return spec;
  }

  public File getImage() {
    return image;
  }

  public ImageAnalysisResult run() {
    try {
      ProcessRunnerResult runResult = runProcess();
      if (!wasSuccessful(runResult)) return newAnalysisResult(runResult);
      return parseProcessOutput(runResult.stdout());
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      return newAnalysisResult("The analysis was interrupted", e);
    }
  }

  protected ProcessRunnerResult runProcess() throws InterruptedException {
    return ProcessRunner.run(createProcessBuilder());
  }

  protected ProcessBuilder createProcessBuilder() {
    ProcessBuilder result = new ProcessBuilder();
    result.command(isForLocalExecution() ? createCommand() : createClusterCommand());
    return result;
  }

  protected ImageAnalysisResult parseProcessOutput(ByteBuffer buffer) {
    ByteArrayInputStream stream = null;
    try {
      stream = new ByteArrayInputStream(buffer.array(), 0, buffer.limit());
      return newAnalysisParser(stream).run();
    } finally {
      closeSilently(stream);
    }
  }

  /**
   * Creates the command to run to analyze the image.
   * <p>
   * If cluster scoring is to be used, the apostrophe character (047) must not
   * appear in any part of the command.
   *
   * @return the command to run
   */
  protected abstract List<String> createCommand();

  protected abstract ImageAnalysisParser newAnalysisParser(InputStream stream);

  private boolean isForLocalExecution() {
    return clusterKeyFile == null || NetworkUtils.isLocalhost(clusterNode);
  }

  private List<String> createClusterCommand() {
    List<String> result = new ArrayList<>();
    result.add(Resources.getInstance().getSshHelper().getPath());
    result.add("-n");
    result.add("-i");
    result.add(expandedClusterKeyFilePath());
    result.add("-o");
    result.add("BatchMode=yes");
    result.add("-o");
    result.add("IdentitiesOnly=yes");
    result.add("-o");
    result.add("LogLevel=ERROR");
    result.add(clusterNode);
    List<String> localCommand = createCommand();
    result.add(encodeAsSingleQuotedShellString((new File(localCommand.get(0))).getName()));
    for (String each : localCommand.subList(1, localCommand.size())) {
      result.add(encodeAsSingleQuotedShellString(each));
    }
    return result;
  }

  private String expandedClusterKeyFilePath() {
    Map<String, String> varNamesToValues = new HashMap<>(1);
    varNamesToValues.put("user", Resources.getUsername());
    return ShellUtils.expand(clusterKeyFile.getPath(), varNamesToValues);
  }

  public static ImageAnalysis newInstance(ImageAnalysisSpec spec, File image) {
    return newInstance(spec, image, null, null);
  }

  public static ImageAnalysis newInstance(ImageAnalysisSpec spec, File image, File clusterKeyFile,
      String clusterNode) {
    switch (spec.getAnalysisPackage()) {
      case LABELIT:
        return new LabelitImageAnalysis(spec, image, clusterKeyFile, clusterNode);
      case DTREK:
        return new DTrekImageAnalysis(spec, image, clusterKeyFile, clusterNode);
    }
    throw new BugException();
  }

  protected static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }

  protected static void closeSilently(Closeable closeable) {
    if (closeable == null) return;
    try {
      closeable.close();
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (IOException | RuntimeException e) {
      /* IGNORED */
    }
  }

  protected static void deleteSilently(File fileOrDirectory) {
    if (fileOrDirectory == null) return;
    try {
      FileUtils.delete(fileOrDirectory.toPath());
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (IOException | RuntimeException e) {
      /* IGNORED */
    }
  }

  protected static File createWorkingDirectory(String prefix) throws IOException {
    return FileUtils.createTempDir(prefix).toFile();
  }

  protected static boolean wasSuccessful(ProcessRunnerResult result) {
    if (result.hasException()) return false;
    if (!result.hasExitStatus()) return false;
    return result.exitStatus().equals(Integer.valueOf(0));
  }

  protected static ImageAnalysisResult newAnalysisResult(ProcessRunnerResult runResult) {
    if (runResult.hasException()) {
      return newAnalysisResult("An error occurred while running the analysis subprocess",
          runResult.exception());
    }
    if (!runResult.hasExitStatus()) {
      return newAnalysisResult("The analysis subprocess did not exit with an exit status");
    }
    if (runResult.exitStatus().intValue() != 0) {
      return newAnalysisResult("The analysis subprocess exited with a nonzero exit status");
    }
    return newAnalysisResult();
  }

  protected static ImageAnalysisResult newAnalysisResult() {
    return newAnalysisResult((String)null);
  }

  protected static ImageAnalysisResult newAnalysisResult(String errorMessage) {
    return newAnalysisResult(errorMessage, null);
  }

  protected static ImageAnalysisResult newAnalysisResult(String errorMessage, Throwable cause) {
    ImageAnalysisResult result = new ImageAnalysisResult();
    result.setError(errorMessage, cause);
    return result;
  }

  /**
   * Encodes the specified string as a Unix (Bourne) shell single-quoted
   * string.
   *
   * @param s the string to encode
   *
   * @return a single-quoted shell string
   *
   * @throws IllegalArgumentException if {@code s} contains an apostrophe
   *           character (047) which is the single-quote opening and closing
   *           character which is impossible to encode in a single-quoted
   *           string
   */
  private static String encodeAsSingleQuotedShellString(String s) {
    if (s.contains("'")) {
      throw new IllegalArgumentException("String contains apostrophe character (\\047)");
    } else {
      return "'" + s + "'";
    }
  }
}
