package imcacat.imcadr;

import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;

/**
 * I am a specification of an image analysis.
 */
public class ImageAnalysisSpec {
  private static final AnalysisPackage DEFAULT_ANALYSIS_PACKAGE = AnalysisPackage.LABELIT;
  private static final Map<AnalysisProgram, List<String>> DEFAULT_ANALYSIS_PROGRAM_ARGUMENTS;
  static {
    Map<AnalysisProgram, List<String>> apArgs = new EnumMap<>(AnalysisProgram.class);
    ImcadrConfiguration c = ImcadrConfiguration.Builder.newWithDefaults().build();
    for (AnalysisProgram each : AnalysisProgram.values()) {
      apArgs.put(each, Collections.unmodifiableList(c.analysisProgramArguments(each)));
    }
    DEFAULT_ANALYSIS_PROGRAM_ARGUMENTS = Collections.unmodifiableMap(apArgs);
  }

  private final AnalysisPackage analysisPackage;
  private final Map<AnalysisProgram, List<String>> analysisProgramArguments;
  private final Double highResolutionCutoff;
  private final Double lowResolutionCutoff;
  private final Integer minimumSpotArea;
  private final Double minimumSignalHeight;

  private ImageAnalysisSpec(Builder b) {
    super();
    analysisPackage = (b.analysisPackage == null) ? DEFAULT_ANALYSIS_PACKAGE : b.analysisPackage;
    analysisProgramArguments = new EnumMap<>(b.analysisProgramArguments);
    highResolutionCutoff = b.highResolutionCutoff;
    lowResolutionCutoff = b.lowResolutionCutoff;
    minimumSpotArea = b.minimumSpotArea;
    minimumSignalHeight = b.minimumSignalHeight;
  }

  public AnalysisPackage getAnalysisPackage() {
    return analysisPackage;
  }

  public List<String> getAnalysisProgramArguments(AnalysisProgram p) {
    List<String> result = analysisProgramArguments.get(p);
    return (result != null) ? result : DEFAULT_ANALYSIS_PROGRAM_ARGUMENTS.get(p);
  }

  public Double getHighResolutionCutoff() {
    return highResolutionCutoff;
  }

  public Double getLowResolutionCutoff() {
    return lowResolutionCutoff;
  }

  public Integer getMinimumSpotArea() {
    return minimumSpotArea;
  }

  public Double getMinimumSignalHeight() {
    return minimumSignalHeight;
  }

  /**
   * I build an {@link ImageAnalysisSpec}.
   */
  public static class Builder {
    private AnalysisPackage analysisPackage;
    private Map<AnalysisProgram, List<String>> analysisProgramArguments = new EnumMap<>(
        AnalysisProgram.class);
    private Double highResolutionCutoff;
    private Double lowResolutionCutoff;
    private Integer minimumSpotArea;
    private Double minimumSignalHeight;

    public Builder() {
      super();
    }

    public AnalysisPackage analysisPackage() {
      return analysisPackage;
    }

    public Builder analysisPackage(AnalysisPackage analysisPackage) {
      this.analysisPackage = analysisPackage;
      return this;
    }

    /**
     * @param p the program whose arguments should be returned
     *
     * @return arguments for the specified program or {@code null} if not set
     */
    public List<String> analysisProgramArguments(AnalysisProgram p) {
      return analysisProgramArguments.get(p);
    }

    /**
     * @param p the program for which to set the arguments
     * @param args the list of program arguments; the client must not modify
     *          the list
     */
    public Builder analysisProgramArguments(AnalysisProgram p, List<String> args) {
      analysisProgramArguments.put(p, args);
      return this;
    }

    public Double highResolutionCutoff() {
      return highResolutionCutoff;
    }

    public Builder highResolutionCutoff(Double highResolutionCutoff) {
      this.highResolutionCutoff = highResolutionCutoff;
      return this;
    }

    public Double lowResolutionCutoff() {
      return lowResolutionCutoff;
    }

    public Builder lowResolutionCutoff(Double lowResolutionCutoff) {
      this.lowResolutionCutoff = lowResolutionCutoff;
      return this;
    }

    public Integer minimumSpotArea() {
      return minimumSpotArea;
    }

    public Builder minimumSpotArea(Integer minimumSpotArea) {
      this.minimumSpotArea = minimumSpotArea;
      return this;
    }

    public Double minimumSignalHeight() {
      return minimumSignalHeight;
    }

    public Builder minimumSignalHeight(Double minimumSignalHeight) {
      this.minimumSignalHeight = minimumSignalHeight;
      return this;
    }

    public ImageAnalysisSpec build() {
      return new ImageAnalysisSpec(this);
    }
  }
}
