package imcacat.imcadr;

import java.io.File;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import imcacat.imcadr.util.NetworkUtils;
import imcacat.imcadr.util.ShellUtils;

/**
 * I am a configuration for the Imcadr program.
 * <p>
 * My fields that are not set have the {@code null} value. This means it is
 * impossible to differentiate a field that has been set to {@code null} from
 * one that has not been set; {@code null} means "not set."
 * <p>
 * My getter methods return the first value that is set, starting with me: if
 * my field is not set, the configuration chain of defaults is traversed until
 * a set value is found or the root configuration is reached at which point the
 * root's value is returned.
 */
public class ImcadrConfiguration {
  private final ImcadrConfiguration defaults;
  private final File clusterKeyFile;
  private final List<ClusterNode> clusterNodes;
  private final Map<AnalysisProgram, List<String>> analysisProgramArguments;

  /**
   * @see Builder#build()
   */
  private ImcadrConfiguration(Builder builder) {
    super();
    defaults = builder.defaults;
    if (defaults == null) {
      clusterKeyFile = builder.clusterKeyFile;
      clusterNodes = Objects.requireNonNull(builder.clusterNodes);
      analysisProgramArguments = Objects.requireNonNull(builder.analysisProgramArguments);
    } else {
      clusterKeyFile = builder.clusterKeyFile;
      clusterNodes = builder.clusterNodes;
      analysisProgramArguments = Objects.requireNonNull(builder.analysisProgramArguments);
    }

    /*
     * Define temporary variables for values to be used below in argument
     * validation that represent the current value (either the specified value
     * or the value obtained from the defaults).
     */
    List<ClusterNode> xClusterNodes;
    if (clusterNodes != null) {
      xClusterNodes = clusterNodes;
    } else {
      xClusterNodes = defaults.clusterNodes();
    }

    /* Validate arguments. */
    if (xClusterNodes.isEmpty()) {
      throw new IllegalArgumentException(
          "\"" + ImcadrConfigurationParser.PROP_CLUSTER_NODES + "\": empty");
    }
    for (AnalysisProgram each : AnalysisProgram.values()) {
      List<String> args = analysisProgramArguments.get(each);
      if (args == null && defaults != null) {
        args = defaults.analysisProgramArguments(each);
      }
      if (args == null) {
        throw new IllegalArgumentException(
            "\"" + ImcadrConfigurationParser.PROP_ANALYSIS_PROG_ARGS.get(each) + "\": not set");
      }
      if (args.isEmpty()) {
        throw new IllegalArgumentException(
            "\"" + ImcadrConfigurationParser.PROP_ANALYSIS_PROG_ARGS.get(each) + "\": empty");
      }
      if (!args.contains("$more") && !args.contains("${more}")) {
        throw new IllegalArgumentException("\"" +
            ImcadrConfigurationParser.PROP_ANALYSIS_PROG_ARGS.get(each) + "\": no \"$more\" word");
      }
      varNamesLoop: for (String varName : Arrays.asList("image")) {
        for (String arg : args) {
          if (ShellUtils.containsVariable(arg, varName)) continue varNamesLoop;
        }
        throw new IllegalArgumentException(
            "\"" + ImcadrConfigurationParser.PROP_ANALYSIS_PROG_ARGS.get(each) + "\": no \"$" +
                varName + "\" variable");
      }
    }
  }

  /**
   * @return the cluster key file or {@code null} if not set
   */
  public File clusterKeyFile() {
    return (clusterKeyFile != null) ? clusterKeyFile
        : ((defaults != null) ? defaults.clusterKeyFile() : null);
  }

  public List<ClusterNode> clusterNodes() {
    return (clusterNodes != null) ? clusterNodes : defaults.clusterNodes();
  }

  public List<String> analysisProgramArguments(AnalysisProgram p) {
    List<String> result = analysisProgramArguments.get(p);
    return (result != null) ? result : defaults.analysisProgramArguments(p);
  }

  public ImcadrConfiguration defaults(ImcadrConfiguration c) {
    Builder b = Builder.newEmpty();
    b.defaults(c);
    b.clusterKeyFile(clusterKeyFile);
    b.clusterNodes(clusterNodes);
    for (Map.Entry<AnalysisProgram, List<String>> each : analysisProgramArguments.entrySet()) {
      b.analysisProgramArguments(each.getKey(), each.getValue());
    }
    return b.build();
  }

  @Override
  public boolean equals(Object o) {
    if (o == this) return true;
    if (!(o instanceof ImcadrConfiguration)) return false;
    ImcadrConfiguration c = (ImcadrConfiguration)o;
    if (!Objects.equals(clusterKeyFile, c.clusterKeyFile)) return false;
    if (!Objects.equals(clusterNodes, c.clusterNodes)) return false;
    if (!analysisProgramArguments.equals(c.analysisProgramArguments)) return false;
    return Objects.equals(defaults, c.defaults);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + ((defaults == null) ? 0 : defaults.hashCode());
    result = 31 * result + ((clusterKeyFile == null) ? 0 : clusterKeyFile.hashCode());
    result = 31 * result + ((clusterNodes == null) ? 0 : clusterNodes.hashCode());
    result = 31 * result + analysisProgramArguments.hashCode();
    return result;
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    if (clusterKeyFile != null) {
      result.append("clusterKeyFile=");
      result.append(clusterKeyFile);
    }
    if (clusterNodes != null) {
      if (result.length() > 0) result.append(", ");
      result.append("clusterNodes=");
      result.append(clusterNodes);
    }
    if (!analysisProgramArguments.isEmpty()) {
      if (result.length() > 0) result.append(", ");
      result.append("analysisProgramArguments=");
      result.append(analysisProgramArguments);
    }
    if (defaults != null) {
      if (result.length() > 0) result.append(", ");
      result.append("defaults={");
      result.append(defaults);
      result.append("}");
    }
    return result.toString();
  }

  /**
   * I build an {@link ImcadrConfiguration}.
   */
  public static class Builder {
    private ImcadrConfiguration defaults;
    private File clusterKeyFile;
    private List<ClusterNode> clusterNodes;
    private Map<AnalysisProgram, List<String>> analysisProgramArguments = new EnumMap<>(
        AnalysisProgram.class);

    public Builder() {
      super();
    }

    public ImcadrConfiguration defaults() {
      return defaults;
    }

    public Builder defaults(ImcadrConfiguration defaults) {
      this.defaults = defaults;
      return this;
    }

    public File clusterKeyFile() {
      return clusterKeyFile;
    }

    public Builder clusterKeyFile(File clusterKeyFile) {
      this.clusterKeyFile = clusterKeyFile;
      return this;
    }

    public List<ClusterNode> clusterNodes() {
      return clusterNodes;
    }

    /**
     * @param clusterNodes the list of cluster nodes; the client must not
     *          modify the list
     */
    public Builder clusterNodes(List<ClusterNode> clusterNodes) {
      this.clusterNodes = clusterNodes;
      return this;
    }

    /**
     * @param p the program whose arguments should be returned
     */
    public List<String> analysisProgramArguments(AnalysisProgram p) {
      return analysisProgramArguments.get(p);
    }

    /**
     * @param p the program for which to set the arguments
     * @param args the list of program arguments; the client must not modify
     *          the list
     */
    public Builder analysisProgramArguments(AnalysisProgram p, List<String> args) {
      analysisProgramArguments.put(p, args);
      return this;
    }

    /**
     * @throws IllegalArgumentException if {@code maxNumScoreJobs} is not
     *           {@code null} and less than 1
     * @throws NullPointerException if {@code clusterNodes} or
     *           {@code maxNumScoreJobs} are {@code null} when {@code defaults}
     *           is {@code null}
     */
    public ImcadrConfiguration build() {
      return new ImcadrConfiguration(this);
    }

    /**
     * Creates a {@code Builder} that has no fields set.
     * <p>
     * This builder is appropriate when creating a root configuration or when
     * creating a configuration that will have its defaults configuration set
     * to another configuration.
     *
     * @see #newWithDefaults()
     */
    public static Builder newEmpty() {
      return new Builder();
    }

    /**
     * Creates a {@code Builder} that has no fields set except for the
     * {@code defaults} field which is set to the default root configuration.
     *
     * @see #newEmpty()
     */
    public static Builder newWithDefaults() {
      ImcadrConfiguration root = newEmpty()
          .clusterKeyFile(null)
          .clusterNodes(
              Arrays.asList(new ClusterNode(NetworkUtils.LOCALHOST_NAME, processorCount())))
          .analysisProgramArguments(AnalysisProgram.DTREK_DTRANKER,
              Arrays.asList("image=$image", "$more"))
          .analysisProgramArguments(AnalysisProgram.LABELIT_DISTL_SIGNAL_STRENGTH,
              Arrays.asList("$image", "$more"))
          .build();
      return newEmpty().defaults(root);
    }

    private static int processorCount() {
      return Runtime.getRuntime().availableProcessors();
    }
  }
}
