package imcacat.imcadr;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.regex.Pattern;

import org.imca_cat.shellwords.Shellwords;

import imcacat.imcadr.util.StringUtils;

/**
 * I am a configuration parser for the Imcadr program.
 */
public class ImcadrConfigurationParser {
  static final String PROP_CLUSTER_KEY_FILE = "ClusterKeyFile";
  static final String PROP_CLUSTER_NODES = "ClusterNodes";
  static final Map<AnalysisProgram, String> PROP_ANALYSIS_PROG_ARGS;
  static {
    Map<AnalysisProgram, String> apProps = new EnumMap<>(AnalysisProgram.class);
    apProps.put(AnalysisProgram.DTREK_DTRANKER, "DTrek.DtrankerArgs");
    apProps.put(AnalysisProgram.LABELIT_DISTL_SIGNAL_STRENGTH, "Labelit.DistlSignalStrengthArgs");
    if (apProps.size() != AnalysisProgram.values().length) {
      throw new BugException();
    }
    PROP_ANALYSIS_PROG_ARGS = Collections.unmodifiableMap(apProps);
  }
  static final String PROP_MAX_NUM_SCORE_JOBS = "MaxNumScoreJobs";

  static final List<String> ALL_PROPS;
  static {
    List<String> allProps = new ArrayList<>();
    allProps.add(PROP_CLUSTER_KEY_FILE);
    allProps.add(PROP_CLUSTER_NODES);
    allProps.addAll(PROP_ANALYSIS_PROG_ARGS.values());
    allProps.add(PROP_MAX_NUM_SCORE_JOBS);
    ALL_PROPS = Collections.unmodifiableList(allProps);
  }

  private static final Pattern COMMA = Pattern.compile("\\s*,\\s*");
  private static final Pattern COLON = Pattern.compile("\\s*:\\s*");

  private final Reader in;

  public ImcadrConfigurationParser(Reader in) {
    super();
    this.in = in;
  }

  public ImcadrConfiguration parse() throws IOException, ConfigurationException {
    return parse(in);
  }

  public static ImcadrConfiguration parse(File confFile)
      throws IOException, ConfigurationException {
    try (BufferedReader confReader = newBufferedReader(confFile)) {
      return (new ImcadrConfigurationParser(confReader)).parse();
    }
  }

  public static ImcadrConfiguration parse(String confFileData)
      throws IOException, ConfigurationException {
    try (BufferedReader confReader = newBufferedReader(confFileData)) {
      return (new ImcadrConfigurationParser(confReader)).parse();
    }
  }

  private static ImcadrConfiguration parse(Reader r) throws IOException, ConfigurationException {
    Properties props = new Properties();
    props.load(r);
    return parse(props);
  }

  private static ImcadrConfiguration parse(Properties p) throws ConfigurationException {
    ImcadrConfiguration.Builder b = ImcadrConfiguration.Builder.newWithDefaults();
    b.clusterKeyFile(parseClusterKeyFile(p));
    /*
     * Backward compatibility: use the deprecated MaxNumScoreJobs as the
     * default cluster node job capacity if it's set.
     */
    Integer maxNumScoreJobs = parseMaxNumScoreJobs(p);
    int defaultJobCapacity = maxNumScoreJobs == null ? 1 : maxNumScoreJobs.intValue();
    List<ClusterNode> nodes = parseClusterNodes(p, defaultJobCapacity);
    /*
     * Backward compatibility: an empty cluster nodes list used to be allowed,
     * but now it's not allowed, so treat an empty list as if it were not set.
     */
    if (nodes != null && nodes.isEmpty()) nodes = null;
    b.clusterNodes(nodes);
    Map<AnalysisProgram, List<String>> analysisProgArgs = parseAnalysisProgramArguments(p);
    for (Entry<AnalysisProgram, List<String>> each : analysisProgArgs.entrySet()) {
      b.analysisProgramArguments(each.getKey(), each.getValue());
    }
    try {
      return b.build();
    } catch (IllegalArgumentException e) {
      throw new ConfigurationException(e.getMessage());
    }
  }

  private static File parseClusterKeyFile(Properties p) throws ConfigurationException {
    String k = PROP_CLUSTER_KEY_FILE;
    String v = parseString(p.getProperty(k), k);
    if (v == null) return null;
    File f = new File(v);
    if (!f.isAbsolute()) {
      throw new ConfigurationException("\"" + k + "\": \"" + v + "\": not an absolute path");
    }
    return f;
  }

  private static List<ClusterNode> parseClusterNodes(Properties p, int defaultJobCapacity)
      throws ConfigurationException {
    String k = PROP_CLUSTER_NODES;
    List<String> nodeStrings = parseCommaSeparatedList(p.getProperty(k), k);
    if (nodeStrings == null) return null;
    if (nodeStrings.size() == 1 && nodeStrings.get(0).isEmpty()) return Collections.emptyList();
    List<ClusterNode> result = new ArrayList<>(nodeStrings.size());
    for (String each : nodeStrings) {
      List<String> parts = parseColonSeparatedList(each, k);
      String hostname;
      int jobCapacity;
      if (parts.size() == 1) {
        /* Backward compatibility. */
        jobCapacity = defaultJobCapacity;
      } else if (parts.size() == 2) {
        jobCapacity = parseInteger(parts.get(1), k);
        if (jobCapacity < 1) {
          throw new ConfigurationException("\"" + k + "\": \"" + jobCapacity + "\": not >= 1");
        }
      } else {
        throw new ConfigurationException(
            "\"" + k + "\": \"" + each + "\": not a colon-separated pair");
      }
      hostname = parts.get(0);
      if (hostname.isEmpty()) {
        throw new ConfigurationException("\"" + k + "\": \"" + each + "\": empty hostname");
      }
      ClusterNode node = null;
      try {
        node = new ClusterNode(hostname, jobCapacity);
      } catch (IllegalArgumentException e) {
        String msg = e.getMessage();
        msg = (msg == null) ? "invalid" : StringUtils.uncapitalize(msg);
        throw new ConfigurationException("\"" + k + "\": \"" + each + "\": " + msg);
      }
      result.add(node);
    }
    return result;
  }

  private static Map<AnalysisProgram, List<String>> parseAnalysisProgramArguments(Properties p)
      throws ConfigurationException {
    Map<AnalysisProgram, List<String>> result = new EnumMap<>(AnalysisProgram.class);
    for (Entry<AnalysisProgram, String> each : PROP_ANALYSIS_PROG_ARGS.entrySet()) {
      String k = each.getValue();
      String v = p.getProperty(k);
      if (v == null) continue;
      List<String> args = Shellwords.split(v);
      if (args.isEmpty()) {
        throw new ConfigurationException("\"" + k + "\": empty");
      }
      result.put(each.getKey(), args);
    }
    return result;
  }

  private static Integer parseMaxNumScoreJobs(Properties p) throws ConfigurationException {
    String k = PROP_MAX_NUM_SCORE_JOBS;
    Integer v = parseInteger(p.getProperty(k), k);
    if (v == null) return null;
    int i = v.intValue();
    if (i < 1) {
      throw new ConfigurationException("\"" + k + "\": \"" + i + "\": not >= 1");
    }
    return v;
  }

  private static Integer parseInteger(String s, String key) throws ConfigurationException {
    if (s == null) return null;
    try {
      return Integer.parseInt(s);
    } catch (NumberFormatException e) {
      throw new ConfigurationException("\"" + key + "\": \"" + s + "\": not an integer");
    }
  }

  private static String parseString(String s, String key) throws ConfigurationException {
    return s;
  }

  private static List<String> parseCommaSeparatedList(String s, String key)
      throws ConfigurationException {
    if (s == null) return null;
    return Arrays.asList(COMMA.split(s));
  }

  private static List<String> parseColonSeparatedList(String s, String key)
      throws ConfigurationException {
    if (s == null) return null;
    return Arrays.asList(COLON.split(s));
  }

  private static BufferedReader newBufferedReader(File f) throws IOException {
    return (f == null) ? null
        : new BufferedReader(new InputStreamReader(new FileInputStream(f), StandardCharsets.UTF_8));
  }

  private static BufferedReader newBufferedReader(String s) throws IOException {
    return (s == null) ? null : new BufferedReader(new StringReader(s));
  }
}
