package imcacat.imcadr;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.Charset;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.regex.Pattern;

import imcacat.imcadr.util.Pair;

/**
 * I am the server interface for this application. I use my standard input and
 * standard output streams for communication with a client.
 * <p>
 * Units: exposure time [s], cell size [micrometers], raster size [cells],
 * position [mm], resolution [angstroms].
 */
public class ImcadrServer implements Runnable {
  public static final long SHUTDOWN_TIMEOUT = 15000L;

  protected static final Charset UTF_8 = Charset.forName("UTF-8");
  protected static final String NIL = "nil";
  protected static final String TRUE = "true";
  protected static final String FALSE = "false";
  protected static final String OK = "+OK";
  protected static final String OK_WITH = OK + " ";
  protected static final String ERR = "-ERR";
  protected static final String ERR_WITH = ERR + " ";
  protected static final Pattern SPACE_RE = Pattern.compile(" ");
  protected static final long STOP_TIMEOUT = 15000L;

  protected static final HashMap<String, AnalysisPackage> DATA_TO_ANALYSIS_PACKAGE_MAP;
  protected static final EnumMap<SpaceGroup, String> SPACE_GROUP_TO_DATA_MAP;

  static {
    DATA_TO_ANALYSIS_PACKAGE_MAP = new HashMap<String, AnalysisPackage>();
    DATA_TO_ANALYSIS_PACKAGE_MAP.put("labelit", AnalysisPackage.LABELIT);
    DATA_TO_ANALYSIS_PACKAGE_MAP.put("dtrek", AnalysisPackage.DTREK);
    if (DATA_TO_ANALYSIS_PACKAGE_MAP.size() != AnalysisPackage.values().length) {
      throw new BugException();
    }
    /*
     * Backward compatibility: accept deprecated name.
     */
    DATA_TO_ANALYSIS_PACKAGE_MAP.put("dstartrek", AnalysisPackage.DTREK);

    SPACE_GROUP_TO_DATA_MAP = new EnumMap<SpaceGroup, String>(SpaceGroup.class);
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.P1, "P1");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.P2, "P2");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.P222, "P222");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.P4, "P4");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.P6, "P6");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.C222, "C222");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.C2, "C2");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.I4, "I4");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.I222, "I222");
    SPACE_GROUP_TO_DATA_MAP.put(SpaceGroup.F222, "F222");
    if (SPACE_GROUP_TO_DATA_MAP.size() != SpaceGroup.values().length) {
      throw new BugException();
    }
  }

  protected Imcadr model;
  protected ImcadrConfiguration conf;
  protected LineReader in;
  protected BufferedWriter out;
  protected ImcadrServerSession session;
  protected boolean closeInputWhenDone;
  protected Exception exception;

  public ImcadrServer() {
    this(new Imcadr(), ImcadrConfiguration.Builder.newWithDefaults().build());
  }

  public ImcadrServer(Imcadr model, ImcadrConfiguration conf) {
    this(model, conf, System.in, System.out);
  }

  public ImcadrServer(Imcadr model, InputStream in, OutputStream out) {
    this(model, ImcadrConfiguration.Builder.newWithDefaults().build(), in, out);
  }

  public ImcadrServer(Imcadr model, ImcadrConfiguration conf, InputStream in, OutputStream out) {
    this.model = model;
    this.conf = conf;
    this.in = new LineReader(new InputStreamReader(in, UTF_8));
    this.in.closePrerequisite(newCloseCallable(in));
    this.out = new BufferedWriter(new OutputStreamWriter(out, UTF_8));
    session = null;
    /*
     * Calling System.in.close() will block on some, if not all, platforms, so
     * don't call it. (It is called when closing the input LineReader.) Since
     * this server is using System.in for input, assume we are about to exit
     * the JVM once this server finishes, so leaving the input LineReader open
     * is an acceptable resource leak.
     */
    closeInputWhenDone = (in == System.in) ? false : true;
    exception = null;
  }

  protected ImcadrServerSession createSession() {
    return new ImcadrServerSession(model.getBeamlineDriver(), conf);
  }

  @Override
  public void run() {
    boolean interrupted = false;
    try {
      session = createSession();
      processCommands();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      exception = e;
      interrupted = true;
    } catch (IOException | RuntimeException e) {
      exception = e;
    } finally {
      try {
        if (Thread.interrupted()) interrupted = true;
        shutDown();
      } catch (ImcadrServerException | RuntimeException e) {
        if (exception == null) exception = e;
      }
      if (closeInputWhenDone) {
        try {
          if (Thread.interrupted()) interrupted = true;
          in.close();
        } catch (InterruptedIOException | ClosedByInterruptException e) {
          if (exception == null) exception = e;
          interrupted = true;
        } catch (IOException e) {
          if (exception == null) exception = e;
        }
      }
      try {
        if (Thread.interrupted()) interrupted = true;
        out.close();
      } catch (InterruptedIOException | ClosedByInterruptException e) {
        if (exception == null) exception = e;
        interrupted = true;
      } catch (IOException e) {
        if (exception == null) exception = e;
      }
      if (interrupted) Thread.currentThread().interrupt();
    }
  }

  public boolean closeInputWhenDone() {
    return closeInputWhenDone;
  }

  public void closeInputWhenDone(boolean close) {
    closeInputWhenDone = close;
  }

  public Exception exception() {
    return exception;
  }

  protected void processCommands() throws IOException {
    for (;;) {
      String line = readLine();
      if (line == null) break;
      line = line.trim();
      if (line.isEmpty()) continue;
      Command command = parseCommand(line);
      if (command.type() == null) {
        writeLine(ERR_WITH + "Do not understand");
        continue;
      }
      switch (command.type()) {
        case STAT:
          handleStatusRequest();
          break;
        case IMAGES_DIR_GET:
          handleGetImagesDirectoryRequest();
          break;
        case IMAGES_DIR_SET:
          handleSetImagesDirectoryRequest(command.argument());
          break;
        case EXP_TIME_GET:
          handleGetExposureTimeRequest();
          break;
        case EXP_TIME_SET:
          handleSetExposureTimeRequest(command.argument());
          break;
        case CELL_SIZE_GET:
          handleGetCellSizeRequest();
          break;
        case CELL_SIZE_SET:
          handleSetCellSizeRequest(command.argument());
          break;
        case ZY_SIZE_GET:
          handleGetZySizeRequest();
          break;
        case ZY_SIZE_SET:
          handleSetZySizeRequest(command.argument());
          break;
        case ZX_SIZE_GET:
          handleGetZxSizeRequest();
          break;
        case ZX_SIZE_SET:
          handleSetZxSizeRequest(command.argument());
          break;
        case FLY_SCAN_GET:
          handleGetFlyScanEnabledRequest();
          break;
        case FLY_SCAN_SET:
          handleSetFlyScanEnabledRequest(command.argument());
          break;
        case CHASING_GET:
          handleGetChasingEnabledRequest();
          break;
        case CHASING_SET:
          handleSetChasingEnabledRequest(command.argument());
          break;
        case ANALYSIS_PACKAGE_GET:
          handleGetAnalysisPackageRequest();
          break;
        case ANALYSIS_PACKAGE_SET:
          handleSetAnalysisPackageRequest(command.argument());
          break;
        case WEAK_DIFF_GET:
          handleGetWeakDiffracterRequest();
          break;
        case WEAK_DIFF_SET:
          handleSetWeakDiffracterRequest(command.argument());
          break;
        case HIGH_RES_CUTOFF_GET:
          handleGetHighResolutionCutoffRequest();
          break;
        case HIGH_RES_CUTOFF_SET:
          handleSetHighResolutionCutoffRequest(command.argument());
          break;
        case LOW_RES_CUTOFF_GET:
          handleGetLowResolutionCutoffRequest();
          break;
        case LOW_RES_CUTOFF_SET:
          handleSetLowResolutionCutoffRequest(command.argument());
          break;
        case MIN_SPOT_AREA_GET:
          handleGetMinimumSpotAreaRequest();
          break;
        case MIN_SPOT_AREA_SET:
          handleSetMinimumSpotAreaRequest(command.argument());
          break;
        case MIN_SIGNAL_HEIGHT_GET:
          handleGetMinimumSignalHeightRequest();
          break;
        case MIN_SIGNAL_HEIGHT_SET:
          handleSetMinimumSignalHeightRequest(command.argument());
          break;
        case START:
          handleStartRequest();
          break;
        case STOP:
          handleStopRequest();
          break;
        case RESET:
          handleResetRequest();
          break;
        case QUIT:
          boolean okToQuit = handleQuitRequest();
          if (okToQuit) return;
          break;
        case ERROR_DETAIL:
          handleErrorDetailRequest();
          break;
      }
    }
  }

  protected void handleStatusRequest() throws IOException {
    ImcadrServerResult result = session.getResultSnapshot();
    writeLine(OK);
    writeLine("STATE " + toData(result.getState(), result.getMessage()));
    writeLine(format("PROGRESS %d", result.getProgress()));
    if (result.getOriginZy() != null) {
      writeLine("ORIGIN_ZY " + toData(result.getOriginZy()));
    }
    if (result.getOriginZx() != null) {
      writeLine("ORIGIN_ZX " + toData(result.getOriginZx()));
    }
    if (result.getScoresZy() != null) {
      writeLine("SCORES_ZY " + toData(result.getScoresZy()));
    }
    if (result.getScoresZx() != null) {
      writeLine("SCORES_ZX " + toData(result.getScoresZx()));
    }
    if (result.getSpotTotalsZy() != null) {
      writeLine("SPOT_TOT_ZY " + toData(result.getSpotTotalsZy()));
    }
    if (result.getSpotTotalsZx() != null) {
      writeLine("SPOT_TOT_ZX " + toData(result.getSpotTotalsZx()));
    }
    if (result.getInResolutionTotalsZy() != null) {
      writeLine("IN_RES_TOT_ZY " + toData(result.getInResolutionTotalsZy()));
    }
    if (result.getInResolutionTotalsZx() != null) {
      writeLine("IN_RES_TOT_ZX " + toData(result.getInResolutionTotalsZx()));
    }
    if (result.getGoodBraggTotalsZy() != null) {
      writeLine("BRAGG_TOT_ZY " + toData(result.getGoodBraggTotalsZy()));
    }
    if (result.getGoodBraggTotalsZx() != null) {
      writeLine("BRAGG_TOT_ZX " + toData(result.getGoodBraggTotalsZx()));
    }
    if (result.getIceRingTotalsZy() != null) {
      writeLine("ICE_RING_TOT_ZY " + toData(result.getIceRingTotalsZy()));
    }
    if (result.getIceRingTotalsZx() != null) {
      writeLine("ICE_RING_TOT_ZX " + toData(result.getIceRingTotalsZx()));
    }
    if (result.getEstimatedResolutions1Zy() != null) {
      writeLine("RES1_ZY " + toData(result.getEstimatedResolutions1Zy(), 2));
    }
    if (result.getEstimatedResolutions1Zx() != null) {
      writeLine("RES1_ZX " + toData(result.getEstimatedResolutions1Zx(), 2));
    }
    if (result.getEstimatedResolutions2Zy() != null) {
      writeLine("RES2_ZY " + toData(result.getEstimatedResolutions2Zy(), 2));
    }
    if (result.getEstimatedResolutions2Zx() != null) {
      writeLine("RES2_ZX " + toData(result.getEstimatedResolutions2Zx(), 2));
    }
    if (result.getTotalIntegratedSignalsZy() != null) {
      writeLine("INT_SIG_TOT_ZY " + toData(result.getTotalIntegratedSignalsZy()));
    }
    if (result.getTotalIntegratedSignalsZx() != null) {
      writeLine("INT_SIG_TOT_ZX " + toData(result.getTotalIntegratedSignalsZx()));
    }
    if (result.getMosaicitiesZy() != null) {
      writeLine("MOSAICITY_ZY " + toData(result.getMosaicitiesZy(), 2));
    }
    if (result.getMosaicitiesZx() != null) {
      writeLine("MOSAICITY_ZX " + toData(result.getMosaicitiesZx(), 2));
    }
    if (result.getSpaceGroupsZy() != null) {
      writeLine("SPACE_GRP_ZY " + toData(result.getSpaceGroupsZy()));
    }
    if (result.getSpaceGroupsZx() != null) {
      writeLine("SPACE_GRP_ZX " + toData(result.getSpaceGroupsZx()));
    }
    if (result.getUnitCellsZy() != null) {
      writeLine("UNIT_CELL_ZY " + toData(result.getUnitCellsZy(), 2, 2));
    }
    if (result.getUnitCellsZx() != null) {
      writeLine("UNIT_CELL_ZX " + toData(result.getUnitCellsZx(), 2, 2));
    }
    if (result.getCenterZy() != null) {
      writeLine("CENTER_ZY " + toData(result.getCenterZy()));
    }
    if (result.getCenterZx() != null) {
      writeLine("CENTER_ZX " + toData(result.getCenterZx()));
    }
    if (result.getCenter() != null) {
      writeLine("CENTER " + toData(result.getCenter()));
    }
    if (result.getTime() != null) {
      writeLine(format("TIME %d", result.getTime()));
    }
    writeLine(".");
  }

  protected void handleSetImagesDirectoryRequest(String data) throws IOException {
    File value = parseFile(data);
    String message = ImcadrServerSession.validateImagesDirectory(value);
    if (message == null) {
      session.setImagesDirectory(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetImagesDirectoryRequest() throws IOException {
    File value = session.getImagesDirectory();
    String valueString = (value == null) ? NIL : value.getAbsolutePath();
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetExposureTimeRequest(String data) throws IOException {
    Double value = null;
    try {
      value = parseDouble(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateExposureTime(value);
    if (message == null) {
      session.setExposureTime(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetExposureTimeRequest() throws IOException {
    Double value = session.getExposureTime();
    String valueString = (value == null) ? NIL : format("%.3f", value);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetCellSizeRequest(String data) throws IOException {
    Rectangle value = null;
    try {
      value = parseCellSize(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateCellSize(value);
    if (message == null) {
      session.setCellSize(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetCellSizeRequest() throws IOException {
    Rectangle value = session.getCellSize();
    String valueString =
        (value == null) ? NIL : format("%.0fx%.0f", value.getWidth(), value.getHeight());
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetZySizeRequest(String data) throws IOException {
    Rectangle value = null;
    try {
      value = parseRasterSize(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateZySize(value);
    if (message == null) {
      session.setZySize(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetZySizeRequest() throws IOException {
    Rectangle value = session.getZySize();
    String valueString =
        (value == null) ? NIL : format("%.0fx%.0f", value.getWidth(), value.getHeight());
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetZxSizeRequest(String data) throws IOException {
    Rectangle value = null;
    try {
      value = parseRasterSize(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateZxSize(value);
    if (message == null) {
      session.setZxSize(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetZxSizeRequest() throws IOException {
    Rectangle value = session.getZxSize();
    String valueString =
        (value == null) ? NIL : format("%.0fx%.0f", value.getWidth(), value.getHeight());
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetFlyScanEnabledRequest(String data) throws IOException {
    Boolean value = null;
    try {
      value = parseBoolean(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    session.setFlyScanEnabled(value);
    writeLine(OK);
  }

  protected void handleGetFlyScanEnabledRequest() throws IOException {
    Boolean value = session.getFlyScanEnabled();
    String valueString = (value == null) ? NIL : (value.booleanValue() ? TRUE : FALSE);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetChasingEnabledRequest(String data) throws IOException {
    Boolean value = null;
    try {
      value = parseBoolean(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    session.setChasingEnabled(value);
    writeLine(OK);
  }

  protected void handleGetChasingEnabledRequest() throws IOException {
    Boolean value = session.getChasingEnabled();
    String valueString = (value == null) ? NIL : (value.booleanValue() ? TRUE : FALSE);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetAnalysisPackageRequest(String data) throws IOException {
    Pair<AnalysisPackage, String> value = null;
    try {
      value = parseAnalysisPackage(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    session.setAnalysisPackage(value.getFirst(), value.getSecond());
    writeLine(OK);
  }

  protected void handleGetAnalysisPackageRequest() throws IOException {
    String valueId = session.getAnalysisPackageId();
    String valueString = (valueId == null) ? NIL : valueId;
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetWeakDiffracterRequest(String data) throws IOException {
    Boolean value = null;
    try {
      value = parseBoolean(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    session.setWeakDiffracter(value);
    writeLine(OK);
  }

  protected void handleGetWeakDiffracterRequest() throws IOException {
    Boolean value = session.getWeakDiffracter();
    String valueString = (value == null) ? NIL : (value.booleanValue() ? TRUE : FALSE);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetHighResolutionCutoffRequest(String data) throws IOException {
    Double value = null;
    try {
      value = parseDouble(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateHighResolutionCutoff(value);
    if (message == null) {
      session.setHighResolutionCutoff(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetHighResolutionCutoffRequest() throws IOException {
    Double value = session.getHighResolutionCutoff();
    String valueString = (value == null) ? NIL : format("%.2f", value);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetLowResolutionCutoffRequest(String data) throws IOException {
    Double value = null;
    try {
      value = parseDouble(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateLowResolutionCutoff(value);
    if (message == null) {
      session.setLowResolutionCutoff(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetLowResolutionCutoffRequest() throws IOException {
    Double value = session.getLowResolutionCutoff();
    String valueString = (value == null) ? NIL : format("%.2f", value);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetMinimumSpotAreaRequest(String data) throws IOException {
    Integer value = null;
    try {
      value = parseInteger(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateMinimumSpotArea(value);
    if (message == null) {
      session.setMinimumSpotArea(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetMinimumSpotAreaRequest() throws IOException {
    Integer value = session.getMinimumSpotArea();
    String valueString = (value == null) ? NIL : format("%d", value);
    writeLine(OK_WITH + valueString);
  }

  protected void handleSetMinimumSignalHeightRequest(String data) throws IOException {
    Double value = null;
    try {
      value = parseDouble(data);
    } catch (ImcadrServerException e) {
      writeLine(ERR_WITH + capitalize(e.getMessage()));
      return;
    }
    String message = ImcadrServerSession.validateMinimumSignalHeight(value);
    if (message == null) {
      session.setMinimumSignalHeight(value);
      writeLine(OK);
    } else {
      writeLine(ERR_WITH + capitalize(message));
    }
  }

  protected void handleGetMinimumSignalHeightRequest() throws IOException {
    Double value = session.getMinimumSignalHeight();
    String valueString = (value == null) ? NIL : format("%.2f", value);
    writeLine(OK_WITH + valueString);
  }

  protected void handleStartRequest() throws IOException {
    if (session.isRunningOrPaused()) {
      writeLine(ERR_WITH + "Already running");
      return;
    }
    String message = session.validate();
    if (message != null) {
      writeLine(ERR_WITH + capitalize(message));
      return;
    }
    ImcadrServerRasterizerThread thread = session.createAndSetRasterizerThread();
    ImcadrServerResult result = session.getResult();
    result.setState(ImcadrServerState.RUNNING);
    thread.start();
    writeLine(OK);
  }

  protected void handleStopRequest() throws IOException {
    if (!session.isRunningOrPaused()) {
      writeLine(OK_WITH + "Nothing to stop");
      return;
    }
    Thread thread = session.getRasterizerThread();
    thread.interrupt();
    try {
      thread.join(STOP_TIMEOUT);
    } catch (InterruptedException e) {
      try {
        writeLine(ERR_WITH + "Interrupted while stopping execution");
      } catch (IOException | RuntimeException ex) {
        /* IGNORED */
      }
      Thread.currentThread().interrupt();
      throw newInterruptedIOException(e);
    }
    if (thread.isAlive()) {
      writeLine(ERR_WITH + "Failed to stop execution");
      return;
    }
    ImcadrServerResult result = session.getResult();
    result.setState(ImcadrServerState.IDLE);
    result.setMessage("Stopped");
    writeLine(OK);
  }

  protected void handleResetRequest() throws IOException {
    if (session.isRunningOrPaused()) {
      writeLine(ERR_WITH + "Cannot while executing");
      return;
    }
    session = createSession();
    writeLine(OK);
  }

  protected boolean handleQuitRequest() throws IOException {
    if (session.isRunningOrPaused()) {
      writeLine(ERR_WITH + "Cannot while executing");
      return false;
    }
    writeLine(OK);
    return true;
  }

  protected void handleErrorDetailRequest() throws IOException {
    ImcadrServerRasterizerThread thread = session.getRasterizerThread();
    if (thread == null || !thread.hasException()) {
      writeLine(OK);
      writeLine(NIL);
      writeLine(".");
      return;
    }
    writeLine(OK);
    ByteArrayOutputStream stackTraceOut = new ByteArrayOutputStream(1024);
    PrintWriter stackTraceWriter = new PrintWriter(new OutputStreamWriter(stackTraceOut, UTF_8));
    thread.getException().printStackTrace(stackTraceWriter);
    stackTraceWriter.flush();
    ByteArrayInputStream stackTraceIn = new ByteArrayInputStream(stackTraceOut.toByteArray());
    BufferedReader stackTraceReader =
        new BufferedReader(new InputStreamReader(stackTraceIn, UTF_8));
    for (String each = stackTraceReader.readLine(); each != null; each =
        stackTraceReader.readLine()) {
      if (each.startsWith(".")) {
        writeLine("." + each);
      } else {
        writeLine(each);
      }
    }
    writeLine(".");
  }

  protected String readLine() throws IOException {
    return in.readLine();
  }

  protected void writeLine(String data) throws IOException {
    out.write(data);
    out.write('\n');
    out.flush();
  }

  protected void shutDown() throws ImcadrServerException {
    if (session == null) return;
    if (!session.isRunningOrPaused()) return;
    Thread thread = session.getRasterizerThread();
    if (thread == null) return;
    thread.interrupt();
    try {
      thread.join(SHUTDOWN_TIMEOUT);
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw new ImcadrServerException("Interrupted while shutting down", e);
    }
    if (thread.isAlive()) {
      throw new ImcadrServerException("Failed to shut down because failed to stop rasterizer");
    }
  }

  protected static Command parseCommand(String data) {
    String[] parts = SPACE_RE.split(data, 2);
    if (parts.length == 1) {
      return new Command(CommandType.valueFor(parts[0], false));
    } else {
      return new Command(CommandType.valueFor(parts[0], true), parts[1]);
    }
  }

  protected static File parseFile(String data) {
    String path = data.trim();
    if (NIL.equals(path)) return null;
    return new File(path);
  }

  protected static Double parseDouble(String data) throws ImcadrServerException {
    String seconds = data.trim();
    if (NIL.equals(seconds)) return null;
    try {
      return Double.valueOf(seconds);
    } catch (NumberFormatException e) {
      throw new ImcadrServerException("Value not a number", e);
    }
  }

  protected static Integer parseInteger(String data) throws ImcadrServerException {
    String integerString = data.trim();
    if (NIL.equals(integerString)) return null;
    try {
      return Integer.valueOf(integerString);
    } catch (NumberFormatException e) {
      throw new ImcadrServerException("Value not an integer", e);
    }
  }

  protected static Rectangle parseCellSize(String data) throws ImcadrServerException {
    String sizeString = data.trim();
    if (NIL.equals(sizeString)) return null;
    String[] parts = sizeString.split("[xX]");
    if (parts.length != 2) throw new ImcadrServerException("Value not a 2D size");
    try {
      double width = Double.parseDouble(parts[0].trim());
      double height = Double.parseDouble(parts[1].trim());
      return Rectangle.um(width, height);
    } catch (NumberFormatException e) {
      throw new ImcadrServerException("Value not a 2D size");
    }
  }

  protected static Rectangle parseRasterSize(String data) throws ImcadrServerException {
    String sizeString = data.trim();
    if (NIL.equals(sizeString)) return null;
    String[] parts = sizeString.split("[xX]");
    if (parts.length != 2) throw new ImcadrServerException("Value not a 2D size");
    try {
      double width = Double.parseDouble(parts[0].trim());
      double height = Double.parseDouble(parts[1].trim());
      return new Rectangle(width, height, LengthUnit.CELL);
    } catch (NumberFormatException e) {
      throw new ImcadrServerException("Value not a 2D size");
    }
  }

  protected static Boolean parseBoolean(String data) throws ImcadrServerException {
    String booleanString = data.trim();
    if (NIL.equals(booleanString)) return null;
    if (TRUE.equalsIgnoreCase(booleanString)) return Boolean.TRUE;
    if (FALSE.equalsIgnoreCase(booleanString)) return Boolean.FALSE;
    throw new ImcadrServerException("Value not a boolean");
  }

  protected static Pair<AnalysisPackage, String> parseAnalysisPackage(String data)
      throws ImcadrServerException {
    String resultPkgId = data.trim();
    if (NIL.equals(resultPkgId)) return new Pair<AnalysisPackage, String>(null, null);
    AnalysisPackage resultPkg = DATA_TO_ANALYSIS_PACKAGE_MAP.get(resultPkgId);
    if (resultPkg != null) return new Pair<AnalysisPackage, String>(resultPkg, resultPkgId);
    throw new ImcadrServerException("Value not an analysis package");
  }

  protected static String toData(ImcadrServerState state, String message) {
    StringBuilder buffer = new StringBuilder(128);
    buffer.append(toData(state));
    if (message == null || message.isEmpty()) return buffer.toString();
    buffer.append(" (");
    buffer.append(message);
    buffer.append(")");
    return buffer.toString();
  }

  protected static String toData(ImcadrServerState state) {
    switch (state) {
      case IDLE:
        return "Idle";
      case RUNNING:
        return "Running";
      case PAUSED:
        return "Paused";
      case ERROR:
        return "Error";
    }
    throw new BugException(state.getClass().getName() + ": " + state);
  }

  protected static String toData(Position2d value) {
    if (value == null) return NIL;
    return format("%.3f,%.3f", value.getX(), value.getY());
  }

  protected static String toData(Position3d value) {
    if (value == null) return NIL;
    return format("%.3f,%.3f,%.3f", value.getX(), value.getY(), value.getZ());
  }

  protected static String toData(Double[] value) {
    if (value == null) return NIL;
    StringBuilder buffer = new StringBuilder(32);
    for (int i = 0; i < value.length; i++) {
      if (i != 0) buffer.append(',');
      if (value[i] != null) buffer.append(format("%.3f", value[i]));
    }
    return buffer.toString();
  }

  protected static String toData(Integer[][] value) {
    if (value == null) return NIL;
    StringBuilder buffer = new StringBuilder(128);
    for (int r = 0; r < value.length; r++) {
      if (r != 0) buffer.append(';');
      for (int c = 0; c < value[0].length; c++) {
        if (c != 0) buffer.append(',');
        if (value[r][c] != null) buffer.append(value[r][c]);
      }
    }
    return buffer.toString();
  }

  protected static String toData(Double[][] value, int precision) {
    if (value == null) return NIL;
    final String format = "%." + precision + "f";
    StringBuilder buffer = new StringBuilder(128);
    for (int r = 0; r < value.length; r++) {
      if (r != 0) buffer.append(';');
      for (int c = 0; c < value[0].length; c++) {
        if (c != 0) buffer.append(',');
        if (value[r][c] != null) buffer.append(format(format, value[r][c]));
      }
    }
    return buffer.toString();
  }

  protected static String toData(SpaceGroup[][] value) {
    if (value == null) return NIL;
    StringBuilder buffer = new StringBuilder(128);
    for (int r = 0; r < value.length; r++) {
      if (r != 0) buffer.append(';');
      for (int c = 0; c < value[0].length; c++) {
        if (c != 0) buffer.append(',');
        if (value[r][c] != null) buffer.append(SPACE_GROUP_TO_DATA_MAP.get(value[r][c]));
      }
    }
    return buffer.toString();
  }

  protected static String toData(UnitCell[][] value, int lengthPrecision, int anglePrecision) {
    if (value == null) return NIL;
    final String lengthFormat = "%." + lengthPrecision + "f";
    final String angleFormat = "%." + anglePrecision + "f";
    StringBuilder buffer = new StringBuilder(128);
    for (int r = 0; r < value.length; r++) {
      if (r != 0) buffer.append(';');
      for (int c = 0; c < value[0].length; c++) {
        if (c != 0) buffer.append(',');
        if (value[r][c] != null) {
          double[] lengths = value[r][c].getLengths();
          for (int i = 0; i < lengths.length; i++) {
            buffer.append(format(lengthFormat, lengths[i]));
            buffer.append('/');
          }
          double[] angles = value[r][c].getAngles();
          for (int i = 0; i < angles.length; i++) {
            if (i != 0) buffer.append('/');
            buffer.append(format(angleFormat, angles[i]));
          }
        }
      }
    }
    return buffer.toString();
  }

  protected static String capitalize(String s) {
    if (s == null) return null;
    if (s.isEmpty()) return s;
    if (Character.isTitleCase(s.charAt(0))) return s;
    return s.substring(0, 1).toUpperCase(Locale.US) + s.substring(1);
  }

  protected static String format(String format, Object... args) {
    return String.format(Locale.US, format, args);
  }

  protected static InterruptedIOException newInterruptedIOException(Throwable cause) {
    try {
      InterruptedIOException result = new InterruptedIOException();
      result.initCause(cause);
      return result;
    } catch (IllegalStateException e) {
      return new InterruptedIOException(cause.toString());
    }
  }

  protected static Callable<Void> newCloseCallable(final Closeable c) {
    return new Callable<Void>() {
      @Override
      public Void call() throws Exception {
        c.close();
        return null;
      }
    };
  }

  private static class Command {
    private final CommandType type;
    private final String argument;

    public Command(CommandType type) {
      this(type, null);
    }

    public Command(CommandType type, String argument) {
      super();
      this.type = type;
      this.argument = argument;
    }

    public CommandType type() {
      return type;
    }

    public String argument() {
      return argument;
    }
  }

  private static enum CommandType {
    STAT("STAT"),
    IMAGES_DIR_GET("IMAGES_DIR", false),
    IMAGES_DIR_SET("IMAGES_DIR", true),
    EXP_TIME_GET("EXP_TIME", false),
    EXP_TIME_SET("EXP_TIME", true),
    CELL_SIZE_GET("CELL_SIZE", false),
    CELL_SIZE_SET("CELL_SIZE", true),
    ZY_SIZE_GET("ZY_SIZE", false),
    ZY_SIZE_SET("ZY_SIZE", true),
    ZX_SIZE_GET("ZX_SIZE", false),
    ZX_SIZE_SET("ZX_SIZE", true),
    FLY_SCAN_GET("FLY_SCAN", false),
    FLY_SCAN_SET("FLY_SCAN", true),
    CHASING_GET("CHASING", false),
    CHASING_SET("CHASING", true),
    ANALYSIS_PACKAGE_GET("ANALYSIS_PACKAGE", false),
    ANALYSIS_PACKAGE_SET("ANALYSIS_PACKAGE", true),
    WEAK_DIFF_GET("WEAK_DIFF", false),
    WEAK_DIFF_SET("WEAK_DIFF", true),
    HIGH_RES_CUTOFF_GET("HIGH_RES_CUTOFF", false),
    HIGH_RES_CUTOFF_SET("HIGH_RES_CUTOFF", true),
    LOW_RES_CUTOFF_GET("LOW_RES_CUTOFF", false),
    LOW_RES_CUTOFF_SET("LOW_RES_CUTOFF", true),
    MIN_SPOT_AREA_GET("MIN_SPOT_AREA", false),
    MIN_SPOT_AREA_SET("MIN_SPOT_AREA", true),
    MIN_SIGNAL_HEIGHT_GET("MIN_SIGNAL_HEIGHT", false),
    MIN_SIGNAL_HEIGHT_SET("MIN_SIGNAL_HEIGHT", true),
    START("START"),
    STOP("STOP"),
    RESET("RESET"),
    QUIT("QUIT"),
    ERROR_DETAIL("ERROR_DETAIL");

    private static final HashMap<String, CommandType> COMMAND_TAKING_ARGUMENT_TO_TYPE_MAP;
    private static final HashMap<String, CommandType> COMMAND_NOT_TAKING_ARGUMENT_TO_TYPE_MAP;

    static {
      COMMAND_TAKING_ARGUMENT_TO_TYPE_MAP = new HashMap<String, CommandType>();
      COMMAND_NOT_TAKING_ARGUMENT_TO_TYPE_MAP = new HashMap<String, CommandType>();
      for (CommandType each : CommandType.values()) {
        if (each.takesArgument()) {
          COMMAND_TAKING_ARGUMENT_TO_TYPE_MAP.put(each.command(), each);
        } else {
          COMMAND_NOT_TAKING_ARGUMENT_TO_TYPE_MAP.put(each.command(), each);
        }
      }
    }

    private final String command;
    private final boolean takesArgument;

    CommandType(String command) {
      this(command, false);
    }

    CommandType(String command, boolean takesArgument) {
      this.command = command;
      this.takesArgument = takesArgument;
    }

    public String command() {
      return command;
    }

    public boolean takesArgument() {
      return takesArgument;
    }

    public static CommandType valueFor(String command, boolean hasArgument) {
      if (hasArgument) {
        return COMMAND_TAKING_ARGUMENT_TO_TYPE_MAP.get(command);
      } else {
        return COMMAND_NOT_TAKING_ARGUMENT_TO_TYPE_MAP.get(command);
      }
    }
  }
}
