package imcacat.imcadr;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * I run a {@code Rasterizer} for {@code ImcadrServer}.
 */
public abstract class ImcadrServerRasterizerThread extends Thread {
  protected Rasterizer rasterizer;
  protected ImcadrServerResult result;
  protected Throwable exception;

  public ImcadrServerRasterizerThread(Rasterizer rasterizer, ImcadrServerResult result) {
    super();
    this.rasterizer = rasterizer;
    this.result = result;
    exception = null;
    setUncaughtExceptionHandler(new UncaughtExceptionHandler() {
      public void uncaughtException(Thread t, Throwable e) {
        exception = e;
      }
    });
  }

  @Override
  public void run() {
    try {
      rasterizer.run();
      if (rasterizer.hasException()) exception = rasterizer.getException();
    } catch (RuntimeException e) {
      result.setState(ImcadrServerState.ERROR);
      result.setMessage(null);
      exception = e;
    }
  }

  public boolean hasException() {
    return exception != null;
  }

  public Throwable getException() {
    return exception;
  }

  public ImcadrServerResult getResult() {
    return result;
  }

  protected void registerResultProgressInterest(BaseRasterizer r) {
    r.addPropertyChangeListener(BaseRasterizer.PROPNAME_PROGRESS, new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        updateResultProgress((Integer)event.getNewValue());
      }
    });
  }

  protected void registerResultStateInterest(final BaseRasterizer r) {
    r.addPropertyChangeListener(BaseRasterizer.PROPNAME_RUNNING, new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        updateResultState((Boolean)event.getNewValue(), r);
      }
    });
  }

  protected void registerResultTimeInterest(BaseRasterizer r) {
    r.addPropertyChangeListener(BaseRasterizer.PROPNAME_WALL_TIME, new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        updateResultTime((Long)event.getNewValue());
      }
    });
  }

  protected void registerResultOriginInterest(final Rasterizer2d r) {
    r.getRaster().addPropertyChangeListener(Raster2d.PROPNAME_ORIGIN,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            updateResultOrigin((Position2d)event.getNewValue(), r);
          }
        });
  }

  protected void registerResultScoresInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_SCORE,
            createScoreChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultSpotTotalsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_SPOT_TOTAL,
            createSpotTotalChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultInResolutionTotalsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_IN_RES_TOTAL,
            createInResolutionTotalChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultGoodBraggTotalsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_GOOD_BRAGG_TOTAL,
            createGoodBraggTotalChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultIceRingTotalsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_ICE_RING_TOTAL,
            createIceRingTotalChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultEstimatedResolutions1Interest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_ESTIMATED_RES1,
            createEstimatedResolution1ChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultEstimatedResolutions2Interest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_ESTIMATED_RES2,
            createEstimatedResolution2ChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultTotalIntegratedSignalsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_TOTAL_INT_SIGNAL,
            createTotalIntegratedSignalChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultMosaicitiesInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_MOSAICITY,
            createMosaicityChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultSpaceGroupsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_SPACE_GROUP,
            createSpaceGroupChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultUnitCellsInterest(Rasterizer2d r) {
    Raster2d raster = r.getRaster();
    int numRows = raster.rowCount();
    int numCols = raster.columnCount();
    Raster2dPlane plane = r.getPlane();
    for (int row = 0; row < numRows; row++) {
      for (int col = 0; col < numCols; col++) {
        raster.getCell(col, row).addPropertyChangeListener(RasterCell.PROPNAME_UNIT_CELL,
            createUnitCellChangeListener(col, row, plane));
      }
    }
  }

  protected void registerResultCenterInterest(final Rasterizer2d r) {
    r.addPropertyChangeListener(Rasterizer2d.PROPNAME_BEST_POSITION,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            updateResultCenter((Position2d)event.getNewValue(), r);
          }
        });
  }

  protected void updateResultProgress(Integer newProgress) {
    result.setProgress(newProgress);
  }

  protected void updateResultState(Boolean running, BaseRasterizer r) {
    if (running != null && running.booleanValue()) {
      result.setState(ImcadrServerState.RUNNING);
      result.setMessage(r.getStatus());
      return;
    }
    if (!r.wasSuccessful()) {
      result.setState(ImcadrServerState.ERROR);
      result.setMessage(null);
      return;
    }
    result.setState(ImcadrServerState.IDLE);
    result.setMessage(null);
  }

  protected void updateResultTime(Long newTime) {
    result.setTime((newTime == null) ? null : new Integer((int)(newTime.longValue() / 1000L)));
  }

  protected void updateResultOrigin(Position2d newOrigin, Rasterizer2d r) {
    Raster2dPlane plane = r.getPlane();
    switch (plane) {
      case ZY:
        result.setOriginZy(newOrigin);
        break;
      case ZX:
        result.setOriginZx(newOrigin);
        break;
    }
  }

  protected void updateResultCenter(Position2d newCenter, Rasterizer2d r) {
    Raster2dPlane plane = r.getPlane();
    switch (plane) {
      case ZY:
        result.setCenterZy(newCenter);
        break;
      case ZX:
        result.setCenterZx(newCenter);
        break;
    }
  }

  protected PropertyChangeListener createScoreChangeListener(final int column, final int row,
      final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer score = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setScoreZy(column, row, score);
            break;
          case ZX:
            result.setScoreZx(column, row, score);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createSpotTotalChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer total = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setSpotTotalZy(column, row, total);
            break;
          case ZX:
            result.setSpotTotalZx(column, row, total);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createInResolutionTotalChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer total = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setInResolutionTotalZy(column, row, total);
            break;
          case ZX:
            result.setInResolutionTotalZx(column, row, total);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createGoodBraggTotalChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer total = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setGoodBraggTotalZy(column, row, total);
            break;
          case ZX:
            result.setGoodBraggTotalZx(column, row, total);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createIceRingTotalChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer total = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setIceRingTotalZy(column, row, total);
            break;
          case ZX:
            result.setIceRingTotalZx(column, row, total);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createEstimatedResolution1ChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Double resolution = (Double)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setEstimatedResolution1Zy(column, row, resolution);
            break;
          case ZX:
            result.setEstimatedResolution1Zx(column, row, resolution);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createEstimatedResolution2ChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Double resolution = (Double)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setEstimatedResolution2Zy(column, row, resolution);
            break;
          case ZX:
            result.setEstimatedResolution2Zx(column, row, resolution);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createTotalIntegratedSignalChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Integer total = (Integer)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setTotalIntegratedSignalZy(column, row, total);
            break;
          case ZX:
            result.setTotalIntegratedSignalZx(column, row, total);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createMosaicityChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        Double mosaicity = (Double)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setMosaicityZy(column, row, mosaicity);
            break;
          case ZX:
            result.setMosaicityZx(column, row, mosaicity);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createSpaceGroupChangeListener(final int column,
      final int row, final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        SpaceGroup spaceGroup = (SpaceGroup)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setSpaceGroupZy(column, row, spaceGroup);
            break;
          case ZX:
            result.setSpaceGroupZx(column, row, spaceGroup);
            break;
        }
      }
    };
  }

  protected PropertyChangeListener createUnitCellChangeListener(final int column, final int row,
      final Raster2dPlane plane) {
    return new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent event) {
        UnitCell unitCell = (UnitCell)event.getNewValue();
        switch (plane) {
          case ZY:
            result.setUnitCellZy(column, row, unitCell);
            break;
          case ZX:
            result.setUnitCellZx(column, row, unitCell);
            break;
        }
      }
    };
  }
}
