package imcacat.imcadr;

import java.io.File;
import java.util.Locale;

/**
 * I am a session for {@code ImcadrServer}.
 */
public class ImcadrServerSession {
  protected BeamlineDriver driver;
  protected ImcadrConfiguration conf;
  protected File imagesDirectory;
  protected Double exposureTime;
  protected Rectangle cellSize;
  protected Rectangle zySize;
  protected Rectangle zxSize;
  protected Boolean flyScanEnabled;
  protected Boolean chasingEnabled;
  protected AnalysisPackage analysisPackage;
  protected String analysisPackageId;
  protected Boolean weakDiffracter;
  protected Double highResolutionCutoff;
  protected Double lowResolutionCutoff;
  protected Integer minimumSpotArea;
  protected Double minimumSignalHeight;
  protected ImcadrServerRasterizerThread rasterizerThread;
  protected ImcadrServerResult resultSnapshot;
  protected ImcadrServerResult resultDummy;

  public ImcadrServerSession(BeamlineDriver driver) {
    this(driver, ImcadrConfiguration.Builder.newWithDefaults().build());
  }

  public ImcadrServerSession(BeamlineDriver driver, ImcadrConfiguration conf) {
    this.driver = driver;
    this.conf = conf;
    imagesDirectory = null;
    exposureTime = null;
    cellSize = null;
    zySize = null;
    zxSize = null;
    flyScanEnabled = null;
    chasingEnabled = null;
    analysisPackage = null;
    analysisPackageId = null;
    weakDiffracter = null;
    highResolutionCutoff = null;
    lowResolutionCutoff = null;
    minimumSpotArea = null;
    minimumSignalHeight = null;
    rasterizerThread = null;
    resultSnapshot = null;
    resultDummy = new ImcadrServerResult();
  }

  public File getImagesDirectory() {
    return imagesDirectory;
  }

  public void setImagesDirectory(File imagesDirectory) {
    this.imagesDirectory = imagesDirectory;
  }

  public Double getExposureTime() {
    return exposureTime;
  }

  public void setExposureTime(Double exposureTime) {
    this.exposureTime = exposureTime;
  }

  public Rectangle getCellSize() {
    return cellSize;
  }

  public void setCellSize(Rectangle cellSize) {
    this.cellSize = cellSize;
  }

  public Rectangle getZySize() {
    return zySize;
  }

  public void setZySize(Rectangle zySize) {
    this.zySize = zySize;
  }

  public Rectangle getZxSize() {
    return zxSize;
  }

  public void setZxSize(Rectangle zxSize) {
    this.zxSize = zxSize;
  }

  public Boolean getFlyScanEnabled() {
    return flyScanEnabled;
  }

  public void setFlyScanEnabled(Boolean flyScanEnabled) {
    this.flyScanEnabled = flyScanEnabled;
  }

  public Boolean getChasingEnabled() {
    return chasingEnabled;
  }

  public void setChasingEnabled(Boolean chasingEnabled) {
    this.chasingEnabled = chasingEnabled;
  }

  public AnalysisPackage getAnalysisPackage() {
    return analysisPackage;
  }

  public String getAnalysisPackageId() {
    return analysisPackageId;
  }

  public void setAnalysisPackage(AnalysisPackage analysisPackage, String id) {
    this.analysisPackage = analysisPackage;
    this.analysisPackageId = id;
  }

  public Boolean getWeakDiffracter() {
    return weakDiffracter;
  }

  public void setWeakDiffracter(Boolean weakDiffracter) {
    this.weakDiffracter = weakDiffracter;
  }

  public Double getHighResolutionCutoff() {
    return highResolutionCutoff;
  }

  public void setHighResolutionCutoff(Double highResolutionCutoff) {
    this.highResolutionCutoff = highResolutionCutoff;
  }

  public Double getLowResolutionCutoff() {
    return lowResolutionCutoff;
  }

  public void setLowResolutionCutoff(Double lowResolutionCutoff) {
    this.lowResolutionCutoff = lowResolutionCutoff;
  }

  public Integer getMinimumSpotArea() {
    return minimumSpotArea;
  }

  public void setMinimumSpotArea(Integer minimumSpotArea) {
    this.minimumSpotArea = minimumSpotArea;
  }

  public Double getMinimumSignalHeight() {
    return minimumSignalHeight;
  }

  public void setMinimumSignalHeight(Double minimumSignalHeight) {
    this.minimumSignalHeight = minimumSignalHeight;
  }

  public ImcadrServerRasterizerThread getRasterizerThread() {
    return rasterizerThread;
  }

  public void setRasterizerThread(ImcadrServerRasterizerThread rasterizerThread) {
    this.rasterizerThread = rasterizerThread;
  }

  public ImcadrServerRasterizerThread createAndSetRasterizerThread() {
    ImcadrServerRasterizerThread newThread =
        is3d() ? new ImcadrServerRasterizer3dThread(createRasterizer3d(),
            new ImcadrServerResult(zySize, zxSize)) : new ImcadrServerRasterizer2dThread(
            createRasterizer2d(), new ImcadrServerResult(zySize, zxSize));
    setRasterizerThread(newThread);
    return newThread;
  }

  public ImcadrServerResult getResult() {
    if (rasterizerThread == null) return resultDummy;
    return rasterizerThread.getResult();
  }

  public ImcadrServerResult getResultSnapshot() {
    resultSnapshot = getResult().snapshot(resultSnapshot);
    return resultSnapshot;
  }

  public boolean isRunningOrPaused() {
    return getResult().getState().isRunningOrPaused();
  }

  public String validate() {
    String message;

    if (imagesDirectory == null) return "Images directory is required";
    message = validateImagesDirectory(imagesDirectory);
    if (message != null) return "Images directory " + message;

    if (exposureTime == null) return "Exposure time is required";
    message = validateExposureTime(exposureTime);
    if (message != null) return "Exposure time " + message;

    if (cellSize == null) return "Cell size is required";
    message = validateCellSize(cellSize);
    if (message != null) return "Cell size " + message;

    if (zySize == null && zxSize == null) return "ZY size or ZX size is required";

    message = validateZySize(zySize);
    if (message != null) return "ZY size " + message;

    message = validateZxSize(zxSize);
    if (message != null) return "ZX size " + message;

    message = validateHighResolutionCutoff(highResolutionCutoff);
    if (message != null) return "High resolution cutoff " + message;

    message = validateLowResolutionCutoff(lowResolutionCutoff);
    if (message != null) return "Low resolution cutoff " + message;

    if (highResolutionCutoff != null && lowResolutionCutoff != null &&
        highResolutionCutoff.doubleValue() >= lowResolutionCutoff.doubleValue()) {
      return "Low resolution cutoff must be greater than high resolution cutoff";
    }

    message = validateMinimumSpotArea(minimumSpotArea);
    if (message != null) return "Minimum spot area " + message;

    message = validateMinimumSignalHeight(minimumSignalHeight);
    if (message != null) return "Minimum signal height " + message;

    return null;
  }

  protected boolean is3d() {
    return zySize != null && zxSize != null;
  }

  protected Rasterizer2d createRasterizer2d() {
    if (zySize != null) return createRasterizer2d(Raster2dPlane.ZY);
    if (zxSize != null) return createRasterizer2d(Raster2dPlane.ZX);
    throw new BugException();
  }

  protected Rasterizer2d createRasterizer2d(Raster2dPlane plane) {
    return new Rasterizer2d(createRaster2dSpec(plane), driver, false);
  }

  protected Raster2dSpec createRaster2dSpec(Raster2dPlane plane) {
    Rectangle rasterSize = null;
    if (plane.isZy()) rasterSize = zySize;
    else if (plane.isZx()) rasterSize = zxSize;
    else throw new BugException();
    return new Raster2dSpec(plane, (int)rasterSize.getWidth(), (int)rasterSize.getHeight(), null,
        null, new Raster2dMask(), cellSize.getWidth(), cellSize.getHeight(), cellSize.getUnit(),
        secondsToMillis(exposureTime.doubleValue()), computeMaxNumConcurrentAnalyzers(),
        (flyScanEnabled == null) ? true : flyScanEnabled.booleanValue(),
        (chasingEnabled == null) ? true : chasingEnabled.booleanValue(), null, imagesDirectory,
        createImageAnalysisSpec(), conf.clusterKeyFile(), conf.clusterNodes());
  }

  protected int computeMaxNumConcurrentAnalyzers() {
    int numProcessors = processorCount();
    if (conf.clusterKeyFile() == null) return numProcessors;
    int numClusterNodes = conf.clusterNodes().size();
    if (numClusterNodes == 0) return numProcessors;
    return numClusterNodes * 4;
  }

  protected ImageAnalysisSpec createImageAnalysisSpec() {
    ImageAnalysisSpec.Builder b = new ImageAnalysisSpec.Builder();
    b.analysisPackage(analysisPackage);
    for (AnalysisProgram each : AnalysisProgram.values()) {
      b.analysisProgramArguments(each, conf.analysisProgramArguments(each));
    }
    if (weakDiffracter != null && weakDiffracter.booleanValue() && highResolutionCutoff == null) {
      b.highResolutionCutoff(Resources.getInstance().getWeakDiffracterHighResolutionCutoff());
    } else {
      b.highResolutionCutoff(highResolutionCutoff);
    }
    b.lowResolutionCutoff(lowResolutionCutoff);
    b.minimumSpotArea(minimumSpotArea);
    b.minimumSignalHeight(minimumSignalHeight);
    return b.build();
  }

  protected Rasterizer3d createRasterizer3d() {
    return new Rasterizer3d(createRaster3dSpec(), driver, false);
  }

  protected Raster3dSpec createRaster3dSpec() {
    return new Raster3dSpec(createRaster2dSpec(Raster2dPlane.ZY),
        createRaster2dSpec(Raster2dPlane.ZX));
  }

  public static String validateImagesDirectory(File dir) {
    if (dir == null) return null;
    if (!dir.exists()) return "does not exist";
    if (!dir.isDirectory()) return "not a directory";
    return null;
  }

  public static String validateExposureTime(Double expTime) {
    if (expTime == null) return null;
    if (expTime.doubleValue() <= 0.0) return "must be greater than 0";
    if (expTime.isInfinite()) return "must not be infinite";
    if (expTime.isNaN()) return "must not be Not-a-Number";
    return null;
  }

  public static String validateCellSize(Rectangle size) {
    if (size == null) return null;
    if (size.getWidth() <= 0.0) return "width must be greater than 0";
    if (size.getHeight() <= 0.0) return "height must be greater than 0";
    if (Double.isInfinite(size.getWidth())) return "width must not be infinite";
    if (Double.isInfinite(size.getHeight())) return "height must not be infinite";
    if (Double.isNaN(size.getWidth())) return "width must not be Not-a-Number";
    if (Double.isNaN(size.getHeight())) return "height must not be Not-a-Number";
    return null;
  }

  public static String validateZySize(Rectangle size) {
    return validateRasterSize(size);
  }

  public static String validateZxSize(Rectangle size) {
    return validateRasterSize(size);
  }

  protected static String validateRasterSize(Rectangle size) {
    if (size == null) return null;
    if (size.getWidth() <= 0.0) return "width must be greater than 0";
    if (size.getHeight() <= 0.0) return "height must be greater than 0";
    if (Double.isInfinite(size.getWidth())) return "width must not be infinite";
    if (Double.isInfinite(size.getHeight())) return "height must not be infinite";
    if (Double.isNaN(size.getWidth())) return "width must not be Not-a-Number";
    if (Double.isNaN(size.getHeight())) return "height must not be Not-a-Number";
    return null;
  }

  public static String validateHighResolutionCutoff(Double highResolutionCutoff) {
    if (highResolutionCutoff == null) return null;
    double cutoffMin = Resources.getInstance().getHighResolutionCutoffMin();
    if (highResolutionCutoff.doubleValue() < cutoffMin) {
      return String.format(Locale.US, "must be greater than or equal to %.2f", cutoffMin);
    }
    double cutoffMax = Resources.getInstance().getHighResolutionCutoffMax();
    if (highResolutionCutoff.doubleValue() > cutoffMax) {
      return String.format(Locale.US, "must be less than or equal to %.2f", cutoffMax);
    }
    if (highResolutionCutoff.isInfinite()) return "must not be infinite";
    if (highResolutionCutoff.isNaN()) return "must not be Not-a-Number";
    return null;
  }

  public static String validateLowResolutionCutoff(Double lowResolutionCutoff) {
    if (lowResolutionCutoff == null) return null;
    double cutoffMin = Resources.getInstance().getLowResolutionCutoffMin();
    if (lowResolutionCutoff.doubleValue() < cutoffMin) {
      return String.format(Locale.US, "must be greater than or equal to %.2f", cutoffMin);
    }
    double cutoffMax = Resources.getInstance().getLowResolutionCutoffMax();
    if (lowResolutionCutoff.doubleValue() > cutoffMax) {
      return String.format(Locale.US, "must be less than or equal to %.2f", cutoffMax);
    }
    if (lowResolutionCutoff.isInfinite()) return "must not be infinite";
    if (lowResolutionCutoff.isNaN()) return "must not be Not-a-Number";
    return null;
  }

  public static String validateMinimumSpotArea(Integer minimumSpotArea) {
    if (minimumSpotArea == null) return null;
    int cutoffMin = Resources.getInstance().getMinimumSpotAreaMin();
    if (minimumSpotArea.intValue() < cutoffMin) {
      return String.format(Locale.US, "must be greater than or equal to %d", cutoffMin);
    }
    int cutoffMax = Resources.getInstance().getMinimumSpotAreaMax();
    if (minimumSpotArea.intValue() > cutoffMax) {
      return String.format(Locale.US, "must be less than or equal to %d", cutoffMax);
    }
    return null;
  }

  public static String validateMinimumSignalHeight(Double minimumSignalHeight) {
    if (minimumSignalHeight == null) return null;
    double cutoffMin = Resources.getInstance().getMinimumSignalHeightMin();
    if (minimumSignalHeight.doubleValue() < cutoffMin) {
      return String.format(Locale.US, "must be greater than or equal to %.2f", cutoffMin);
    }
    double cutoffMax = Resources.getInstance().getMinimumSignalHeightMax();
    if (minimumSignalHeight.doubleValue() > cutoffMax) {
      return String.format(Locale.US, "must be less than or equal to %.2f", cutoffMax);
    }
    if (minimumSignalHeight.isInfinite()) return "must not be infinite";
    if (minimumSignalHeight.isNaN()) return "must not be Not-a-Number";
    return null;
  }

  protected static long secondsToMillis(double seconds) {
    return (long)(seconds * 1000.0);
  }

  protected static int processorCount() {
    return Runtime.getRuntime().availableProcessors();
  }
}
