package imcacat.imcadr;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.channels.ClosedByInterruptException;
import java.nio.file.AccessDeniedException;
import java.nio.file.NoSuchFileException;
import java.util.Collections;
import java.util.LinkedList;
import java.util.logging.FileHandler;
import java.util.logging.Filter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.event.QueuedEventDispatcher;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.imcadr.util.LoggingUtils;
import imcacat.imcadr.util.StringUtils;
import imcacat.imcadr.util.Version;

import javax.swing.Action;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import com.jgoodies.looks.plastic.PlasticLookAndFeel;

/**
 * I am the core of the imcadr application.
 *
 * I handle starting the imcadr application as well as shutting it down.
 */
public class ImcadrSystem {
  private static final Logger LOG = Logger.getLogger(ImcadrSystem.class.getName());

  protected static Action specialMenuBarExitAction;
  protected static Action specialMenuBarAboutAction;

  public static void main(String[] args) {
    setHeadless(true);
    initializeLogging();
    initializeEpicsca4jEventDispatcherName();
    CliArguments arguments = parseArguments(args);
    if (arguments.isEmpty() || arguments.onlyHasImagesDirectory()) {
      setHeadless(false);
      startGui(arguments);
    } else {
      if (arguments.help()) {
        printHelpAndExit();
      } else if (arguments.version()) {
        printVersionAndExit();
      }
      initializeDebugLogging(arguments.debugFile());
      ImcadrConfiguration conf = parseConfiguration(arguments);
      if (arguments.server()) {
        startServer(arguments, conf);
      } else {
        startTui(arguments, conf);
      }
    }
  }

  protected static void initializeLogging() {
    try (InputStream s = ImcadrSystem.class.getResourceAsStream("/logging.properties")) {
      LogManager.getLogManager().readConfiguration(s);
    } catch (IOException e) {
      System.err.println("error: fatal exception");
      e.printStackTrace();
      basicExit(ExitStatus.FAILURE.getStatus());
    }
  }

  protected static void initializeDebugLogging(File debugFile) {
    if (debugFile == null) return;
    File debugFileDir = debugFile.getParentFile();
    if (debugFileDir != null && !debugFileDir.isDirectory()) {
      System.err.println("error: " + debugFileDir + ": not a directory");
      basicExit(ExitStatus.FAILURE.getStatus());
      return; /* NOTREACHED */
    }
    String pattern = debugFile.getPath();
    pattern.replace(File.separatorChar, '/');
    pattern.replace("%", "%%");
    Handler h;
    try {
      h = new FileHandler(pattern);
    } catch (IOException e) {
      System.err.println("error: failed to initialize debug file logging");
      e.printStackTrace();
      basicExit(ExitStatus.FAILURE.getStatus());
      return; /* NOTREACHED */
    }
    h.setFormatter(new SimpleFormatter());
    h.setFilter(new Filter() {
      @Override
      public boolean isLoggable(LogRecord record) {
        String loggerName = record.getLoggerName();
        if (loggerName == null) return false;
        return loggerName.startsWith("imcacat.imcadr.");
      }
    });
    Logger.getLogger("").addHandler(h);
  }

  protected static void initializeEpicsca4jEventDispatcherName() {
    imcacat.epicsca4j.util.Resources.getDefaultProperties().setProperty(
        imcacat.epicsca4j.util.Resources.EPICSCA4J_JCA_CONTEXT_EVENT_DISPATCHER_KEY,
        QueuedEventDispatcher.class.getCanonicalName());
  }

  protected static CliArguments parseArguments(String[] args) {
    try {
      return CliArgumentParser.parse(args);
    } catch (CliArgumentException e) {
      String message = e.getMessage();
      if (message == null) message = "invalid command line argument(s)";
      System.err.println("error: " + message);
      System.err.println("Invoke with --help option for a help message");
      basicExit(ExitStatus.FAILURE.getStatus());
      return null; /* NOTREACHED */
    }
  }

  protected static void setHeadless(boolean headless) {
    System.setProperty("java.awt.headless", Boolean.toString(headless));
  }

  protected static ImcadrConfiguration parseConfiguration(CliArguments a) {
    ImcadrConfiguration cliConf = a.toConfiguration();
    ImcadrConfiguration otherConf = null;
    ImcadrConfiguration userConf = null;
    ImcadrConfiguration sysConf = null;
    if (a.configurationFile() != null) {
      otherConf = parseConfFile(a.configurationFile());
    } else {
      File userConfFile = findUserConfFile();
      if (userConfFile != null) {
        userConf = parseConfFile(userConfFile);
      }
      File sysConfFile = Resources.getInstance().getSystemConfigurationFile();
      if (sysConfFile.exists()) {
        sysConf = parseConfFile(sysConfFile);
      }
    }
    LinkedList<ImcadrConfiguration> confs = new LinkedList<>();
    confs.add(cliConf);
    if (otherConf != null) {
      confs.add(otherConf);
    } else {
      if (userConf != null) confs.add(userConf);
      if (sysConf != null) confs.add(sysConf);
    }
    Collections.reverse(confs);
    ImcadrConfiguration result = confs.removeFirst();
    for (ImcadrConfiguration each : confs) {
      result = each.defaults(result);
    }
    return result;
  }

  protected static ImcadrConfiguration parseConfFile(File f) {
    try {
      return ImcadrConfigurationParser.parse(f);
    } catch (ConfigurationException e) {
      System.err.printf("%s: %s%n", f, e.getMessage());
      basicExit(ExitStatus.FAILURE.getStatus());
      return null; /* NOTREACHED */
    } catch (AccessDeniedException e) {
      System.err.printf("%s: access denied%n", f);
      basicExit(ExitStatus.FAILURE.getStatus());
      return null; /* NOTREACHED */
    } catch (FileNotFoundException | NoSuchFileException e) {
      System.err.printf("%s: file not found%n", f);
      basicExit(ExitStatus.FAILURE.getStatus());
      return null; /* NOTREACHED */
    } catch (IOException e) {
      String msg = e.getMessage();
      if (msg == null || msg.trim().isEmpty()) {
        msg = "I/O error";
      }
      System.err.printf("%s: %s%n", f, msg);
      basicExit(ExitStatus.FAILURE.getStatus());
      return null; /* NOTREACHED */
    }
  }

  protected static File findUserConfFile() {
    String envVarPath = System.getenv("IMCADR_CONFIG");
    if (envVarPath != null) {
      return new File(envVarPath);
    }
    File result = new File(Resources.getUserHomeDirectory(), ".imcadr.conf");
    if (result.exists()) return result;
    envVarPath = System.getenv("XDG_CONFIG_HOME");
    if (envVarPath != null) {
      result = new File(StringUtils.join(File.separator, envVarPath, "imcadr", "config"));
      if (result.exists()) return result;
    } else {
      result = new File(Resources.getUserHomeDirectory(),
          StringUtils.join(File.separator, ".config", "imcadr", "config"));
      if (result.exists()) return result;
    }
    return null;
  }

  protected static void printHelpAndExit() {
    printHelpAndExit(System.out, ExitStatus.SUCCESS.getStatus());
  }

  protected static void printHelpAndExit(OutputStream sink, int exitStatus) {
    try {
      CliArgumentParser.printHelp(sink);
      sink.flush();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.interrupted();
      basicExit(ExitStatus.FAILURE.getStatus());
    } catch (IOException e) {
      LOG.log(Level.SEVERE, "Fatal exception", e);
      basicExit(ExitStatus.FAILURE.getStatus());
    }
    basicExit(exitStatus);
  }

  protected static void printVersionAndExit() {
    printVersionAndExit(System.out);
  }

  protected static void printVersionAndExit(OutputStream sink) {
    try {
      printVersion(sink);
      sink.flush();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.interrupted();
      basicExit(ExitStatus.FAILURE.getStatus());
    } catch (IOException e) {
      LOG.log(Level.SEVERE, "Fatal exception", e);
      basicExit(ExitStatus.FAILURE.getStatus());
    }
    basicExit(ExitStatus.SUCCESS.getStatus());
  }

  protected static void printVersionAndExit(Writer sink) {
    try {
      printVersion(sink);
      sink.flush();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.interrupted();
      basicExit(ExitStatus.FAILURE.getStatus());
    } catch (IOException e) {
      LOG.log(Level.SEVERE, "Fatal exception", e);
      basicExit(ExitStatus.FAILURE.getStatus());
    }
    basicExit(ExitStatus.SUCCESS.getStatus());
  }

  protected static void printVersion(OutputStream sink) throws IOException {
    Writer sinkWriter = new OutputStreamWriter(sink);
    try {
      printVersion(sinkWriter);
    } finally {
      sinkWriter.close();
    }
  }

  protected static void printVersion(Writer sink) throws IOException {
    sink.write(Resources.getApplicationName() + " " + Version.getId());
    sink.write(Resources.getLineSeparator());
  }

  public static void basicExit(int exitStatus) {
    System.exit(exitStatus);
  }

  protected static void startGui(final CliArguments arguments) {
    initializeUiLookAndFeel();
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        try {
          ImcadrFrame frame = ImcadrFrame.onImagesDirectory(arguments.imagesDirectory());
          frame.pack();
          frame.setLocationRelativeTo(null);
          frame.requestPreferredInitialFocus();
          frame.setVisible(true);
        } catch (InterruptedException e) {
          basicExit(ExitStatus.FAILURE.getStatus());
        } catch (EpicsCaException | RuntimeException e) {
          LOG.log(Level.SEVERE, "Fatal exception", e);
          basicExit(ExitStatus.FAILURE.getStatus());
        }
      }
    });
  }

  protected static void initializeUiLookAndFeel() {
    PlasticLookAndFeel.setSelectTextOnKeyboardFocusGained(true);
    try {
      UIManager.setLookAndFeel(new PlasticLookAndFeel());
    } catch (UnsupportedLookAndFeelException e) {
      LOG.log(Level.SEVERE, "Failed to set Look & Feel", e);
      throw new RuntimeException(e);
    }
  }

  protected static void startTui(CliArguments arguments, ImcadrConfiguration conf) {
    try {
      disableFiringEpicsPvChangeEventsFromEdt();
      runTui(arguments, conf);
    } catch (RuntimeException e) {
      LOG.log(Level.SEVERE, "Fatal exception", e);
      basicExit(ExitStatus.FAILURE.getStatus());
    }
    basicExit(ExitStatus.SUCCESS.getStatus());
  }

  protected static void runTui(CliArguments arguments, ImcadrConfiguration conf) {
    if (arguments.imagesDirectory() == null) arguments.imagesDirectory(Resources
        .getCurrentWorkingDirectory());
    Imcadr model = new Imcadr();
    model.setRasterizerWorkingDirectory(arguments.imagesDirectory());
    ImcadrTui tui = new ImcadrTui(model, arguments, conf);
    final Thread tuiThread = new Thread(tui, "ImcadrTui");

    try {
      Runtime.getRuntime().addShutdownHook(new Thread() {
        @Override
        public void run() {
          try {
            tuiThread.interrupt();
            tuiThread.join(ImcadrTui.SHUTDOWN_TIMEOUT);
          } catch (InterruptedException e) {
            /* IGNORED */
          }
        }
      });
    } catch (IllegalStateException e) {
      basicExit(ExitStatus.FAILURE.getStatus());
    }

    tuiThread.start();
    try {
      tuiThread.join();
    } catch (InterruptedException e) {
      tuiThread.interrupt();
      try {
        tuiThread.join(ImcadrTui.SHUTDOWN_TIMEOUT);
      } catch (InterruptedException ex) {
        /* IGNORED */
      }
      basicExit(ExitStatus.FAILURE.getStatus());
    }

    Exception exception = tui.exception();
    if (exception == null) {
      basicExit(ExitStatus.SUCCESS.getStatus());
    } else {
      try {
        throw exception;
      } catch (BestScoreCutoffException e) {
        StringBuilder sb = new StringBuilder();
        sb.append("notice: best score ");
        if (e.score() != null) sb.append("(" + e.score() + ") ");
        sb.append("<= cutoff");
        if (e.cutoff() != null) sb.append(" (" + e.cutoff() + ")");
        System.err.println(sb.toString());
        basicExit(ExitStatus.BEST_SCORE_CUTOFF.getStatus());
      } catch (InterruptedException | InterruptedIOException | ClosedByInterruptException e) {
        Thread.interrupted();
        basicExit(ExitStatus.FAILURE.getStatus());
      } catch (Exception e) {
        LOG.log(Level.SEVERE, "Fatal exception", e);
        basicExit(ExitStatus.FAILURE.getStatus());
      }
    }
  }

  protected static void startServer(CliArguments arguments, ImcadrConfiguration conf) {
    try {
      disableConsoleLogging();
      disableFiringEpicsPvChangeEventsFromEdt();
      runServer(conf);
    } catch (RuntimeException e) {
      System.err.println("error: fatal exception");
      e.printStackTrace();
      basicExit(ExitStatus.FAILURE.getStatus());
    }
  }

  protected static void disableConsoleLogging() {
    LoggingUtils.disableConsoleLogging();
  }

  protected static void disableFiringEpicsPvChangeEventsFromEdt() {
    EpicsPv.setFirePropertyChangeEventsFromEdt(false);
  }

  protected static void runServer(ImcadrConfiguration conf) {
    ImcadrServer server = new ImcadrServer(new Imcadr(), conf);
    final Thread serverThread = new Thread(server, "ImcadrServer");

    try {
      Runtime.getRuntime().addShutdownHook(new Thread() {
        @Override
        public void run() {
          try {
            serverThread.interrupt();
            serverThread.join(ImcadrServer.SHUTDOWN_TIMEOUT);
          } catch (InterruptedException e) {
            /* IGNORED */
          }
        }
      });
    } catch (IllegalStateException e) {
      basicExit(ExitStatus.FAILURE.getStatus());
    }

    serverThread.start();
    try {
      serverThread.join();
    } catch (InterruptedException e) {
      serverThread.interrupt();
      try {
        serverThread.join(ImcadrServer.SHUTDOWN_TIMEOUT);
      } catch (InterruptedException ex) {
        /* IGNORED */
      }
      basicExit(ExitStatus.FAILURE.getStatus());
    }

    Exception exception = server.exception();
    if (exception == null) {
      basicExit(ExitStatus.SUCCESS.getStatus());
    } else {
      try {
        throw exception;
      } catch (InterruptedException | InterruptedIOException | ClosedByInterruptException e) {
        Thread.interrupted();
        basicExit(ExitStatus.FAILURE.getStatus());
      } catch (Exception e) {
        System.err.println("error: fatal exception");
        e.printStackTrace();
        basicExit(ExitStatus.FAILURE.getStatus());
      }
    }
  }
}
