package imcacat.imcadr;

/**
 * I am the text user interface for this application.
 */
public class ImcadrTui implements Runnable {
  public static final long SHUTDOWN_TIMEOUT = 15000L;

  protected Imcadr model;
  protected CliArguments arguments;
  protected ImcadrConfiguration conf;
  protected Exception exception;

  public ImcadrTui(Imcadr model, CliArguments arguments, ImcadrConfiguration conf) {
    this.model = model;
    this.arguments = arguments;
    this.conf = conf;
    exception = null;
  }

  @Override
  public void run() {
    BaseRasterizer rasterizer = null;
    try {
      rasterizer = createRasterizer();
      rasterizer.run();
      if (rasterizer.hasException()) {
        exception = toException(rasterizer.getException());
      } else if (!rasterizer.wasSuccessful()) {
        exception = new UnsuccessfulRasterizeException(rasterizer);
      }
    } catch (RuntimeException e) {
      exception = e;
    }
  }

  public Exception exception() {
    return exception;
  }

  protected BaseRasterizer createRasterizer() {
    BaseRasterizer result;
    if (arguments.zySize() != null && arguments.zxSize() != null) {
      result = new Rasterizer3d(create3dRasterSpec(), model.getBeamlineDriver(), false);
    } else if (arguments.zySize() != null) {
      result = new Rasterizer2d(createZy2dRasterSpec(), model.getBeamlineDriver(), false);
    } else if (arguments.zxSize() != null) {
      result = new Rasterizer2d(createZx2dRasterSpec(), model.getBeamlineDriver(), false);
    } else {
      throw new BugException("Neither ZY nor ZX raster size specified");
    }
    result.doNotLogException();
    return result;
  }

  protected Raster3dSpec create3dRasterSpec() {
    return new Raster3dSpec(createZy2dRasterSpec(), createZx2dRasterSpec(), arguments.zxFirst());
  }

  protected Raster2dSpec createZy2dRasterSpec() {
    return new Raster2dSpec(Raster2dPlane.ZY, arguments.zySize().width,
        arguments.zySize().height, (arguments.zyPreClipSize() == null) ? null
            : arguments.zyPreClipSize().width, (arguments.zyPreClipSize() == null) ? null
            : arguments.zyPreClipSize().height, (arguments.zyMask() == null)
            ? new Raster2dMask() : arguments.zyMask(), arguments.cellSize().width,
        arguments.cellSize().height, LengthUnit.MICROMETER,
        secondsToMillis(arguments.exposureTime()), computeMaxNumConcurrentAnalyzers(),
        arguments.flyScanEnabled(), arguments.chasingEnabled(), arguments.bestScoreCutoff(),
        arguments.imagesDirectory(), createImageAnalysisSpec(), conf.clusterKeyFile(),
        conf.clusterNodes());
  }

  protected Raster2dSpec createZx2dRasterSpec() {
    return new Raster2dSpec(Raster2dPlane.ZX, arguments.zxSize().width,
        arguments.zxSize().height, (arguments.zxPreClipSize() == null) ? null
            : arguments.zxPreClipSize().width, (arguments.zxPreClipSize() == null) ? null
            : arguments.zxPreClipSize().height, (arguments.zxMask() == null)
            ? new Raster2dMask() : arguments.zxMask(), arguments.cellSize().width,
        arguments.cellSize().height, LengthUnit.MICROMETER,
        secondsToMillis(arguments.exposureTime()), computeMaxNumConcurrentAnalyzers(),
        arguments.flyScanEnabled(), arguments.chasingEnabled(), arguments.bestScoreCutoff(),
        arguments.imagesDirectory(), createImageAnalysisSpec(), conf.clusterKeyFile(),
        conf.clusterNodes());
  }

  protected int computeMaxNumConcurrentAnalyzers() {
    int numProcessors = processorCount();
    if (conf.clusterKeyFile() == null) return numProcessors;
    int numClusterNodes = conf.clusterNodes().size();
    if (numClusterNodes == 0) return numProcessors;
    return numClusterNodes * 4;
  }

  protected ImageAnalysisSpec createImageAnalysisSpec() {
    ImageAnalysisSpec.Builder b = new ImageAnalysisSpec.Builder();
    b.analysisPackage(arguments.analysisPackage());
    for (AnalysisProgram each : AnalysisProgram.values()) {
      b.analysisProgramArguments(each, conf.analysisProgramArguments(each));
    }
    if (arguments.weakDiffracter() && arguments.highResolutionCutoff() == null) {
      b.highResolutionCutoff(Resources.getInstance().getWeakDiffracterHighResolutionCutoff());
    } else {
      b.highResolutionCutoff(arguments.highResolutionCutoff());
    }
    b.lowResolutionCutoff(arguments.lowResolutionCutoff());
    b.minimumSpotArea(arguments.minimumSpotArea());
    b.minimumSignalHeight(arguments.minimumSignalHeight());
    return b.build();
  }

  private static long secondsToMillis(double seconds) {
    return (long)(seconds * 1000.0);
  }

  private static int processorCount() {
    return Runtime.getRuntime().availableProcessors();
  }

  private static Exception toException(Throwable t) {
    return (t instanceof Exception) ? (Exception)t : new Exception(t);
  }
}
