package imcacat.imcadr;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.charset.Charset;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * I parse the output of the <a
 * href="http://adder.lbl.gov/labelit/">LABELIT</a>
 * <tt>distl.signal_strength</tt> program.
 */
public class LabelitDistSignalStrengthParser extends ImageAnalysisParser {
  private static final Charset UTF8 = Charset.forName("UTF-8");

  private static final Pattern SPOT_TOTAL_RE = newPattern("^\\s*Spot Total\\s*:\\s*(\\d+).*");
  private static final Pattern IN_RES_TOTAL_RE =
      newPattern("^\\s*In-Resolution Total\\s*:\\s*(\\d+).*");
  private static final Pattern GOOD_BRAGG_TOTAL_RE =
      newPattern("^\\s*Good Bragg Candidates\\s*:\\s*(\\d+).*");
  private static final Pattern ICE_RING_TOTAL_RE = newPattern("^\\s*Ice Rings\\s*:\\s*(\\d+).*");
  private static final Pattern ESTIMATED_RES1_RE =
      newPattern("^\\s*Method 1 Resolution\\s*:\\s*([0-9\\.a-zA-Z]+).*");
  private static final Pattern ESTIMATED_RES2_RE =
      newPattern("^\\s*Method 2 Resolution\\s*:\\s*([0-9\\.a-zA-Z]+).*");
  private static final Pattern TOTAL_INT_SIGNAL_RE =
      newPattern("^\\s*Total integrated signal[^a-zA-Z].*\\s(\\d+)\\s*$");

  private BufferedReader reader = null;
  private InputStreamReader inputStreamReader = null;

  public LabelitDistSignalStrengthParser(InputStream in) {
    super(in);
  }

  @Override
  public ImageAnalysisResult run() {
    try {
      initializeReader();
      return parse();
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
      return newAnalysisResult("Interrupted", e);
    } catch (IOException | RuntimeException e) {
      return newAnalysisResult("Failed to parse analysis program output", e);
    } finally {
      closeReaderSilently();
    }
  }

  private void initializeReader() {
    inputStreamReader = new InputStreamReader(in, UTF8);
    reader = new BufferedReader(inputStreamReader);
  }

  private void closeReaderSilently() {
    closeSilently(reader);
    closeSilently(inputStreamReader);
  }

  private ImageAnalysisResult parse() throws IOException {
    ImageAnalysisResult result = new ImageAnalysisResult();
    result.setSpotTotal(parseSpotTotal());
    result.setInResolutionTotal(parseInResolutionTotal());
    result.setGoodBraggTotal(parseGoodBraggTotal());
    result.setIceRingTotal(parseIceRingTotal());
    result.setEstimatedResolution1(parseEstimatedResolution1());
    result.setEstimatedResolution2(parseEstimatedResolution2());
    result.setTotalIntegratedSignal(parseTotalIntegratedSignal());
    result.setScore(computeScore(result.getGoodBraggTotal()));
    return result;
  }

  private Integer parseSpotTotal() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = SPOT_TOTAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private Integer parseInResolutionTotal() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = IN_RES_TOTAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private Integer parseGoodBraggTotal() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = GOOD_BRAGG_TOTAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private Integer parseIceRingTotal() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = ICE_RING_TOTAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private Double parseEstimatedResolution1() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = ESTIMATED_RES1_RE.matcher(line);
      if (m.matches()) return getDouble(m.group(1));
    }
    return null;
  }

  private Double parseEstimatedResolution2() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = ESTIMATED_RES2_RE.matcher(line);
      if (m.matches()) return getDouble(m.group(1));
    }
    return null;
  }

  private Integer parseTotalIntegratedSignal() throws IOException {
    for (String line = reader.readLine(); line != null; line = reader.readLine()) {
      Matcher m = TOTAL_INT_SIGNAL_RE.matcher(line);
      if (m.matches()) return getInteger(m.group(1));
    }
    return null;
  }

  private static Integer getInteger(String value) {
    try {
      return Integer.valueOf(value);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  private static Double getDouble(String value) {
    try {
      return Double.valueOf(value);
    } catch (NumberFormatException e) {
      return null;
    }
  }

  private static Integer computeScore(Integer goodBraggTotal) {
    return goodBraggTotal;
  }
}
