/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.imcadr;

import java.io.Closeable;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.Reader;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * A stream for reading lines from a {@link Reader} with a timeout, with memory
 * limits, and without blocking.
 * <p>
 * It is best if the {@code Reader} reads from an {@code InterruptibleChannel}
 * so that reads will be interruptible. Unfortunately, most of the time this
 * won't be the case. To ensure {@code Reader} reads can be interrupted, supply
 * a close prerequisite.
 * <p>
 * A close prerequisite is a {@code Callable} that will end a blocking read.
 * The {@code Callable} can do anything, but the most common thing it will do
 * is close a stream or terminate a process. For example, if an instance of
 * this class is reading lines from a {@code Process} instance's standard
 * output stream, the close prerequisite should close the {@code InputStream}
 * attached to the process' standard output stream or terminate the process, so
 * that the blocking read will return.
 * <p>
 * Instances of this class are not safe for use by multiple threads.
 *
 * @version j6utils 0.5.0
 */
public class LineReader implements Closeable {
  private final LineWorker worker;
  private boolean workerStartNeeded;
  private Callable<?> closePrerequisite;
  private boolean closed;

  /**
   * Constructs a new instance with the specified {@code Reader} and an initial
   * line buffer size of <tt>80</tt> characters, a maximum line buffer size of
   * {@code Integer.MAX_VALUE} characters, and a read buffer size of
   * <tt>8192</tt> characters.
   *
   * @param in source from which lines will be read
   */
  public LineReader(Reader in) {
    this(in, 80, Integer.MAX_VALUE, 8192);
  }

  /**
   * Constructs a new instance with the specified {@code Reader}.
   *
   * @param in source from which lines will be read
   * @param initialLineBufferSize initial size of the line buffer
   * @param maxLineBufferSize maximum size of the line buffer
   * @param readBufferSize size of buffer used for reading from {@code in}
   *
   * @throws IllegalArgumentException if {@code initialLineBufferSize} is
   *           negative, {@code maxLineBufferSize} is non-positive, or
   *           {@code readBufferSize} is non-positive
   */
  public LineReader(Reader in, int initialLineBufferSize, int maxLineBufferSize,
      int readBufferSize) {
    super();
    worker = new LineWorker(in, initialLineBufferSize, maxLineBufferSize, readBufferSize);
    workerStartNeeded = true;
    closePrerequisite = null;
    closed = false;
  }

  /**
   * Determines whether a line is ready to be read without blocking. In some
   * cases, data will be read from the underlying {@code Reader} to make the
   * determination.
   *
   * @return {@code true} if {@code readLine()} will not block; {@code false}
   *         otherwise
   *
   * @throws IOException if an I/O error occurs while making the determination
   * @throws InterruptedIOException if interrupted while making the
   *           determination
   */
  public boolean lineReady() throws IOException {
    if (isClosed()) throw newClosedException();
    startWorkerIfNeeded();
    try {
      return worker.hasLine();
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw newInterruptedIOException(e);
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
      throw e;
    }
  }

  /**
   * Reads a line from the underlying {@code Reader}, waiting if necessary
   * until a line is available. A line is terminated by a line feed character
   * ('\n'), a carriage return character ('\r'), a carriage return character
   * followed by a line feed character ("\r\n"), or the end of the stream.
   *
   * @return the line read without the line terminator, or {@code null} if
   *         reached the end of the stream
   *
   * @throws IOException if an I/O error occurs while reading the line
   * @throws InterruptedIOException if interrupted while reading or waiting to
   *           read
   */
  public String readLine() throws IOException {
    if (isClosed()) throw newClosedException();
    startWorkerIfNeeded();
    try {
      return worker.takeLine();
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw newInterruptedIOException(e);
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
      throw e;
    }
  }

  /**
   * Reads a line from the underlying {@code Reader}. A line is terminated by a
   * line feed character ('\n'), a carriage return character ('\r'), a carriage
   * return character followed by a line feed character ("\r\n"), or the end of
   * the stream.
   *
   * @param timeout milliseconds to wait for line to be read; a timeout of 0 or
   *          less does not wait at all
   *
   * @return line read without the line terminator, or {@code null} if reached
   *         the end of the stream
   *
   * @throws IOException if an I/O error occurs while reading the line
   * @throws TimeoutException if the line could not be read within the
   *           specified timeout
   * @throws InterruptedIOException if interrupted while reading or waiting to
   *           read
   */
  public String readLine(long timeout) throws IOException, TimeoutException {
    if (isClosed()) throw newClosedException();
    startWorkerIfNeeded();
    try {
      return worker.takeLine(timeout);
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw newInterruptedIOException(e);
    } catch (InterruptedIOException e) {
      Thread.currentThread().interrupt();
      throw e;
    }
  }

  /**
   * Closes this stream and releases any system resources associated with it.
   * Once closed, further {@code lineRead()} or {@code readLine()} invocations
   * will throw an {@code IOException}. Closing a previously closed stream has
   * no effect.
   *
   * @throws IOException if an I/O error occurs
   * @throws InterruptedIOException if interrupted while closing
   *
   * @see #closeReaderOnClose()
   */
  @Override
  public void close() throws IOException {
    if (closed) return;
    workerStartNeeded = false;
    if (closePrerequisite != null) {
      try {
        closePrerequisite.call();
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        throw newInterruptedIOException(e);
      } catch (InterruptedIOException e) {
        Thread.currentThread().interrupt();
        throw e;
      } catch (IOException e) {
        throw e;
      } catch (Exception e) {
        throw new IOException(e);
      }
    }
    worker.interrupt();
    try {
      worker.join();
    } catch (InterruptedException e) {
      Thread.currentThread().interrupt();
      throw newInterruptedIOException(e);
    }
    worker.close();
    closed = true;
  }

  /**
   * Answers whether this stream is open.
   *
   * @return {@code true} if this stream is open; {@code false} otherwise
   */
  public boolean isOpen() {
    return !isClosed();
  }

  private boolean isClosed() {
    return closed;
  }

  /**
   * Gets the close prerequisite that is called as the first step in closing
   * this stream. The default is {@code null}.
   *
   * @return close prerequisite or {@code null} if none
   */
  public Callable<?> closePrerequisite() {
    return closePrerequisite;
  }

  /**
   * Sets the close prerequisite that is called as the first step in closing
   * this stream. The default is {@code null}.
   *
   * @param prerequisite called as the first step in closing this stream or
   *          {@code null} for none
   */
  public void closePrerequisite(Callable<?> prerequisite) {
    closePrerequisite = prerequisite;
  }

  /**
   * Answers whether the {@code Reader} should be closed when this stream is
   * closed. If {@code true}, the {@code Reader} will be closed when this
   * stream's {@code close()} method is invoked. If {@code false}, the
   * {@code Reader} will not be closed. The default is {@code true}.
   *
   * @return {@code true} if the {@code Reader} should be closed when this
   *         stream is closed; {@code false} otherwise
   *
   * @see #closeReaderOnClose(boolean)
   */
  public boolean closeReaderOnClose() {
    return worker.closeReaderOnClose();
  }

  /**
   * Sets whether the {@code Reader} should be closed when this stream is
   * closed. If {@code true}, the {@code Reader} will be closed when this
   * stream's {@code close()} method is invoked. If {@code false}, the
   * {@code Reader} will not be closed. The default is {@code true}.
   *
   * @param should whether the {@code Reader} should be closed when this stream
   *          is closed
   *
   * @see #closeReaderOnClose()
   */
  public void closeReaderOnClose(boolean should) {
    worker.closeReaderOnClose(should);
  }

  private void startWorkerIfNeeded() {
    if (!workerStartNeeded) return;
    worker.start();
    workerStartNeeded = false;
  }

  private static class LineWorker extends Thread implements Closeable {
    private static final String END_SENTINEL = "End Sentinel";
    private static final String EXCEPTION_SENTINEL = "Exception Sentinel";

    private Reader in;
    private StringBuilder lineBuffer;
    private final int lineBufferMaxSize;
    private boolean lineBufferContainsData;
    private boolean lineBufferContainsLine;
    private boolean lineBufferTerminatorIsCarriageReturn;
    private char[] readBuffer;
    private int readBufferOffset;
    private int readBufferLength;
    private boolean haveReadToEnd;
    private boolean closed;
    private boolean closeReaderOnClose;

    private boolean atEnd;
    private volatile Throwable exception;
    private final ArrayBlockingQueue<String> queue;

    public LineWorker(Reader in, int initialLineBufferSize, int maxLineBufferSize,
        int readBufferSize) {
      super("LineWorker");
      if (in == null) {
        throw new IllegalArgumentException("Null \"in\"");
      }
      if (initialLineBufferSize < 0) {
        throw new IllegalArgumentException("Negative \"initialLineBufferSize\"");
      }
      if (maxLineBufferSize <= 0) {
        throw new IllegalArgumentException("Non-positive \"maxLineBufferSize\"");
      }
      if (readBufferSize <= 0) {
        throw new IllegalArgumentException("Non-positive \"readBufferSize\"");
      }
      this.in = in;
      lineBuffer = new StringBuilder(initialLineBufferSize);
      lineBufferMaxSize = maxLineBufferSize;
      lineBufferContainsData = false;
      lineBufferContainsLine = false;
      lineBufferTerminatorIsCarriageReturn = false;
      readBuffer = new char[readBufferSize];
      readBufferOffset = 0;
      readBufferLength = 0;
      haveReadToEnd = false;
      closed = false;
      closeReaderOnClose = true;

      atEnd = false;
      exception = null;
      queue = new ArrayBlockingQueue<String>(1);
    }

    @Override
    public void run() {
      try {
        for (;;) {
          String line = readLine();
          if (line == null) {
            queue.put(END_SENTINEL);
            break;
          } else {
            queue.put(line);
          }
        }
      } catch (Throwable e) {
        exception = e;
        if (queue.isEmpty()) {
          try {
            queue.put(EXCEPTION_SENTINEL);
          } catch (Throwable ex) {
            /* IGNORED */
          }
        }
      }
    }

    public boolean hasLine() throws IOException, InterruptedException {
      if (closed) throw newClosedException();
      if (atEnd) return true;
      if (exception != null) rethrow(exception);
      String result = queue.peek();
      if (exception != null) rethrow(exception);
      assert result != EXCEPTION_SENTINEL;
      return result != null;
    }

    public String takeLine() throws IOException, InterruptedException {
      if (closed) throw newClosedException();
      if (atEnd) return null;
      if (exception != null) rethrow(exception);
      String result = queue.take();
      if (exception != null) rethrow(exception);
      assert result != EXCEPTION_SENTINEL;
      if (result == END_SENTINEL) {
        result = null;
        atEnd = true;
      }
      return result;
    }

    /**
     * @param timeout milliseconds
     */
    public String takeLine(long timeout) throws IOException, InterruptedException,
        TimeoutException {
      if (closed) throw newClosedException();
      if (atEnd) return null;
      if (exception != null) rethrow(exception);
      String result = queue.poll(timeout, TimeUnit.MILLISECONDS);
      if (exception != null) rethrow(exception);
      assert result != EXCEPTION_SENTINEL;
      if (result == END_SENTINEL) {
        atEnd = true;
        return null;
      } else if (result != null) {
        return result;
      } else {
        throw new TimeoutException();
      }
    }

    @Override
    public void close() throws IOException {
      if (closed) return;
      if (closeReaderOnClose) in.close();
      in = null;
      lineBuffer = null;
      readBuffer = null;
      closed = true;
    }

    public boolean closeReaderOnClose() {
      return closeReaderOnClose;
    }

    public void closeReaderOnClose(boolean should) {
      closeReaderOnClose = should;
    }

    private String readLine() throws IOException {
      if (lineBufferContainsLine || (lineBufferContainsData && haveReadToEnd)) {
        String result = lineBuffer.toString();
        lineBuffer.setLength(0);
        lineBufferContainsData = false;
        lineBufferContainsLine = false;
        return result;
      } else if (haveReadToEnd) {
        return null;
      } else {
        basicReadLine();
        return readLine();
      }
    }

    private void basicReadLine() throws IOException {
      if (haveReadToEnd) return;
      for (;;) {
        consumeReadBufferStartingLineFeedOfCarriageReturnLineFeedPair();
        if (readBufferContainsLineTerminator()) {
          moveReadBufferLineToLineBuffer();
          return;
        }
        if (readBufferContainsData()) {
          moveReadBufferToLineBuffer();
        }
        int numCharsRead = readIntoReadBuffer();
        if (numCharsRead == -1) return;
      }
    }

    private void consumeReadBufferStartingLineFeedOfCarriageReturnLineFeedPair() {
      if (!lineBufferTerminatorIsCarriageReturn) return;
      if (readBufferLength <= 0) return;
      if (readBuffer[readBufferOffset] != '\n') return;
      lineBufferTerminatorIsCarriageReturn = false;
      readBufferOffset++;
      readBufferLength--;
    }

    private boolean readBufferContainsLineTerminator() {
      return containsLineTerminator(readBuffer, readBufferOffset, readBufferLength);
    }

    private boolean readBufferContainsData() {
      return readBufferLength != 0;
    }

    private void moveReadBufferToLineBuffer() throws IOException {
      if ((long)lineBuffer.length() + (long)readBufferLength > (long)lineBufferMaxSize) {
        throw new IOException("Line buffer would exceed max size");
      }
      lineBuffer.append(readBuffer, readBufferOffset, readBufferLength);
      lineBufferContainsData = true;
      lineBufferContainsLine = false;

      readBufferOffset = 0;
      readBufferLength = 0;
    }

    private void moveReadBufferLineToLineBuffer() throws IOException {
      int terminatorIndex = findLineTerminator(readBuffer, readBufferOffset, readBufferLength);
      if (terminatorIndex == -1) return;

      int lineLength = terminatorIndex - readBufferOffset;
      if ((long)lineBuffer.length() + (long)lineLength > (long)lineBufferMaxSize) {
        throw new IOException("Line buffer would exceed max size");
      }
      lineBuffer.append(readBuffer, readBufferOffset, lineLength);
      lineBufferContainsData = true;
      lineBufferContainsLine = true;
      lineBufferTerminatorIsCarriageReturn = readBuffer[terminatorIndex] == '\r';

      int lineAndTerminatorLength = lineLength + 1;
      readBufferOffset += lineAndTerminatorLength;
      readBufferLength -= lineAndTerminatorLength;
    }

    private int readIntoReadBuffer() throws IOException {
      int numCharsRead = in.read(readBuffer);
      readBufferOffset = 0;
      if (numCharsRead == -1) {
        readBufferLength = 0;
        haveReadToEnd = true;
      } else {
        readBufferLength = numCharsRead;
      }
      return numCharsRead;
    }

    private static boolean containsLineTerminator(char[] buffer, int offset, int length) {
      return findLineTerminator(buffer, offset, length) != -1;
    }

    private static int findLineTerminator(char[] buffer, int offset, int length) {
      int maxIndex = add(offset, length) - 1;
      for (int i = offset; i <= maxIndex; i++) {
        if (buffer[i] == '\n' || buffer[i] == '\r') return i;
      }
      return -1;
    }

    private static int add(int a, int b) {
      long result = (long)a + (long)b;
      if (result <= (long)Integer.MAX_VALUE) return (int)result;
      throw new ArithmeticException("Integer overflow");
    }

    private static void rethrow(Throwable e) throws IOException, InterruptedException {
      if (e instanceof InterruptedException) {
        throw (InterruptedException)e;
      } else if (e instanceof IOException) {
        throw (IOException)e;
      } else {
        throw new IOException(e);
      }
    }
  }

  private static InterruptedIOException newInterruptedIOException(Throwable cause) {
    try {
      InterruptedIOException result = new InterruptedIOException();
      result.initCause(cause);
      return result;
    } catch (IllegalStateException e) {
      return new InterruptedIOException(cause.toString());
    }
  }

  private static IOException newClosedException() {
    return new IOException("Closed");
  }
}
