package imcacat.imcadr;

import gov.aps.jca.TimeoutException;
import imcacat.epicsca4j.ca.EpicsCaException;
import imcacat.epicsca4j.convert.ShortToBooleanEpicsCaValueConverter;
import imcacat.epicsca4j.convert.ShortToEnumEpicsCaValueConverter;
import imcacat.epicsca4j.model.EpicsPv;
import imcacat.imcadr.util.MathUtils;
import imcacat.imcadr.util.PhysicsUtils;

/**
 * I am a driver for a DECTRIS PILATUS detector.
 */
public class PilatusDetectorDriver extends ExternalSoftwareTriggerDetectorDriver {
  /* Readout time of 6M. */
  private static final double READOUT_TIME = 2.3e-3;
  /*
   * This timeout should be greater than the timeout in the areaDetector
   * PILATUS driver to avoid aborting the acquisition at roughly the same time
   * that the PILATUS driver would abort in order to avoid any bugs in the
   * PILATUS driver or underlying DECTRIS software (e.g., camserver) that
   * might be triggered by doing so.  The CAMSERVER_ACQUIRE_TIMEOUT macro in
   * areaDetector's pilatusDetector.cpp is set to 10 seconds, so set this to
   * 5 seconds more than that.
   */
  private static final double EXPOSURE_TIMEOUT = 15.0;

  protected static final String ARMED_PV_NAME = "Armed";
  protected static final String THRESHOLD_AUTO_APPLY_PV_NAME = "ThresholdAutoApply";
  protected static final String THRESHOLD_GAIN_PV_NAME = "GainMenu";
  protected static final String THRESHOLD_GAIN_RBV_PV_NAME = "GainMenu_RBV";
  protected static final String THRESHOLD_ENERGY_PV_NAME = "ThresholdEnergy";
  protected static final String THRESHOLD_ENERGY_RBV_PV_NAME = "ThresholdEnergy_RBV";
  protected static final String THRESHOLD_APPLY_PV_NAME = "ThresholdApply";
  protected static final String DETECTOR_DISTANCE_PV_NAME = "DetDist";
  protected static final String DETECTOR_2THETA_PV_NAME = "Det2theta";
  protected static final String WAVELENGTH_PV_NAME = "Wavelength";
  protected static final String FILTER_TRANSM_PV_NAME = "FilterTransm";
  protected static final String START_ANGLE_PV_NAME = "StartAngle";
  protected static final String ANGLE_INCREMENT_PV_NAME = "AngleIncr";
  protected static final String OMEGA_PV_NAME = "Omega";
  protected static final String OMEGA_INCREMENT_PV_NAME = "OmegaIncr";

  protected static final long ARM_DISARM_TIMEOUT = 5000;
  protected static final long THRESHOLD_CHANGE_TIMEOUT = 110000;
  protected static final double INCOMING_ENERGY_TO_THRESHOLD_SCALE = 0.61;
  protected static final double THRESHOLD_ENERGY_MINIMUM = 4.187;
  protected static final double THRESHOLD_ENERGY_DEADBAND =
      0.3 * INCOMING_ENERGY_TO_THRESHOLD_SCALE;
  protected static final int THRESHOLD_ENERGY_DECIMAL_PLACES = 3;

  private volatile boolean released = false;
  protected EpicsPv armedPv;
  protected EpicsPv thresholdAutoApplyPv;
  protected EpicsPv thresholdGainPv;
  protected EpicsPv thresholdGainRbvPv;
  protected EpicsPv thresholdEnergyPv;
  protected EpicsPv thresholdEnergyRbvPv;
  protected EpicsPv thresholdApplyPv;
  protected EpicsPv detectorDistancePv;
  protected EpicsPv detector2thetaPv;
  protected EpicsPv wavelengthPv;
  protected EpicsPv filterTransmPv;
  protected EpicsPv startAnglePv;
  protected EpicsPv angleIncrementPv;
  protected EpicsPv omegaPv;
  protected EpicsPv omegaIncrementPv;

  public PilatusDetectorDriver(String pvName) {
    super(pvName);
    armedPv =
        new EpicsPv(pvName + ARMED_PV_NAME, false, true,
            new ShortToBooleanEpicsCaValueConverter());
    thresholdAutoApplyPv =
        new EpicsPv(pvName + THRESHOLD_AUTO_APPLY_PV_NAME, false, false,
            new ShortToBooleanEpicsCaValueConverter());
    thresholdGainPv =
        new EpicsPv(pvName + THRESHOLD_GAIN_PV_NAME, false, false,
            new ShortToEnumEpicsCaValueConverter(PilatusThresholdGain.class));
    thresholdGainRbvPv =
        new EpicsPv(pvName + THRESHOLD_GAIN_RBV_PV_NAME, false, false,
            new ShortToEnumEpicsCaValueConverter(PilatusThresholdGain.class));
    thresholdEnergyPv = new EpicsPv(pvName + THRESHOLD_ENERGY_PV_NAME, false, false);
    thresholdEnergyRbvPv = new EpicsPv(pvName + THRESHOLD_ENERGY_RBV_PV_NAME, false, false);
    thresholdApplyPv =
        new EpicsPv(pvName + THRESHOLD_APPLY_PV_NAME, false, false,
            new ShortToBooleanEpicsCaValueConverter());
    detectorDistancePv = new EpicsPv(pvName + DETECTOR_DISTANCE_PV_NAME, false, false);
    detector2thetaPv = new EpicsPv(pvName + DETECTOR_2THETA_PV_NAME, false, false);
    wavelengthPv = new EpicsPv(pvName + WAVELENGTH_PV_NAME, false, false);
    filterTransmPv = new EpicsPv(pvName + FILTER_TRANSM_PV_NAME, false, false);
    startAnglePv = new EpicsPv(pvName + START_ANGLE_PV_NAME, false, false);
    angleIncrementPv = new EpicsPv(pvName + ANGLE_INCREMENT_PV_NAME, false, false);
    omegaPv = new EpicsPv(pvName + OMEGA_PV_NAME, false, false);
    omegaIncrementPv = new EpicsPv(pvName + OMEGA_INCREMENT_PV_NAME, false, false);
  }

  @Override
  protected final void finalize() throws Throwable {
    try {
      release();
    } finally {
      super.finalize();
    }
  }

  public void release() {
    if (released) return;
    released = true;

    try {
      if (armedPv != null) armedPv.release();
      if (thresholdAutoApplyPv != null) thresholdAutoApplyPv.release();
      if (thresholdGainPv != null) thresholdGainPv.release();
      if (thresholdGainRbvPv != null) thresholdGainRbvPv.release();
      if (thresholdEnergyPv != null) thresholdEnergyPv.release();
      if (thresholdEnergyRbvPv != null) thresholdEnergyRbvPv.release();
      if (thresholdApplyPv != null) thresholdApplyPv.release();
      if (detectorDistancePv != null) detectorDistancePv.release();
      if (detector2thetaPv != null) detector2thetaPv.release();
      if (wavelengthPv != null) wavelengthPv.release();
      if (filterTransmPv != null) filterTransmPv.release();
      if (startAnglePv != null) startAnglePv.release();
      if (angleIncrementPv != null) angleIncrementPv.release();
      if (omegaPv != null) omegaPv.release();
      if (omegaIncrementPv != null) omegaIncrementPv.release();
    } finally {
      super.release();
    }
  }

  @Override
  public double getReadoutTime(boolean binningEnabled) {
    return READOUT_TIME;
  }

  @Override
  public double getExposureTimeout() {
    return EXPOSURE_TIMEOUT;
  }

  @Override
  public void startExposing() throws InterruptedException, DriverException {
    waitToBeDisarmed();
    super.startExposing();
    waitToBeArmed();
  }

  public void lastImage() throws InterruptedException, DriverException {
  }

  public void setDetector2theta(double position) throws InterruptedException, DriverException {
    try {
      detector2thetaPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setDetectorDistance(double position) throws InterruptedException, DriverException {
    try {
      detectorDistancePv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setImageWidth(double width) throws InterruptedException, DriverException {
    try {
      angleIncrementPv.setValueAndWait(width);
      omegaIncrementPv.setValueAndWait(width);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setPhi(double position) throws InterruptedException, DriverException {
    try {
      startAnglePv.setValueAndWait(position);
      omegaPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  public void setRotationAxis(CrystalRotationAxis axis) throws InterruptedException,
      DriverException {
  }

  public void setWavelength(double position) throws InterruptedException, DriverException {
    /*
     * WORKAROUND: Must set threshold (energy and gain) *before* any other MX
     * parameters (wavelength in this case) because the Pilatus control server
     * (camserver) resets MX parameters when the threshold energy or gain are
     * changed.
     */
    setThresholdFor(PhysicsUtils.convertToEnergy(position) / 1e3);
    basicSetWavelength(position);
  }

  protected void basicSetWavelength(double position) throws InterruptedException,
      DriverException {
    try {
      wavelengthPv.setValueAndWait(position);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void setThresholdFor(double incomingEnergyInKeV) throws InterruptedException,
      DriverException {
    setThresholdAutoApply(false);
    boolean shouldApply = false;

    double thresholdEnergyRbv = getThresholdEnergy();
    double thresholdEnergy = incomingEnergyInKeV * INCOMING_ENERGY_TO_THRESHOLD_SCALE;
    if (thresholdEnergy < THRESHOLD_ENERGY_MINIMUM) {
      thresholdEnergy = THRESHOLD_ENERGY_MINIMUM;
    }
    setThresholdEnergy(thresholdEnergy);
    double thresholdEnergyDelta = Math.abs(thresholdEnergy - thresholdEnergyRbv);
    if (thresholdEnergyDelta > THRESHOLD_ENERGY_DEADBAND) shouldApply = true;

    PilatusThresholdGain gainRbv = getThresholdGain();
    PilatusThresholdGain gain;
    if (incomingEnergyInKeV < 9.0) {
      gain = PilatusThresholdGain.HIGH;
    } else if (incomingEnergyInKeV < 12.3) {
      gain = PilatusThresholdGain.MEDIUM;
    } else {
      gain = PilatusThresholdGain.LOW;
    }
    setThresholdGain(gain);
    if (!gain.equals(gainRbv)) shouldApply = true;

    if (shouldApply) applyThreshold();
  }

  protected void setThresholdAutoApply(boolean autoApply) throws InterruptedException,
      DriverException {
    try {
      thresholdAutoApplyPv.setValueAndWait(autoApply);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected PilatusThresholdGain getThresholdGain() throws InterruptedException, DriverException {
    try {
      return (PilatusThresholdGain)thresholdGainRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void setThresholdGain(PilatusThresholdGain gain) throws InterruptedException,
      DriverException {
    try {
      thresholdGainPv.setValueAndWait(gain);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  /**
   * Returns threshold energy in KeV.
   */
  protected double getThresholdEnergy() throws InterruptedException, DriverException {
    try {
      return (Double)thresholdEnergyRbvPv.getValue();
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void setThresholdEnergy(double thresholdEnergyInKeV) throws InterruptedException,
      DriverException {
    try {
      thresholdEnergyPv.setValueAndWait(MathUtils.round(thresholdEnergyInKeV,
          THRESHOLD_ENERGY_DECIMAL_PLACES));
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected void applyThreshold() throws InterruptedException, DriverException {
    try {
      thresholdApplyPv.setValueAndWait(Boolean.TRUE, THRESHOLD_CHANGE_TIMEOUT);
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    } catch (TimeoutException e) {
      throw new DriverException(e);
    }
  }

  public void setAttenuation(double attenuation) throws InterruptedException, DriverException {
    try {
      filterTransmPv.setValueAndWait(1.0 - (attenuation / 100.0));
    } catch (EpicsCaException e) {
      throw new DriverException(e);
    }
  }

  protected boolean waitToBeArmed() throws InterruptedException, DriverException {
    return waitToBeArmed(ARM_DISARM_TIMEOUT);
  }

  protected boolean waitToBeArmed(long timeout) throws InterruptedException, DriverException {
    return waitForArmedToBe(Boolean.TRUE, timeout);
  }

  protected boolean waitToBeDisarmed() throws InterruptedException, DriverException {
    return waitToBeDisarmed(ARM_DISARM_TIMEOUT);
  }

  protected boolean waitToBeDisarmed(long timeout) throws InterruptedException, DriverException {
    return waitForArmedToBe(Boolean.FALSE, timeout);
  }

  protected boolean waitForArmedToBe(Boolean armed, long timeout) throws InterruptedException,
      DriverException {
    try {
      armedPv.waitForValue(armed, timeout);
      return true;
    } catch (EpicsCaException e) {
      throw new DriverException("Failed to wait for detector to " + (armed ? "arm" : "disarm"));
    } catch (TimeoutException e) {
      return false;
    }
  }
}
