package imcacat.imcadr;

/**
 * I represent a position in two-dimensional space.
 */
public class Position2d {
  protected double x;
  protected double y;
  protected LengthUnit unit;

  public Position2d() {
    this(LengthUnit.ARBITRARY);
  }

  public Position2d(LengthUnit unit) {
    this(0.0, 0.0, unit);
  }

  public Position2d(double x, double y) {
    this(x, y, LengthUnit.ARBITRARY);
  }

  public Position2d(double x, double y, LengthUnit unit) {
    this.x = x;
    this.y = y;
    this.unit = unit;
  }

  /**
   * Determines whether I am equal to the specified <tt>Object</tt> instance.
   * <p>
   * Does not attempt to convert to the same units; that is, if my unit is
   * millimeters and the unit of the <tt>Object</tt> instance is micrometers,
   * we are not considered equal even if our values would be equal after a
   * conversion to a common unit.
   */
  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    if (!(obj instanceof Position2d)) return false;
    Position2d pos = (Position2d)obj;
    if (x != pos.x) return false;
    if (y != pos.y) return false;
    return unit.equals(pos.unit);
  }

  public boolean equals(Position2d pos, Position2d delta) {
    if (pos == null) return false;
    Position2d deltaInSameUnit = delta.convertTo(unit);
    if (Math.abs(x - pos.x) > deltaInSameUnit.getX()) return false;
    if (Math.abs(y - pos.y) > deltaInSameUnit.getY()) return false;
    return true;
  }

  @Override
  public int hashCode() {
    return (int)Double.doubleToLongBits(x) ^ (int)Double.doubleToLongBits(y) ^ unit.hashCode();
  }

  public double getX() {
    return x;
  }

  public double getY() {
    return y;
  }

  public LengthUnit getUnit() {
    return unit;
  }

  @Override
  public String toString() {
    return String.format("(%.3f, %.3f) %s", x, y, unit);
  }

  public String toDisplayString() {
    return toDisplayString(1);
  }

  public String toDisplayString(int precision) {
    String unitAbbreviation = unit.getAbbreviation();
    if (unitAbbreviation.length() != 0) unitAbbreviation = " " + unitAbbreviation;
    String xySpecifier = "%." + precision + "f";
    return String.format(xySpecifier + "%s, " + xySpecifier + "%s", x, unitAbbreviation, y,
        unitAbbreviation);
  }

  public Position2d convertTo(LengthUnit newUnit) {
    if (unit.equals(newUnit)) return this;
    Double ratioObject = LengthUnit.conversionRatio(unit, newUnit);
    if (ratioObject == null) {
      throw new RuntimeException("Do not know how to convert from " + unit + " to " + newUnit);
    }
    double ratio = ratioObject.doubleValue();
    return new Position2d(x * ratio, y * ratio, newUnit);
  }

  public static Position2d mm(double x, double y) {
    return new Position2d(x, y, LengthUnit.MILLIMETER);
  }

  public static Position2d um(double x, double y) {
    return new Position2d(x, y, LengthUnit.MICROMETER);
  }
}
