package imcacat.imcadr;

/**
 * I represent a position in three-dimensional space.
 */
public class Position3d extends Position2d {
  protected double z;

  public Position3d() {
    this(LengthUnit.ARBITRARY);
  }

  public Position3d(LengthUnit unit) {
    this(0.0, 0.0, 0.0, unit);
  }

  public Position3d(double x, double y, double z) {
    this(x, y, z, LengthUnit.ARBITRARY);
  }

  public Position3d(double x, double y, double z, LengthUnit unit) {
    super(x, y, unit);
    this.z = z;
  }

  /**
   * Determines whether I am equal to the specified <tt>Object</tt> instance.
   * <p>
   * Does not attempt to convert to the same units; that is, if my unit is
   * millimeters and the unit of the <tt>Object</tt> instance is micrometers,
   * we are not considered equal even if our values would be equal after a
   * conversion to a common unit.
   */
  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    if (!(obj instanceof Position3d)) return false;
    Position3d pos = (Position3d)obj;
    if (x != pos.x) return false;
    if (y != pos.y) return false;
    if (z != pos.z) return false;
    return unit.equals(pos.unit);
  }

  public boolean equals(Position3d pos, Position3d delta) {
    if (pos == null) return false;
    Position3d deltaInSameUnit = delta.convertTo(unit);
    if (Math.abs(x - pos.x) > deltaInSameUnit.getX()) return false;
    if (Math.abs(y - pos.y) > deltaInSameUnit.getY()) return false;
    if (Math.abs(z - pos.z) > deltaInSameUnit.getZ()) return false;
    return true;
  }

  @Override
  public int hashCode() {
    return (int)Double.doubleToLongBits(x) ^ (int)Double.doubleToLongBits(y) ^
        (int)Double.doubleToLongBits(z) ^ unit.hashCode();
  }

  public double getZ() {
    return z;
  }

  @Override
  public String toString() {
    return String.format("(%.3f, %.3f, %.3f) %s", x, y, z, unit);
  }

  @Override
  public String toDisplayString(int precision) {
    String unitAbbreviation = unit.getAbbreviation();
    if (unitAbbreviation.length() != 0) unitAbbreviation = " " + unitAbbreviation;
    String xyzSpecifier = "%." + precision + "f";
    return String.format(xyzSpecifier + "%s, " + xyzSpecifier + "%s, " + xyzSpecifier + "%s", x,
        unitAbbreviation, y, unitAbbreviation, z, unitAbbreviation);
  }

  @Override
  public Position3d convertTo(LengthUnit newUnit) {
    if (unit.equals(newUnit)) return this;
    Double ratioObject = LengthUnit.conversionRatio(unit, newUnit);
    if (ratioObject == null) {
      throw new RuntimeException("Do not know how to convert from " + unit + " to " + newUnit);
    }
    double ratio = ratioObject.doubleValue();
    return new Position3d(x * ratio, y * ratio, z * ratio, newUnit);
  }

  public static Position3d mm(double x, double y, double z) {
    return new Position3d(x, y, z, LengthUnit.MILLIMETER);
  }

  public static Position3d um(double x, double y, double z) {
    return new Position3d(x, y, z, LengthUnit.MICROMETER);
  }
}
