package imcacat.imcadr;

import imcacat.imcadr.util.MathUtils;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Objects;

/**
 * I model a two-dimensional raster.
 */
public class Raster2d extends EdtAwareModel {
  private static final long serialVersionUID = 1;

  public static final String PROPNAME_ORIGIN = "origin";

  protected Raster2dSpec spec;
  protected RasterCell[][] cells;
  protected int lowScore;
  protected int highScore;
  protected Position2d origin;
  protected Raster2dMask mask;

  public Raster2d(Raster2dSpec spec, String imageFileNameExtension,
      boolean firePropertyChangesOnEdtEnabled) {
    super(firePropertyChangesOnEdtEnabled);

    if (!LengthUnit.MILLIMETER.equals(spec.getCellDimensionUnit()) &&
        !LengthUnit.MICROMETER.equals(spec.getCellDimensionUnit())) {
      throw new IllegalArgumentException("Unsupported cell dimension unit from Raster2dSpec: " +
          spec.getCellDimensionUnit());
    }
    int numColumns = spec.getWidth();
    int numRows = spec.getHeight();
    if (numColumns < 1 || numRows < 1) throw new IllegalArgumentException("Minimum size is 1x1");
    this.spec = spec;
    cells = new RasterCell[numRows][numColumns];
    for (int c = 0; c < cells[0].length; c++)
      for (int r = 0; r < cells.length; r++)
        cells[r][c] =
            new RasterCell(c, r, new Position2d(spec.getCellDimensionUnit()), newImageFile(
                spec.getImagesDirectory(), imageFileNameExtension, spec.getPlane(), c, r),
                firePropertyChangesOnEdtEnabled);
    int minimumScore = getScoreMinimumRangeLow();
    lowScore = minimumScore;
    highScore = getScoreMinimumRangeHigh();
    origin = computeOrigin(cells[0][0].getPosition(), spec.getCellSize());
    mask = Objects.requireNonNull(spec.getMask());
    if (mask.createsGap()) {
      throw new IllegalArgumentException("Mask creates a gap");
    }

    cells[0][0].addPropertyChangeListener(RasterCell.PROPNAME_POSITION,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            setOrigin(computeOrigin((Position2d)event.getNewValue()));
          }
        });
  }

  public Raster2dSpec getSpec() {
    return spec;
  }

  public Raster2dPlane getPlane() {
    return spec.getPlane();
  }

  public boolean isClipped() {
    return spec.getPreClipWidth() != null || spec.getPreClipHeight() != null;
  }

  public RasterCell getCell(Raster2dIndex index) {
    return getCell(index.getColumn(), index.getRow());
  }

  public RasterCell getCell(int column, int row) {
    return cells[row][column];
  }

  public File getImageFilePath(Raster2dIndex index) {
    return getCell(index).getImageFilePath();
  }

  public File getImageFilePath(int column, int row) {
    return getCell(column, row).getImageFilePath();
  }

  public double getCellHeight() {
    return spec.getCellHeight();
  }

  public double getCellWidth() {
    return spec.getCellWidth();
  }

  public LengthUnit getCellDimensionUnit() {
    return spec.getCellDimensionUnit();
  }

  public int columnCount() {
    return cells[0].length;
  }

  public int rowCount() {
    return cells.length;
  }

  public int cellCount() {
    return cellCount(null);
  }

  public int cellCount(Raster2dEdge edge) {
    if (edge == null) return columnCount() * rowCount();
    int cellCount = 0;
    if (edge.involvesTop() || edge.involvesBottom()) cellCount = columnCount();
    if (edge.involvesLeft() || edge.involvesRight()) cellCount +=
        ((cellCount == 0) ? 0 : -1) + rowCount();
    return cellCount;
  }

  protected int unmaskedUnscoredCellCount(Raster2dEdge edge, Raster2dIndex edgeIndex) {
    int count = 0;
    for (Iterator<?> i = unmaskedUnscoredIterator(edge, edgeIndex); i.hasNext(); i.next())
      count++;
    return count;
  }

  public Position2d getOrigin() {
    return origin;
  }

  protected void setOrigin(Position2d newOrigin) {
    if (origin == null && newOrigin == null) return;
    if (origin != null && origin.equals(newOrigin)) return;
    Position2d oldOrigin = origin;
    origin = newOrigin;
    notifyChanged(PROPNAME_ORIGIN, oldOrigin, newOrigin);
  }

  public boolean isMasked(Raster2dIndex index) {
    return isMasked(index.getColumn(), index.getRow());
  }

  public boolean isMasked(int column, int row) {
    return mask.get(column, row);
  }

  /**
   * Assigns positions of cells such that the raster is centered about the
   * specified center position.
   */
  public void assignPositionsCenteredOn(double unscaledCenterHorizPos,
      double unscaledCenterVertPos) {
    Raster2dPositionCalculator calc = new Raster2dPositionCalculator(spec);
    calc.center(unscaledCenterHorizPos, unscaledCenterVertPos);
    assignPositions(calc.unscaledTopLeftHorizPos(), calc.unscaledTopLeftVertPos());
  }

  /**
   * Assigns positions of cells such that the top left cell of the raster has
   * the specified position.
   */
  public void assignPositionsRelativeToCell(double unscaledTopLeftCellHorizPos,
      double unscaledTopLeftCellVertPos) {
    Raster2dPositionCalculator calc = new Raster2dPositionCalculator(spec);
    calc.topLeftCell(unscaledTopLeftCellHorizPos, unscaledTopLeftCellVertPos);
    assignPositions(calc.unscaledTopLeftHorizPos(), calc.unscaledTopLeftVertPos());
  }

  /**
   * Assigns positions of cells such that the top left corner of the raster is
   * at the specified position. Note: The specified position is <i>not</i> the
   * center of the top left cell.
   */
  protected void assignPositions(double unscaledTopLeftHorizPos, double unscaledTopLeftVertPos) {
    assignPositions(unscaledTopLeftHorizPos, unscaledTopLeftVertPos, 0, 0, columnCount(),
        rowCount());
  }

  /**
   * Assigns positions of cells, within the specified start and end range, such
   * that the top left corner of the raster is at the specified position. Note:
   * The specified position is <i>not</i> the center of the top left cell.
   */
  protected void assignPositions(double unscaledTopLeftHorizPos, double unscaledTopLeftVertPos,
      int startColumn, int startRow, int endColumn, int endRow) {
    Raster2dPositionCalculator calc = new Raster2dPositionCalculator(spec);
    calc.topLeft(unscaledTopLeftHorizPos, unscaledTopLeftVertPos);
    long horizPosStartOffset = startColumn * calc.cellWidth();
    long vertPosStartOffset = startRow * calc.cellHeight();

    long horizPos = calc.topLeftCellHorizPos() + horizPosStartOffset;
    for (int c = startColumn; c < endColumn; c++) {
      long vertPos = calc.topLeftCellVertPos() + vertPosStartOffset;
      for (int r = startRow; r < endRow; r++) {
        setPosition(c, r, horizPos / calc.scale(), vertPos / calc.scale());
        vertPos += calc.cellHeight();
      }
      horizPos += calc.cellWidth();
    }
  }

  /**
   * Assigns positions of cells along the specified edge such that the top left
   * corner of the raster is at the specified position. Note: The specified
   * position is <i>not</i> the center of the top left cell.
   */
  protected void assignPositionsAlong(double unscaledTopLeftHorizPos,
      double unscaledTopLeftVertPos, Raster2dEdge edge) {
    if (edge.involvesTop()) assignPositions(unscaledTopLeftHorizPos, unscaledTopLeftVertPos, 0,
        0, columnCount(), 1);
    else if (edge.involvesBottom()) assignPositions(unscaledTopLeftHorizPos,
        unscaledTopLeftVertPos, 0, rowCount() - 1, columnCount(), rowCount());

    if (edge.involvesLeft()) assignPositions(unscaledTopLeftHorizPos, unscaledTopLeftVertPos, 0,
        0, 1, rowCount());
    else if (edge.involvesRight()) assignPositions(unscaledTopLeftHorizPos,
        unscaledTopLeftVertPos, columnCount() - 1, 0, columnCount(), rowCount());
  }

  public void setPosition(int column, int row, double horizontalPosition, double verticalPosition) {
    getCell(column, row).setPosition(horizontalPosition, verticalPosition);
  }

  public synchronized void loadResult(int column, int row, ImageAnalysisResult result) {
    RasterCell cell = getCell(column, row);
    Integer score = result.getScore();
    cell.setScore(score);
    updateLowOrHighScoreIfNeeded(score);
    cell.setNormalizedScore(normalizeScore(score));
    cell.setSpotTotal(result.getSpotTotal());
    cell.setInResolutionTotal(result.getInResolutionTotal());
    cell.setGoodBraggTotal(result.getGoodBraggTotal());
    cell.setIceRingTotal(result.getIceRingTotal());
    cell.setEstimatedResolution1(result.getEstimatedResolution1());
    cell.setEstimatedResolution2(result.getEstimatedResolution2());
    cell.setTotalIntegratedSignal(result.getTotalIntegratedSignal());
    cell.setMosaicity(result.getMosaicity());
    cell.setSpaceGroup(result.getSpaceGroup());
    cell.setUnitCell(result.getUnitCell());
  }

  private synchronized void updateLowOrHighScoreIfNeeded(Integer score) {
    if (score == null) return;
    if (score < getLowScore()) {
      setLowScore(score);
    } else if (score > getHighScore()) {
      setHighScore(score);
    }
  }

  protected synchronized int getLowScore() {
    return lowScore;
  }

  protected synchronized void setLowScore(int newLowScore) {
    lowScore = newLowScore;
    updateScores();
  }

  protected synchronized int getHighScore() {
    return highScore;
  }

  protected synchronized void setHighScore(int newHighScore) {
    highScore = newHighScore;
    updateScores();
  }

  protected synchronized void updateScores() {
    for (int r = 0; r < cells.length; r++) {
      for (int c = 0; c < cells[0].length; c++) {
        cells[r][c].setNormalizedScore(normalizeScore(cells[r][c].getScore()));
      }
    }
  }

  protected Double normalizeScore(Integer score) {
    if (score == null) return null;
    return MathUtils.scale(score, getLowScore(), getHighScore(),
        RasterCell.MIN_NORMALIZED_SCORE, RasterCell.MAX_NORMALIZED_SCORE);
  }

  public Raster2dIterator iterator() {
    return new Raster2dAllIterator(this);
  }

  public Raster2dIterator iterator(Raster2dEdge edge) {
    return (edge == null) ? iterator() : new Raster2dEdgeIterator(this, edge);
  }

  public Raster2dIterator unmaskedUnscoredIterator() {
    return new Raster2dUnscoredIterator(new Raster2dUnmaskedIterator(this));
  }

  public Raster2dIterator unmaskedUnscoredIterator(Raster2dEdge edge, Raster2dIndex edgeIndex) {
    return (edge == null) ? unmaskedUnscoredIterator() : new Raster2dUnscoredIterator(
        new Raster2dUnmaskedIterator(this, new Raster2dEdgeIterator(this, edge, edgeIndex)));
  }

  public Iterator<Raster2dFlyScan> unmaskedUnscoredFlyScanIterator(Raster2dEdge edge,
      Raster2dIndex edgeIndex) {
    if (edge == null) {
      double rowMajorWeight;
      double colMajorWeight;
      int numCols = spec.getWidth();
      int numRows = spec.getHeight();
      if (numCols == numRows) {
        /* Minimize the fly-scanned distance. */
        double cellWidth = spec.getCellWidth();
        double cellHeight = spec.getCellHeight();
        /* Overflows to infinity which is OK. */
        double rowWidth = (double)numCols * cellWidth;
        /* Overflows to infinity which is OK. */
        double colHeight = (double)numRows * cellHeight;
        rowMajorWeight = colHeight;
        colMajorWeight = rowWidth;
      } else {
        /* Minimize the number of fly scans. */
        rowMajorWeight = numCols;
        colMajorWeight = numRows;
      }
      rowMajorWeight *= getAxisFlyScanWeight(spec.getPlane(), Direction2d.HORIZONTAL);
      colMajorWeight *= getAxisFlyScanWeight(spec.getPlane(), Direction2d.VERTICAL);
      return (rowMajorWeight >= colMajorWeight) ? new Raster2dRowFlyScanIterator(this)
          : new Raster2dColumnFlyScanIterator(this);
    } else {
      return new Raster2dEdgeFlyScanIterator(this, edge, edgeIndex);
    }
  }

  public Position2d computeBestPosition() {
    return computeAveragePosition(computeBestPositionIndex());
  }

  protected Position2d computeAveragePosition(Raster2dIndex index) {
    if (!shouldAveragePositions(index)) return getCell(index).getPosition();
    java.awt.Rectangle region = adjacentsRegion(index);
    int maxRegionX = region.x + region.width - 1;
    int maxRegionY = region.y + region.height - 1;
    double normalizedScoreSum = computeNormalizedScoreSum(region);
    if (normalizedScoreSum == 0.0) return getCell(index).getPosition();
    BigDecimal avgX = BigDecimal.ZERO;
    BigDecimal avgY = BigDecimal.ZERO;
    for (int c = region.x; c <= maxRegionX; c++) {
      for (int r = region.y; r <= maxRegionY; r++) {
        Double normalizedScore = cells[r][c].getNormalizedScore();
        if (normalizedScore == null) continue;
        double weight = normalizedScore.doubleValue() / normalizedScoreSum;
        avgX = avgX.add(BigDecimal.valueOf(cells[r][c].getHorizontalPosition() * weight));
        avgY = avgY.add(BigDecimal.valueOf(cells[r][c].getVerticalPosition() * weight));
      }
    }
    return new Position2d(avgX.doubleValue(), avgY.doubleValue(), spec.getCellDimensionUnit());
  }

  protected boolean shouldAveragePositions(Raster2dIndex index) {
    if (isOnEdge(index)) return false;
    if (isOnUnmaskedEdge(index)) return false;
    int columnCount = columnCount();
    int rowCount = rowCount();
    if (columnCount == 1 && rowCount >= 3) return true;
    if (rowCount == 1 && columnCount >= 3) return true;
    if (columnCount >= 3 && rowCount >= 3) return true;
    return false;
  }

  protected java.awt.Rectangle adjacentsRegion(Raster2dIndex index) {
    java.awt.Rectangle region =
        new java.awt.Rectangle(index.getColumn() - 1, index.getRow() - 1, 3, 3);
    return region.intersection(new java.awt.Rectangle(0, 0, columnCount(), rowCount()));
  }

  protected double computeNormalizedScoreSum(java.awt.Rectangle region) {
    int maxRegionX = region.x + region.width - 1;
    int maxRegionY = region.y + region.height - 1;
    BigDecimal sum = BigDecimal.ZERO;
    for (int c = region.x; c <= maxRegionX; c++) {
      for (int r = region.y; r <= maxRegionY; r++) {
        Double each = cells[r][c].getNormalizedScore();
        if (each != null) sum = sum.add(BigDecimal.valueOf(each.doubleValue()));
      }
    }
    return sum.doubleValue();
  }

  protected RasterCell getCellWithBestScore() {
    Raster2dIndex index = computeBestPositionIndex();
    return getCell(index.getColumn(), index.getRow());
  }

  public Raster2dIndex computeBestPositionIndex() {
    int bestCol = -1;
    int bestRow = -1;
    for (int c = 0; c < cells[0].length; c++) {
      for (int r = 0; r < cells.length; r++) {
        if (mask.get(c, r)) continue;
        if (bestCol == -1) {
          bestCol = c;
          bestRow = r;
        }
        Integer score = cells[r][c].getScore();
        Integer bestScore = cells[bestRow][bestCol].getScore();
        if (score == null) continue;
        /* @formatter:off */
        if (bestScore == null ||
            score > bestScore ||
            (score == bestScore &&
              ((isOnEdge(c, r) && !isOnEdge(bestCol, bestRow)) ||
               (isOnUnmaskedEdge(c, r) && !isOnUnmaskedEdge(bestCol, bestRow))))) {
          bestCol = c;
          bestRow = r;
        }
        /* @formatter:on */
      }
    }
    if (bestCol == -1) {
      bestCol = 0;
      bestRow = 0;
    }
    return new Raster2dIndex(bestCol, bestRow);
  }

  public Integer getScoreAtBestPosition() {
    Raster2dIndex best = computeBestPositionIndex();
    return cells[best.getRow()][best.getColumn()].getScore();
  }

  public boolean isOnEdge(Raster2dIndex index) {
    return isOnEdge(index.getColumn(), index.getRow());
  }

  public boolean isOnEdge(int column, int row) {
    final int numCols = columnCount();
    final int numRows = rowCount();

    if (numCols > 1) {
      if (column == 0) return true;
      if (column == numCols - 1) return true;
    }
    if (numRows > 1) {
      if (row == 0) return true;
      if (row == numRows - 1) return true;
    }
    if (numCols == 1 && numRows == 1) return true;
    return false;
  }

  public Raster2dEdge edge(Raster2dIndex index) {
    final int numCols = columnCount();
    final int numRows = rowCount();

    if (numCols == 1 && numRows == 1) return Raster2dEdge.ALL;

    Raster2dEdge result = null;
    if (index.getColumn() == 0) {
      if (index.getRow() == 0) result = Raster2dEdge.TOP_LEFT;
      else if (index.getRow() == numRows - 1) result = Raster2dEdge.BOTTOM_LEFT;
      else result = Raster2dEdge.LEFT;
    } else if (index.getColumn() == numCols - 1) {
      if (index.getRow() == 0) result = Raster2dEdge.TOP_RIGHT;
      else if (index.getRow() == numRows - 1) result = Raster2dEdge.BOTTOM_RIGHT;
      else result = Raster2dEdge.RIGHT;
    } else {
      if (index.getRow() == 0) result = Raster2dEdge.TOP;
      else if (index.getRow() == numRows - 1) result = Raster2dEdge.BOTTOM;
    }
    if (result == null) return result;
    if (numCols <= 1) result = result.minusLeftAndRight();
    if (result == null) return result;
    if (numRows <= 1) result = result.minusTopAndBottom();
    return result;
  }

  public boolean isOnUnmaskedEdge(Raster2dIndex index) {
    return unmaskedEdge(index) != null;
  }

  public boolean isOnUnmaskedEdge(int column, int row) {
    return isOnUnmaskedEdge(new Raster2dIndex(column, row));
  }

  /*
   * The unmasked edge is the edge on which the cell at the specified index
   * resides, taking into account the raster mask. The idea is to determine the
   * edge like the edge(Raster2dIndex) method, but treat the masked out region
   * as if it were the edge of the raster rather than the actual edge of the
   * raster.
   *
   * We inspect four cells relative to the specified cell: above, to the right,
   * below, and to the left. These cells, along with the specified cell's
   * location in the raster, determine the unmasked edge.
   */
  public Raster2dEdge unmaskedEdge(Raster2dIndex index) {
    final int numCols = columnCount();
    final int numRows = rowCount();
    final int col = index.getColumn();
    final int row = index.getRow();

    if (isMasked(col, row)) return null;

    EnumSet<Raster2dEdge> edges = EnumSet.noneOf(Raster2dEdge.class);

    int c;
    int r;

    /* Top */
    c = col;
    r = row - 1;
    if (r < 0 || isMasked(c, r)) {
      edges.add(Raster2dEdge.TOP);
    }

    /* Right */
    c = col + 1;
    r = row;
    if (c > (numCols - 1) || isMasked(c, r)) {
      edges.add(Raster2dEdge.RIGHT);
    }

    /* Bottom */
    c = col;
    r = row + 1;
    if (r > (numRows - 1) || isMasked(c, r)) {
      edges.add(Raster2dEdge.BOTTOM);
    }

    /* Left */
    c = col - 1;
    r = row;
    if (c < 0 || isMasked(c, r)) {
      edges.add(Raster2dEdge.LEFT);
    }

    switch (edges.size()) {
      case 0:
        return null;
      case 1:
        return edges.iterator().next();
      case 2:
        if (edges.contains(Raster2dEdge.TOP) && edges.contains(Raster2dEdge.RIGHT)) {
          return Raster2dEdge.TOP_RIGHT;
        } else if (edges.contains(Raster2dEdge.TOP) && edges.contains(Raster2dEdge.LEFT)) {
          return Raster2dEdge.TOP_LEFT;
        } else if (edges.contains(Raster2dEdge.BOTTOM) && edges.contains(Raster2dEdge.RIGHT)) {
          return Raster2dEdge.BOTTOM_RIGHT;
        } else if (edges.contains(Raster2dEdge.BOTTOM) && edges.contains(Raster2dEdge.LEFT)) {
          return Raster2dEdge.BOTTOM_LEFT;
        } else {
          /* TOP and BOTTOM, or RIGHT and LEFT */
          return null;
        }
      case 3:
        if (edges.contains(Raster2dEdge.TOP) && edges.contains(Raster2dEdge.RIGHT) &&
            edges.contains(Raster2dEdge.BOTTOM)) {
          return Raster2dEdge.RIGHT;
        } else if (edges.contains(Raster2dEdge.TOP) && edges.contains(Raster2dEdge.LEFT) &&
            edges.contains(Raster2dEdge.BOTTOM)) {
          return Raster2dEdge.LEFT;
        } else if (edges.contains(Raster2dEdge.LEFT) && edges.contains(Raster2dEdge.TOP) &&
            edges.contains(Raster2dEdge.RIGHT)) {
          return Raster2dEdge.TOP;
        } else if (edges.contains(Raster2dEdge.LEFT) && edges.contains(Raster2dEdge.BOTTOM) &&
            edges.contains(Raster2dEdge.RIGHT)) {
          return Raster2dEdge.BOTTOM;
        } else {
          throw new BugException();
        }
      case 4:
        return Raster2dEdge.ALL;
      default:
        throw new BugException();
    }
  }

  public boolean hasMaskedOrthogonalNeighbor(Raster2dIndex index, Raster2dEdge edge) {
    for (Raster2dIndex each : orthogonalNeighborIndices(index, edge)) {
      if (isMasked(each)) return true;
    }
    return false;
  }

  public void unmaskOrthogonalNeighbors(Raster2dIndex index, Raster2dEdge edge) {
    for (Raster2dIndex each : orthogonalNeighborIndices(index, edge)) {
      mask.set(each, false);
    }
  }

  private List<Raster2dIndex> orthogonalNeighborIndices(Raster2dIndex index, Raster2dEdge edge) {
    final int col = index.getColumn();
    final int row = index.getRow();

    /* @formatter:off */
    int[][] indexCandidates;
    switch (edge) {
      case ALL:
        indexCandidates = new int[][] {
            {col - 1, row - 1},
            {col,     row - 1},
            {col + 1, row - 1},
            {col - 1, row},
            {col + 1, row},
            {col - 1, row + 1},
            {col,     row + 1},
            {col + 1, row + 1}
        };
        break;
      case TOP:
      case BOTTOM:
        indexCandidates = new int[][] {
            {col - 1, row},
            {col + 1, row}
        };
        break;
      case TOP_RIGHT:
      case BOTTOM_LEFT:
        indexCandidates = new int[][] {
            {col - 1, row - 1},
            {col + 1, row + 1}
        };
        break;
      case RIGHT:
      case LEFT:
        indexCandidates = new int[][] {
            {col, row - 1},
            {col, row + 1}
        };
        break;
      case BOTTOM_RIGHT:
      case TOP_LEFT:
        indexCandidates = new int[][] {
            {col + 1, row - 1},
            {col - 1, row + 1}
        };
        break;
      default:
        throw new BugException();
    }
    /* @formatter:on */

    final int colMin = Math.max(0, col - 1);
    final int colMax = Math.min(columnCount() - 1, col + 1);
    final int rowMin = Math.max(0, row - 1);
    final int rowMax = Math.min(rowCount() - 1, row + 1);

    List<Raster2dIndex> result = new ArrayList<>();
    for (int[] each : indexCandidates) {
      if (each[0] < colMin || each[0] > colMax) continue;
      if (each[1] < rowMin || each[1] > rowMax) continue;
      result.add(new Raster2dIndex(each[0], each[1]));
    }

    return result;
  }

  public synchronized void shiftToward(Raster2dEdge edge) throws FileDeleteException,
      FileRenameException {
    int maxColumn = columnCount() - 1;
    int maxRow = rowCount() - 1;

    Raster2dPositionCalculator calc = new Raster2dPositionCalculator(spec);
    Position2d topLeftCellPos = getCell(0, 0).getPosition();
    calc.topLeftCell(topLeftCellPos.getX(), topLeftCellPos.getY());
    Position2d shiftedTopLeftPos =
        computePositionShiftedToward(
            new Position2d(calc.unscaledTopLeftHorizPos(), calc.unscaledTopLeftVertPos(),
                topLeftCellPos.getUnit()), edge);

    if (edge.involvesTop()) {
      for (int c = 0; c <= maxColumn; c++)
        for (int r = maxRow; r > 0; r--)
          moveCell(c, r - 1, c, r);
      for (int c = 0; c <= maxColumn; c++)
        resetCell(c, 0);
    } else if (edge.involvesBottom()) {
      for (int c = 0; c <= maxColumn; c++)
        for (int r = 0; r < maxRow; r++)
          moveCell(c, r + 1, c, r);
      for (int c = 0; c <= maxColumn; c++)
        resetCell(c, maxRow);
    }

    if (edge.involvesLeft()) {
      for (int c = maxColumn; c > 0; c--)
        for (int r = 0; r <= maxRow; r++)
          moveCell(c - 1, r, c, r);
      for (int r = 0; r <= maxRow; r++)
        resetCell(0, r);
    } else if (edge.involvesRight()) {
      for (int c = 0; c < maxColumn; c++)
        for (int r = 0; r <= maxRow; r++)
          moveCell(c + 1, r, c, r);
      for (int r = 0; r <= maxRow; r++)
        resetCell(maxColumn, r);
    }

    assignPositionsAlong(shiftedTopLeftPos.getX(), shiftedTopLeftPos.getY(), edge);
  }

  protected Position2d computePositionShiftedToward(Position2d position, Raster2dEdge edge) {
    Raster2dPositionCalculator calc = new Raster2dPositionCalculator(spec);
    calc.topLeftCell(position.getX(), position.getY());
    calc.shiftToward(edge);
    return new Position2d(calc.unscaledTopLeftCellHorizPos(), calc.unscaledTopLeftCellVertPos(),
        position.getUnit());
  }

  protected synchronized void moveCell(int fromColumn, int fromRow, int toColumn, int toRow)
      throws FileDeleteException, FileRenameException {
    RasterCell from = cells[fromRow][fromColumn];
    RasterCell to = cells[toRow][toColumn];
    to.setScore(from.getScore());
    to.setNormalizedScore(from.getNormalizedScore());
    to.setSpotTotal(from.getSpotTotal());
    to.setInResolutionTotal(from.getInResolutionTotal());
    to.setGoodBraggTotal(from.getGoodBraggTotal());
    to.setIceRingTotal(from.getIceRingTotal());
    to.setEstimatedResolution1(from.getEstimatedResolution1());
    to.setEstimatedResolution2(from.getEstimatedResolution2());
    to.setTotalIntegratedSignal(from.getTotalIntegratedSignal());
    to.setMosaicity(from.getMosaicity());
    to.setSpaceGroup(from.getSpaceGroup());
    to.setUnitCell(from.getUnitCell());
    to.setPosition(from.getPosition());
    renameCellImageFile(from.getImageFilePath(), to.getImageFilePath());
  }

  protected void renameCellImageFile(File from, File to) throws FileDeleteException,
      FileRenameException {
    boolean ok;
    if (to.exists()) {
      try {
        ok = to.delete();
      } catch (SecurityException e) {
        throw new FileDeleteException(to, e);
      }
      if (!ok) throw new FileDeleteException(to);
    }
    if (from.exists()) {
      try {
        ok = from.renameTo(to);
      } catch (SecurityException e) {
        throw new FileRenameException(from, to, e);
      }
      if (!ok) throw new FileRenameException(from, to);
    }
  }

  protected synchronized void resetCell(int column, int row) throws FileDeleteException {
    RasterCell cell = cells[row][column];
    cell.reset();
    File imageFilePath = cell.getImageFilePath();
    if (imageFilePath.exists()) {
      try {
        boolean ok = imageFilePath.delete();
        if (!ok) throw new FileDeleteException(imageFilePath);
      } catch (SecurityException e) {
        throw new FileDeleteException(imageFilePath, e);
      }
    }
  }

  protected int getScoreMinimumRangeLow() {
    return getScoreMinimumRange().low();
  }

  protected int getScoreMinimumRangeHigh() {
    return getScoreMinimumRange().high();
  }

  protected Range<Integer> getScoreMinimumRange() {
    return spec.getAnalysisSpec().getAnalysisPackage().getScoreMinimumRange();
  }

  protected double getAxisFlyScanWeight(Raster2dPlane plane, Direction2d direction) {
    switch (getAxis(plane, direction)) {
      case X:
        return Resources.getInstance().getXMotorFlyScanWeight();
      case Y:
        return Resources.getInstance().getYMotorFlyScanWeight();
      case Z:
        return Resources.getInstance().getZMotorFlyScanWeight();
      default:
        throw new BugException();
    }
  }

  protected Position2d computeOrigin(Position2d topLeftCellCenter) {
    return computeOrigin(topLeftCellCenter, spec.getCellSize());
  }

  protected Position2d computeOrigin(Position2d topLeftCellCenter, Rectangle cellSize) {
    return new Position2d(topLeftCellCenter.getX() - (cellSize.getWidth() / 2.0),
        topLeftCellCenter.getY() - (cellSize.getHeight() / 2.0), topLeftCellCenter.getUnit());
  }

  protected static BeamlineAxis getAxis(Raster2dPlane plane, Direction2d direction) {
    if (direction.equals(Direction2d.HORIZONTAL)) return BeamlineAxis.Z;
    if (plane.equals(Raster2dPlane.ZY)) return BeamlineAxis.Y;
    if (plane.equals(Raster2dPlane.ZX)) return BeamlineAxis.X;
    throw new BugException();
  }

  protected static File newImageFile(File imagesDirectory, String imageFileNameExtension,
      Raster2dPlane plane, int column, int row) {
    return new File(imagesDirectory, imageFileName(imageFileNameExtension, plane, column, row));
  }

  protected static String imageFileName(String imageFileNameExtension, Raster2dPlane plane,
      int column, int row) {
    return String.format(Locale.US, "cell_%s_%02d_%02d%s", imageFileNamePlaneId(plane),
        column + 1, row + 1, imageFileNameExtension);
  }

  public static String imageFileNamePlaneId(Raster2dPlane plane) {
    if (Raster2dPlane.ZX.equals(plane)) return "zx";
    if (Raster2dPlane.ZY.equals(plane)) return "zy";
    throw new BugException();
  }

  public static String bestImageFileName(String imageFileNameExtension, Raster2dPlane plane) {
    return String.format(Locale.US, "cell_%s_best%s", imageFileNamePlaneId(plane),
        imageFileNameExtension);
  }
}
