package imcacat.imcadr;

import imcacat.imcadr.util.MathUtils;

import java.util.NoSuchElementException;

/**
 * I am an iterator that iterates over all cells in a {@link Raster2d}.
 * <p>
 * I iterate in row-major or column-major order according to whichever order
 * minimizes the total distance moved: if the row width in cell dimension units
 * is less than or equal to the column height in cell dimension units, then the
 * order will be row-major; otherwise it will be column-major.
 *
 */
public class Raster2dAllIterator implements Raster2dIterator {
  private final Raster2d raster;
  private final int minCol;
  private final int maxCol;
  private final int minRow;
  private final int maxRow;
  private final boolean rowMajor;
  private int col;
  private int row;
  private boolean atEnd;

  public Raster2dAllIterator(Raster2d raster) {
    this.raster = raster;
    minCol = 0;
    maxCol = raster.columnCount() - 1;
    minRow = 0;
    maxRow = raster.rowCount() - 1;
    int numCols = maxCol - minCol + 1;
    int numRows = maxRow - minRow + 1;
    double cellWidth = raster.getCellWidth();
    double cellHeight = raster.getCellHeight();
    double rowWidth = (double)numCols * cellWidth;
    if (rowWidth == Double.POSITIVE_INFINITY) {
      throw new IllegalArgumentException("row width in cell dimension units overflowed");
    }
    double colHeight = (double)numRows * cellHeight;
    if (colHeight == Double.POSITIVE_INFINITY) {
      throw new IllegalArgumentException("column height in cell dimension units overflowed");
    }
    rowMajor = rowWidth <= colHeight;
    col = minCol;
    row = minRow;
    atEnd = false;
  }

  @Override
  public boolean hasNext() {
    return !atEnd;
  }

  @Override
  public RasterCell next() {
    if (atEnd) throw new NoSuchElementException();
    RasterCell result = raster.getCell(col, row);
    step();
    return result;
  }

  @Override
  public void remove() {
    throw new UnsupportedOperationException();
  }

  @Override
  public int nextColumn() {
    if (atEnd) throw new NoSuchElementException();
    return col;
  }

  @Override
  public int nextRow() {
    if (atEnd) throw new NoSuchElementException();
    return row;
  }

  @Override
  public Raster2dIndex nextIndex() {
    if (atEnd) throw new NoSuchElementException();
    return new Raster2dIndex(col, row);
  }

  private void step() {
    if (rowMajor) {
      if (MathUtils.even(row)) {
        if (col == maxCol) {
          row++;
          if (row > maxRow) atEnd = true;
        } else {
          col++;
        }
      } else {
        if (col == minCol) {
          row++;
          if (row > maxRow) atEnd = true;
        } else {
          col--;
        }
      }
    } else {
      if (MathUtils.even(col)) {
        if (row == maxRow) {
          col++;
          if (col > maxCol) atEnd = true;
        } else {
          row++;
        }
      } else {
        if (row == minRow) {
          col++;
          if (col > maxCol) atEnd = true;
        } else {
          row--;
        }
      }
    }
  }
}
