package imcacat.imcadr;

import java.util.NoSuchElementException;
import java.util.Objects;

/**
 * I am an iterator for iterating along an edge of a {@link Raster2d}.
 */
public class Raster2dEdgeIterator implements Raster2dIterator {
  private final Raster2d raster;
  private final Raster2dEdge edge;
  private final int minCol;
  private final int maxCol;
  private final int minRow;
  private final int maxRow;
  private int col;
  private int row;
  private Raster2dEdge allIterationEdge;
  private boolean atEnd;

  /**
   * Creates an iterator with the specified raster and edge.
   *
   * @param raster raster to iterate over
   * @param edge outer edge of raster to iterator over
   *
   * @throws NullPointerException if {@code raster} or {@code edge} are
   *           {@code null}
   */
  public Raster2dEdgeIterator(Raster2d raster, Raster2dEdge edge) {
    this(raster, edge, null);
  }

  /**
   * Creates an iterator with the specified raster, edge, and edge index.
   *
   * @param raster raster to iterate over
   * @param edge edge of raster to iterator over
   * @param edgeIndex index in {@code raster} that {@code edge} goes through,
   *          and if {@code edge} is a composite edge, then the index is such
   *          that all component edges go through it, or {@code null} to
   *          indicate the outer edge of the raster
   *
   * @throws NullPointerException if {@code raster} or {@code edge} are
   *           {@code null}
   */
  public Raster2dEdgeIterator(Raster2d raster, Raster2dEdge edge, Raster2dIndex edgeIndex) {
    this.raster = Objects.requireNonNull(raster);
    this.edge = Objects.requireNonNull(edge);

    allIterationEdge = null;

    switch (edge) {
      case ALL:
        minCol = (edgeIndex != null) ? edgeIndex.getColumn() : 0;
        maxCol = (edgeIndex != null) ? edgeIndex.getColumn() : raster.columnCount() - 1;
        minRow = (edgeIndex != null) ? edgeIndex.getRow() : 0;
        maxRow = (edgeIndex != null) ? edgeIndex.getRow() : raster.rowCount() - 1;
        col = minCol;
        row = minRow;
        allIterationEdge = Raster2dEdge.TOP;
        break;
      case TOP:
        minCol = 0;
        maxCol = raster.columnCount() - 1;
        minRow = (edgeIndex != null) ? edgeIndex.getRow() : 0;
        maxRow = minRow;
        col = minCol;
        row = minRow;
        break;
      case TOP_RIGHT:
        minCol = 0;
        maxCol = (edgeIndex != null) ? edgeIndex.getColumn() : raster.columnCount() - 1;
        minRow = (edgeIndex != null) ? edgeIndex.getRow() : 0;
        maxRow = raster.rowCount() - 1;
        col = minCol;
        row = minRow;
        break;
      case RIGHT:
        minCol = (edgeIndex != null) ? edgeIndex.getColumn() : raster.columnCount() - 1;
        maxCol = minCol;
        minRow = 0;
        maxRow = raster.rowCount() - 1;
        col = maxCol;
        row = minRow;
        break;
      case BOTTOM_RIGHT:
        minCol = 0;
        maxCol = (edgeIndex != null) ? edgeIndex.getColumn() : raster.columnCount() - 1;
        minRow = 0;
        maxRow = (edgeIndex != null) ? edgeIndex.getRow() : raster.rowCount() - 1;
        col = maxCol;
        row = minRow;
        break;
      case BOTTOM:
        minCol = 0;
        maxCol = raster.columnCount() - 1;
        minRow = (edgeIndex != null) ? edgeIndex.getRow() : raster.rowCount() - 1;
        maxRow = minRow;
        col = maxCol;
        row = maxRow;
        break;
      case BOTTOM_LEFT:
        minCol = (edgeIndex != null) ? edgeIndex.getColumn() : 0;
        maxCol = raster.columnCount() - 1;
        minRow = 0;
        maxRow = (edgeIndex != null) ? edgeIndex.getRow() : raster.rowCount() - 1;
        col = maxCol;
        row = maxRow;
        break;
      case LEFT:
        minCol = (edgeIndex != null) ? edgeIndex.getColumn() : 0;
        maxCol = minCol;
        minRow = 0;
        maxRow = raster.rowCount() - 1;
        col = minCol;
        row = maxRow;
        break;
      case TOP_LEFT:
        minCol = (edgeIndex != null) ? edgeIndex.getColumn() : 0;
        maxCol = raster.columnCount() - 1;
        minRow = (edgeIndex != null) ? edgeIndex.getRow() : 0;
        maxRow = raster.rowCount() - 1;
        col = minCol;
        row = maxRow;
        break;
      default:
        throw new BugException();
    }

    atEnd = false;
  }

  @Override
  public boolean hasNext() {
    return !atEnd;
  }

  @Override
  public RasterCell next() {
    if (atEnd) throw new NoSuchElementException();
    RasterCell result = raster.getCell(col, row);
    step();
    return result;
  }

  @Override
  public void remove() {
    throw new UnsupportedOperationException();
  }

  @Override
  public int nextColumn() {
    if (atEnd) throw new NoSuchElementException();
    return col;
  }

  @Override
  public int nextRow() {
    if (atEnd) throw new NoSuchElementException();
    return row;
  }

  @Override
  public Raster2dIndex nextIndex() {
    if (atEnd) throw new NoSuchElementException();
    return new Raster2dIndex(col, row);
  }

  private void step() {
    switch (edge) {
      case ALL:
        if (minCol == maxCol && minRow == maxRow) {
          /* There is only a single cell, and we were on it. */
          atEnd = true;
          break;
        }
        loop: for (;;) {
          switch (allIterationEdge) {
            case TOP:
              col++;
              if (col <= maxCol) break loop;
              if (minRow == maxRow) {
                atEnd = true;
                break loop;
              }
              col = maxCol;
              allIterationEdge = Raster2dEdge.RIGHT;
              break;
            case RIGHT:
              row++;
              if (row <= maxRow) break loop;
              if (minCol == maxCol) {
                atEnd = true;
                break loop;
              }
              row = maxRow;
              allIterationEdge = Raster2dEdge.BOTTOM;
              break;
            case BOTTOM:
              col--;
              if (col >= minCol) break loop;
              col = minCol;
              allIterationEdge = Raster2dEdge.LEFT;
              break;
            case LEFT:
              row--;
              if (row <= minRow) atEnd = true;
              break loop;
            default:
              throw new BugException();
          }
        }
        break;
      case TOP:
        col++;
        if (col > maxCol) atEnd = true;
        break;
      case TOP_RIGHT:
        if (col == maxCol) {
          row++;
          if (row > maxRow) atEnd = true;
        } else {
          col++;
        }
        break;
      case RIGHT:
        row++;
        if (row > maxRow) atEnd = true;
        break;
      case BOTTOM_RIGHT:
        if (row == maxRow) {
          col--;
          if (col < minCol) atEnd = true;
        } else {
          row++;
        }
        break;
      case BOTTOM:
        col--;
        if (col < minCol) atEnd = true;
        break;
      case BOTTOM_LEFT:
        if (col == minCol) {
          row--;
          if (row < minRow) atEnd = true;
        } else {
          col--;
        }
        break;
      case LEFT:
        row--;
        if (row < minRow) atEnd = true;
        break;
      case TOP_LEFT:
        if (row == minRow) {
          col++;
          if (col > maxCol) atEnd = true;
        } else {
          row--;
        }
        break;
      default:
        throw new BugException();
    }
  }
}
