package imcacat.imcadr;

/**
 * I perform calculations for a fly scan.
 * <p>
 * Position adjustments for shutter and detector latencies are placed in the
 * constant-velocity part of the move profile to avoid a dependence on the
 * shape of the acceleration and deceleration parts. For example, the
 * acceleration part might be linear or an S-curve. And if it's an S-curve, the
 * exact position function for it would need to be known. By placing the
 * adjustments in the constant-velocity part of the move profile, we don't need
 * to know what type of acceleration is used.
 */
public class Raster2dFlyScanCalculator {
  private double exposureTime;
  private double exposurePeriod;
  private double velocity;
  private double shutterOpenPos;
  private double shutterClosePos;
  private double detectorTriggerPos;
  private double startPos;
  private double endPos;
  private long cellCount;

  /**
   * Constructs an instance with the specified parameters.
   * <p>
   * The unit of length is arbitrary but must be the same across the
   * parameters; it is referred to as
   * <q>engineering units.</q>
   *
   * @param startCellCenterPos the position of the center of the start cell in
   *          engineering units
   * @param endCellCenterPos the position of the center of the end cell in
   *          engineering units
   * @param cellLength the length of a cell in engineering units
   * @param reqExposureTime the requested exposure time for each cell in
   *          seconds
   * @param minVelo the minimum velocity in engineering units per second
   * @param maxVelo the maximum velocity in engineering units per second; 0
   *          means no maximum velocity
   * @param accelTime the time to accelerate from rest to constant velocity in
   *          seconds
   * @param detectorReadoutTime the time it takes for the detector to read out
   *          a frame in seconds
   * @param shutterLatency the time it takes for the exposure shutter to
   *          actually open or close after being commanded to do so in seconds
   * @param detectorLatency the time it takes for the detector exposure to
   *          actually start after being commanded to do so in seconds
   * @param cvSettleTime the time to spend at constant velocity before the
   *          exposure starts, and again after the exposure ends before
   *          decelerating, in seconds
   *
   * @throws IllegalArgumentException if {@code cellLength} or
   *           {@code reqExposureTime} are nonpositive, if {@code minVelo},
   *           {@code maxVelo}, {@code accelTime}, {@code detectorReadoutTime},
   *           {@code shutterLatency}, {@code detectorLatency}, or
   *           {@code cvSettleTime} are negative, or if {@code minVelo} is
   *           greater than {@code maxVelo} when {@code maxVelo} is nonzero
   */
  public Raster2dFlyScanCalculator(double startCellCenterPos, double endCellCenterPos,
      double cellLength, double reqExposureTime, double minVelo, double maxVelo, double accelTime,
      double detectorReadoutTime, double shutterLatency, double detectorLatency,
      double cvSettleTime) {
    super();

    if (cellLength <= 0.0) {
      throw new IllegalArgumentException("Nonpositive cell length");
    }
    if (reqExposureTime <= 0.0) {
      throw new IllegalArgumentException("Nonpositive requested exposure time");
    }
    if (minVelo < 0.0) {
      throw new IllegalArgumentException("Negative minimum velocity");
    }
    if (maxVelo < 0.0) {
      throw new IllegalArgumentException("Negative maximum velocity");
    }
    if (maxVelo != 0.0 && minVelo > maxVelo) {
      throw new IllegalArgumentException("Minimum velocity greater than maximum velocity");
    }
    if (accelTime < 0.0) {
      throw new IllegalArgumentException("Negative acceleration time");
    }
    if (detectorReadoutTime < 0.0) {
      throw new IllegalArgumentException("Negative detector readout time");
    }
    if (shutterLatency < 0.0) {
      throw new IllegalArgumentException("Negative shutter latency");
    }
    if (detectorLatency < 0.0) {
      throw new IllegalArgumentException("Negative detector latency");
    }
    if (cvSettleTime < 0.0) {
      throw new IllegalArgumentException("Negative constant-velocity settle time");
    }

    /*
     * The displacement, x, equals the constant velocity times the change in
     * time:
     *
     *   x = vt
     *   v = x/t
     *   t = x/v
     */
    exposureTime = reqExposureTime;
    velocity = cellLength / exposureTime;
    if (velocity < minVelo) {
      velocity = minVelo;
      exposureTime = cellLength / velocity;
    } else if (maxVelo != 0.0 && velocity > maxVelo) {
      velocity = maxVelo;
      exposureTime = cellLength / velocity;
    }
    exposurePeriod = exposureTime;
    exposureTime = exposurePeriod - detectorReadoutTime;

    boolean positiveMove = startCellCenterPos <= endCellCenterPos;

    double halfCellLength = cellLength / 2.0;
    double offset = positiveMove ? halfCellLength : -halfCellLength;
    double zeroLatencyStartExposurePos = startCellCenterPos - offset;
    double zeroLatencyEndExposurePos = endCellCenterPos + offset;

    int cvPositionsIndex = 0;
    double cvStartPositions[] = new double[3];
    double cvEndPositions[] = new double[cvStartPositions.length];

    double shutterLatencyDisplacement = velocity * shutterLatency;
    offset = positiveMove ? shutterLatencyDisplacement : -shutterLatencyDisplacement;
    shutterOpenPos = zeroLatencyStartExposurePos - offset;
    shutterClosePos = zeroLatencyEndExposurePos - offset;
    cvStartPositions[cvPositionsIndex] = shutterOpenPos;
    cvEndPositions[cvPositionsIndex] = shutterClosePos;
    cvPositionsIndex++;

    double detectorLatencyDisplacement = velocity * detectorLatency;
    offset = positiveMove ? detectorLatencyDisplacement : -detectorLatencyDisplacement;
    detectorTriggerPos = zeroLatencyStartExposurePos - offset;
    cvStartPositions[cvPositionsIndex] = detectorTriggerPos;
    cvEndPositions[cvPositionsIndex] = zeroLatencyEndExposurePos;
    cvPositionsIndex++;

    double cvSettleTimeDisplacement = velocity * cvSettleTime;
    offset = positiveMove ? cvSettleTimeDisplacement : -cvSettleTimeDisplacement;
    double cvSettleStartPos = zeroLatencyStartExposurePos - offset;
    double cvSettleEndPos = zeroLatencyEndExposurePos + offset;
    cvStartPositions[cvPositionsIndex] = cvSettleStartPos;
    cvEndPositions[cvPositionsIndex] = cvSettleEndPos;

    double cvStartPos;
    double cvEndPos;
    if (positiveMove) {
      double minStartPos = cvStartPositions[0];
      double maxEndPos = cvEndPositions[0];
      for (int i = 1; i < cvStartPositions.length; i++) {
        if (cvStartPositions[i] < minStartPos) minStartPos = cvStartPositions[i];
        if (cvEndPositions[i] > maxEndPos) maxEndPos = cvEndPositions[i];
      }
      cvStartPos = minStartPos;
      cvEndPos = maxEndPos;
    } else {
      double maxStartPos = cvStartPositions[0];
      double minEndPos = cvEndPositions[0];
      for (int i = 1; i < cvStartPositions.length; i++) {
        if (cvStartPositions[i] > maxStartPos) maxStartPos = cvStartPositions[i];
        if (cvEndPositions[i] < minEndPos) minEndPos = cvEndPositions[i];
      }
      cvStartPos = maxStartPos;
      cvEndPos = minEndPos;
    }

    double accelerationDisplacement = velocity * accelTime;
    offset = positiveMove ? accelerationDisplacement : -accelerationDisplacement;
    startPos = cvStartPos - offset;
    endPos = cvEndPos + offset;

    double exposureDisplacement = zeroLatencyEndExposurePos - zeroLatencyStartExposurePos;
    cellCount = Math.round(Math.abs(exposureDisplacement) / cellLength);
  }

  /**
   * Gets the exposure time for each cell.
   *
   * @return the exposure time in seconds
   */
  public double exposureTime() {
    return exposureTime;
  }

  /**
   * Gets the exposure period for each cell.
   *
   * @return the exposure period in seconds
   */
  public double exposurePeriod() {
    return exposurePeriod;
  }

  /**
   * Gets the constant velocity at which the fly scan axis should be moved.
   *
   * @return the velocity in engineering units per second
   */
  public double velocity() {
    return velocity;
  }

  /**
   * Gets the position at which the axis should start for the fly scan move.
   * This position includes an offset for shutter and detector latency and for
   * acceleration to constant velocity.
   *
   * @return the start position in engineering units
   */
  public double startPosition() {
    return startPos;
  }

  /**
   * Gets the position at which the axis should end for the fly scan move. This
   * position includes an offset for shutter latency and for deceleration from
   * constant velocity.
   *
   * @return the end position in engineering units
   */
  public double endPosition() {
    return endPos;
  }

  /**
   * Gets the position at which the exposure shutter should be commanded to
   * open.
   *
   * @return the command-shutter-open position in engineering units
   */
  public double shutterOpenPosition() {
    return shutterOpenPos;
  }

  /**
   * Gets the position at which the exposure shutter should be commanded to
   * close.
   *
   * @return the command-shutter-close position in engineering units
   */
  public double shutterClosePosition() {
    return shutterClosePos;
  }

  /**
   * Gets the position at which the detector should be commanded to start the
   * exposure.
   *
   * @return the command-detector-start-exposure position in engineering units
   */
  public double detectorTriggerPosition() {
    return detectorTriggerPos;
  }

  /**
   * Gets the number of raster cells in the fly scan.
   *
   * @return the number of cells
   */
  public long cellCount() {
    return cellCount;
  }
}
