package imcacat.imcadr;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

/**
 * I am a binary mask for a {@link Raster2d}. When I am applied to a
 * {@code Raster2d}, the resulting polygon must be convex. My
 * {@link #makeConcave()} method can be called to ensure that I will always
 * produce a convex polygon when applied to a {@code Raster2d}.
 */
public class Raster2dMask {
  private final Map<Integer, Set<Integer>> mask;

  /**
   * Constructs a new instance with an empty backing mask.
   */
  public Raster2dMask() {
    this(new HashMap<Integer, Set<Integer>>());
  }

  /**
   * Constructs a new instance with the specified backing mask.
   *
   * @param mask backing mask of rows of columns; each key is a nonnegative
   *          {@code Raster2d} row index mapped to a row mask; each row mask is
   *          a nonnull nonempty set of nonnegative column indices marked as
   *          {@code true} ({@code 1}); the mask may be sparse (that is, an
   *          entry is not required for every row nor column); the caller must
   *          not modify {@code mask} nor any objects in it once this call is
   *          made
   *
   * @throws NullPointerException if {@code mask} is {@code null}
   */
  public Raster2dMask(Map<Integer, Set<Integer>> mask) {
    this.mask = Objects.requireNonNull(mask);
  }

  /**
   * Gets the value of the mask at the specified index.
   *
   * @param index index of the value to get
   *
   * @return value at the specified index or {@code false} if the backing mask
   *         does not contain a value at the specified index
   *
   * @throws IllegalArgumentException if {@code index} contains a column or row
   *           that is less than zero
   */
  public boolean get(Raster2dIndex index) {
    return get(index.getColumn(), index.getRow());
  }

  /**
   * Gets the value of the mask at the specified index.
   *
   * @param column column index of the value to get
   * @param row row index of the value to get
   *
   * @return value at the specified index or {@code false} if the backing mask
   *         does not contain a value at the specified index
   */
  public boolean get(int column, int row) {
    if (column < 0 || row < 0) {
      throw new IllegalArgumentException("Negative index");
    }
    Set<Integer> rowMask = mask.get(row);
    return (rowMask == null) ? false : rowMask.contains(column);
  }

  /**
   * Sets the value of the mask at the specified index.
   *
   * @param index index of the value to set
   * @param value value to set
   *
   * @throws IllegalArgumentException if {@code index} contains a column or row
   *           that is less than zero
   */
  public void set(Raster2dIndex index, boolean value) {
    set(index.getColumn(), index.getRow(), value);
  }

  /**
   * Sets the value of the mask at the specified index.
   *
   * @param column column index of the value to set
   * @param row row index of the value to set
   * @param value value to set
   *
   * @throws IllegalArgumentException if {@code column} or {@code row} are less
   *           than zero
   */
  public void set(int column, int row, boolean value) {
    if (column < 0 || row < 0) {
      throw new IllegalArgumentException("Negative index");
    }
    Set<Integer> rowMask = mask.get(row);
    if (value) {
      if (rowMask == null) {
        rowMask = new HashSet<Integer>();
        mask.put(row, rowMask);
      }
      rowMask.add(column);
    } else {
      if (rowMask == null) return;
      rowMask.remove(column);
      if (rowMask.isEmpty()) mask.remove(row);
    }
  }

  /**
   * Gets the index of the column farthest to the right that is set to
   * {@code true}.
   *
   * @return the maximum column index or {@code -1} if empty
   */
  public int maxColumn() {
    int max = -1;
    for (Integer row : mask.keySet()) {
      for (Integer col : mask.get(row)) {
        if (col > max) max = col;
      }
    }
    return max;
  }

  /**
   * Gets the index of the row farthest down that is set to {@code true}.
   *
   * @return the maximum row index or {@code -1} if empty
   */
  public int maxRow() {
    int max = -1;
    for (Integer row : mask.keySet()) {
      if (row > max) max = row;
    }
    return max;
  }

  /**
   * Gets the number of indices set to {@code true}.
   *
   * @return number of masked indices
   */
  public int count() {
    int result = 0;
    for (Integer row : mask.keySet()) {
      result += mask.get(row).size();
    }
    return result;
  }

  /**
   * Answers whether this mask creates a gap in a {@code Raster2d}. A gap is a
   * discontinuity (that is, a masked cell) in a sequence of unmasked column or
   * row cells.
   *
   * @return {@code true} if creates a gap; {@code false} otherwise
   */
  public boolean createsGap() {
    int maxCol = maxColumn();
    int maxRow = maxRow();
    /* Find gaps in unmasked rows */
    for (int r = 0; r <= maxRow; r++) {
      int startCol = findMinNonmaskedCol(r);
      if (startCol < 0) continue;
      int endCol = findMaxNonmaskedCol(r);
      for (int c = startCol; c <= endCol; c++) {
        if (get(c, r)) return true;
      }
    }
    /* Find gaps in unmasked columns */
    for (int c = 0; c <= maxCol; c++) {
      int startRow = findMinNonmaskedRow(c);
      if (startRow < 0) continue;
      int endRow = findMaxNonmaskedRow(c);
      for (int r = startRow; r <= endRow; r++) {
        if (get(c, r)) return true;
      }
    }
    return false;
  }

  /**
   * Modifies this mask such that no gaps are created in a {@code Raster2d}.
   *
   * @see #createsGap()
   */
  public void removeGaps() {
    int maxCol = maxColumn();
    int maxRow = maxRow();
    /* Find gaps in unmasked rows and remove them */
    for (int r = 0; r <= maxRow; r++) {
      int startCol = findMinNonmaskedCol(r);
      if (startCol < 0) continue;
      int endCol = findMaxNonmaskedCol(r);
      for (int c = startCol; c <= endCol; c++) {
        if (get(c, r)) set(c, r, false);
      }
    }
    /* Find gaps in unmasked columns and remove them */
    for (int c = 0; c <= maxCol; c++) {
      int startRow = findMinNonmaskedRow(c);
      if (startRow < 0) continue;
      int endRow = findMaxNonmaskedRow(c);
      for (int r = startRow; r <= endRow; r++) {
        if (get(c, r)) set(c, r, false);
      }
    }
  }

  private int findMinNonmaskedCol(int row) {
    final int maxCol = maxColumn();
    for (int c = 0; c <= maxCol; c++) {
      if (!get(c, row)) {
        return c;
      }
    }
    return -1;
  }

  private int findMaxNonmaskedCol(int row) {
    for (int c = maxColumn(); c >= 0; c--) {
      if (!get(c, row)) {
        return c;
      }
    }
    return -1;
  }

  private int findMinNonmaskedRow(int col) {
    final int maxRow = maxRow();
    for (int r = 0; r <= maxRow; r++) {
      if (!get(col, r)) {
        return r;
      }
    }
    return -1;
  }

  private int findMaxNonmaskedRow(int col) {
    for (int r = maxRow(); r >= 0; r--) {
      if (!get(col, r)) {
        return r;
      }
    }
    return -1;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof Raster2dMask)) return false;
    return mask.equals(((Raster2dMask)o).mask);
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + mask.hashCode();
    return result;
  }

  @Override
  public String toString() {
    StringBuilder s = new StringBuilder();
    s.append("{");
    List<Integer> rowIndices = new ArrayList<>(mask.keySet());
    Collections.sort(rowIndices);
    for (Integer r : rowIndices) {
      s.append(r).append(":{");
      List<Integer> columnIndices = new ArrayList<>(mask.get(r));
      Collections.sort(columnIndices);
      for (Integer c : columnIndices) {
        s.append(c).append(",");
      }
      if (!columnIndices.isEmpty()) s.deleteCharAt(s.length() - 1);
      s.append("},");
    }
    if (!rowIndices.isEmpty()) s.deleteCharAt(s.length() - 1);
    s.append("}");
    return s.toString();
  }
}
