package imcacat.imcadr;

import java.math.BigInteger;

/**
 * I perform calculations related to generating the positions of cells in a
 * raster.
 */
public class Raster2dPositionCalculator {
  protected Raster2dSpec spec;
  protected double scale;
  protected int columnCount;
  protected int rowCount;
  protected long cellWidth;
  protected long cellHeight;
  protected long width;
  protected long height;
  protected long centerHorizPos;
  protected long centerVertPos;
  protected long topLeftHorizPos;
  protected long topLeftVertPos;
  protected long topLeftCellHorizPos;
  protected long topLeftCellVertPos;

  public Raster2dPositionCalculator(Raster2dSpec spec) {
    super();
    if (!LengthUnit.MILLIMETER.equals(spec.getCellDimensionUnit()) &&
        !LengthUnit.MICROMETER.equals(spec.getCellDimensionUnit())) {
      throw new IllegalArgumentException("Unsupported cell dimension unit from Raster2dSpec: " +
          spec.getCellDimensionUnit());
    }
    this.spec = spec;
    calculate();
  }

  protected void calculate() {
    scale = BigInteger.valueOf(10).pow(fractionalDigitsCount()).longValue();
    columnCount = spec.getWidth();
    rowCount = spec.getHeight();
    cellWidth = (long)(spec.getCellWidth() * scale);
    cellHeight = (long)(spec.getCellHeight() * scale);
    width = columnCount * cellWidth;
    height = rowCount * cellHeight;
  }

  protected int fractionalDigitsCount() {
    LengthUnit unit = spec.getCellDimensionUnit();
    if (LengthUnit.MILLIMETER.equals(unit)) return 6;
    else if (LengthUnit.MICROMETER.equals(unit)) return 3;
    else throw new BugException("Unsupported unit: " + unit);
  }

  public void center(double unscaledCenterHorizPos, double unscaledCenterVertPos) {
    centerHorizPos = (long)(unscaledCenterHorizPos * scale);
    centerVertPos = (long)(unscaledCenterVertPos * scale);

    topLeftHorizPos = centerHorizPos - (width / 2);
    topLeftVertPos = centerVertPos - (height / 2);

    topLeftCellHorizPos = topLeftHorizPos + (cellWidth / 2);
    topLeftCellVertPos = topLeftVertPos + (cellHeight / 2);
  }

  public void topLeft(double unscaledTopLeftHorizPos, double unscaledTopLeftVertPos) {
    topLeftHorizPos = (long)(unscaledTopLeftHorizPos * scale);
    topLeftVertPos = (long)(unscaledTopLeftVertPos * scale);

    topLeftCellHorizPos = topLeftHorizPos + (cellWidth / 2);
    topLeftCellVertPos = topLeftVertPos + (cellHeight / 2);

    centerHorizPos = topLeftHorizPos + (width / 2);
    centerVertPos = topLeftVertPos + (height / 2);
  }

  public void topLeftCell(double unscaledTopLeftCellHorizPos, double unscaledTopLeftCellVertPos) {
    topLeftCellHorizPos = (long)(unscaledTopLeftCellHorizPos * scale);
    topLeftCellVertPos = (long)(unscaledTopLeftCellVertPos * scale);

    topLeftHorizPos = topLeftCellHorizPos - (cellWidth / 2);
    topLeftVertPos = topLeftCellVertPos - (cellHeight / 2);

    centerHorizPos = topLeftHorizPos + (width / 2);
    centerVertPos = topLeftVertPos + (height / 2);
  }

  public void shiftToward(Raster2dEdge edge) {
    long horizShift = 0;
    long vertShift = 0;

    if (edge.involvesTop()) vertShift = -cellHeight;
    else if (edge.involvesBottom()) vertShift = cellHeight;

    if (edge.involvesLeft()) horizShift = -cellWidth;
    else if (edge.involvesRight()) horizShift = cellWidth;

    topLeftHorizPos += horizShift;
    topLeftVertPos += vertShift;

    topLeftCellHorizPos += horizShift;
    topLeftCellVertPos += vertShift;

    centerHorizPos += horizShift;
    centerVertPos += vertShift;
  }

  public double unscaledCenterHorizPos() {
    return centerHorizPos / scale;
  }

  public double unscaledCenterVertPos() {
    return centerVertPos / scale;
  }

  public double unscaledTopLeftHorizPos() {
    return topLeftHorizPos / scale;
  }

  public double unscaledTopLeftVertPos() {
    return topLeftVertPos / scale;
  }

  public double unscaledTopLeftCellHorizPos() {
    return topLeftCellHorizPos / scale;
  }

  public double unscaledTopLeftCellVertPos() {
    return topLeftCellVertPos / scale;
  }

  public double scale() {
    return scale;
  }

  public int columnCount() {
    return columnCount;
  }

  public int rowCount() {
    return rowCount;
  }

  public long cellWidth() {
    return cellWidth;
  }

  public long cellHeight() {
    return cellHeight;
  }

  public long width() {
    return width;
  }

  public long height() {
    return height;
  }

  public long centerHorizPos() {
    return centerHorizPos;
  }

  public long centerVertPos() {
    return centerVertPos;
  }

  public long topLeftHorizPos() {
    return topLeftHorizPos;
  }

  public long topLeftVertPos() {
    return topLeftVertPos;
  }

  public long topLeftCellHorizPos() {
    return topLeftCellHorizPos;
  }

  public long topLeftCellVertPos() {
    return topLeftCellVertPos;
  }
}
