package imcacat.imcadr;

import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import imcacat.imcadr.util.NetworkUtils;

/**
 * I am a specification for a two-dimensional raster.
 */
public class Raster2dSpec implements Cloneable {
  protected Raster2dPlane plane;
  protected int width;
  protected int height;
  protected Integer preClipWidth;
  protected Integer preClipHeight;
  protected Raster2dMask mask;
  protected double cellWidth;
  protected double cellHeight;
  protected LengthUnit cellDimensionUnit;
  protected long exposureTime;
  protected int maxNumConcurrentAnalyzers;
  protected boolean flyScanEnabled;
  protected boolean chasingEnabled;
  protected Integer bestScoreCutoff;
  protected File imagesDirectory;
  protected ImageAnalysisSpec analysisSpec;
  protected File clusterKeyFile;
  protected List<ClusterNode> clusterNodes;

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, File imagesDirectory) {
    this(plane, rasterWidth, rasterHeight, cellWidth, cellHeight, cellDimensionUnit,
        exposureTime, maxNumConcurrentAnalyzers, false, true, imagesDirectory, null);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean chasingEnabled,
      File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, cellWidth, cellHeight, cellDimensionUnit,
        exposureTime, maxNumConcurrentAnalyzers, false, chasingEnabled, null,
        imagesDirectory, analysisSpec);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean flyScanEnabled, boolean chasingEnabled,
      File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, cellWidth, cellHeight, cellDimensionUnit,
        exposureTime, maxNumConcurrentAnalyzers, flyScanEnabled, chasingEnabled, null,
        imagesDirectory, analysisSpec);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean chasingEnabled, Integer bestScoreCutoff,
      File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, null, null, new Raster2dMask(), cellWidth,
        cellHeight, cellDimensionUnit, exposureTime, maxNumConcurrentAnalyzers, false,
        chasingEnabled, bestScoreCutoff, imagesDirectory, analysisSpec);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean flyScanEnabled, boolean chasingEnabled,
      Integer bestScoreCutoff, File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, null, null, new Raster2dMask(), cellWidth,
        cellHeight, cellDimensionUnit, exposureTime, maxNumConcurrentAnalyzers, flyScanEnabled,
        chasingEnabled, bestScoreCutoff, imagesDirectory, analysisSpec);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight,
      Integer rasterPreClipWidth, Integer rasterPreClipHeight, Raster2dMask mask,
      double cellWidth, double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean chasingEnabled, Integer bestScoreCutoff,
      File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, rasterPreClipWidth, rasterPreClipHeight, mask,
        cellWidth, cellHeight, cellDimensionUnit, exposureTime, maxNumConcurrentAnalyzers,
        false, chasingEnabled, bestScoreCutoff, imagesDirectory, analysisSpec);
  }

  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight,
      Integer rasterPreClipWidth, Integer rasterPreClipHeight, Raster2dMask mask, double cellWidth,
      double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean flyScanEnabled, boolean chasingEnabled,
      Integer bestScoreCutoff, File imagesDirectory, ImageAnalysisSpec analysisSpec) {
    this(plane, rasterWidth, rasterHeight, rasterPreClipWidth, rasterPreClipHeight, mask, cellWidth,
        cellHeight, cellDimensionUnit, exposureTime, maxNumConcurrentAnalyzers, flyScanEnabled,
        chasingEnabled, bestScoreCutoff, imagesDirectory, analysisSpec, null,
        Arrays.asList(new ClusterNode(NetworkUtils.LOCALHOST_NAME, 1)));
  }

  /**
   * @param clusterNodes the list of cluster nodes; this list must not be
   *          modified during or after this call
   */
  public Raster2dSpec(Raster2dPlane plane, int rasterWidth, int rasterHeight,
      Integer rasterPreClipWidth, Integer rasterPreClipHeight, Raster2dMask mask,
      double cellWidth, double cellHeight, LengthUnit cellDimensionUnit, long exposureTime,
      int maxNumConcurrentAnalyzers, boolean flyScanEnabled, boolean chasingEnabled,
      Integer bestScoreCutoff, File imagesDirectory, ImageAnalysisSpec analysisSpec,
      File clusterKeyFile, List<ClusterNode> clusterNodes) {
    this.plane = plane;
    this.width = rasterWidth;
    this.height = rasterHeight;
    if ((rasterPreClipWidth == null && rasterPreClipHeight != null) ||
        (rasterPreClipWidth != null && rasterPreClipHeight == null)) {
      throw new IllegalArgumentException(
          "Raster pre-clip width and height not both null or both nonnull");
    }
    this.preClipWidth = rasterPreClipWidth;
    this.preClipHeight = rasterPreClipHeight;
    this.mask = Objects.requireNonNull(mask);
    this.cellWidth = cellWidth;
    this.cellHeight = cellHeight;
    this.cellDimensionUnit = cellDimensionUnit;
    this.exposureTime = exposureTime;
    this.maxNumConcurrentAnalyzers = Math.max(1, maxNumConcurrentAnalyzers);
    this.flyScanEnabled = flyScanEnabled;
    this.chasingEnabled = chasingEnabled;
    this.bestScoreCutoff = bestScoreCutoff;
    this.imagesDirectory = imagesDirectory;
    this.analysisSpec = (analysisSpec == null) ? (new ImageAnalysisSpec.Builder()).build()
        : analysisSpec;
    this.clusterKeyFile = clusterKeyFile;
    this.clusterNodes = Collections.unmodifiableList(clusterNodes);
    if (this.clusterNodes.isEmpty()) {
      throw new IllegalArgumentException("clusterNodes is empty");
    }
  }

  public Raster2dPlane getPlane() {
    return plane;
  }

  public int getWidth() {
    return width;
  }

  public int getHeight() {
    return height;
  }

  public Integer getPreClipWidth() {
    return preClipWidth;
  }

  public Integer getPreClipHeight() {
    return preClipHeight;
  }

  public Raster2dMask getMask() {
    return mask;
  }

  public double getCellWidth() {
    return cellWidth;
  }

  public double getCellHeight() {
    return cellHeight;
  }

  public double getCellLength(Direction2d d) {
    return d.isHorizontal() ? getCellWidth() : getCellHeight();
  }

  public LengthUnit getCellDimensionUnit() {
    return cellDimensionUnit;
  }

  public Rectangle getCellSize() {
    return new Rectangle(cellWidth, cellHeight, cellDimensionUnit);
  }

  public long getExposureTime() {
    return exposureTime;
  }

  public int getMaxNumConcurrentAnalyzers() {
    return maxNumConcurrentAnalyzers;
  }

  public boolean isFlyScanEnabled() {
    return flyScanEnabled;
  }

  public boolean isChasingEnabled() {
    return chasingEnabled;
  }

  public Integer getBestScoreCutoff() {
    return bestScoreCutoff;
  }

  public File getImagesDirectory() {
    return imagesDirectory;
  }

  public ImageAnalysisSpec getAnalysisSpec() {
    return analysisSpec;
  }

  public File getClusterKeyFile() {
    return clusterKeyFile;
  }

  public boolean hasClusterNodes() {
    return !clusterNodes.isEmpty();
  }

  public List<ClusterNode> getClusterNodes() {
    return clusterNodes;
  }

  /**
   * I create a copy of myself with my cell dimensions in millimeters.
   *
   * @return copy of myself with cell dimensions in millimeters
   *
   * @throws UnsupportedConversionException if my cell dimension unit is
   *           neither millimeters nor micrometers
   */
  public Raster2dSpec inMillimeters() throws UnsupportedConversionException {
    Raster2dSpec mmSpec;
    try {
      mmSpec = (Raster2dSpec)clone();
    } catch (CloneNotSupportedException e) {
      throw new BugException(e);
    }
    if (LengthUnit.MILLIMETER.equals(cellDimensionUnit)) return mmSpec;
    if (!LengthUnit.MICROMETER.equals(cellDimensionUnit)) throw new UnsupportedConversionException(
        cellDimensionUnit + " to " + LengthUnit.MILLIMETER);
    mmSpec.cellWidth = cellWidth / 1000.0;
    mmSpec.cellHeight = cellHeight / 1000.0;
    mmSpec.cellDimensionUnit = LengthUnit.MILLIMETER;
    return mmSpec;
  }

  /**
   * I create a copy of myself with my cell dimensions in micrometers.
   *
   * @return copy of myself with cell dimensions in micrometers
   *
   * @throws UnsupportedConversionException if my cell dimension unit is
   *           neither millimeters nor micrometers
   */
  public Raster2dSpec inMicrometers() throws UnsupportedConversionException {
    Raster2dSpec mmSpec;
    try {
      mmSpec = (Raster2dSpec)clone();
    } catch (CloneNotSupportedException e) {
      throw new BugException(e);
    }
    if (LengthUnit.MICROMETER.equals(cellDimensionUnit)) return mmSpec;
    if (!LengthUnit.MILLIMETER.equals(cellDimensionUnit)) throw new UnsupportedConversionException(
        cellDimensionUnit + " to " + LengthUnit.MICROMETER);
    mmSpec.cellWidth = cellWidth * 1000.0;
    mmSpec.cellHeight = cellHeight * 1000.0;
    mmSpec.cellDimensionUnit = LengthUnit.MICROMETER;
    return mmSpec;
  }
}
