package imcacat.imcadr;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.nio.channels.ClosedByInterruptException;
import java.util.logging.Level;
import java.util.logging.Logger;

import imcacat.imcadr.util.DurationUtils;

/**
 * I analyze a raster cell.
 */
public class RasterCellAnalyzer implements ClusterRunnable {
  private static final Logger LOG = Logger.getLogger(RasterCellAnalyzer.class.getName());

  protected static final long IMAGE_EXISTS_TEST_DELAY = 1000;
  protected static final long IMAGE_COMPLETE_TEST_DELAY = 2000;
  protected static final long TIME_JITTER_MAX = 2000;

  protected final int column;
  protected final int row;
  protected final Raster2d raster;
  protected final File initialImageFilePath;
  protected final File imageFilePath;
  protected final long imageAcquiredTime;
  protected final long imageAcquiredMonotonicTime;
  protected Exception exception;

  private volatile String hostname;
  private volatile long imageWaitTime;
  private volatile long scoreTime;

  public RasterCellAnalyzer(Raster2dIndex index, Raster2d raster) {
    this(index.getColumn(), index.getRow(), raster, -1L);
  }

  public RasterCellAnalyzer(Raster2dIndex index, Raster2d raster, long earliestImageAcquiredTime) {
    this(index.getColumn(), index.getRow(), raster, earliestImageAcquiredTime);
  }

  public RasterCellAnalyzer(Raster2dIndex index, Raster2d raster, long earliestImageAcquiredTime,
      String initialImageFileName) {
    this(index.getColumn(), index.getRow(), raster, earliestImageAcquiredTime,
        initialImageFileName);
  }

  public RasterCellAnalyzer(Raster2dIndex index, Raster2d raster, long earliestImageAcquiredTime,
      long earliestImageAcquiredMonotonicTime, String initialImageFileName) {
    this(index.getColumn(), index.getRow(), raster, earliestImageAcquiredTime,
        earliestImageAcquiredMonotonicTime, initialImageFileName);
  }

  public RasterCellAnalyzer(int column, int row, Raster2d raster) {
    this(column, row, raster, -1L);
  }

  public RasterCellAnalyzer(int column, int row, Raster2d raster, long earliestImageAcquiredTime) {
    this(column, row, raster, earliestImageAcquiredTime, null);
  }

  public RasterCellAnalyzer(int column, int row, Raster2d raster, long earliestImageAcquiredTime,
      String initialImageFileName) {
    this(column, row, raster, earliestImageAcquiredTime, System.nanoTime(), initialImageFileName);
  }

  public RasterCellAnalyzer(int column, int row, Raster2d raster, long earliestImageAcquiredTime,
      long earliestImageAcquireMonotonicTime, String initialImageFileName) {
    super();
    this.column = column;
    this.row = row;
    this.raster = raster;
    imageFilePath = raster.getImageFilePath(column, row);
    initialImageFilePath = (initialImageFileName == null) ? imageFilePath
        : new File(imageFilePath.getParentFile(), initialImageFileName);
    imageAcquiredTime = earliestImageAcquiredTime;
    imageAcquiredMonotonicTime = earliestImageAcquireMonotonicTime;
    hostname = null;
    imageWaitTime = 0L;
    scoreTime = 0L;
  }

  @Override
  public void run() {
    try {
      go();
    } catch (InterruptedException | InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
      exception = e;
    } catch (Exception e) {
      LOG.log(Level.SEVERE, "Failed to score image \"" + imageFilePath + "\"", e);
      exception = e;
    }
  }

  @Override
  public void setClusterNode(String hostname) {
    this.hostname = hostname;
  }

  public long getImageWaitTime() {
    return imageWaitTime;
  }

  public long getScoreTime() {
    return scoreTime;
  }

  protected void go() throws Exception {
    waitForImageToBeAcquired();
    long startTime = System.nanoTime();
    waitForImageToExist();
    imageWaitTime = nanosToMillis(System.nanoTime() - startTime);
    renameImage();
    startTime = System.nanoTime();
    ImageAnalysisResult result = analyzeImage();
    scoreTime = nanosToMillis(System.nanoTime() - startTime);
    raster.loadResult(column, row, result);
  }

  protected void waitForImageToBeAcquired() throws InterruptedException {
    long delta = imageAcquiredMonotonicTime - System.nanoTime();
    if (delta > 0) Thread.sleep(nanosToMillis(delta));
  }

  protected void waitForImageToExist() throws InterruptedException {
    for (;;) {
      if (initialImageFilePath.exists()) break;
      Thread.sleep(IMAGE_EXISTS_TEST_DELAY);
    }

    if (imageAcquiredTime >= 0L) {
      long acquiredTimeWithMaxJitter = imageAcquiredTime - TIME_JITTER_MAX;
      for (;;) {
        if (initialImageFilePath.lastModified() >= acquiredTimeWithMaxJitter) break;
        Thread.sleep(IMAGE_EXISTS_TEST_DELAY);
      }
    }

    for (;;) {
      long lastModified = initialImageFilePath.lastModified();
      Thread.sleep(IMAGE_COMPLETE_TEST_DELAY);
      if (lastModified != initialImageFilePath.lastModified()) continue;
      if (initialImageFilePath.length() != 0) break;
    }
  }

  protected void renameImage() throws IOException {
    if (initialImageFilePath.getName().equals(imageFilePath.getName())) return;
    imageFilePath.delete();
    boolean ok = initialImageFilePath.renameTo(imageFilePath);
    if (!ok) {
      throw new IOException("Could not rename \"" + initialImageFilePath +
          "\" to \"" + imageFilePath + "\"");
    }
  }

  protected ImageAnalysisResult analyzeImage() {
    long startTime = System.nanoTime();
    ImageAnalysisResult result = ImageAnalysis.newInstance(getAnalysisSpec(), imageFilePath,
        raster.getSpec().getClusterKeyFile(), hostname).run();
    long endTime = System.nanoTime();
    logAnalysisResult(result, imageFilePath, hostname, startTime, endTime);
    return result;
  }

  protected ImageAnalysisSpec getAnalysisSpec() {
    return raster.getSpec().getAnalysisSpec();
  }

  protected static long nanosToMillis(long nanoseconds) {
    return nanoseconds / 1000000L;
  }

  protected static void logAnalysisResult(ImageAnalysisResult result, File image, String hostname,
      long startNanoTime, long endNanoTime) {
    StringBuilder b = new StringBuilder();
    b.append("Scored on ");
    b.append(hostname);
    b.append(" in ");
    b.append(DurationUtils.createTimeString(nanosToMillis(endNanoTime - startNanoTime)));
    b.append(" (status=");
    b.append(result.wasSuccessful() ? "OK" : "FAIL");
    b.append(", file=");
    b.append(image);
    if (!result.wasSuccessful()) {
      if (result.getErrorMessage() != null) {
        b.append(", errorMessage=");
        b.append(result.getErrorMessage());
      }
      if (result.getErrorCause() != null && result.getErrorCause().getMessage() != null) {
        b.append(", errorCause=");
        b.append(result.getErrorCause().getMessage());
      }
    }
    b.append(")");
    LOG.fine(b.toString());
  }
}
