package imcacat.imcadr;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.PrintWriter;
import java.nio.channels.ClosedByInterruptException;
import java.nio.file.Files;
import java.nio.file.InvalidPathException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import gov.aps.jca.TimeoutException;
import imcacat.imcadr.util.DurationUtils;
import imcacat.imcadr.util.MathUtils;

/**
 * I rasterize a two-dimensional raster.
 */
public class Rasterizer2d extends BaseRasterizer {
  private static final long serialVersionUID = 1;
  private static final Logger LOG = Logger.getLogger(Rasterizer2d.class.getName());

  protected static final long ANALYSIS_TIMEOUT = 3600;

  public static final String PROPNAME_BEST_POSITION = "bestPosition";
  public static final String PROPNAME_BEST_POSITION_STRING = "bestPositionString";
  public static final String PROPNAME_TIME_STRING = "timeString";

  protected Raster2dSpec spec;
  protected BeamlineDriver driver;
  protected Raster2d raster;
  protected List<RasterCellAnalyzer> analyzers;
  protected ClusterService clusterService;
  protected ScheduledExecutorService imageTakenPublisherExecutor;
  protected boolean okToSetExperimentState = true;
  protected boolean okToEnableOrDisableAutobp = true;
  protected long startAcquireTime = -1;
  protected long startAnalyzeTime = -1;
  protected Position2d bestPosition;
  protected String bestPositionString = "";
  protected String timeString = "";

  public Rasterizer2d(Raster2dSpec spec, BeamlineDriver driver,
      boolean firePropertyChangesOnEdtEnabled) {
    super(firePropertyChangesOnEdtEnabled);
    try {
      this.spec = spec.inMillimeters();
    } catch (UnsupportedConversionException e) {
      throw new IllegalArgumentException(e);
    }
    this.driver = driver;
    raster = new Raster2d(this.spec, driver.getImageFileNameExtension(),
        firePropertyChangesOnEdtEnabled);
    analyzers = new ArrayList<>();
    bestPosition = new Position2d(spec.getCellDimensionUnit());
  }

  protected ClusterService createClusterService() {
    return new ClusterService(this.spec.getClusterNodes());
  }

  protected ScheduledExecutorService createImageTakenPublisherExecutor() {
    return Executors.newScheduledThreadPool(1);
  }

  public void doNotSetExperimentState() {
    okToSetExperimentState = false;
  }

  public void doNotEnableNorDisableAutobp() {
    okToEnableOrDisableAutobp = false;
  }

  public Raster2dPlane getPlane() {
    return spec.getPlane();
  }

  @Override
  protected void go() throws Exception {
    long startTime = System.nanoTime();

    String planeName = getPlane().toString();

    setStatus("Configuring beamline for " + planeName + " rasterization");
    if (okToSetExperimentState) driver.enterCollectDataExperimentState();
    if (okToEnableOrDisableAutobp) driver.disableAutobp();
    driver.movePhi(computePhiFor(spec.getPlane()));
    setProgress(0);

    BeamlineAxis verticalAxis = computeVerticalAxisFor(spec.getPlane());
    raster.assignPositionsCenteredOn(driver.zRbv(), driver.rbv(verticalAxis));

    rasterizeAndAnalyze(verticalAxis, planeName + " plane", 95);

    setStatus("Computing best " + planeName + " raster position");
    Raster2dIndex bestPositionIndex = raster.computeBestPositionIndex();

    Integer cutoff = spec.getBestScoreCutoff();
    if (cutoff != null) {
      Integer bestScore = raster.getCell(bestPositionIndex).getScore();
      if (bestScore == null || bestScore <= cutoff) {
        createBestCellSymlink(bestPositionIndex);
        throw new BestScoreCutoffException(bestScore, cutoff);
      }
    }

    if (spec.isChasingEnabled() && (raster.columnCount() > 1 || raster.rowCount() > 1)) {
      Set<Raster2dEdge> previousOppositeEdgeComponents =
          EnumSet.<Raster2dEdge>noneOf(Raster2dEdge.class);
      int shiftHorizontalCount = 0;
      int shiftVerticalCount = 0;
      loop: for (;;) {
        Raster2dEdge edge;
        Raster2dIndex edgeIndex = null;
        boolean needToUnmaskOrthogonalNeighbors = false;
        if (raster.isOnUnmaskedEdge(bestPositionIndex)) {
          edge = raster.unmaskedEdge(bestPositionIndex);
          edgeIndex = bestPositionIndex;
          if (raster.hasMaskedOrthogonalNeighbor(bestPositionIndex, edge)) {
            needToUnmaskOrthogonalNeighbors = true;
          }
        } else if (raster.isOnEdge(bestPositionIndex)) {
          edge = raster.edge(bestPositionIndex);
        } else {
          break;
        }

        for (Raster2dEdge each : edge.components()) {
          if (previousOppositeEdgeComponents.contains(each)) break loop;
        }
        previousOppositeEdgeComponents.addAll(edge.opposite().components());

        if (edge.involvesLeft() || edge.involvesRight()) {
          if (shiftHorizontalCount >= raster.columnCount()) break;
          shiftHorizontalCount++;
        }
        if (edge.involvesTop() || edge.involvesBottom()) {
          if (shiftVerticalCount >= raster.rowCount()) break;
          shiftVerticalCount++;
        }

        if (needToUnmaskOrthogonalNeighbors) {
          raster.unmaskOrthogonalNeighbors(bestPositionIndex, edge);
        }
        raster.shiftToward(edge);
        rasterizeAndAnalyze(verticalAxis, planeName + " plane edge", 97, edge, edgeIndex);

        setStatus("Computing best " + planeName + " raster position");
        bestPositionIndex = raster.computeBestPositionIndex();
      }
    }

    createBestCellSymlink(bestPositionIndex);

    setProgress(99);
    setBestPosition(raster.computeBestPosition());

    setStatus("Moving to best " + planeName + " raster position");
    driver.moveZ(getBestHorizontalPosition());
    driver.move(verticalAxis, getBestVerticalPosition());

    setStatus("Configuring beamline for use");
    if (okToSetExperimentState) driver.enterIdleExperimentState();
    if (okToEnableOrDisableAutobp) driver.enableAutobp();

    setWallTime(nanosToMillis(System.nanoTime() - startTime));

    setStatus(SUCCESSFUL_COMPLETION_STATUS);
    setProgress(100);
  }

  protected void rasterizeAndAnalyze(BeamlineAxis verticalAxis, String rasterName, int maxProgress)
      throws InterruptedException, DriverException, TimeoutException, CellAnalysisTimeoutException {
    rasterizeAndAnalyze(verticalAxis, rasterName, maxProgress, null, null);
  }

  protected void rasterizeAndAnalyze(BeamlineAxis verticalAxis, String rasterName, int maxProgress,
      Raster2dEdge edge, Raster2dIndex edgeIndex) throws InterruptedException, DriverException,
          TimeoutException, CellAnalysisTimeoutException {
    rasterize(verticalAxis, rasterName, maxProgress, edge, edgeIndex);
    waitForAnalysisToComplete(rasterName);
  }

  protected void rasterize(BeamlineAxis verticalAxis, String rasterName, int maxProgress,
      Raster2dEdge edge, Raster2dIndex edgeIndex)
          throws InterruptedException, DriverException, TimeoutException {
    startAcquireTime = nanosToMillis(System.nanoTime());
    setStatus("Rasterizing " + rasterName);
    clusterService = createClusterService();
    int minProgress = getProgress();
    int numCells = raster.unmaskedUnscoredCellCount(edge, edgeIndex);
    if (spec.isFlyScanEnabled()) {
      imageTakenPublisherExecutor = createImageTakenPublisherExecutor();
      BeamlineAxis horizontalAxis = BeamlineAxis.Z;
      Iterator<Raster2dFlyScan> flyScanIterator =
          raster.unmaskedUnscoredFlyScanIterator(edge, edgeIndex);
      String scanImageFileNameTemplate = "cell_%s_fly_%02d_%04d%s";
      String imageTakenFileNameTemplateTemplate = "cell_%s_fly_%02d_%%04d%s";
      for (int i = 0; flyScanIterator.hasNext(); i++) {
        Raster2dFlyScan scan = flyScanIterator.next();
        RasterCell startCell = scan.startCell();
        RasterCell endCell = scan.endCell();
        String scanImageFileName = String.format(Locale.US, scanImageFileNameTemplate,
            Raster2d.imageFileNamePlaneId(spec.getPlane()), i + 1, 1,
            driver.getImageFileNameExtension());
        BeamlineAxis scanAxis = scan.isHorizontal() ? horizontalAxis : verticalAxis;
        double accelTime = driver.getAccelTime(scanAxis);
        Raster2dFlyScanCalculator c = new Raster2dFlyScanCalculator(
            startCell.getPosition(scan.direction()), endCell.getPosition(scan.direction()),
            spec.getCellLength(scan.direction()), millisToSeconds(spec.getExposureTime()),
            driver.getMinVelo(scanAxis), driver.getMaxVelo(scanAxis), accelTime,
            driver.getDetectorReadoutTime(), Resources.getInstance().getExposureShutterLatency(),
            Resources.getInstance().getDetectorLatency(), flyScanCvSettleTime(scanAxis));
        driver.beginDetectorDataset(c.exposureTime(), c.exposurePeriod(), scanAxis,
            scan.imageCount(), new File(spec.getImagesDirectory(), scanImageFileName));
        driver.move(horizontalAxis,
            scan.isHorizontal() ? c.startPosition() : startCell.getHorizontalPosition());
        driver.move(verticalAxis,
            scan.isVertical() ? c.startPosition() : startCell.getVerticalPosition());
        Double origVelo = driver.getVelo(scanAxis);
        boolean inFlyScan = false;
        ImageTakenPublisher publisher = null;
        ScheduledFuture<?> publisherFuture = null;
        try {
          driver.setVelo(scanAxis, c.velocity());
          inFlyScan = true;
          driver.beginFlyScan(c.shutterOpenPosition(), c.shutterClosePosition(),
              c.detectorTriggerPosition());
          scan.setAcquiring(true);
          if (i == 0) startAnalyzeTime = nanosToMillis(System.nanoTime());
          long earliestImageAcquiredTime =
              System.currentTimeMillis() + secondsToMillis(c.exposurePeriod());
          long flyScanStartNanoTime = System.nanoTime() + millisToNanos(secondsToMillis(accelTime));
          long exposurePeriodNanos = millisToNanos(secondsToMillis(c.exposurePeriod()));
          int seqNum = 1;
          for (RasterCell each : scan) {
            String flyImageName = String.format(Locale.US, scanImageFileNameTemplate,
                Raster2d.imageFileNamePlaneId(spec.getPlane()), i + 1, seqNum,
                driver.getImageFileNameExtension());
            RasterCellAnalyzer analyzer = createAnalyzer(
                new Raster2dIndex(each.getColumn(), each.getRow()), earliestImageAcquiredTime,
                flyScanStartNanoTime + (exposurePeriodNanos * seqNum), flyImageName);
            analyzers.add(analyzer);
            clusterService.execute(analyzer);
            seqNum++;
          }
          String imageTakenFileNameTemplate = String.format(Locale.US,
              imageTakenFileNameTemplateTemplate, Raster2d.imageFileNamePlaneId(spec.getPlane()),
              i + 1, driver.getImageFileNameExtension());
          long imageTakenPeriod = secondsToMillis(c.exposurePeriod());
          publisher = createImageTakenPublisher(imageTakenFileNameTemplate, 1, scan.imageCount());
          publisherFuture = imageTakenPublisherExecutor.scheduleAtFixedRate(publisher,
              secondsToMillis(accelTime) + imageTakenPeriod, imageTakenPeriod,
              TimeUnit.MILLISECONDS);
          driver.move(scanAxis, c.endPosition());
          driver.setVelo(scanAxis, origVelo);
          origVelo = null;
          driver.endFlyScan();
          inFlyScan = false;
        } finally {
          boolean interrupted = false;
          if (origVelo != null) {
            try {
              driver.setVelo(scanAxis, origVelo);
            } catch (InterruptedException e) {
              interrupted = true;
            } catch (DriverException e) {
              /* IGNORED */
            }
          }
          if (inFlyScan) {
            try {
              driver.endFlyScanNow();
            } catch (InterruptedException e) {
              interrupted = true;
            } catch (DriverException e) {
              /* IGNORED */
            }
          }
          scan.setAcquiring(false);
          if (interrupted) {
            Thread.currentThread().interrupt();
          }
        }
        boolean ok = publisher.await(5, TimeUnit.SECONDS);
        if (!ok) LOG.warning("Fly-scan image-taken publisher did not finish");
        publisherFuture.cancel(true);
        if (edge == null) {
          setProgress(
              MathUtils.scale((i + 1) * scan.imageCount(), 0, numCells, minProgress, maxProgress));
        }
        driver.endDetectorDataset();
      }
    } else {
      long exposureTime = spec.getExposureTime();
      driver.beginDetectorDataset(millisToSeconds(exposureTime));
      Raster2dIterator rasterIterator = raster.unmaskedUnscoredIterator(edge, edgeIndex);
      for (int i = 0; rasterIterator.hasNext(); i++) {
        Raster2dIndex index = rasterIterator.nextIndex();
        RasterCell cell = rasterIterator.next();
        driver.moveZ(cell.getHorizontalPosition());
        driver.move(verticalAxis, cell.getVerticalPosition());
        cell.setAcquiring(true);
        long earliestImageAcquiredTime = System.currentTimeMillis() + exposureTime;
        try {
          driver.takeDetectorImage(cell.getImageFilePath(), !rasterIterator.hasNext());
        } finally {
          cell.setAcquiring(false);
        }
        if (edge == null) {
          setProgress(MathUtils.scale(i + 1, 0, numCells, minProgress, maxProgress));
        }
        if (i == 0) startAnalyzeTime = nanosToMillis(System.nanoTime());
        RasterCellAnalyzer analyzer = createAnalyzer(index, earliestImageAcquiredTime);
        analyzers.add(analyzer);
        clusterService.execute(analyzer);
      }
      driver.endDetectorDataset();
    }
    setAcquireTime(
        getAcquireTime() + (nanosToMillis(System.nanoTime()) - startAcquireTime));
  }

  protected RasterCellAnalyzer createAnalyzer(Raster2dIndex index, long earliestImageAcquiredTime) {
    return createAnalyzer(index, earliestImageAcquiredTime, null);
  }

  protected RasterCellAnalyzer createAnalyzer(Raster2dIndex index, long earliestImageAcquiredTime,
      String flyScanImageName) {
    return createAnalyzer(index, earliestImageAcquiredTime, System.nanoTime(), flyScanImageName);
  }

  protected RasterCellAnalyzer createAnalyzer(Raster2dIndex index, long earliestImageAcquiredTime,
      long earliestImageAcquiredMonotonicTime, String flyScanImageName) {
    return new RasterCellAnalyzer(index, raster, earliestImageAcquiredTime,
        earliestImageAcquiredMonotonicTime, flyScanImageName);
  }

  protected ImageTakenPublisher createImageTakenPublisher(String fileNameTemplate, int seqNum,
      int count) {
    return new ImageTakenPublisher(driver, spec.getImagesDirectory(), fileNameTemplate, seqNum,
        count);
  }

  protected void waitForAnalysisToComplete(String rasterName)
      throws InterruptedException, CellAnalysisTimeoutException {
    setStatus("Waiting for " + rasterName + " scoring to complete");
    clusterService.shutDown();
    boolean ok = clusterService.awaitTermination(ANALYSIS_TIMEOUT, TimeUnit.SECONDS);
    if (!ok) throw new CellAnalysisTimeoutException();
    setAnalyzeTime(getAnalyzeTime() + (nanosToMillis(System.nanoTime()) - startAnalyzeTime));
    for (RasterCellAnalyzer each : analyzers) {
      cellAnalyzeImageWaitTimes.add(each.getImageWaitTime());
      cellAnalyzeScoreTimes.add(each.getScoreTime());
    }
  }

  @Override
  protected void cleanUp() {
    boolean interrupted = false;
    Exception caughtException = null;

    analyzers.clear();
    if (clusterService != null) clusterService.shutDownNow();
    if (imageTakenPublisherExecutor != null) imageTakenPublisherExecutor.shutdownNow();

    try {
      driver.endDetectorDataset();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      if (okToSetExperimentState) driver.enterIdleExperimentState();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      if (okToEnableOrDisableAutobp) driver.enableAutobp();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }

    if (interrupted) Thread.currentThread().interrupt();
    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  protected double computePhiFor(Raster2dPlane plane) {
    if (plane.equals(Raster2dPlane.ZY)) return 0.0;
    if (plane.equals(Raster2dPlane.ZX)) return 90.0;
    throw new BugException();
  }

  protected BeamlineAxis computeVerticalAxisFor(Raster2dPlane plane) {
    if (plane.equals(Raster2dPlane.ZY)) return BeamlineAxis.Y;
    if (plane.equals(Raster2dPlane.ZX)) return BeamlineAxis.X;
    throw new BugException();
  }

  public Raster2d getRaster() {
    return raster;
  }

  public Position2d getBestPosition() {
    return bestPosition;
  }

  public double getBestHorizontalPosition() {
    return bestPosition.getX();
  }

  public double getBestVerticalPosition() {
    return bestPosition.getY();
  }

  protected void setBestPosition(Position2d newBestPosition) {
    if (bestPosition == null && newBestPosition == null) return;
    if (bestPosition != null && bestPosition.equals(newBestPosition)) return;
    Position2d oldBestPosition = bestPosition;
    bestPosition = newBestPosition;
    notifyChanged(PROPNAME_BEST_POSITION, oldBestPosition, newBestPosition);
    setBestPositionString(newBestPosition.toDisplayString(3));
  }

  public Integer getScoreAtBestPosition() {
    return raster.getScoreAtBestPosition();
  }

  @Override
  protected void setAcquireTime(long newAcquireTime) {
    super.setAcquireTime(newAcquireTime);
    updateTimeString();
  }

  @Override
  protected void setAnalyzeTime(long newAnalyzeTime) {
    super.setAnalyzeTime(newAnalyzeTime);
    updateTimeString();
  }

  @Override
  protected void setWallTime(long newWallTime) {
    super.setWallTime(newWallTime);
    updateTimeString();
  }

  protected void updateTimeString() {
    setTimeString(createTimeString());
  }

  protected String createTimeString() {
    return DurationUtils.createTimeString(getWallTime()) + ", " +
        DurationUtils.createTimeString(getAcquireTime()) + ", " +
        DurationUtils.createTimeString(getAnalyzeTime());
  }

  public String getBestPositionString() {
    return bestPositionString;
  }

  protected void setBestPositionString(String newBestPositionString) {
    if (bestPositionString.equals(newBestPositionString)) return;
    String oldBestPositionString = bestPositionString;
    bestPositionString = newBestPositionString;
    notifyChanged(PROPNAME_BEST_POSITION_STRING, oldBestPositionString, newBestPositionString);
  }

  public String getTimeString() {
    return timeString;
  }

  protected void setTimeString(String newTimeString) {
    if (timeString.equals(newTimeString)) return;
    String oldTimeString = timeString;
    timeString = newTimeString;
    notifyChanged(PROPNAME_TIME_STRING, oldTimeString, newTimeString);
  }

  @Override
  protected void writeResults(PrintWriter writer) {
    (new Rasterizer2dPrinter(writer)).print(this);
  }

  @Override
  protected File createResultsFile() {
    return new File(spec.getImagesDirectory(),
        "imcadr-" + spec.getPlane().toString() + "-result.html");
  }

  protected void createBestCellSymlink(Raster2dIndex index) {
    /*
     * It's not critical that this succeed, so just log failures.
     *
     * Also, this is a bit confusing because the Java "link" is the Unix ln
     * command "target", and the Java "target" is the Unix ln command "source".
     */
    try {
      Path target = raster.getImageFilePath(index).toPath();
      Path relativeTarget = target.getFileName();

      String linkName =
          Raster2d.bestImageFileName(driver.getImageFileNameExtension(), raster.getPlane());

      Path absoluteLink = target.resolveSibling(linkName);
      Files.deleteIfExists(absoluteLink);

      Files.createSymbolicLink(absoluteLink, relativeTarget);
    } catch (ClosedByInterruptException | InterruptedIOException e) {
      Thread.currentThread().interrupt();
    } catch (InvalidPathException | IOException | SecurityException e) {
      LOG.log(Level.FINE, "Failed to create best-cell symlink", e);
    }
  }

  private static double millisToSeconds(long ms) {
    return ((double)ms) / 1e3;
  }

  private static long secondsToMillis(double s) {
    return (long)(s * 1e3);
  }

  private static double flyScanCvSettleTime(BeamlineAxis axis) {
    switch (axis) {
      case X: return Resources.getInstance().getXMotorFlyScanCvSettleTime();
      case Y: return Resources.getInstance().getYMotorFlyScanCvSettleTime();
      case Z: return Resources.getInstance().getZMotorFlyScanCvSettleTime();
      default: throw new BugException();
    }
  }

  /**
   * Publishes that images in a sequence have been taken. Publishes that an
   * image in the sequence has been taken each time the {@code run()} method is
   * called.
   */
  private static class ImageTakenPublisher implements Runnable {
    private final BeamlineDriver driver;
    private final File imagesDir;
    private final String fileNameTemplate;
    private final CountDownLatch finishedLatch;
    private final int endSeqNum;
    private int seqNum;
    private boolean finished;

    /**
     * Creates an instance with the specified parameters.
     *
     * @param imagesDir the directory where the image sequence images are to be
     *          written
     * @param fileNameTemplate the template for generating each file name in
     *          the sequence; it must have only one format specifier which is
     *          the sequence number
     * @param startSeqNum the starting sequence number of the image sequence
     * @param count the total number of image sequence images to publish
     *
     * @throws IllegalArgumentException if {@code startSeqNum} is less than 0
     *           or if {@code count} is less than or equal to 0
     */
    private ImageTakenPublisher(BeamlineDriver driver, File imagesDir, String fileNameTemplate,
        int startSeqNum, int count) {
      if (startSeqNum < 0) {
        throw new IllegalArgumentException("Start sequence number must be nonnegative");
      }
      if (count <= 0) {
        throw new IllegalArgumentException("Count must be greater than 0");
      }
      this.driver = driver;
      this.imagesDir = imagesDir;
      this.fileNameTemplate = fileNameTemplate;
      finishedLatch = new CountDownLatch(1);
      endSeqNum = startSeqNum + count - 1;
      seqNum = startSeqNum;
      finished = false;
    }

    @Override
    public void run() {
      if (finished) return;
      String fileName = String.format(Locale.US, fileNameTemplate, seqNum);
      try {
        driver.publishDetectorImageTaken(new File(imagesDir, fileName));
      } catch (DriverException e) {
        throw new RuntimeException(e);
      } catch (InterruptedException e) {
        Thread.currentThread().interrupt();
        return;
      }
      seqNum++;
      if (seqNum > endSeqNum) {
        finishedLatch.countDown();
        finished = true;
      }
    }

    /**
     * Causes the current thread to wait until publishing finishes, unless the
     * thread is interrupted or the specified waiting time elapses.
     *
     * @param timeout the time to wait
     * @param unit the unit of the {@code timeout} value
     *
     * @return {@code true} if publishing finished; {@code false} if the
     *         waiting time elapsed before publishing finished
     */
    public boolean await(long timeout, TimeUnit unit) throws InterruptedException {
      return finishedLatch.await(timeout, unit);
    }
  }
}
