package imcacat.imcadr;

import imcacat.imcadr.util.ColorUtils;
import imcacat.imcadr.util.DurationUtils;
import imcacat.imcadr.util.GeometryUtils;
import imcacat.imcadr.util.SimpleStatistics;

import java.awt.Color;
import java.io.PrintWriter;

/**
 * I print a <code>Rasterizer2d</code> to a file as HTML.
 */
public class Rasterizer2dPrinter extends BaseRasterizerPrinter {
  protected static final double MIN_CELL_WIDTH_IN_EM = 2.5;
  protected static final double MIN_CELL_HEIGHT_IN_EM = 1.5;

  public Rasterizer2dPrinter(PrintWriter writer) {
    super(writer);
  }

  public void print(Rasterizer2d rasterizer) {
    printHeader();
    printWarningsIfAny(rasterizer);
    printRasterizer(rasterizer);
    printDivider();
    printDiagnostics(rasterizer);
    printFooter();
  }

  public void printWarningsIfAny(Rasterizer2d rasterizer) {
    if (rasterizer.wasSuccessful() && !rasterizer.getRaster().isClipped()) {
      return;
    }
    printWarningHeader();
    if (!rasterizer.wasSuccessful()) {
      printUnsuccessfulWarning(rasterizer);
    }
    if (rasterizer.getRaster().isClipped()) {
      printClippedWarning(rasterizer.getRaster());
    }
    printWarningFooter();
  }

  protected void printClippedWarning(Raster2d r) {
    writer.print("<li>");
    writer.format("Note: %s raster was clipped from %dx%d cells to %dx%d cells.", escape(r
        .getPlane().toString()), r.getSpec().getPreClipWidth(), r.getSpec().getPreClipHeight(),
        r.getSpec().getWidth(), r.getSpec().getHeight());
    writer.println("</li>");
  }

  protected void printRasterizer(Rasterizer2d rasterizer) {
    printRasterizerHeading(rasterizer);
    printRaster(rasterizer.getRaster());
    printSolution(rasterizer);
  }

  protected void printRasterizerHeading(Rasterizer2d rasterizer) {
    writer.println();
    writer.format("<h1>%s Raster</h1>%n", escape(rasterizer.getPlane().toString()));
  }

  protected void printRaster(Raster2d raster) {
    int numRows = raster.rowCount();
    int numColumns = raster.columnCount();
    double rowHeightPercent = 100.0 / (double)numRows;
    double columnWidthPercent = 100.0 / (double)numColumns;
    Rectangle cellSizeInEm =
        GeometryUtils.scale(raster.getCellWidth(), raster.getCellHeight(), MIN_CELL_WIDTH_IN_EM,
            MIN_CELL_HEIGHT_IN_EM);
    writer.println();
    writer
        .format("<table class=\"Raster\" " + "style=\"width: %.1fem; height: %.1fem\">%n",
            cellSizeInEm.getWidth() * (double)numColumns, cellSizeInEm.getHeight() *
                (double)numRows);
    for (int r = 0; r < numRows; r++) {
      writer.println("  <tr>");
      for (int c = 0; c < numColumns; c++) {
        RasterCell cell = raster.getCell(c, r);
        Integer score = cell.getScore();
        Double normalizedScore = cell.getNormalizedScore();
        if (normalizedScore == null) normalizedScore = RasterCell.MIN_NORMALIZED_SCORE;
        Color background =
            (score == null) ? BACKGROUND_COLOR : ColorUtils.getTemperature(normalizedScore,
                RasterCell.MIN_NORMALIZED_SCORE, RasterCell.MAX_NORMALIZED_SCORE);
        writer.format("    <td title=\"Cell: (%d, %d) Position: (%s)\"%n", cell.getColumn() + 1,
            cell.getRow() + 1, escape(cell.getPosition().toDisplayString(3)));
        writer.format("        style=\"background: #%02x%02x%02x; "
            + "width: %.1f%%; height: %.1f%%\">%s</td>%n", background.getRed(),
            background.getGreen(), background.getBlue(), columnWidthPercent, rowHeightPercent,
            (score == null) ? "" : score.toString());
      }
      writer.println("  </tr>");
    }
    writer.println("</table>");
  }

  protected void printSolution(Rasterizer2d rasterizer) {
    Raster2dPlane plane = rasterizer.getPlane();
    writer.println();
    writer.println("<table>");
    writer.println("  <tr>");
    writer.format("    <td>Center (%s, %s):</td>%n", escape(plane.getFirstAxisName()),
        escape(plane.getSecondAxisName()));
    writer.format("    <td>%s</td>%n", escape(rasterizer.getBestPositionString()));
    writer.println("  </tr>");
    writer.println("  <tr>");
    writer.println("    <td>Time (total, acquire, score):</td>");
    writer.format("    <td>%s</td>%n", escape(rasterizer.getTimeString()));
    writer.println("  </tr>");
    writer.println("</table>");
  }

  protected void printDiagnostics(Rasterizer2d rasterizer) {
    printDiagnosticsHeader();
    printDiagnosticsFor(rasterizer);
    printDiagnosticsFooter();
  }

  protected void printDiagnosticsFor(Rasterizer2d rasterizer) {
    Raster2dPlane plane = rasterizer.getPlane();
    String escapedPlaneName = escape(plane.getFirstAxisName() + plane.getSecondAxisName());
    SimpleStatistics s = SimpleStatistics.compute(rasterizer.getCellAnalyzeImageWaitTimes());
    /* @formatter:off */
    writer.format("%s raster cell score image-wait time (N, min, max, mean, stddev): " +
          "%d, %s, %s, %s, %s<br>%n",
        escapedPlaneName,
        s.count(),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.min()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.max()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.mean()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.stddev()))));
    /* @formatter:on */
    s = SimpleStatistics.compute(rasterizer.getCellAnalyzeScoreTimes());
    /* @formatter:off */
    writer.format("%s raster cell score compute time (N, min, max, mean, stddev): " +
          "%d, %s, %s, %s, %s<br>%n",
        escapedPlaneName,
        s.count(),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.min()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.max()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.mean()))),
        escape(DurationUtils.createTimeString((long)Math.ceil(s.stddev()))));
    /* @formatter:on */
  }
}
