package imcacat.imcadr;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import imcacat.imcadr.util.DurationUtils;
import imcacat.imcadr.util.MathUtils;

/**
 * I rasterize a three-dimensional raster.
 */
public class Rasterizer3d extends BaseRasterizer {
  private static final long serialVersionUID = 1;

  public static final String PROPNAME_ZY_RASTERIZER = "zyRasterizer";
  public static final String PROPNAME_ZX_RASTERIZER = "zxRasterizer";
  public static final String PROPNAME_BEST_POSITION = "bestPosition";
  public static final String PROPNAME_BEST_POSITION_STRING = "bestPositionString";
  public static final String PROPNAME_TIME_STRING = "timeString";

  protected Raster3dSpec spec;
  protected Rasterizer2d zyRasterizer;
  protected Rasterizer2d zxRasterizer;
  protected BeamlineDriver driver;
  protected Position3d bestPosition;
  protected String bestPositionString = "";
  protected String timeString = "";
  protected Range<Integer> zyProgressRange;
  protected Range<Integer> zxProgressRange;

  public Rasterizer3d(Raster3dSpec spec, BeamlineDriver driver,
      boolean firePropertyChangesOnEdtEnabled) {
    this(spec, driver, new Rasterizer2d(spec.getZySpec(), driver,
        firePropertyChangesOnEdtEnabled), new Rasterizer2d(spec.getZxSpec(), driver,
        firePropertyChangesOnEdtEnabled), firePropertyChangesOnEdtEnabled);
  }

  protected Rasterizer3d(Raster3dSpec spec, BeamlineDriver driver, Rasterizer2d zyRasterizer,
      Rasterizer2d zxRasterizer, boolean firePropertyChangesOnEdtEnabled) {
    super(firePropertyChangesOnEdtEnabled);
    this.spec = spec;
    this.driver = driver;
    this.zyRasterizer = zyRasterizer;
    zyRasterizer.addPropertyChangeListener(Rasterizer2d.PROPNAME_PROGRESS,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            zyRasterizerProgressChanged((Integer)event.getNewValue());
          }
        });
    this.zxRasterizer = zxRasterizer;
    zxRasterizer.addPropertyChangeListener(Rasterizer2d.PROPNAME_PROGRESS,
        new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent event) {
            zxRasterizerProgressChanged((Integer)event.getNewValue());
          }
        });
    bestPosition = new Position3d(LengthUnit.MILLIMETER);
  }

  @Override
  protected void go() throws Exception {
    long startTime = System.nanoTime();

    setStatus("Configuring beamline for rasterizations");
    driver.enterCollectDataExperimentState();
    driver.disableAutobp();
    setProgress(1);

    List<Raster2dPlane> planes = new ArrayList<Raster2dPlane>(2);
    planes.add(spec.isZxFirst() ? Raster2dPlane.ZX : Raster2dPlane.ZY);
    planes.add(spec.isZxFirst() ? Raster2dPlane.ZY : Raster2dPlane.ZX);
    List<Range<Integer>> ranges = new ArrayList<Range<Integer>>(2);
    ranges.add(new Range<Integer>(1, 49));
    ranges.add(new Range<Integer>(50, 98));
    for (int i = 0; i < planes.size(); i++) {
      boolean ok = rasterize(planes.get(i), ranges.get(i));
      if (!ok) {
        Rasterizer2d r = getRasterizer(planes.get(i));
        if (r.hasException()) {
          throw toException(r.getException());
        } else {
          throw new UnsuccessfulRasterizeException(r);
        }
      }
      setProgress(ranges.get(i).high() + 1);
    }

    Integer zyBestScore = zyRasterizer.getScoreAtBestPosition();
    if (zyBestScore == null) zyBestScore = Integer.valueOf(Integer.MIN_VALUE);
    Integer zxBestScore = zxRasterizer.getScoreAtBestPosition();
    if (zxBestScore == null) zxBestScore = Integer.valueOf(Integer.MIN_VALUE);

    setStatus("Merging rasterization results");
    setBestPosition(new Position3d(zxRasterizer.getBestVerticalPosition(),
        zyRasterizer.getBestVerticalPosition(), ((zyBestScore > zxBestScore) ? zyRasterizer
            : zxRasterizer).getBestHorizontalPosition(), bestPosition.getUnit()));
    setAcquireTime(zyRasterizer.getAcquireTime() + zxRasterizer.getAcquireTime());
    setAnalyzeTime(zyRasterizer.getAnalyzeTime() + zxRasterizer.getAnalyzeTime());

    if (spec.isZxFirst()) {
      if (zxBestScore > zyBestScore) {
        setStatus("Moving to optimal Z position from ZX raster");
        driver.moveZ(zxRasterizer.getBestHorizontalPosition());
      }
    } else {
      if (zyBestScore > zxBestScore) {
        setStatus("Moving to optimal Z position from ZY raster");
        driver.moveZ(zyRasterizer.getBestHorizontalPosition());
      }
    }

    setStatus("Configuring beamline for use");
    driver.enterIdleExperimentState();
    driver.enableAutobp();

    setWallTime(nanosToMillis(System.nanoTime() - startTime));

    setStatus(SUCCESSFUL_COMPLETION_STATUS);
    setProgress(100);
  }

  protected boolean rasterize(Raster2dPlane plane, Range<Integer> progressRange) {
    setStatus("Rasterizing " + plane + " plane");
    setProgressRange(plane, progressRange);
    final Rasterizer2d rasterizer = getRasterizer(plane);
    rasterizer.doNotLogException();
    rasterizer.doNotSetExperimentState();
    rasterizer.doNotEnableNorDisableAutobp();
    rasterizer.doNotWriteResultsFile();
    rasterizer.run();
    boolean ok = rasterizer.wasSuccessful();
    if (!ok) setStatus(rasterizer.getStatus());
    return ok;
  }

  @Override
  protected void cleanUp() {
    boolean interrupted = false;
    Exception caughtException = null;

    try {
      driver.enterIdleExperimentState();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }
    try {
      driver.enableAutobp();
    } catch (InterruptedException e) {
      interrupted = true;
      if (caughtException == null) caughtException = e;
    } catch (DriverException | RuntimeException e) {
      if (caughtException == null) caughtException = e;
    }

    if (interrupted) Thread.currentThread().interrupt();
    if (caughtException != null) throw new RuntimeException(caughtException);
  }

  public boolean isZxFirst() {
    return spec.isZxFirst();
  }

  public Rasterizer2d getZyRasterizer() {
    return zyRasterizer;
  }

  public Rasterizer2d getZxRasterizer() {
    return zxRasterizer;
  }

  public Rasterizer2d getRasterizer(Raster2dPlane plane) {
    switch (plane) {
      case ZY:
        return getZyRasterizer();
      case ZX:
        return getZxRasterizer();
      default:
        throw new BugException();
    }
  }

  public Position3d getBestPosition() {
    return bestPosition;
  }

  protected void setBestPosition(Position3d newBestPosition) {
    if (bestPosition == null && newBestPosition == null) return;
    if (bestPosition != null && bestPosition.equals(newBestPosition)) return;
    Position3d oldBestPosition = bestPosition;
    bestPosition = newBestPosition;
    notifyChanged(PROPNAME_BEST_POSITION, oldBestPosition, newBestPosition);
    setBestPositionString(newBestPosition.toDisplayString(3));
  }

  @Override
  protected void setAcquireTime(long newAcquireTime) {
    super.setAcquireTime(newAcquireTime);
    updateTimeString();
  }

  @Override
  protected void setAnalyzeTime(long newScoreTime) {
    super.setAnalyzeTime(newScoreTime);
    updateTimeString();
  }

  @Override
  protected void setWallTime(long newWallTime) {
    super.setWallTime(newWallTime);
    updateTimeString();
  }

  protected void updateTimeString() {
    setTimeString(createTimeString());
  }

  protected String createTimeString() {
    return DurationUtils.createTimeString(getWallTime()) + ", " +
        DurationUtils.createTimeString(getAcquireTime()) + ", " +
        DurationUtils.createTimeString(getAnalyzeTime());
  }

  public String getBestPositionString() {
    return bestPositionString;
  }

  protected void setBestPositionString(String newBestPositionString) {
    if (bestPositionString.equals(newBestPositionString)) return;
    String oldBestPositionString = bestPositionString;
    bestPositionString = newBestPositionString;
    notifyChanged(PROPNAME_BEST_POSITION_STRING, oldBestPositionString, newBestPositionString);
  }

  public String getTimeString() {
    return timeString;
  }

  protected void setTimeString(String newTimeString) {
    if (timeString.equals(newTimeString)) return;
    String oldTimeString = timeString;
    timeString = newTimeString;
    notifyChanged(PROPNAME_TIME_STRING, oldTimeString, newTimeString);
  }

  protected synchronized Range<Integer> getZyProgressRange() {
    return zyProgressRange;
  }

  protected synchronized void setZyProgressRange(Range<Integer> r) {
    zyProgressRange = r;
  }

  protected synchronized void setZyProgressRange(int low, int high) {
    setZyProgressRange(new Range<Integer>(low, high));
  }

  protected synchronized Range<Integer> getZxProgressRange() {
    return zxProgressRange;
  }

  protected synchronized void setZxProgressRange(Range<Integer> r) {
    zxProgressRange = r;
  }

  protected synchronized void setZxProgressRange(int low, int high) {
    setZxProgressRange(new Range<Integer>(low, high));
  }

  protected synchronized void setProgressRange(Raster2dPlane plane, Range<Integer> r) {
    switch (plane) {
      case ZY:
        setZyProgressRange(r);
        break;
      case ZX:
        setZxProgressRange(r);
        break;
      default:
        throw new BugException();
    }
  }

  protected synchronized void setProgressRange(Raster2dPlane plane, int low, int high) {
    setProgressRange(plane, new Range<Integer>(low, high));
  }

  @Override
  protected void writeResults(PrintWriter writer) {
    (new Rasterizer3dPrinter(writer)).print(this);
  }

  @Override
  protected File createResultsFile() {
    return new File(spec.getImagesDirectory(), "imcadr-result.html");
  }

  protected void zyRasterizerProgressChanged(Integer zyProgress) {
    if (zyProgress == null) return;
    Range<Integer> range = getZyProgressRange();
    if (range == null) return;
    int oldP = getProgress();
    int newP = MathUtils.scale(zyProgress, 0, 100, range.low(), range.high());
    if (newP <= oldP) return;
    setProgress(newP);
  }

  protected void zxRasterizerProgressChanged(Integer zxProgress) {
    if (zxProgress == null) return;
    Range<Integer> range = getZxProgressRange();
    if (range == null) return;
    int oldP = getProgress();
    int newP = MathUtils.scale(zxProgress, 0, 100, range.low(), range.high());
    if (newP <= oldP) return;
    setProgress(newP);
  }

  private static Exception toException(Throwable t) {
    return (t instanceof Exception) ? (Exception)t : new Exception(t);
  }
}
