package imcacat.imcadr;

/**
 * I represent a rectangle.
 */
public class Rectangle {
  protected double width;
  protected double height;
  protected LengthUnit unit;

  public Rectangle() {
    this(LengthUnit.ARBITRARY);
  }

  public Rectangle(LengthUnit unit) {
    this(0.0, 0.0, unit);
  }

  public Rectangle(double width, double height) {
    this(width, height, LengthUnit.ARBITRARY);
  }

  public Rectangle(double width, double height, LengthUnit unit) {
    this.width = width;
    this.height = height;
    this.unit = unit;
  }

  /**
   * Determines whether I am equal to the specified <tt>Object</tt> instance.
   * <p>
   * Does not attempt to convert to the same units; that is, if my unit is
   * millimeters and the unit of the <tt>Object</tt> instance is micrometers,
   * we are not considered equal even if our values would be equal after a
   * conversion to a common unit.
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) return true;
    if (!(obj instanceof Rectangle)) return false;
    Rectangle r = (Rectangle)obj;
    if (Double.compare(width, r.width) != 0) return false;
    if (Double.compare(height, r.height) != 0) return false;
    return unit.equals(r.unit);
  }

  @Override
  public int hashCode() {
    int result = 17;
    long value = Double.doubleToLongBits(width);
    result = 31 * result + (int)(value ^ (value >>> 32));
    value = Double.doubleToLongBits(height);
    result = 31 * result + (int)(value ^ (value >>> 32));
    result = 31 * result + unit.hashCode();
    return result;
  }

  public double getWidth() {
    return width;
  }

  public double getHeight() {
    return height;
  }

  public LengthUnit getUnit() {
    return unit;
  }

  public String toString() {
    return String.format("%.1f x %.1f with unit %s", width, height, unit);
  }

  public String toDisplayString() {
    return toDisplayString(1);
  }

  public String toDisplayString(int precision) {
    String unitAbbreviation = unit.getAbbreviation();
    if (unitAbbreviation.length() != 0) unitAbbreviation = " " + unitAbbreviation;
    String lengthSpecifier = "%." + precision + "f";
    return String.format(lengthSpecifier + "%s x " + lengthSpecifier + "%s", width,
        unitAbbreviation, height, unitAbbreviation);
  }

  public Rectangle convertTo(LengthUnit newUnit) {
    if (unit.equals(newUnit)) return this;
    Double ratioObject = LengthUnit.conversionRatio(unit, newUnit);
    if (ratioObject == null) {
      throw new RuntimeException("Do not know how to convert from " + unit + " to " + newUnit);
    }
    double ratio = ratioObject.doubleValue();
    return new Rectangle(width * ratio, height * ratio, newUnit);
  }

  public static Rectangle mm(double width, double height) {
    return new Rectangle(width, height, LengthUnit.MILLIMETER);
  }

  public static Rectangle um(double width, double height) {
    return new Rectangle(width, height, LengthUnit.MICROMETER);
  }
}
