package imcacat.imcadr;

import imcacat.imcadr.util.Version;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.channels.ClosedByInterruptException;
import java.util.EnumMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.ImageIcon;

/**
 * I provide resources (e.g. icons, PV names, properties) for this application.
 */
public class Resources {
  private static final Logger LOG = Logger.getLogger(Resources.class.getName());

  private static final EnumMap<AnalysisPackage, String> ANALYSIS_PACKAGE_TO_KEY_MAP;

  protected static Resources INSTANCE;
  protected static String PROPERTIES_RESOURCE_KEY = "imcadr.configuration";

  static {
    ANALYSIS_PACKAGE_TO_KEY_MAP = new EnumMap<AnalysisPackage, String>(AnalysisPackage.class);
    for (AnalysisPackage each : AnalysisPackage.values()) {
      switch (each) {
        case LABELIT:
          ANALYSIS_PACKAGE_TO_KEY_MAP.put(each, "labelit");
          break;
        case DTREK:
          ANALYSIS_PACKAGE_TO_KEY_MAP.put(each, "dtrek");
          break;
      }
    }
  }

  protected Properties properties;

  protected Map<Operation, String> operationToDisplayStringMap;

  protected Resources() {
    super();

    properties = new Properties();

    String defaultPropertiesResource = getDefaultPropertiesResourceName();
    try {
      properties.load(Resources.class.getResourceAsStream(defaultPropertiesResource));
    } catch (InterruptedIOException | ClosedByInterruptException e) {
      Thread.currentThread().interrupt();
      throw new RuntimeException("Interrupted", e);
    } catch (IOException e) {
      LOG.log(Level.SEVERE,
          "Failed to load default properties resource \"" + defaultPropertiesResource + "\"", e);
      throw new RuntimeException("Failed to load default properties " + "resource \"" +
          defaultPropertiesResource + "\"", e);
    }

    String propertiesResource = System.getProperty(PROPERTIES_RESOURCE_KEY);
    if (propertiesResource != null) {
      try {
        properties.load((new URI(propertiesResource)).toURL().openStream());
      } catch (InterruptedIOException | ClosedByInterruptException e) {
        Thread.currentThread().interrupt();
        throw new RuntimeException("Interrupted", e);
      } catch (IOException | URISyntaxException e) {
        LOG.log(Level.SEVERE, "Failed to load properties resource \"" + propertiesResource + "\"",
            e);
        throw new RuntimeException("Failed to load properties resource \"" + propertiesResource +
            "\"", e);
      }
    }

    operationToDisplayStringMap = new EnumMap<Operation, String>(Operation.class);
    operationToDisplayStringMap.put(Operation.IDLE, "Idle");
    operationToDisplayStringMap.put(Operation.RASTERIZE, "Rasterize");
    for (Operation each : Operation.values())
      if (!operationToDisplayStringMap.containsKey(each)) throw new BugException(
          "Missing display string mapping for " + each + " Operation");
  }

  protected String getDefaultPropertiesResourceName() {
    return "/imcadr.properties";
  }

  public String getDisplayString(Operation operation) {
    return operationToDisplayStringMap.get(operation);
  }

  public File getDTrekDtrankerHelper() {
    return new File(properties.getProperty("dtrekDtrankerHelperPath"));
  }

  public File getLabelitDistlSignalStrengthHelper() {
    return new File(properties.getProperty("labelitDistlSignalStrengthHelperPath"));
  }

  public File getSshHelper() {
    return new File(properties.getProperty("sshHelperPath"));
  }

  public File getSystemConfigurationFile() {
    return new File(properties.getProperty("sysConfFilePath"));
  }

  public String getAutobpRequestExternalPausePvName() {
    return properties.getProperty("autobpRequestExternalPausePvName");
  }

  public short getAutobpRequestExternalPauseValue() {
    return Short.parseShort(properties.getProperty("autobpRequestExternalPauseValue"));
  }

  public String getAutobpReleaseExternalPausePvName() {
    return properties.getProperty("autobpReleaseExternalPausePvName");
  }

  public short getAutobpReleaseExternalPauseValue() {
    return Short.parseShort(properties.getProperty("autobpReleaseExternalPauseValue"));
  }

  public String getExperimentStatePvName() {
    return properties.getProperty("experimentStatePvName");
  }

  public String getExperimentStateRbvPvName() {
    return properties.getProperty("experimentStateRbvPvName");
  }

  public String getExperimentStateSetPvName() {
    return properties.getProperty("experimentStateSetPvName");
  }

  public String getExperimentStateStatusPvName() {
    return properties.getProperty("experimentStateStatusPvName");
  }

  public String getExperimentStateAbortPvName() {
    return properties.getProperty("experimentStateAbortPvName");
  }

  public String getAuxiliaryImageFilePathPvName() {
    return properties.getProperty("auxiliaryImageFilePathPvName");
  }

  public String getAreaDetectorPvName() {
    return properties.getProperty("areaDetectorPvName");
  }

  public String getExposyncPhiPvName() {
    return properties.getProperty("exposyncPhiPvName");
  }

  public String getExposyncXPvName() {
    return properties.getProperty("exposyncXPvName");
  }

  public String getExposyncYPvName() {
    return properties.getProperty("exposyncYPvName");
  }

  public String getExposyncZPvName() {
    return properties.getProperty("exposyncZPvName");
  }

  public String getExposureShutterRbvPvName() {
    return properties.getProperty("exposureShutterRbvPvName");
  }

  /**
   * Gets the exposure shutter latency. The latency is the time from when the
   * exposure shutter is commanded to open or close and when it actually opens
   * or closes.
   *
   * @return the exposure shutter latency in seconds
   */
  public double getExposureShutterLatency() {
    return Double.parseDouble(properties.getProperty("exposureShutterLatency"));
  }

  /**
   * Gets the detector latency. The latency is the time from when the detector
   * is commanded to start the exposure and when it actually starts the
   * exposure.
   *
   * @return the detector latency in seconds
   */
  public double getDetectorLatency() {
    return Double.parseDouble(properties.getProperty("detectorLatency"));
  }

  public String getPhiMotorPvName() {
    return properties.getProperty("phiMotorPvName");
  }

  public String getXMotorPvName() {
    return properties.getProperty("xMotorPvName");
  }

  public String getYMotorPvName() {
    return properties.getProperty("yMotorPvName");
  }

  public String getZMotorPvName() {
    return properties.getProperty("zMotorPvName");
  }

  /**
   * Gets the Phi motor's maximum position error. The maximum position error is
   * the maximum magnitude of the difference of the actual position and the
   * commanded position that will be tolerated for a move; anything greater
   * than this is considered a fault.
   *
   * @return the maximum position error in degrees or {@code null} if no
   *         maximum is set
   */
  public Double getPhiMotorPositionErrorMax() {
    String sValue = properties.getProperty("phiMotorPosErrorMax");
    if (sValue == null) return null;
    sValue = sValue.trim();
    if (sValue.isEmpty()) return null;
    return Double.valueOf(sValue);
  }

  /**
   * Gets the X motor's maximum position error. The maximum position error is
   * the maximum magnitude of the difference of the actual position and the
   * commanded position that will be tolerated for a move; anything greater
   * than this is considered a fault.
   *
   * @return the maximum position error in millimeters or {@code null} if no
   *         maximum is set
   */
  public Double getXMotorPositionErrorMax() {
    String sValue = properties.getProperty("xMotorPosErrorMax");
    if (sValue == null) return null;
    sValue = sValue.trim();
    if (sValue.isEmpty()) return null;
    return Double.valueOf(sValue);
  }

  /**
   * Gets the Y motor's maximum position error. The maximum position error is
   * the maximum magnitude of the difference of the actual position and the
   * commanded position that will be tolerated for a move; anything greater
   * than this is considered a fault.
   *
   * @return the maximum position error in millimeters or {@code null} if no
   *         maximum is set
   */
  public Double getYMotorPositionErrorMax() {
    String sValue = properties.getProperty("yMotorPosErrorMax");
    if (sValue == null) return null;
    sValue = sValue.trim();
    if (sValue.isEmpty()) return null;
    return Double.valueOf(sValue);
  }

  /**
   * Gets the Z motor's maximum position error. The maximum position error is
   * the maximum magnitude of the difference of the actual position and the
   * commanded position that will be tolerated for a move; anything greater
   * than this is considered a fault.
   *
   * @return the maximum position error in millimeters or {@code null} if no
   *         maximum is set
   */
  public Double getZMotorPositionErrorMax() {
    String sValue = properties.getProperty("zMotorPosErrorMax");
    if (sValue == null) return null;
    sValue = sValue.trim();
    if (sValue.isEmpty()) return null;
    return Double.valueOf(sValue);
  }

  /**
   * Gets the X motor's fly-scan constant-velocity settle time. The time is the
   * time to spend at constant velocity before the exposure starts, and again
   * after the exposure ends before decelerating. In other words, half of this
   * value is not spent before the exposure starts and the other half after the
   * exposure ends, rather this whole time is spent before and then this whole
   * time is spent again after.
   *
   * @return the fly-scan constant-velocity settle time in seconds
   */
  public double getXMotorFlyScanCvSettleTime() {
    return Double.parseDouble(properties.getProperty("xMotorFlyScanCvSettleTime"));
  }

  /**
   * Gets the Y motor's fly-scan constant-velocity settle time. The time is the
   * time to spend at constant velocity before the exposure starts, and again
   * after the exposure ends before decelerating. In other words, half of this
   * value is not spent before the exposure starts and the other half after the
   * exposure ends, rather this whole time is spent before and then this whole
   * time is spent again after.
   *
   * @return the fly-scan constant-velocity settle time in seconds
   */
  public double getYMotorFlyScanCvSettleTime() {
    return Double.parseDouble(properties.getProperty("yMotorFlyScanCvSettleTime"));
  }

  /**
   * Gets the Z motor's fly-scan constant-velocity settle time. The time is the
   * time to spend at constant velocity before the exposure starts, and again
   * after the exposure ends before decelerating. In other words, half of this
   * value is not spent before the exposure starts and the other half after the
   * exposure ends, rather this whole time is spent before and then this whole
   * time is spent again after.
   *
   * @return the fly-scan constant-velocity settle time in seconds
   */
  public double getZMotorFlyScanCvSettleTime() {
    return Double.parseDouble(properties.getProperty("zMotorFlyScanCvSettleTime"));
  }

  /**
   * Gets the X motor's fly-scan weight. It is a floating-point value greater
   * than or equal to 0. If the X axis is the axis to be scanned for a row or
   * column fly-scan, the row-major or column-major weight for determining a
   * row-major or column-major fly-scanning order will be multiplied by this
   * motor fly-scan weight. This can be used to prefer fly-scanning along one
   * axis over another.
   *
   * @return the X motor fly-scan weight in arbitrary units
   */
  public double getXMotorFlyScanWeight() {
    double result = Double.parseDouble(properties.getProperty("xMotorFlyScanWeight"));
    if (result >= 0.0) return result;
    throw new IllegalStateException("xMotorFlyScanWeight < 0.0");
  }

  /**
   * Gets the Y motor's fly-scan weight. It is a floating-point value greater
   * than or equal to 0. If the Y axis is the axis to be scanned for a row or
   * column fly-scan, the row-major or column-major weight for determining a
   * row-major or column-major fly-scanning order will be multiplied by this
   * motor fly-scan weight. This can be used to prefer fly-scanning along one
   * axis over another.
   *
   * @return the Y motor fly-scan weight in arbitrary units
   */
  public double getYMotorFlyScanWeight() {
    double result = Double.parseDouble(properties.getProperty("yMotorFlyScanWeight"));
    if (result >= 0.0) return result;
    throw new IllegalStateException("yMotorFlyScanWeight < 0.0");
  }

  /**
   * Gets the Z motor's fly-scan weight. It is a floating-point value greater
   * than or equal to 0. If the Z axis is the axis to be scanned for a row or
   * column fly-scan, the row-major or column-major weight for determining a
   * row-major or column-major fly-scanning order will be multiplied by this
   * motor fly-scan weight. This can be used to prefer fly-scanning along one
   * axis over another.
   *
   * @return the Z motor fly-scan weight in arbitrary units
   */
  public double getZMotorFlyScanWeight() {
    double result = Double.parseDouble(properties.getProperty("zMotorFlyScanWeight"));
    if (result >= 0.0) return result;
    throw new IllegalStateException("zMotorFlyScanWeight < 0.0");
  }

  public String getDetectorDistanceRbvPvName() {
    return properties.getProperty("detectorDistanceRbvPvName");
  }

  public String getDetector2thetaRbvPvName() {
    return properties.getProperty("detector2thetaRbvPvName");
  }

  public String getWavelengthRbvPvName() {
    return properties.getProperty("wavelengthRbvPvName");
  }

  public String getAttenuationRbvPvName() {
    return properties.getProperty("attenuationRbvPvName");
  }

  public File getRasterizerWorkingDirectory() {
    String path = properties.getProperty("rasterizerWorkingDirectory");
    if (path == null) return getUserHomeDirectory();
    path = path.replace("%u", getUsername());
    File result = new File(path);
    if (result.isDirectory()) return result;
    return getUserHomeDirectory();
  }

  public String getDisplayImageFileProgramName() {
    return properties.getProperty("displayImageFileProgramName");
  }

  public String getDisplayImageFileCommand() {
    return properties.getProperty("displayImageFileCommand");
  }

  public double getWeakDiffracterHighResolutionCutoff() {
    return Double.parseDouble(properties.getProperty("weakDiffracterHighResolutionCutoff"));
  }

  public double getHighResolutionCutoffMin() {
    return Double.parseDouble(properties.getProperty("highResolutionCutoffMin"));
  }

  public double getHighResolutionCutoffMax() {
    return Double.parseDouble(properties.getProperty("highResolutionCutoffMax"));
  }

  public double getLowResolutionCutoffMin() {
    return Double.parseDouble(properties.getProperty("lowResolutionCutoffMin"));
  }

  public double getLowResolutionCutoffMax() {
    return Double.parseDouble(properties.getProperty("lowResolutionCutoffMax"));
  }

  public int getMinimumSpotAreaMin() {
    return Integer.parseInt(properties.getProperty("minimumSpotAreaMin"));
  }

  public int getMinimumSpotAreaMax() {
    return Integer.parseInt(properties.getProperty("minimumSpotAreaMax"));
  }

  public double getMinimumSignalHeightMin() {
    return Double.parseDouble(properties.getProperty("minimumSignalHeightMin"));
  }

  public double getMinimumSignalHeightMax() {
    return Double.parseDouble(properties.getProperty("minimumSignalHeightMax"));
  }

  public Range<Integer> getScoreMinimumRange(AnalysisPackage analysisPackage) {
    return getMinimumRange(analysisPackage, "score");
  }

  public Range<Integer> getSpotTotalMinimumRange(AnalysisPackage analysisPackage) {
    return getMinimumRange(analysisPackage, "spotTotal");
  }

  public Range<Integer> getGoodBraggTotalMinimumRange(AnalysisPackage analysisPackage) {
    return getMinimumRange(analysisPackage, "goodBraggTotal");
  }

  public Range<Integer> getTotalIntegratedSignalMinimumRange(AnalysisPackage analysisPackage) {
    return getMinimumRange(analysisPackage, "totalIntegratedSignal");
  }

  private Range<Integer> getMinimumRange(AnalysisPackage analysisPackage,
      String subpropertyPrefix) {
    String prefix = ANALYSIS_PACKAGE_TO_KEY_MAP.get(analysisPackage);
    if (prefix == null) return null;
    prefix += ".";
    String lowString = properties.getProperty(prefix + subpropertyPrefix + "MinimumRangeLow");
    if (lowString == null) return null;
    String highString = properties.getProperty(prefix + subpropertyPrefix + "MinimumRangeHigh");
    if (highString == null) return null;
    return new Range<Integer>(Integer.parseInt(lowString), Integer.parseInt(highString));
  }

  public static ImageIcon getAppIcon() {
    String resourcePath = "resources/app-icon-16.png";
    URL url = Resources.class.getResource("/" + resourcePath);
    return url == null ? new ImageIcon() : new ImageIcon(url);
  }

  public static ImageIcon getLargeAppIcon() {
    String resourcePath = "resources/app-icon-48.png";
    URL url = Resources.class.getResource("/" + resourcePath);
    return (url == null) ? new ImageIcon() : new ImageIcon(url);
  }

  public static String getApplicationName() {
    return "Imcadr";
  }

  public static String getApplicationLongName() {
    return "IMCA-CAT Diffraction Rasterizer";
  }

  public static String getApplicationVersion() {
    return Version.getId();
  }

  public static boolean isPlatformMacOsX() {
    return System.getProperty("os.name").toLowerCase().startsWith("mac os x");
  }

  public static boolean isPlatformWindows() {
    return System.getProperty("os.name").toLowerCase().startsWith("windows");
  }

  public static String getUsername() {
    return System.getProperty("user.name");
  }

  public static File getUserHomeDirectory() {
    return new File(System.getProperty("user.home"));
  }

  public static File getCurrentWorkingDirectory() {
    return new File(System.getProperty("user.dir"));
  }

  public static String getLineSeparator() {
    return System.getProperty("line.separator");
  }

  /**
   * Gets an instance of <code>Resources</code>.
   */
  public static synchronized Resources getInstance() {
    if (INSTANCE == null) INSTANCE = new Resources();
    return INSTANCE;
  }
}
