package imcacat.imcadr;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;

/**
 * I am a two state button <code>Action</code>.
 * <p>
 * Say, for example, there is a GUI for moving a motor with a button that says
 * "Start". When you click the button, its label changes to "Stop". If you
 * click the button again while it says "Stop", the motor will stop and its
 * label will change to "Start" again. This two state button behavior is what I
 * provide.
 */
public class TwoStateButtonAction extends AbstractAction {
  private static final long serialVersionUID = 1;

  public static final int STATE_1 = 1;
  public static final int STATE_2 = 2;

  public static final String STATE_1_COMMAND = "1";
  public static final String STATE_2_COMMAND = "2";

  protected String state1Label;
  protected String state2Label;
  protected boolean onePressPerStateEnabled;
  protected List<ActionListener> actionListeners;
  protected String lastCommandPerformed;

  public TwoStateButtonAction(String state1Label, String state2Label) {
    this(state1Label, state2Label, true);
  }

  public TwoStateButtonAction(String state1Label, String state2Label,
      boolean onePressPerStateEnabled) {
    super();
    this.state1Label = state1Label;
    this.state2Label = state2Label;
    this.onePressPerStateEnabled = onePressPerStateEnabled;
    actionListeners = new ArrayList<ActionListener>();
    setStateTo1();
  }

  public void actionPerformed(ActionEvent event) {
    String command = event.getActionCommand();
    if (onePressPerStateEnabled && lastCommandPerformed.equals(command)) return;
    lastCommandPerformed = command;
    notifyActionListeners(event);
  }

  protected void basicSetStateTo1() {
    putValue(Action.ACTION_COMMAND_KEY, STATE_1_COMMAND);
    putValue(Action.NAME, state1Label);
  }

  /**
   * Sets my state to 1.
   */
  public void setStateTo1() {
    basicSetStateTo1();
    lastCommandPerformed = STATE_2_COMMAND;
  }

  protected void basicSetStateTo2() {
    putValue(Action.ACTION_COMMAND_KEY, STATE_2_COMMAND);
    putValue(Action.NAME, state2Label);
  }

  /**
   * Sets my state to 2.
   */
  public void setStateTo2() {
    basicSetStateTo2();
    lastCommandPerformed = STATE_1_COMMAND;
  }

  protected synchronized void notifyActionListeners(ActionEvent event) {
    for (ActionListener each : actionListeners)
      each.actionPerformed(event);
  }

  public synchronized void addActionListener(ActionListener listener) {
    actionListeners.add(listener);
  }

  public synchronized void removeActionListener(ActionListener listener) {
    if (listener == null) return;
    actionListeners.remove(listener);
  }
}
