package imcacat.imcadr.util;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileAttribute;
import java.nio.file.attribute.PosixFilePermissions;

/**
 * I provide utilities related to files and directories.
 */
public class FileUtils {
  private FileUtils() {
    throw new AssertionError();
  }

  /**
   * Creates a temporary directory in the default location with the the
   * specified prefix and with POSIX mode <tt>0700</tt>.
   *
   * Invokes {@link Files#createTempDirectory(String, FileAttribute...)}, and
   * may throw any of the exceptions it may throw.
   *
   * @param prefix prefix for temporary directory name; may be {@code null}
   *
   * @return path to the newly created temporary directory
   */
  public static Path createTempDir(String prefix) throws IOException {
    return Files.createTempDirectory(prefix,
        PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rwx------")));
  }

  /**
   * Deletes the specified file or directory. Symbolic links are not followed,
   * and if the path refers to a directory, the directory will be deleted
   * recursively (i.e. the entire tree will be deleted).
   *
   * @param path file or directory to delete
   */
  public static void delete(Path path) throws IOException {
    if (path == null) return;
    Files.walkFileTree(path, new DeleteFileVisitor());
  }

  /**
   * A {@code FileVisitor} that deletes a file or recursively deletes a
   * directory.
   */
  private static class DeleteFileVisitor extends SimpleFileVisitor<Path> {
    @Override
    public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
      Files.delete(file);
      return FileVisitResult.CONTINUE;
    }

    @Override
    public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
      if (exc != null) throw exc;
      Files.delete(dir);
      return FileVisitResult.CONTINUE;
    }
  }
}
