package imcacat.imcadr.util;

import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * I provide utility methods related to the Unix Bourne shell.
 */
public class ShellUtils {
  /* @formatter:off */
  private static final Pattern INTERPOLATION = Pattern.compile(
      "(?>" +
        "\\\\(?<escaped>.)|" +
        "\\$" +
          "(?>" +
            "(?<bareVarName>[a-zA-Z_][a-zA-Z0-9_]*+)|" +
            "\\{(?<braceVarName>[a-zA-Z_][a-zA-Z0-9_]*+)\\}" +
          ")" +
      ")",
      Pattern.DOTALL);
  /* @formatter:on */

  private ShellUtils() {
    throw new AssertionError();
  }

  /**
   * Performs string interpolation by expanding the variables in the specified
   * input with their values from the specified mapping.
   * <p>
   * Variables are of the Bourne shell form: <code>$&lt;var_name&gt;</code> or
   * <code>${&lt;var_name&gt;}</code>. For example: <code>$color</code> or
   * <code>${color}</code>. Variable names are composed of ASCII letters,
   * digits, and underscores ([<code>a-zA-Z0-9_</code>]) but must not start
   * with a digit. Any character, include the dollar sign and the backslash,
   * can be escaped by preceding it with a backslash. Since the input is a
   * string, not Bourne shell code, single quotes and double quotes have no
   * special meaning that changes the behavior of the interpolation.
   * <p>
   * The keys of the mapping are <code>&lt;var_name&gt;</code>. For example, if
   * the input string was "<code>Favorite color: $color</code>", the key for
   * the variable would need to be "<code>color</code>" (i.e., without the
   * dollar sign). If no mapping is provided for a variable, the variable is
   * expanded to the empty string.
   *
   * @param input the string to be interpolated
   * @param varNamesToValues the mapping of variable names to their values
   *
   * @return the expansion of the specified input
   *
   * @throws NullPointerException if either {@code input} or
   *           {@code varNamesToValues} is {@code null}
   */
  public static String expand(String input, Map<String, String> varNamesToValues) {
    /*
     * Use StringBuilder instead of StringBuffer if increase minimum Java
     * version to 9 where Matcher has methods that accept a StringBuilder.  Or
     * if performance is an issue, the equivalent of the append-and-replace
     * loop can be performed with a StringBuilder and existing Matcher methods.
     */
    if (input.isEmpty()) return input;
    StringBuffer result = new StringBuffer(input.length());
    Matcher m = INTERPOLATION.matcher(input);
    while (m.find()) {
      String escaped = m.group("escaped");
      if (escaped != null) {
        m.appendReplacement(result, Matcher.quoteReplacement(escaped));
        continue;
      }
      String varName = m.group("bareVarName");
      if (varName == null) varName = m.group("braceVarName");
      String value = varNamesToValues.get(varName);
      m.appendReplacement(result, (value == null) ? "" : Matcher.quoteReplacement(value));
    }
    m.appendTail(result);
    return result.toString();
  }

  /**
   * Returns {@code true} if the specified input contains the specified
   * variable. The rules for the input and variable are the same as those
   * detailed in {@link #expand(String, Map)}.
   *
   * @param input the string in which to search for the variable
   * @param varName the name of the variable (excluding the dollar sign) to
   *          find
   *
   * @return {@code true} if the variable occurs in the input; {@code false}
   *         otherwise
   *
   * @throws NullPointerException if either {@code input} or {@code varName} is
   *           {@code null}
   */
  public static boolean containsVariable(String input, String varName) {
    if (input.isEmpty()) return false;
    Matcher m = INTERPOLATION.matcher(input);
    while (m.find()) {
      if (m.group("escaped") != null) continue;
      if (varName.equals(m.group("bareVarName"))) return true;
      if (varName.equals(m.group("braceVarName"))) return true;
    }
    return false;
  }
}
