package imcacat.imcadr.util;

import java.util.List;

/**
 * I represent statistics of a collection of numbers.
 */
public class SimpleStatistics {
  private final int count;
  private final double min;
  private final double max;
  private final double sum;
  private final double mean;
  private final double stddev;

  public SimpleStatistics() {
    this(0, 0.0, 0.0, 0.0, 0.0, 0.0);
  }

  public SimpleStatistics(int count, double min, double max, double sum, double mean,
      double stddev) {
    this.count = count;
    this.min = min;
    this.max = max;
    this.sum = sum;
    this.mean = mean;
    this.stddev = stddev;
  }

  public int count() {
    return count;
  }

  public double min() {
    return min;
  }

  public double max() {
    return max;
  }

  public double sum() {
    return sum;
  }

  public double mean() {
    return mean;
  }

  public double stddev() {
    return stddev;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof SimpleStatistics)) return false;
    SimpleStatistics s = (SimpleStatistics)o;
    if (count != s.count) return false;
    if (min != s.min) return false;
    if (max != s.max) return false;
    if (sum != s.sum) return false;
    if (mean != s.mean) return false;
    if (stddev != s.stddev) return false;
    return true;
  }

  @Override
  public int hashCode() {
    int result = 17;
    result = 31 * result + count;
    long minBits = Double.doubleToLongBits(min);
    result = 31 * result + (int)(minBits ^ (minBits >>> 32));
    long maxBits = Double.doubleToLongBits(max);
    result = 31 * result + (int)(maxBits ^ (maxBits >>> 32));
    long sumBits = Double.doubleToLongBits(sum);
    result = 31 * result + (int)(sumBits ^ (sumBits >>> 32));
    long meanBits = Double.doubleToLongBits(mean);
    result = 31 * result + (int)(meanBits ^ (meanBits >>> 32));
    long stddevBits = Double.doubleToLongBits(stddev);
    result = 31 * result + (int)(stddevBits ^ (stddevBits >>> 32));
    return result;
  }

  @Override
  public String toString() {
    return "count=" + count + ", min=" + min + ", max=" + max + ", sum=" + sum + ", mean=" + mean +
        ", stddev=" + stddev;
  }

  /*
   * Since the statistics to be computed include the sum, the algorithms
   * used here of course compute the sum and use it.  This will obviously
   * only be suitable for values that will not overflow the sum (i.e. too
   * many or too big).  If the sum can be omitted, then an iterative mean
   * algorithm could be used to compute the mean without computing the sum.
   * Similarly for computing the variance.  One such algorithm is given at
   * <http://stackoverflow.com/a/1934266> which, according to a comment by
   * user "j.p.", "is numerically stable and recommended in Knuth, The Art of
   * Computer Programming Vol 2, section 4.2.2."
   */
  public static SimpleStatistics compute(List<? extends Number> values) {
    if (values.isEmpty()) return new SimpleStatistics();

    int count = values.size();

    double min = Double.POSITIVE_INFINITY;
    for (Number each : values) {
      if (each.doubleValue() < min) min = each.doubleValue();
    }

    double max = Double.NEGATIVE_INFINITY;
    for (Number each : values) {
      if (each.doubleValue() > max) max = each.doubleValue();
    }

    double sum = 0.0;
    for (Number each : values) {
      sum += each.doubleValue();
    }

    double mean = sum / (double)count;

    double[] sqMeanDiffs = new double[count];
    int i = 0;
    for (Number each : values) {
      sqMeanDiffs[i] = Math.pow(each.doubleValue() - mean, 2);
      i++;
    }

    double sqMeanDiffsSum = 0.0;
    for (double each : sqMeanDiffs) {
      sqMeanDiffsSum += each;
    }

    double variance = sqMeanDiffsSum / count;

    double stddev = Math.sqrt(variance);

    return new SimpleStatistics(count, min, max, sum, mean, stddev);
  }
}
